:-prolog_flag(compiling,X,compactcode).
:-prolog_flag(redefine_warnings,X,off).
:-prolog_flag(single_var_warnings,X,off).

go:-
    compile(benches),
    open(res_sics,write,S),
    run(S).

run(S):-
    bench_mark(Name, Iterations, Action, Control),
    write(user,runing(Name)),nl(user),
    get_cpu_time(T0),
    (repeat(Iterations), call(Action), fail
      ;   get_cpu_time(T1)
    ),
    (   repeat(Iterations), call(Control), fail
      ;   get_cpu_time(T2)
    ),
    report(S,Name,Iterations, T0, T1, T2),
    fail.
run(S):-close(S).

ratios(F1,F2):-
    readall(F1,L1),
    readall(F2,L2),
    write(ratios),nl,
    compute_ratios(L1,L2,0,0).

readall(F,L):-
    see(F),
    read(X),
    read_data(X,L).

read_data(end_of_file,L):-!,L=[],seen.
read_data(X,L):-
    L=[X|L1],
    read(NewX),
    read_data(NewX,L1).

compute_ratios([],[],Ps,Mean):-write('number of programs:'),write(Ps),nl,
    Temp is Mean/Ps, write('mean ratio:'), write(Temp).
compute_ratios([X|Xs],[Y|Ys],Ps,Mean):-
    functor(X,F,N),
    functor(Y,F,N),
    arg(1,X,Ax),
    arg(1,Y,Ay),
    (Ay=:=0->R=1;R is Ax/Ay),
    write(F),write('('),write(R),write(')'),write('.'),nl,
    Ps1 is Ps+1,
    Mean1 is Mean+R,
    compute_ratios(Xs,Ys,Ps1,Mean1).

%   repeat(N)
%   succeeds precisely N times.

repeat(N) :-
        N > 0,
        from(1, N).

from(I, I) :- !.
from(L, U) :- M is (L+U)>>1,   from(L, M).
from(L, U) :- M is (L+U)>>1+1, from(M, U).

get_cpu_time(T) :-
        statistics(runtime, [T,_]).

report(Stream,Name,N, T0, T1, T2) :-
    TestTime is T1-T0,
    OverHead is T2-T1,
    Average  is (TestTime-OverHead)/N,
    put(Stream,39),
    write(Stream,Name),
    put(Stream,39),
    write(Stream,'('), write(Stream,Average), write(Stream,').'),nl(Stream).

