%Subject: source program of Boyer (boyer.pro)
/*
:-mode(main,0,[]).
:-mode(wff,1,[f]).
:-mode(tautology,1,[c]).
:-mode(rewrite,2,[c,d]).
:-mode(rewrite_args,3,[c,c,d]).
:-mode('boyer.pl10001',3,[f,d,d]).
:-mode(equal,2,[c,f]).
:-mode(difference,3,[c,c,f]).
:-mode(eq,3,[c,c,f]).
:-mode(exp,3,[c,c,f]).
:-mode(gcd,3,[c,c,f]).
:-mode(mylength,2,[c,f]).
:-mode(lessp,3,[c,c,f]).
:-mode(meaning,3,[c,c,f]).
:-mode(mymember,3,[c,c,f]).
:-mode(nth,3,[c,c,f]).
:-mode(plus,3,[c,c,f]).
:-mode(power_eval,3,[c,c,f]).
:-mode(quotient,3,[c,c,f]).
:-mode(remainder,3,[c,c,f]).
:-mode(reverse_loop,3,[c,c,f]).
:-mode(times,3,[c,c,f]).
:-mode(tautology,3,[c,c,c]).
:-mode(truep,2,[c,c]).
:-mode(falsep,2,[c,c]).
:-mode('boyer.pl10000',5,[c,c,c,c,c]).
:-mode(member,2,[c,c]).
*/
/* Copyright Herve' Touati, Aquarius Project, UC Berkeley */

%------------------------------------------------------------------------------
%	Benchmark Program - Boyer
%	Lisp vs. Prolog Study
%
%	Copyright by Evan Tick
%	Date: November 12 1985
%
%------------------------------------------------------------------------------
/*
:- public main/0.

:- mode equal(c,d).
:- mode rewrite([c,d],[c,c]).
*/

go:- true :
    cputime(Start),
    main,
    cputime(End),
    T is End-Start,
    write('execution time is '),write(T), write(milliseconds).

top :-	true : wff(X),tautology(X).

wff(WFF):-
    true :
    WFF:=implies(and(implies(X,Y),
		and(implies(Y,Z),
		    and(implies(Z,U),
			implies(U,W)))),
		 implies(X,W)),
	X = f(plus(plus(a,b),plus(c,zero))),
	Y = f(times(times(a,b),plus(c,d))),
	Z = f(reverse(append(append(a,b),[]))),
	U = equal(plus(a,b),difference(x,y)),
	W = lessp(remainder(a,b),member(a,length(b))).

tautology(Wff) :-
    true :
    write('rewriting...'),nl,
    rewrite(Wff,NewWff),
    write('proving...'),nl,
%    write(NewWff),nl,
    tautology(NewWff,[],[]).


tautology(Wff,Tlist,Flist) :-
    truep(Wff,Tlist) : true.
tautology(Wff,Tlist,Flist) :-
    falsep(Wff,Flist) : fail.
tautology(Wff,Tlist,Flist) :-
    if(If,Then,Else)<=Wff,
    truep(If,Tlist) : tautology(Then,Tlist,Flist).
tautology(Wff,Tlist,Flist) :-
    if(If,Then,Else)<=Wff,
    falsep(If,Flist) : tautology(Else,Tlist,Flist).
tautology(Wff,Tlist,Flist) :-
    if(If,Then,Else)<=Wff :
    tautology(Then,[If|Tlist],Flist),	% both must hold
    tautology(Else,Tlist,[If|Flist]).



rewrite(Atom,Atom1) :-
    atomic(Atom) :
    Atom1 = Atom.
rewrite(Old,New) :-
    true :
    functor(Old,F,N),
    functor(Mid,F,N),
%	write('Old ='),write(F),nl,
    rewrite_args(N,Old,Mid),
    rewrite_1(Mid,Next,New).

rewrite_1(Mid,Next,New) :-
    equal(Mid,Next) :	% should be ->, but is compiler smart enough?
    rewrite(Next,New).	% to generate cut for ->?
rewrite_1(Mid,Next,New) :-
    true :
    New=Mid.

rewrite_args(0,_,_) :- 
    true : true.
rewrite_args(N,Old,Mid) :- 
    true :
    arg(N,Old,OldArg),
    arg(N,Mid,MidArg),
    rewrite(OldArg,MidArg),
    N1 is N-1,
    rewrite_args(N1,Old,Mid).


truep(t,_) :- 
    true : true.
truep(Wff,Tlist) :- 
    true : member(Wff,Tlist).

falsep(f,_) :- 
    true : true.
falsep(Wff,Flist) :- 
    true : member(Wff,Flist).

member(X,[X|_]) :- 
    true : true.
member(X,[_|T]) :- 
    true :
    member(X,T).


equal(	and(P,Q),			% 106 rules
	Wff):-
    true : Wff:=if(P,if(Q,t,f),f).
equal(	append(append(X,Y),Z),
	Wff):-
    true : Wff:=append(X,append(Y,Z)).
equal(	assignment(X,append(A,B)),
	Wff):-
    true : Wff:=if(assignedp(X,A),
		   assignment(X,A),
		   assignment(X,B)).
equal(	assume_false(Var,Alist),
	Wff):-
    true : Wff:=cons(cons(Var,f),Alist).
equal(	assume_true(Var,Alist),
	Wff):-
    true : Wff:=cons(cons(Var,t),Alist).
equal(	boolean(X),
	Wff):-
    true : Wff:=or(equal(X,t),equal(X,f)).
equal(	car(gopher(X)),
	Wff):-
    true : Wff:=if(listp(X),
	   car(flatten(X)),
	   zero).
equal(	compile(Form),
	Wff):-
    true : Wff:=reverse(codegen(optimize(Form),[])).
equal(	count_list(Z,sort_lp(X,Y)),
	Wff):-
    true : Wff:=plus(count_list(Z,X),
	      count_list(Z,Y)).
equal(	countps_(L,Pred),
      	Wff):-
    true : Wff:=countps_loop(L,Pred,zero).
equal(	difference(A,B),
	C
     ) :- 
    true : difference(A,B,C).
equal(	divides(X,Y),
	Wff):-
    true : Wff:=zerop(remainder(Y,X)).
equal(	dsort(X),
	Wff):-
    true : Wff:=sort2(X).
equal(	eqp(X,Y),
	Wff):-
    true : Wff:=equal(fix(X),fix(Y)).
equal(	equal(A,B),C) :-
    true : eq(A,B,C).
equal(	even1(X),
	Wff):-
    true : Wff:=if(zerop(X),t,odd(decr(X))).
equal(	exec(append(X,Y),Pds,Envrn),
	Wff):-
    true : Wff:=exec(Y,exec(X,Pds,Envrn),Envrn).
equal(	exp(A,B),
	C) :- 
    true : exp(A,B,C).
equal(	fact_(I),
	Wff):-
    true : Wff:=fact_loop(I,1).
equal(	falsify(X),
	Wff):-
    true : Wff:=falsify1(normalize(X),[]).
equal(	fix(X),
	Wff):-
    true : Wff:=if(numberp(X),X,zero).
equal(	flatten(cdr(gopher(X))),
	Wff):-
    true : Wff:=if(listp(X),
	   cdr(flatten(X)),
	   cons(zero,[])).
equal(  gcd(A,B),
	C
	) :- 
%	write('==> gcd '), write(A), nl,write(B),nl,
    true :
    gcd(A,B,C).
%	write('<== gcd '), write(A), nl,write(B),nl,write(C),nl.
equal(	get(J,set(I,Val,Mem)),
	Wff):-
    true : Wff:=if(eqp(J,I),Val,get(J,Mem)).
equal(	greatereqp(X,Y),
	Wff):-
    true : Wff:=not(lessp(X,Y)).
equal(	greatereqpr(X,Y),
	Wff):-
	true : Wff:=not(lessp(X,Y)).
equal(	greaterp(X,Y),
	Wff):-
    true : Wff:=lessp(Y,X).
equal(	if(if(A,B,C),D,E),
	Wff):-
    true : Wff:=if(A,if(B,D,E),if(C,D,E)).
equal(	iff(X,Y),
	Wff):-
    true : Wff:=and(implies(X,Y),implies(Y,X)).
equal(	implies(P,Q),
	Wff):-
    true : Wff:=if(P,if(Q,t,f),t).
equal(	last(append(A,B)),
	Wff):-
    true : Wff:= if(listp(B),
		    last(B),
		    if(listp(A),
		       cons(car(last(A))),
		       B)).
equal(	length(A),
	B
	) :- 
%	write('==> mylength '), write(A), nl,
    true :
	mylength(A,B).
%	write('<== mylength '), write(A), nl,write(B),nl.

equal(	lesseqp(X,Y),
	Wff):-
    true :
	Wff:=not(lessp(Y,X)).
equal(	lessp(A,B),
	C
	) :- 
%	write('==> lessp '), write(A), nl,write(B),nl,
    true :
	lessp(A,B,C).
%	write('<== lessp '), write(A), nl,write(B),nl,write(C),nl.

equal(	listp(gopher(X)),
	Wff):-
    true : Wff:=listp(X).
equal(	mc_flatten(X,Y),
	Wff):-
    true :
	Wff:=append(flatten(X),Y).
equal(	meaning(A,B),
	C
	) :- 
%	write('==> meaning '), write(A), nl,write(B),nl,
    true :
	meaning(A,B,C).
%	write('<== meaning '), write(A), nl,write(B),nl,write(C),nl.

equal(  member(A,B),
	C
	) :- 
%	write('==> mymember '), write(A), nl,write(B),nl,
    true :
	mymember(A,B,C).
%	write('<== mymeber '), write(A), nl,write(B),nl,write(C),nl.
	
equal(	not(P),
	Wff):-
    true : Wff:=if(P,f,t).
equal(  nth(A,B),
	C
	) :- 
%	write('==> nth '), write(A), nl,write(B),nl,
    true :
	nth(A,B,C).
%	write('<== nth '), write(A), nl,write(B),nl,write(C),nl.

equal(	numberp(greatest_factor(X,Y)),
	Wff):-
    true :
	Wff:=not(and(or(zerop(Y),equal(Y,1)),
		 not(numberp(X)))).
equal(	or(P,Q),
	Wff):-
    true : 
    Wff:=if(P,t,if(Q,t,f),f).
equal(  plus(A,B),
	C
	) :- 
%	write('==> plus '), write(A), nl,write(B),nl,
    true :
	plus(A,B,C).
%	write('<== plus '), write(A), nl,write(B),nl,write(C),nl.
equal(  power_eval(A,B),
	C
	) :- 
%	write('==> power_eval '), write(A), nl,write(B),nl,
    true :
	power_eval(A,B,C).
%	write('<== power_eval '), write(A), nl,write(B),nl,write(C),nl.
equal(	prime(X),
	Wff):-
    true : Wff:= and(not(zerop(X)),
		     and(not(equal(X,add1(zero))),
			 prime1(X,decr(X)))).
equal(	prime_list(append(X,Y)),
	Wff):-
    true : Wff:=and(prime_list(X),prime_list(Y)).
equal(  quotient(A,B),
	C
	) :- 
%	write('==> quotient '), write(A), nl,write(B),nl,
    true :
	quotient(A,B,C).
%	write('<== quotient '), write(A), nl,write(B),nl,write(C),nl.
equal(  remainder(A,B),
	C
	) :- 
%	write('==> remainder '), write(A), nl,write(B),nl,
    true :
	remainder(A,B,C).
%	write('<== remainder '), write(A), nl,write(B),nl,write(C),nl.
equal(	reverse_(X),
	Wff):-
    true : Wff:=reverse_loop(X,[]).
equal(	reverse(append(A,B)),
	Wff):-
    true : Wff:=append(reverse(B),reverse(A)).
equal(	reverse_loop(A,B),
	C
	) :- 
    true : reverse_loop(A,B,C).
equal(	samefringe(X,Y),
	Wff):-
    true : Wff:=equal(flatten(X),flatten(Y)).
equal(	sigma(zero,I),
	Wff):-
    true : Wff:=quotient(times(I,add1(I)),2).
equal(	sort2(delete(X,L)),
	Wff):-
    true : Wff:=delete(X,sort2(L)).
equal(	tautology_checker(X),
	Wff):-
    true : Wff:=tautologyp(normalize(X),[]).
equal(  times(A,B),C) :- 
    true : times(A,B,C).
equal(	times_list(append(X,Y)),
	Wff):-
    true : Wff:=times(times_list(X),times_list(Y)).
equal(	value(normalize(X),A),
	Wff):-
    true : Wff:=value(X,A).
equal(	zerop(X),
	Wff):-
    true : Wff:=or(equal(X,zero),not(numberp(X))).

difference(X,			X,		Wff):-
    true : Wff=zero.
difference(plus(X,Y),		X,		Wff):-
    true : Wff:=fix(Y).
difference(plus(Y,X),		X,		Wff):-
    true : Wff:=fix(Y).
difference(plus(X,Y),		plus(X,Z),	Wff):-
    true : Wff:=difference(Y,Z).
difference(plus(B,plus(A,C)),	A,		Wff):-
    true : Wff:=plus(B,C).
difference(add1(plus(Y,Z)),	Z,		Wff):-
    true : Wff:=add1(Y).
difference(add1(add1(X)),	2,		Wff):-
    true : Wff:=fix(X).

eq(plus(A,B),		zero,		Wff):-
    true : Wff:=and(zerop(A),zerop(B)).
eq(plus(A,B),		plus(A,C),	Wff):-
    true : Wff:=equal(fix(B),fix(C)).
eq(zero,		difference(X,Y),Wff):-
    true : Wff:=not(lessp(Y,X)).
eq(X,			difference(X,Y),Wff):-
    true : Wff:=and(numberp(X),
		    and(or(equal(X,zero),
			   zerop(Y)))).
eq(times(X,Y),		zero,		Wff):-
    true : Wff:=or(zerop(X),zerop(Y)).
eq(append(A,B),		append(A,C),	Wff):-
    true : Wff:=equal(B,C).
eq(flatten(X),		cons(Y,[]),	Wff):-
    true : Wff:=and(nlistp(X),equal(X,Y)).
eq(greatest_factor(X,Y),zero,		Wff):-
    true : Wff:=and(or(zerop(Y),equal(Y,1)),
		    equal(X,zero)).
eq(greatest_factor(X,_),1,		Wff):-
    true : Wff:=equal(X,1).
eq(Z,			times(W,Z),	Wff):-
    true : Wff:=and(numberp(Z),
		    or(equal(Z,zero),
		       equal(W,1))).
eq(X,			times(X,Y),	Wff):-
    true : Wff:=or(equal(X,zero),
		   and(numberp(X),equal(Y,1))).
eq(times(A,B),		1,		Wff):-
    true : Wff:=and(not(equal(A,zero)),
		    and(not(equal(B,zero)),
			and(numberp(A),
			    and(numberp(B),
				and(equal(decr(A),zero),
				    equal(decr(B),zero)))))).
eq(difference(X,Y),	difference(Z,Y),Wff):-
    true : Wff:=if(lessp(X,Y),
		   not(lessp(Y,Z)),
		   if(lessp(Z,Y),
		      not(lessp(Y,X)),
		      equal(fix(X),fix(Z)))).
eq(lessp(X,Y),		Z,		Wff):-
    true : Wff:=if(lessp(X,Y),
		   equal(t,Z),
		   equal(f,Z)).

exp(I,		plus(J,K),	Wff):-
    true : Wff:=times(exp(I,J),exp(I,K)).
exp(I,		times(J,K),	Wff):-
    true : Wff:=exp(exp(I,J),K).

gcd(X,		Y,		Wff):-
    true : Wff:=gcd(Y,X).
gcd(times(X,Z),	times(Y,Z),	Wff):-
    true : Wff:=times(Z,gcd(X,Y)).

mylength(reverse(X),Wff):-
    true : Wff:=length(X).
mylength(cons(_,cons(_,cons(_,cons(_,cons(_,cons(_,X7)))))),
	 Wff):-
    true : Wff:=plus(6,length(X7)).

lessp(remainder(_,Y),	Y,	Wff):-
    true : Wff:=not(zerop(Y)).
lessp(quotient(I,J),	I,	Wff):-
    true : Wff:=and(not(zerop(I)),
		    or(zerop(J),
		       not(equal(J,1)))).
lessp(remainder(X,Y),	X,	Wff):-
    true : Wff:=and(not(zerop(Y)),
	       and(not(zerop(X)),
		   not(lessp(X,Y)))).
lessp(plus(X,Y),	plus(X,Z),	Wff):-
    true : Wff:=lessp(Y,Z).
lessp(times(X,Z),	times(Y,Z),	Wff):-
    true : Wff:=and(not(zerop(Z)),
		    lessp(X,Y)).
lessp(Y,		plus(X,Y),	Wff):-
    true : Wff:=not(zerop(X)).
lessp(length(delete(X,L)),      length(L),	Wff):-
    true : Wff:=member(X,L).


meaning(plus_tree(append(X,Y)),A,
	Wff):-
    true : Wff:=plus(meaning(plus_tree(X),A),
		     meaning(plus_tree(Y),A)).
meaning(plus_tree(plus_fringe(X)),A,
	Wff):-
    true : Wff:=fix(meaning(X,A)).
meaning(plus_tree(delete(X,Y)),A,
	Wff):-
    true : Wff:=if(member(X,Y),
	      difference(meaning(plus_tree(Y),A),
			 meaning(X,A)),
	      meaning(plus_tree(Y),A)).

mymember(X,append(A,B),Wff):-
    true : Wff:=or(member(X,A),member(X,B)).
mymember(X,reverse(Y),Wff):-
    true : Wff:=member(X,Y).
mymember(A,intersect(B,C),Wff):-
    true : Wff:=and(member(A,B),member(A,C)).

nth(zero,_,Wff):-true : Wff:=zero.
nth([],I,Wff):-
    true : Wff:=if(zerop(I),[],zero).
nth(append(A,B),I,Wff):-
    true : Wff:=append(nth(A,I),nth(B,difference(I,length(A)))).

plus(plus(X,Y),Z,
     Wff):-
    true : Wff:=plus(X,plus(Y,Z)).
plus(remainder(X,Y),
     times(Y,quotient(X,Y)),
     Wff):-
    true : Wff:=fix(X).
plus(X,add1(Y),
     Wff):-
    true : Wff:=if(numberp(Y),
		   add1(plus(X,Y)),
		   add1(X)).

power_eval(big_plus1(L,I,Base),Base,
	   Wff):-
    true : Wff:=plus(power_eval(L,Base),I).
power_eval(power_rep(I,Base),Base,
	   Wff):-
    true : Wff:=fix(I).
power_eval(big_plus(X,Y,I,Base),Base,
	   Wff):-
    true : Wff:=plus(I,plus(power_eval(X,Base),
			    power_eval(Y,Base))).
power_eval(big_plus(power_rep(I,Base),
			    power_rep(J,Base),
			    zero,
			    Base),
	   Base,
	   Wff):-
    true : Wff:=plus(I,J).


quotient(plus(X,plus(X,Y)),2,Wff):-
    true : Wff:=plus(X,quotient(Y,2)).
quotient(times(Y,X),Y,Wff):-
    true : Wff:=if(zerop(Y),zero,fix(X)).

remainder(_,		1,Wff):-
    true : Wff=zero.
remainder(X,		X,Wff):-
    true : Wff=zero.
remainder(times(_,Z),	Z,Wff):-
    true : Wff=zero.
remainder(times(Y,_),	Y,Wff):-
    true : Wff=zero.

reverse_loop(X,Y,	Wff):-
    true : Wff:=append(reverse(X),Y).
reverse_loop(X,[],	Wff):-
    true : Wff:=reverse(X).

times(X,	 plus(Y,Z),	 Wff):-
    true : Wff:=plus(times(X,Y),times(X,Z)).
times(times(X,Y),Z,		 Wff):-
    true : Wff:=times(X,times(Y,Z)).
times(X,	 difference(C,W),Wff):-
    true : Wff:=difference(times(C,X),times(W,X)).
times(X,	 add1(Y),        Wff):-
    true : Wff:=if(numberp(Y),
		   plus(X,times(X,Y)),
		   fix(X)).


