/* Sterling and Shapiro's N-queens algorithm */
/*
:- public go/0.
:-mode(go,1,[c]).
:-mode(queens,2,[c,f]).
:-mode(queens,3,[c,c,f]).
:-mode(select1,3,[f,c,f]).
:-mode('queens.pl10000',2,[c,c]).
:-mode(attack,2,[c,c]).
:-mode(attack,3,[c,c,c]).
*/
go:-
    statistics(runtime,[_,Start]),
    allqueens,
    statistics(runtime,[_,T]),
    write('execution time is '),write(T), write(milliseconds).
    
top:-allqueens.

allqueens:-
    queens(8,Q),
    fail.
allqueens.

queens(N,Qs):-
	range(1,N,Ns),
	queens(Ns,[],Qs).

queens(UnplacedQs,SafeQs,Qs):-
	select1(Q,UnplacedQs,UnplacedQs1),
        not_attack(Q,SafeQs),
	queens(UnplacedQs1,[Q|SafeQs],Qs).
queens([],Qs,Qs).

not_attack(Q,SafeQs):-
	attack(Q,SafeQs),!,
	fail.
not_attack(_,_).


/*:-mode range(c,c,f).*/
range(M,N,[M|Ns]):-
	M<N,
	M1 is M+1,
	range(M1,N,Ns).
range(M,N,[N]):-
	M =:= N.

select1(X,[X|Xs],Xs).
select1(X,[Y|Ys],[Y|Xs]):-
	select1(X,Ys,Xs).

attack(X,Xs):-
	attack(X,1,Xs).

attack(X,N,[Y|Ys]):-
	X=:=Y+N,!.
attack(X,N,[Y|Ys]):-
	X=:=Y-N,!.
attack(X,N,[Y|Ys]):-
	N1 is N+1,
	attack(X,N1,Ys).

