/*
 * plc_java.c: JIPL native interface for both Java and K-Prolog
 *
 * Copyright (c) 1997 Nobukuni Kino, ISAC,Inc.
 *
 * Permission to use, copy, modify, and distribute this
 * release of the software for RESEARCH/EVALUATION
 * purposes and without fee is hereby granted provided
 * that this copyright notice appears in all copies.
 * Notice that EDUCATIONAL purpose usage is NOT granted.
 *
 * The Java package name, CO.jp.isac.plc, is reserved for
 * ISAC, Inc. So you should rename the package when you
 * apply JIPL to other language processors.
 *
 * Nobukuni Kino and ISAC, Inc. make no warranties
 * about the suitability of the software, either
 * expressed or implied, fitness for a particular
 * purpose.
 * Nobukuni Kino and ISAC, Inc. shall not be liable for
 * any damages suffered by licensee as a result of using,
 * modifying or distributing this software or its
 * derivatives.
 */

#include <native.h>
#include "cpreds.h"
#include "kapi.h"
extern SYM_REC_PTR  objectRef;

static jclass class_Object;
static jclass class_ArrayOfObject;
static jclass class_String;
static jmethodID conv_String;
static jmethodID bytes_String;
static jclass class_Integer;
static jmethodID init_Integer;
static jclass class_BigInteger;
static jmethodID init_BigInteger;
static jclass class_Double;
static jmethodID init_Double;
static jclass class_Number;
static jclass class_Boolean;
static jclass class_Character;
static jclass class_Exception;
static jclass class_Plc;
static jmethodID result_Plc;
static jmethodID instance_Plc;
static jmethodID static_Plc;
static jmethodID set_Plc;
static jmethodID get_Plc;
static jmethodID new_Plc;
static JNIEnv	*Env;
static jobject  encoding;
double PvalueOfReal();

static void
java2prolog(JNIEnv *env, jobject val, CELL arg) {
	const char *str;
	if (!val) return;
	if ((*env)->IsInstanceOf(env, val, class_Number)) {
		jobject obj = (*env)->CallObjectMethod(env, val,
			(*env)->GetMethodID(env, (*env)->GetObjectClass(env,val),
				"toString", "()Ljava/lang/String;"),
			0);
		if (obj) str = (*env)->GetStringUTFChars(env,obj,NULL);
		else str = "invalid argument";
		/* !!! */
		PnameToTerm(str, arg);
		if (obj) {
			(*env)->ReleaseStringUTFChars(env, obj, str);
			(*env)->DeleteLocalRef(env, obj);
		}
	} else if ((*env)->IsInstanceOf(env, val, class_Boolean)) {
		jboolean b = (*env)->CallBooleanMethod(env, val,
			(*env)->GetMethodID(env, class_Boolean,
				"booleanValue","()Z"),
			0);
		PuInt(arg, (int)b);
	} else if ((*env)->IsInstanceOf(env, val, class_String) ||
				(*env)->IsInstanceOf(env, val, class_Character)) {
		if ((*env)->IsInstanceOf(env, val, class_Character)) {
			val = (*env)->CallObjectMethod(env, val,
				(*env)->GetMethodID(env, class_Character,
					"toString", "()Ljava/lang/String;"),0);
		}
#if !defined(PlatformCString)
		{
			jarray array = (*env)->CallObjectMethod(env,
				val, bytes_String, encoding); 
			jint len = (*env)->GetArrayLength(env,array);
			str = (char*)calloc(len*2+1,sizeof(jbyte));
			(*env)->GetByteArrayRegion(env, array, 0, len, (jbyte*)str);
/*			PuStr(arg, str); */
			PuAtom(arg, str); 
			free((void*)str);
		}
		/* Or simply try this. Sufficient for ASCII users...
		/* str = (*env)->GetStringUTFChars(env,val,NULL);
		/* PuStr(arg, str);
		/* (*env)->ReleaseStringUTFChars(env, val, str); */
#else
		/* This works in Solaris, faster */
		str = (jbyte*)makePlatformCString(DeRef(env, val));
/*		PuStr(arg, str); */
		PuAtom(arg, str); 
		free((void*)str);
#endif
	} else if ((*env)->IsInstanceOf(env, val,class_ArrayOfObject)) {
		jsize i;
		CELL list;
		jsize len = (*env)->GetArrayLength(env, val);
		list = Plistn(len);
		Punify(arg,list);
		for(i=0; i<len; i++) {
			CELL car = PlistCar(arg);
			java2prolog(env, (*env)->GetObjectArrayElement(env,val, i), car);
			arg = PlistCdr(arg);
		}
	} else /* if ((*env)->IsInstanceOf(env, val,class_Object)) */ {
#if defined(DEBUG)
		jobject mess = (*env)->CallObjectMethod(env,val,
			(*env)->GetMethodID(env,(*env)->GetObjectClass(env,val),
				"toString","()Ljava/lang/String;"), 0);
		const char *str = (*env)->GetStringUTFChars(env,mess,NULL);
		(*env)->DeleteLocalRef(env, mess);
#endif
		PuAddr(arg, (*env)->NewGlobalRef(env, val)); /* +1 */
	}
}

jref 
addr2ref(JNIEnv *env, CELL term) {
  register unsigned int w1,w2;

  UNTAG(term);
  w1 = INTVAL(*((LONG_PTR)term+1));
  w2 = INTVAL(*((LONG_PTR)term+2));
  return (jref)(w2<<16 | w1);
}


static jref
prolog2java(JNIEnv *env, CELL arg) {
	const char *str;
	register LONG_PTR top;
	DEREF(arg);
	if (PisInt(arg)) {
		if (PisLong(arg)) {
			jref res;
			jref bignum;
			str = PnumberToStr(arg);
			res = (*env)->NewStringUTF(env, str);
			bignum = (*env)->NewObject(env,class_BigInteger, init_BigInteger,res);
			(*env)->DeleteLocalRef(env, res);
			return bignum;
		} else {
			return (*env)->NewObject(env,class_Integer, init_Integer, 
				(jint)PvalueOfInt(arg));
		}
	} else if (PisReal(arg)) {
		return (*env)->NewObject(env,class_Double, init_Double, 
			(jdouble)PvalueOfReal(arg));
	} else if (PisAddr(arg)) {
		return addr2ref(env, arg);
	} else if (PisAtom(arg)){
	       str=get_NAME(arg);
		/* Or simply try this. Sufficient for ASCII users...*/
	       return (*env)->NewStringUTF(env, str);
/*
#if !defined(PlatformCString)
		 converting EUCJIS/SJIS string to a String object 
		jsize len = (jsize)strlen(str);
		jarray array = (*env)->NewByteArray(env,len);
		jstring strobj;
		(*env)->SetByteArrayRegion(env, array, 0, len, (jbyte*)str);
		strobj = (*env)->NewObject(env,class_String,conv_String, array, encoding);
		(*env)->DeleteLocalRef(env, array);
		return strobj;
#else
		/* This works in Solaris, faster 
		return MkRefLocal(env,
			(void*)makeJavaStringFromPlatformCString((char*)str, strlen(str)),
			JNI_REF_HANDLE_TAG);
#endif
*/
	} else if (PisList(arg)) {
		int len = PlistLength(arg);
		jref array = (*env)->NewObjectArray(env, len, class_Object, 0);
		int i;
		for(i=0; i<len; i++) {
			jref res = prolog2java(env, PlistCar(arg));
			(*env)->SetObjectArrayElement(env, array, i, res);
			(*env)->DeleteLocalRef(env, res);
			arg = PlistCdr(arg);
		}
		return array;
	} else if (PisCompound(arg)) {
	  LONG_PTR  ptr= UNTAGGED(arg);
	  int len = GET_ARITY((SYM_REC_PTR)FOLLOW(ptr));
	  jref array = (*env)->NewObjectArray(env, len, class_Object, 0);
	  int i;
	  for(i=0; i<len; i++) {
	    jref res = prolog2java(env, FOLLOW(ptr+i+1));
	    (*env)->SetObjectArrayElement(env, array, i, res);
	    (*env)->DeleteLocalRef(env, res);
	  }
	  return array;
	} else {
	  return (*env)->NewStringUTF(env, "?");
	}
      }

static jref
prolog2javaArray(JNIEnv *env, CELL arg) {
  register LONG_PTR top;
  DEREF(arg);
  if (PisCompound(arg)) return prolog2java(env,arg);
  else return (*env)->NewObjectArray(env, 0, class_Object, 0);
}

static void
initialize(JNIEnv *env, jref arrayobj) {
	class_Object = (*env)->NewGlobalRef(env,
		(*env)->FindClass(env, "java/lang/Object"));
	class_String = (*env)->NewGlobalRef(env,
		(*env)->FindClass(env, "java/lang/String"));
	class_Integer = (*env)->NewGlobalRef(env,
		(*env)->FindClass(env, "java/lang/Integer"));
	class_BigInteger = (*env)->NewGlobalRef(env,
		(*env)->FindClass(env, "java/math/BigInteger"));
	class_Double = (*env)->NewGlobalRef(env,
		(*env)->FindClass(env, "java/lang/Double"));
	class_Number = (*env)->NewGlobalRef(env,
		(*env)->FindClass(env, "java/lang/Number"));
	class_Boolean = (*env)->NewGlobalRef(env,
		(*env)->FindClass(env, "java/lang/Boolean"));
	class_Character = (*env)->NewGlobalRef(env,
		(*env)->FindClass(env, "java/lang/Character"));
	class_Exception = (*env)->NewGlobalRef(env,
		(*env)->FindClass(env, "java/lang/Exception"));
#if FindClassResolved 
	class_ArrayOfObject = (*env)->FindClass(env, "[Ljava/lang/Object;");
#else
	class_ArrayOfObject =(*env)->NewGlobalRef(env,
		(*env)->GetObjectClass(env,arrayobj));
#endif
	class_Plc = (*env)->NewGlobalRef(env,
		(*env)->FindClass(env,"bprolog/plc/Plc"));
	{
		Env = env;
		init_Integer = (*env)->GetMethodID(env,class_Integer, "<init>", "(I)V");
		init_BigInteger = (*env)->GetMethodID(env,class_BigInteger, "<init>", "(Ljava/lang/String;)V");
		init_Double = (*env)->GetMethodID(env,class_Double, "<init>", "(D)V");
		conv_String = (*env)->GetMethodID(env,class_String, "<init>", "([BLjava/lang/String;)V");
		bytes_String = (*env)->GetMethodID(env,class_String, "getBytes", "(Ljava/lang/String;)[B");
		result_Plc = (*env)->GetMethodID(env,class_Plc,"setResult",
			"(ILjava/lang/Object;)V");
		instance_Plc = (*env)->GetStaticMethodID(env,class_Plc,"instanceMethod",
			"(Ljava/lang/Object;Ljava/lang/String;[Ljava/lang/Object;)Ljava/lang/Object;");
		static_Plc = (*env)->GetStaticMethodID(env,class_Plc,"staticMethod",
			"(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Object;)Ljava/lang/Object;");
		set_Plc = (*env)->GetStaticMethodID(env,class_Plc,"setField",
			"(Ljava/lang/Class;Ljava/lang/Object;Ljava/lang/String;Ljava/lang/Object;)V");
		get_Plc = (*env)->GetStaticMethodID(env,class_Plc,"getField",
			"(Ljava/lang/Class;Ljava/lang/Object;Ljava/lang/String;)Ljava/lang/Object;");
		new_Plc = (*env)->GetStaticMethodID(env,class_Plc,"javaConstructor",
			"(Ljava/lang/Class;[Ljava/lang/Object;)Ljava/lang/Object;");
	}

#if defined(DEBUG) 
	printf("%8x,%8x,%8x,%8x\n",class_ArrayOfObject,class_Plc,class_String,bytes_String);
	printf("%8x,%8x,%8x,%8x, env=%x\n",instance_Plc,static_Plc,new_Plc, env);
#endif
}
static int
callJavaException(JNIEnv *env) {
	const char *str;
	jobject mess = NULL;
	jobject e = (*env)->ExceptionOccurred(env);
	CELL command = Pfunctorn(Patom("jiplException"),1);
	if (e) {
		jclass exclass = (*env)->GetObjectClass(env,e);
		(*env)->CallObjectMethod(env,e,
			(*env)->GetMethodID(env,exclass,"printStackTrace","()V"),0);
		mess = (*env)->CallObjectMethod(env,e,
			(*env)->GetMethodID(env,exclass,"toString","()Ljava/lang/String;"),
			0);
	} 
	if (mess) str = (*env)->GetStringUTFChars(env,mess,NULL);
	else str = "unknown";
	/* !!! */
/*	PuStr(PargOfCompound(command, 0),str); */
	PuAtom(PargOfCompound(command, 0),str); 
	if (mess) (*env)->ReleaseStringUTFChars(env, mess, str);
	call_bprolog_term(command);
	return FALSE;
}
static jclass
findClass(JNIEnv *env, const char *name) {
	jclass clazz = (*env)->FindClass(env,name);
	if (!clazz) {
		char *lname = (char*)malloc(strlen(name)+32);
		sprintf(lname,"java/lang/%s", name);
		clazz = (*env)->FindClass(env,lname);
		free(lname);
	}
	if (!clazz) {
		(*env)->ThrowNew(env, (*env)->FindClass(env,
			"java/lang/ClassNotFoundException"),name);
	}
	return clazz;
}
int
javaMethod(){
  TERM obj = ARG(1,3);
  TERM method = ARG(2,3);
  TERM result = ARG(3,3);

	JNIEnv *env = Env;
	const char *functor = (const char*)PascOfFunc(method);
	/* !!! */
	jstring funcname = (*env)->NewStringUTF(env,functor);
	jref args;
	jobject res;
	jref prolog2javaArray();
	register LONG_PTR top;
	DEREF(obj);

	args = prolog2javaArray(env,method);
	
	if ((*env)->ExceptionOccurred(env)) {
		return callJavaException(env);
	}

#if defined(DEBUG)
	{ static int count;
	fprintf(stderr,"count = %d\n",++count);
	fprintf(stderr,"globalRefTable.size = %d\n",globalRefTable.size);
	fprintf(stderr,"localRefTable.size = %d\n",JNIEnvGetLocalRefs(env)->size);
	}
#endif

	if (PisAddr(obj)) { /* assuming, obj is an Instance */
	  res = (*env)->CallStaticObjectMethod(env,class_Plc,instance_Plc,
			addr2ref(env, obj), funcname, args);
	} else {
		const char *class = (const char*)PascOfAtom(obj);
		/* !!! */
		jclass clazz = findClass(env,class);
		if (!clazz) return callJavaException(env);
		res =  (*env)->CallStaticObjectMethod(env,class_Plc,static_Plc,
			clazz, funcname, args);
		(*env)->DeleteLocalRef(env, clazz);
	}
	(*env)->DeleteLocalRef(env, funcname);
	(*env)->DeleteLocalRef(env, args);

	if ((*env)->ExceptionOccurred(env)) {
/*		return callJavaException(env); */
	  callJavaException(env);
		exit(1);

	}
	if (res) {
		java2prolog(env,res,result);
		(*env)->DeleteLocalRef(env, res);
	}
	return TRUE;
}

int
javaGetField(){
  TERM obj, field, value;
  obj = ARG(1,3);
  field = ARG(2,3);
  value = ARG(3,3);
  DEREF(field);
  if (!ISATOM(field)){
    exception = illegal_arguments;
    return -1;
  }
  return javaGetField1(obj, get_NAME(field), value);
}

int
javaGetField1(CELL obj, const char*field, CELL value) {
	JNIEnv *env = Env;
	/* !!! */
	jstring fieldname = (*env)->NewStringUTF(env,field);
	jobject val;
	register LONG_PTR top;
	DEREF(obj);

	if (PisAddr(obj)) { /* assuming, obj is an Instance */
		val = (*env)->CallStaticObjectMethod(env,class_Plc,get_Plc,
			NULL,addr2ref(env, obj), fieldname);
	} else {
		char *class = (char*)PascOfAtom(obj);
		/* !!! */
		jclass clazz = findClass(env,class);
		if (!clazz) return callJavaException(env);
		val = (*env)->CallStaticObjectMethod(env,class_Plc,get_Plc,
			clazz,NULL, fieldname);
		(*env)->DeleteLocalRef(env, clazz);
	}
	(*env)->DeleteLocalRef(env, fieldname);
	if ((*env)->ExceptionOccurred(env)) {
		return callJavaException(env);
	}
	java2prolog(env,val,value);
	return TRUE;
}

int
javaSetField(){
  TERM obj,field,value;

  obj = ARG(1,3);
  field = ARG(2,3);
  value = ARG(3,3);
  DEREF(field);
  if (!ISATOM(field)){
    exception = illegal_arguments;
    return -1;
  }
  return javaSetField1(obj, get_NAME(field), value);
}

int
javaSetField1(CELL obj, const char*field, CELL value) {
	JNIEnv *env = Env;
	/* !!! */
	jstring fieldname = (*env)->NewStringUTF(env,field);
	register LONG_PTR top;
	DEREF(obj);

	if (PisAddr(obj)) { /* assuming, obj is an Instance */
		jref res = prolog2java(env,value);
		(*env)->CallStaticObjectMethod(env,class_Plc,set_Plc,
			NULL,addr2ref(env, obj), fieldname, res);
		(*env)->DeleteLocalRef(env, res);
	} else {
		char *class = (char*)PascOfAtom(obj);
		/* !!! */
		jclass clazz = findClass(env,class);
		jref res = prolog2java(env,value);
		if (!clazz) return callJavaException(env);
		(*env)->CallStaticObjectMethod(env,class_Plc,set_Plc,
			clazz,NULL, fieldname, res);
		(*env)->DeleteLocalRef(env, res);
		(*env)->DeleteLocalRef(env, clazz);
	}
	(*env)->DeleteLocalRef(env, fieldname);
	if ((*env)->ExceptionOccurred(env)) {
		return callJavaException(env);
	}
	return TRUE;
}

int
javaConstructor(){
  TERM init, objaddr;
  init = ARG(1,2);
  objaddr = ARG(2,2);
  return javaConstructor1(init,objaddr);
}

int
javaConstructor1(CELL init, CELL objaddr) {
        JNIEnv *env = Env;
	jref args = prolog2javaArray(env,init); 
	const char *class = (const char*)PascOfFunc(init);
	jclass clazz;
	jobject obj;
	char fullname[64];

	/* !!!*/
	clazz = findClass(env,class);
	if (!clazz) return callJavaException(env);
#if defined(DEBUG)
	fprintf(stderr, "clazz = %x\n",clazz);
#endif
	obj =  (*env)->CallStaticObjectMethod(env, class_Plc, new_Plc, clazz, args);
	(*env)->DeleteLocalRef(env, clazz);
	(*env)->DeleteLocalRef(env, args);
#if defined(DEBUG)
	fprintf(stderr, "obj = %x\n",obj);
	fprintf(stderr,"globalRefTable.size = %d\n",globalRefTable.size);
#endif
	if ((*env)->ExceptionOccurred(env)) {
		return callJavaException(env);
	}
	if (obj) {
		obj = (*env)->NewGlobalRef(env,obj); /* +1 */
		return PuAddr(objaddr, (void*)obj);
	}
	else return FALSE; /* fail */
}
void javaDeleteRef(CELL objaddr) {
	JNIEnv *env = Env;
	jref ref = (jref)PvalueOfAddr(objaddr); 
	(*env)->DeleteGlobalRef(env,ref);	/* -1 */
}

#include "bprolog_plc_Plc.h"
JNIEXPORT void JNICALL Java_bprolog_plc_Plc_startPlc(JNIEnv *env,jclass clazz,
		jarray args, jarray oa) {
	int i;
	int argc = 0;
	const char* argv[12];
	static jboolean once = JNI_FALSE;

	if (once) return;
	once = JNI_TRUE;

	argv[argc++] = "bprolog";
	for (i=(int)((*env)->GetArrayLength(env,args)); --i>=0;) {
		jobject arg = (*env)->GetObjectArrayElement(env, args, argc-1);
		/* !!! */
		argv[argc++] = (*env)->GetStringUTFChars(env, arg, NULL);
	}
	initialize_bprolog(argc,argv);
/*	PexecP("true"); */
	for (i=(int)((*env)->GetArrayLength(env,args)); --i>=0;) {
		jobject arg = (*env)->GetObjectArrayElement(env, args, i);
		(*env)->ReleaseStringUTFChars(env, arg, argv[i+1]);
	}

	/* initialize compiled Prolog modules */
	jni_interface();
	plc_sup();
	initialize(env,oa);
	{
		/* K-Prolog knows the locale, but does not exports it. Shall be fixed.*/
#		if defined(WIN32)
		char *encSJIS = "SJIS";
		encoding = (*env)->NewGlobalRef(env,
			(*env)->NewStringUTF(env,encSJIS)); 
#		else
		char *encEUC = "EUCJIS";
		encoding = (*env)->NewGlobalRef(env,
			(*env)->NewStringUTF(env,encEUC)); 
#		endif
	}
}

JNIEXPORT jboolean JNICALL Java_bprolog_plc_Plc_exec(JNIEnv *env, jclass clazz,
		jstring cmd) {
	jboolean result;
	const char *ucmd;

	/* !!!*/
	result = call_bprolog_string(ucmd=((*env)->GetStringUTFChars(env,cmd, NULL)));
	(*env)->ReleaseStringUTFChars(env, cmd, ucmd);
	return result;
}
JNIEXPORT jboolean JNICALL Java_bprolog_plc_Plc_call(JNIEnv *env,
		jobject self, jstring functor, jarray args){
	int i;
	int rc;
	/* !!! */
	const char *name = (*env)->GetStringUTFChars(env, functor, NULL);
	int arity = (int)((*env)->GetArrayLength(env,args));
	CELL command = Pfunctorn(Patom(name),arity);

	(*env)->ReleaseStringUTFChars(env, functor, name);
	for (i=0; i<arity; i++) {
		jobject arg = (*env)->GetObjectArrayElement(env, args, i);
		java2prolog(env, arg, PargOfCompound(command,i));
	}
	rc = PcallX(command);
	if (rc) {
		jobject arg;
		for (i=0; i<arity; i++) {
			arg = (*env)->GetObjectArrayElement(env, args, i);
			if (arg == NULL) {
				jref res = prolog2java(env,PargOfCompound(command,i));
				(*env)->CallVoidMethod(env,self,result_Plc, i,res);
				(*env)->DeleteLocalRef(env, res);
			}
			(*env)->DeleteLocalRef(env,arg);
		}
	} 
	return (jboolean)rc;
}

int
Cboot_plc(){
  SYM_REC_PTR insert();
  extern int atomToTerm();

  objectRef = insert("$objectAddr",11,2);
  insert_cpred("javaConstructor",2,javaConstructor);
  insert_cpred("javaMethod",3,javaMethod);
  insert_cpred("javaGetField",3,javaGetField);
  insert_cpred("javaSetField",3,javaSetField);
  insert_cpred("atomToTerm",2,atomToTerm);
}
