#include "basic.h"
#include "term.h"
#include "macro.h"

  SYM_REC_PTR 
  abs1,
  add2,
  and2,
  atan1,
  ceiling1,
  cos1,
  div2,
  div_ge2,
  div_le2,
  exp1,
  float1,
  floor1,
  idiv2,
  log1,
  shiftl2,
  mod2,
  mul2,
  negation1,
  or2,
  ppow2,
  rem2,
  round1,
  shiftr2,
  sign1,
  sin1,
  sqrt1,
  sub1,
  sub2,
  tan1,
  truncate1,cputime1;

void init_arith_sym(){
  SYM_REC_PTR insert();
  
  abs1 = insert("abs",3,1);
  add2 = insert("+",1,2);
  and2 = insert("/\\",2,2);
  atan1 = insert("atan",4,1);
  ceiling1 = insert("ceiling",7,1);
  cos1 = insert("cos",3,1);
  div2 = insert("/",1,2);
  div_ge2 = insert("/>",2,2);
  div_le2 = insert("/<",2,2);
  exp1 = insert("exp",3,1);
  float1 = insert("float",5,1);
  floor1 = insert("floor",5,1);
  idiv2 = insert("//",2,2);
  log1 = insert("log",3,1);
  shiftl2 = insert("<<",2,2);
  mod2 = insert("mod",3,2);
  mul2 = insert("*",1,2);
  negation1 = insert("\\",1,1);
  or2 = insert("\\/",2,2);
  ppow2 = insert("**",2,2);
  rem2 = insert("rem",3,2);
  round1 = insert("round",5,1);
  shiftr2 = insert(">>",2,2);
  sign1 = insert("sign",4,1);
  sin1 = insert("sin",3,1);
  sqrt1 = insert("sqrt",4,1);
  sub1 = insert("-",1,1);
  sub2 = insert("-",1,2);
  tan1 = insert("tan",3,1);
  truncate1 = insert("truncate",8,1);
  cputime1 = insert("cputime",7,0);
}

/* evaluate an arithmetic exression */
  
int b_EVAL_ARITH_cf(ex,res)
register LONG ex,res;
{
  LONG res1;
  int eval_arith();

  if (eval_arith(ex,&res1)<0) return -1;
  ASSIGN_f(res,res1);
  return 1;
}

int eval_arith(ex,res)
register LONG ex,*res;
{
  register char *ch_ptr;
  SYM_REC_PTR ptr;
  int arity;
  register LONG ex1,ex2;
  LONG res1,res2;
  
  exception = number_expected;
 /* l1: */
      
  SWITCH_OP(ex,l1,
	    {return -1;},

	    {if (ISINT(ex)){
	      FOLLOW(res) = ex;
	      return 1;
	    }
	    else  return -1;},

	    {GET_CAR(ex,ex);
	     goto l1;},

	    {if (ISFLOAT(ex)) {
	      FOLLOW(res) = ex;
	      return 1;
	    }
	    else goto eval_ex;},
	    {return -1;});

 eval_ex:  /* ex is an exression */
  ptr = (SYM_REC_PTR)FOLLOW(UNTAGGED(ex));
  ch_ptr = GET_NAME(ptr);

  switch (*ch_ptr) {
  case '+':
    if (ptr==add2) {
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      GET_ARG(ex2,ex,2);
      if (eval_arith(ex2,&res2)<0) return -1;
      return ADD(res1,res2,res);
    }
    else return -1;

  case '-':
    if (ptr==sub2){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      GET_ARG(ex2,ex,2);
      if (eval_arith(ex2,&res2)<0) return -1;
      return SUB(res1,res2,res);
    }
    else if (ptr==sub1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return SUB(MAKEINT(0),res1,res);
    }
    else return -1;
    
  case '*':
    if (ptr==ppow2){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      GET_ARG(ex2,ex,2);
      if (eval_arith(ex2,&res2)<0) return -1;
      return b_FLOAT_POW_ccf(res1,res2,res);
    }
    else if (ptr==mul2) {
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      GET_ARG(ex2,ex,2);
      if (eval_arith(ex2,&res2)<0) return -1;
      return MUL(res1,res2,res);
    }
    else return -1;
    
  case '/':
    if (ptr==idiv2){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      GET_ARG(ex2,ex,2);
      if (eval_arith(ex2,&res2)<0) return -1;
      return IDIV(res1,res2,res);
    }
    else if (ptr==and2){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      GET_ARG(ex2,ex,2);
      if (eval_arith(ex2,&res2)<0) return -1;
      return AND(res1,res2,res);
    }
    else if (ptr==div2){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      GET_ARG(ex2,ex,2);
      if (eval_arith(ex2,&res2)<0) return -1;
      return DIV(res1,res2,res);
    }
    else if (ptr==div_ge2){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      GET_ARG(ex2,ex,2);
      if (eval_arith(ex2,&res2)<0) return -1;
      return DIV_GE(res1,res2,res);
    }
    else if (ptr==div_le2){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      GET_ARG(ex2,ex,2);
      if (eval_arith(ex2,&res2)<0) return -1;
      return DIV_LE(res1,res2,res);
    }
    else return -1;
    
  case '<':
    if (ptr==shiftl2){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      GET_ARG(ex2,ex,2);
      if (eval_arith(ex2,&res2)<0) return -1;
      return SHIFTL(res1,res2,res);
    }
    else return -1;
    
  case '>':
    if (ptr==shiftr2){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      GET_ARG(ex2,ex,2);
      if (eval_arith(ex2,&res2)<0) return -1;
      return SHIFTR(res1,res2,res);
    }
    else return -1;
    
  case '\\':
    if (ptr==or2){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      GET_ARG(ex2,ex,2);
      if (eval_arith(ex2,&res2)<0) return -1;
      return OR(res1,res2,res);
    }
    else if (ptr==negation1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return NEG(res1,res);
    }
    else return -1;
    
  case 'a':
    if (ptr==abs1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return b_FLOAT_ABS_cf(res1,res);
    }
    else if (ptr==atan1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return b_FLOAT_ATAN_cf(res1,res);
    }
    
  case 'c':
    if (ptr==cos1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return b_FLOAT_COS_cf(res1,res);
    }
    else if (ptr==ceiling1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return b_FLOAT_CEILING_cf(res1,res);
    }
    else if (ptr == cputime1)
      return b_CPUTIME_f(res);
    else return -1;
    
  case 'e':
    if (ptr==exp1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return b_FLOAT_EXP_cf(res1,res);
    }
    else return -1;

  case 'f':
    if (ptr==float1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return b_FLOAT_FLOAT_cf(res1,res);
    }
    else if (ptr==floor1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return b_FLOAT_FLOOR_cf(res1,res);
    }
    else return -1;
    
  case 'l':
    if (ptr==log1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return b_FLOAT_LOG_cf(res1,res);
    }
    else return -1;
    
  case 'm':
    if (ptr==mod2){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      GET_ARG(ex2,ex,2);
      if (eval_arith(ex2,&res2)<0) return -1;
      return MOD(res1,res2,res);
    }
    else return -1;

  case 'r':
    if (ptr==rem2){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      GET_ARG(ex2,ex,2);
      if (eval_arith(ex2,&res2)<0) return -1;
      return REM(res1,res2,res);
    }
    else if (ptr==round1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return b_FLOAT_ROUND_cf(res1,res);
    }
    else return -1;
      
  case 's':
    if (ptr==sin1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return b_FLOAT_SIN_cf(res1,res);
    }
    else if (ptr==sqrt1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return b_FLOAT_SQRT_cf(res1,res);
    }
    else if (ptr==sign1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return SIGN(res1,res);
    }
    else return -1;
  
  case 't':
    if (ptr==tan1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return b_FLOAT_TAN_cf(res1,res);
    }
    else if (ptr==truncate1){
      GET_ARG(ex1,ex,1);
      if (eval_arith(ex1,&res1)<0) return -1;
      return b_FLOAT_TRUNCATE_cf(res1,res);
    }
    else return -1;

  default:
    return -1;
  }
}
    

ADD(op1,op2,op3)
LONG op1,op2,op3;
{
  double floatval();
  LONG top1;
  
  DEREF(op1);
  DEREF(op2);
  
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (!ISNUM(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }

  if (ISINT(op1)) 
    if (ISINT(op2))
      ASSIGN_f(op3,MAKEINT(INTVAL(op1) + INTVAL(op2)));
    else {
      ASSIGN_f(op3,encodefloat((double)INTVAL(op1) + floatval(op2)));
    }
  else 
    if (ISINT(op2)) {
      ASSIGN_f(op3,encodefloat(floatval(op1) + (double)INTVAL(op2)));}
    else 
      ASSIGN_f(op3,encodefloat(floatval(op1) + floatval(op2)));
  return 1;
}
  

SUB(op1,op2,op3)
LONG op1,op2,op3;
{
  double floatval();
  LONG top1;

  DEREF(op1);
  DEREF(op2);
  
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (!ISNUM(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }

  if (ISINT(op1))
    if (ISINT(op2))
      ASSIGN_f(op3,MAKEINT(INTVAL(op1) - INTVAL(op2)));
    else {
      ASSIGN_f(op3,encodefloat((double)INTVAL(op1) - floatval(op2)));
    } 
  else 
    if (ISINT(op2)) {
      ASSIGN_f(op3,encodefloat(floatval(op1) - (double)INTVAL(op2)));}
    else {
      ASSIGN_f(op3,encodefloat(floatval(op1) - floatval(op2)));
    } 
  return 1;
}


MUL(op1,op2,op3)
LONG op1,op2,op3;
{
  double floatval();
  LONG top1;

  DEREF(op1);
  DEREF(op2);
  
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (!ISNUM(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }

  if (ISINT(op1))
    if (ISINT(op2))
      ASSIGN_f(op3,MAKEINT(INTVAL(op1) * INTVAL(op2)));
    else {
      ASSIGN_f(op3,encodefloat((double)INTVAL(op1) * floatval(op2)));
    } 
  else 
    if (ISINT(op2)) {
      ASSIGN_f(op3,encodefloat(floatval(op1) * (double)INTVAL(op2)));
    }
    else {
      ASSIGN_f(op3,encodefloat(floatval(op1) * floatval(op2)));
    }
  return 1;
}

DIV_GE(op1,op2,op3)
LONG op1,op2,op3;
{
  LONG top1;

  DEREF(op1);
  DEREF(op2);
  
  op1 = INTVAL(op1); op2 = INTVAL(op2);
  if ((op1>=0 && op2 >=0) || (op1<=0 && op2 <=0)){
    top1 = op1/op2;
    if (top1*op2!=op1) top1++;
  } else {
    top1 = op1/op2;
  }
  FOLLOW(op3) = MAKEINT(top1);
  return 1;
}

DIV_LE(op1,op2,op3)
LONG op1,op2,op3;
{
  double floatval();
  LONG top1;

  DEREF(op1);
  DEREF(op2);
  
  op1 = INTVAL(op1); op2 = INTVAL(op2);
  if ((op1>=0 && op2 >=0) || (op1<=0 && op2 <=0)){
    top1 = op1/op2;
  } else {
    top1 = op1/op2;
    if (top1*op2!=op1) top1--;
  }
  FOLLOW(op3) = MAKEINT(top1);
  return 1;
}

DIV(op1,op2,op3)
LONG op1,op2,op3;
{
  double floatval();
  LONG top1;

  DEREF(op1);
  DEREF(op2);
  
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (!ISNUM(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }
  
  if (ISINT(op1))
    if (ISINT(op2)){
      op2 = INTVAL(op2);
      if (op2==0){
	exception = divide_by_zero;
	return -1;
      }
      ASSIGN_f(op3,encodefloat((double)INTVAL(op1) / (double)op2));
    }
    else {
      if (floatval(op2)==0.0){
	exception = divide_by_zero;
	return -1;
      }
      ASSIGN_f(op3,encodefloat((double)INTVAL(op1) / floatval(op2)));
    } 
  else 
    if (ISINT(op2)){
      op2 = INTVAL(op2);
      if (op2==0){
	exception = divide_by_zero;
	return -1;
      }
      ASSIGN_f(op3,encodefloat(floatval(op1) / (double)op2));
    }
    else {
      if (floatval(op2)==0.0){
	exception = divide_by_zero;
	return -1;
      }
      ASSIGN_f(op3,encodefloat(floatval(op1) / floatval(op2)));
    }
  return 1;
}

IDIV(op1,op2,op3)
LONG op1,op2,op3;
{
  double floatval();
  LONG top1;

  DEREF(op1);
  DEREF(op2);
  
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (!ISNUM(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }
  
  if (ISINT(op1))
    if (ISINT(op2)){
      op2 = INTVAL(op2);
      if (op2==0){
	exception = divide_by_zero;
	return -1;
      }
      FOLLOW(op3)=MAKEINT((int)(INTVAL(op1) / op2));
    }
    else {
      if (floatval(op2)==0){
	exception = divide_by_zero;
	return -1;
      }
      FOLLOW(op3)=MAKEINT((int)((double)INTVAL(op1) / floatval(op2)));
    }
  else 
    if (ISINT(op2)){
      op2 = INTVAL(op2);
      if (op2==0){
	exception = divide_by_zero;
	return -1;
      }
      FOLLOW(op3)=MAKEINT((int)(floatval(op1) / (double)op2));
    }
    else {
      if (floatval(op2)==0){
	exception = divide_by_zero;
	return -1;
      }
      FOLLOW(op3)= MAKEINT((int)(floatval(op1) / floatval(op2)));
    }
  return 1;
}

MOD(op1,op2,op3)
LONG op1,op2,op3;
{
  LONG top1;

  DEREF(op1);
  DEREF(op2);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (!ISNUM(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }
  
  if (ISINT(op1) && ISINT(op2)){
    op2 = INTVAL(op2);
    if (op2==0){
      exception = divide_by_zero;
      return -1;
    }
    FOLLOW(op3)=MAKEINT((int)(INTVAL(op1) % op2));
  }
  else {
    exception = integer_expected;
    return -1;
  }
  return 1;
}
  
REM(op1,op2,op3)
LONG op1,op2,op3;
{
  LONG top1;
  int temp;
  
  DEREF(op1);
  DEREF(op2);
  
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (!ISNUM(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }
  
  if (ISINT(op1) && ISINT(op2)) {
    FOLLOW(op3) =MAKEINT((int)(INTVAL(op1) % INTVAL(op2)));
  } else {
    exception = integer_expected;
    return -1;
  }
  return 1;
}

AND(op1,op2,op3)
LONG op1,op2,op3;
{
  LONG top1;

  DEREF(op1);
  DEREF(op2);
  
  if (!ISINT(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (!ISINT(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }
  
  if (ISINT(op1) && ISINT(op2)){
    FOLLOW(op3)=MAKEINT(INTVAL(op2) & INTVAL(op1));
  }
  else {
    exception = integer_expected;
    return -1;
  }
  return 1;
}
    
OR(op1,op2,op3)
LONG op1,op2,op3;
{
  LONG top1;

  DEREF(op1);
  DEREF(op2);
  
  if (!ISINT(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (!ISINT(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }
  
  if (ISINT(op1) && ISINT(op2)) {
    FOLLOW(op3)=MAKEINT(INTVAL(op2) | INTVAL(op1));
  }
  else {
    exception = integer_expected;
    return -1;
  }
  return 1;
}

SHIFTL(op1,op2,op3)
LONG op1,op2,op3;
{
  LONG top1;
  
  DEREF(op1);
  DEREF(op2);
  
  if (!ISINT(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (!ISINT(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }
  
  if (ISINT(op1) && ISINT(op2)) 
    FOLLOW(op3)=MAKEINT(INTVAL(op1) << INTVAL(op2));
  else {
    exception = integer_expected;
    return -1;
  }
  return 1;
}


SHIFTR(op1,op2,op3)
LONG op1,op2,op3;
{
  LONG top1;

  DEREF(op1);
  DEREF(op2);
  
  if (!ISINT(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (!ISINT(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }
  
  if (ISINT(op1) && ISINT(op2)) 
    FOLLOW(op3)=MAKEINT(INTVAL(op1) >> INTVAL(op2));
  else {
    exception = integer_expected;
    return -1;
  }
  return 1;
}


NEG(op1,op2)
LONG op1,op2;
{
  LONG top1;

  DEREF(op1);
  
  if (!ISINT(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  
  if (ISINT(op1)) 
    FOLLOW(op2)=MAKEINT(~INTVAL(op1));
  else {
    exception = integer_expected;
    return -1;
  }
  return 1;
}
  
SIGN(op1,op2)
LONG op1,op2;
{
  LONG top1;
  double f,floatval();

  DEREF(op1);

  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  
  if (ISINT(op1)){
    op1 = INTVAL(op1);
    if (op1>0) ASSIGN_f(op2,MAKEINT(1));
    else if (op1==0) ASSIGN_f(op2,MAKEINT(0));
    else ASSIGN_f(op2,MAKEINT(-1));
  }
  else {
    f = floatval(op1);
    if (f>0) ASSIGN_f(op2,MAKEINT(1));
    else if (f==0.0) ASSIGN_f(op2,MAKEINT(0));
    else ASSIGN_f(op2,MAKEINT(-1));
  }
  return 1;
}
