/* sim.h */

#include <stdio.h>
#include <math.h>

#define MAXREGS       256  
#define WRITEFLAG     1
#define READFLAG      0
#define MAX_LEN       1280
#define BUCKET_CHAIN  2047
#define TRUE          1
#define FALSE         0
#define MAXTRIGGERS   10000

#define T_ORDI        0    /* constant-type: no ep definition */
#define T_DYNA        1    /* constant-type: dynamic */
#define T_PRED        2    /* constant-type: ep points to compiled code */
#define C_PRED        3    /* c function */
#define T_BOOT        4    /* boolean table */
#define T_TEMP_PRED  15    /* psc entry for predicate containing offset
                              entry point */

/* ---------- Type Specifiers ----------------------------------------------- */

typedef char                CHAR;     /*  8 bits */
typedef unsigned char       BYTE;     /*  8 bits */
typedef unsigned short int  WORD;     /* 16 bits */
typedef long int            LONG;     /* 32 bits */

typedef CHAR  *CHAR_PTR;              /* pointers to memory types */
typedef BYTE  *BYTE_PTR;
typedef WORD  *WORD_PTR;
typedef LONG  *LONG_PTR;

typedef struct sym_rec *SYM_REC_PTR;

typedef struct sym_rec {
  BYTE      etype;
  BYTE      arity;
  CHAR_PTR  nameptr;
  WORD      length;
  SYM_REC_PTR next;
  int       (*ep)();   /* entry point to byte code,global var, c function, booltable */
} SYM_REC;

typedef struct toam_status {
  int entry; /* from main() in main.c or user-defined main() */
  LONG_PTR arreg;
  LONG_PTR local_top;
  LONG_PTR breg;
  LONG_PTR hreg;
  LONG_PTR hbreg;
  LONG_PTR sfreg;
  LONG_PTR treg;
  LONG_PTR tareg; /* table area */
} TOAM;


extern BYTE     use_tabling;
extern LONG_PTR tableArea;
extern LONG_PTR tableAreaBottom;
extern LONG     tableAreaSize;
extern LONG_PTR tareg;
extern LONG_PTR subgoalTable;
extern LONG subgoalTableBucketSize;
extern LONG answerTableBucketSize;

extern TOAM toam0;

extern LONG      maxmem, maxpspace, maxtrail, maxmemb; /* declared in init.c */

extern LONG_PTR  memory;          /* heap, local stack   */
extern LONG_PTR  pspace;          /* psc records, instructions, p-names */
extern LONG_PTR  tstack;          /* trail stack */
extern LONG      reg[MAXREGS];    /* general purpose registers */
extern LONG_PTR  rreg;            /* register pointer             */

extern LONG_PTR  local_bottom;
extern LONG_PTR  heap_bottom;
extern LONG_PTR  trail_bottom;

extern CHAR_PTR  curr_fence;      /* ptr to next free byte in perm space */
extern LONG_PTR  parea_top;       
extern CHAR_PTR  max_fence;       /* ptr to last+1 free byte in perm space */
extern LONG_PTR  inst_begin;      /* ptr to the beginning of inst. array */
extern LONG_PTR  addr_rest_indomain;
extern LONG_PTR  addr_halt;
extern LONG_PTR  addr_halt0;

extern LONG_PTR  gc_low;
extern LONG_PTR  gc_upper;
                                  /* ptr to the failb inst */
extern LONG_PTR  top;
extern LONG_PTR  arreg;         /* last activation record       */
extern LONG_PTR  local_top;
extern LONG_PTR  breg;            /* last choice point            */
extern LONG_PTR  hreg;            /* top of heap                  */
extern LONG_PTR  treg;           /* top of trail stack           */
extern LONG_PTR  hbreg;           /* heap backtrack point         */
extern LONG_PTR  sreg;            /* current build or unify field */
extern LONG_PTR  sfreg;           /* latest suspension frame      */
extern LONG_PTR  triggers[MAXTRIGGERS];
extern BYTE  event_flag[MAXTRIGGERS];
extern long int  trigger_no;      /* list of woken frames         */
extern SYM_REC_PTR  hash_table[BUCKET_CHAIN];
extern int       toam_signal_vec;

extern BYTE      disassem;
extern int       num_line;        /* print instruction addresses on trace
				     and disassem */
extern BYTE      trace_toam;

extern LONG      exception;
extern SYM_REC_PTR  list_psc;
extern SYM_REC_PTR  interrupt_psc;

extern SYM_REC_PTR  comma_psc;
extern SYM_REC_PTR  float_psc;
extern SYM_REC_PTR  enter_dyn_call;
extern SYM_REC_PTR  damon_load;
extern SYM_REC_PTR  slash_psc;
extern SYM_REC_PTR  domain_psc;
extern SYM_REC_PTR handle_exception_psc;
extern SYM_REC_PTR colon_sym,question_sym,imply_sym,clause4_sym,clause2_sym,clause_no_sym,domain_delta;

extern LONG  nil_sym, period_sym,varid;
extern LONG  atom_expected;
extern LONG  integer_expected;
extern LONG  number_expected;
extern LONG  arith_expr_expected;
extern LONG  list_expected;
extern LONG  structure_expected;
extern LONG  char_code_expected;
extern LONG  float_format_expected;
extern LONG  operator_type_expected;
extern LONG  variable_expected;
extern LONG  nonvariable_expected;
extern LONG  could_not_open_stream;
extern LONG  file_does_not_exist;
extern LONG  divide_by_zero;
extern LONG  illegal_expression;
extern LONG  illegal_list;
extern LONG  illegal_mode_string;
extern LONG  illegal_predicate;
extern LONG  not_modifiable;
extern LONG  out_of_range;
extern LONG  undefined_predicate;
extern LONG  illegal_file;
extern LONG  code_area_overflow;
extern LONG  permission_error;
extern LONG  control_stack_overflow;
extern LONG  trail_stack_overflow;
extern LONG  illegal_arguments;

extern LONG main_args;
extern SYM_REC_PTR insert_cpred();
