/*
*   File   : clpfd.c
*   Author : Neng-Fa ZHOU Copyright (C) All rights reserved
*   Updated: Augest 1999
*   Purpose: Primitives on domain variables
*/
#include "cpreds.h"
#include "event.h"
#include "frame.h"

dvar_bv(op)
register LONG op;
{
  LONG_PTR dv_ptr;

  DEREF(op);
  if (IS_SUSP_VAR(op)){
    dv_ptr = (LONG_PTR)UNTAGGED(op);
    if (IS_BV_DOMAIN(dv_ptr)) return 1; 
  } 
  return 0;
}

dvar(op)
register LONG op;
{
  LONG_PTR dv_ptr;

  DEREF(op);
  if (IS_SUSP_VAR(op)){
    dv_ptr = (LONG_PTR)UNTAGGED(op);
    if (!IS_UN_DOMAIN(dv_ptr)) return 1; 
  } 
  return 0;
}

dvar_or_int(op)
register LONG op;
{
  LONG_PTR dv_ptr;

  DEREF(op);
  if (IS_SUSP_VAR(op)){
    dv_ptr = (LONG_PTR)UNTAGGED(op);
    if (!IS_UN_DOMAIN(dv_ptr)) return 1; 
  } else if (ISINT(op)) 
    return 1;
  return 0;
}

int count_vars(op)
LONG op;
{
  int count,i,arity;
  count = 0;DEREF(op);
  SWITCH_OP(op,n_count_vars,
	    {count++;},

	    {},

	    {UNTAG(op);
	     count += count_vars(FOLLOW(op))+count_vars(FOLLOW((LONG_PTR)op+1));},
	    
	    {UNTAG(op);
	     arity = GET_ARITY((SYM_REC_PTR)FOLLOW(op));
	     for (i = 1; i<=arity; i++)
	       count +=  count_vars(FOLLOW(((LONG_PTR)op+i)));},
	    
	    {count++;});
  return count;
}

int trigger_vars_ins(op)
LONG op;
{
  int i,arity;
  LONG_PTR top,dv_ptr;
  LONG top1;

  SWITCH_OP(op,n_trigger_vars,
	    {my_CREATE_SUSP_VAR_ins(op,hreg,arreg);},

	    {},

	    {UNTAG(op);
	     trigger_vars_ins(FOLLOW(op));
	     trigger_vars_ins(FOLLOW((LONG_PTR)op+1));},
	    
	    {UNTAG(op);
	     arity = GET_ARITY((SYM_REC_PTR)FOLLOW(op));
	     for (i = 1; i<=arity; i++)
	       trigger_vars_ins(FOLLOW(((LONG_PTR)op+i)));},
	    
	    {my_INSERT_SUSP_CALL_ins(op,hreg,arreg);});
  return 1;
}

int trigger_vars_min(op)
LONG op;
{
  int i,arity;
  LONG_PTR top,dv_ptr;
  LONG top1;

  SWITCH_OP(op,n_trigger_vars,
	    {my_CREATE_SUSP_VAR_min(op,hreg,arreg);},

	    {},

	    {UNTAG(op);
	     trigger_vars_min(FOLLOW(op));
	     trigger_vars_min(FOLLOW((LONG_PTR)op+1));},
	    
	    {UNTAG(op);
	     arity = GET_ARITY((SYM_REC_PTR)FOLLOW(op));
	     for (i = 1; i<=arity; i++)
	       trigger_vars_min(FOLLOW(((LONG_PTR)op+i)));},
	    
	    {my_INSERT_SUSP_CALL_min(op,hreg,arreg);});
  return 1;
}

int trigger_vars_max(op)
LONG op;
{
  int i,arity;
  LONG_PTR top,dv_ptr;
  LONG top1;

  SWITCH_OP(op,n_trigger_vars,
	    {my_CREATE_SUSP_VAR_max(op,hreg,arreg);},

	    {},

	    {UNTAG(op);
	     trigger_vars_max(FOLLOW(op));
	     trigger_vars_max(FOLLOW((LONG_PTR)op+1));},
	    
	    {UNTAG(op);
	     arity = GET_ARITY((SYM_REC_PTR)FOLLOW(op));
	     for (i = 1; i<=arity; i++)
	       trigger_vars_max(FOLLOW(((LONG_PTR)op+i)));},
	    
	    {my_INSERT_SUSP_CALL_max(op,hreg,arreg);});
  return 1;
}

int trigger_vars_dom(op)
LONG op;
{
  int i,arity;
  LONG_PTR top,dv_ptr;
  LONG top1;

  SWITCH_OP(op,n_trigger_vars,
	    {my_CREATE_SUSP_VAR_dom(op,hreg,arreg);},

	    {},

	    {UNTAG(op);
	     trigger_vars_dom(FOLLOW(op));
	     trigger_vars_dom(FOLLOW((LONG_PTR)op+1));},
	    
	    {UNTAG(op);
	     arity = GET_ARITY((SYM_REC_PTR)FOLLOW(op));
	     for (i = 1; i<=arity; i++)
	       trigger_vars_dom(FOLLOW(((LONG_PTR)op+i)));},
	    
	    {my_INSERT_SUSP_CALL_dom(op,hreg,arreg);});
  return 1;
}

exclude_elm_dvars(){
  LONG P_elm,P_list1,P_list2;
  long int elm;

  P_elm = ARG(1,3);
  P_list1 = ARG(2,3);
  P_list2 = ARG(3,3);
  
  DEREF_NONVAR(P_elm);
  elm = INTVAL(P_elm);
  
  if (exclude_elm_dvars2(elm,P_list1)==0) return 0;
  if (exclude_elm_dvars2(elm,P_list2)==0) return 0;
  return 1;
}

exclude_elm_dvars2(elm,P_list)
register long int elm;
register LONG P_list;
{
  register LONG_PTR dv_ptr,top;
  LONG_PTR ptr;
  LONG P_v;
  
  DEREF(P_list);
  while (P_list!=nil_sym){
    ptr = (LONG_PTR)UNTAGGED(P_list);
    P_v = FOLLOW(ptr);

    DEREF(P_v);
    if (IS_SUSP_VAR(P_v)){
      dv_ptr = (LONG_PTR)UNTAGGED(P_v);
      domain_set_false_aux(dv_ptr,elm);
    } else if (P_v == MAKEINT(elm)) return 0;

    P_list = FOLLOW(ptr+1);
    DEREF(P_list);
  }
  return 1;
}

c_notin(){
  LONG Var,Low,Up;
  LONG_PTR dv_ptr;
  
  Var = ARG(1,3);
  Low = ARG(2,3);
  Up = ARG(3,3);
  DEREF(Var);  DEREF(Low);  DEREF(Up);
  Low = INTVAL(Low);
  Up = INTVAL(Up);
  if (!IS_SUSP_VAR(Var)){
    if (ISINT(Var))
      return (INTVAL(Var)<Low || INTVAL(Var)>Up);
    else
      return -1;
  }
  dv_ptr = (LONG_PTR)UNTAGGED(Var);
  while (Low<=Up){
    domain_set_false_aux(dv_ptr,Low);
    Low++;
  }
  return 1;
}
  
b_deleteffc(BestVar,Vars,Rest)
TERM BestVar,Vars,Rest;
{
  register TERM Var;
  register int size,cno;
  register LONG_PTR dv_ptr;
  
  DEREF(Vars);
  while (ISLIST(Vars)){
    Var = FETCH_CAR(Vars);
    Vars = FETCH_CDR(Vars);
    DEREF(Var);DEREF(Vars);
    if (IS_SUSP_VAR(Var)){
      dv_ptr = (LONG_PTR)UNTAGGED(Var);
      size = DV_size(dv_ptr);
      cno = DV_dcs_no(dv_ptr);
      return deleteffc_6(dv_ptr,Vars,Rest,BestVar,size,cno);
    }
  }
  return 0;
}

deleteffc_6(dv_ptr0,Vars,Rest,BestVar,size0,cno0)
register LONG Vars,Rest,BestVar;
LONG_PTR dv_ptr0;
int size0,cno0;
{
  register LONG Var;
  register LONG_PTR dv_ptr;
  register int size,cno;

  DEREF(Vars);
  while (ISLIST(Vars)){
    Var = FETCH_CAR(Vars);
    Vars = FETCH_CDR(Vars);
    DEREF(Var);
    DEREF(Vars);
    if (IS_SUSP_VAR(Var)){
      dv_ptr = (LONG_PTR)UNTAGGED(Var);
      size = DV_size(dv_ptr);
      cno = DV_dcs_no(dv_ptr);
      if (size<size0 || (size==size0 && cno>cno0)) {
	FOLLOW(Rest) = ADDTAG(hreg,LST);
	NEW_HEAP_NODE((LONG)dv_ptr0);
	Rest = (LONG)hreg++;
	dv_ptr0 = dv_ptr;size0 = size; cno0=cno;
      } else {
	FOLLOW(Rest) = ADDTAG(hreg,LST);
	NEW_HEAP_NODE((LONG)dv_ptr);
	Rest = (LONG)hreg++;
      }
    }
  }
  if (ISNIL(Vars)){
    FOLLOW(Rest) = nil_sym;
    FOLLOW(BestVar) = (LONG)dv_ptr0;
    return 1;
  }
  else {
    exception = illegal_arguments;
    return -1;
  }
}


b_deleteff(BestVar,Vars,Rest)
register TERM BestVar,Vars,Rest;

{
  TERM Var;
  register int size;
  register LONG_PTR dv_ptr;
  
  DEREF(Vars);
  while (ISLIST(Vars)){
    Var = FETCH_CAR(Vars);
    Vars = FETCH_CDR(Vars);
    DEREF(Var);DEREF(Vars);
    if (IS_SUSP_VAR(Var)){
      dv_ptr = (LONG_PTR)UNTAGGED(Var);
      size = DV_size(dv_ptr);
      return deleteff_5(dv_ptr,Vars,Rest,BestVar,size);
    }
  }
  return 0;
}

deleteff_5(dv_ptr0,Vars,Rest,BestVar,size0)
register LONG Vars,Rest,BestVar;
LONG_PTR dv_ptr0;
int size0;
{
  register LONG Var;
  register LONG_PTR dv_ptr;
  register int size;

  DEREF(Vars);
  while (ISLIST(Vars)){
    Var = FETCH_CAR(Vars);
    Vars = FETCH_CDR(Vars);
    DEREF(Var); DEREF(Vars);
    if (IS_SUSP_VAR(Var)){
      dv_ptr = (LONG_PTR)UNTAGGED(Var);
      size = DV_size(dv_ptr);
      if (size<size0){
	FOLLOW(Rest) = ADDTAG(hreg,LST);
	NEW_HEAP_NODE((LONG)dv_ptr0);
	Rest = (LONG)hreg++;
	dv_ptr0 = dv_ptr;size0 = size; 
      } else {
	FOLLOW(Rest) = ADDTAG(hreg,LST);
	NEW_HEAP_NODE((LONG)dv_ptr);
	Rest = (LONG)hreg++;
      }
    }
  }
  if (ISNIL(Vars)){
    FOLLOW(Rest) = nil_sym;
    FOLLOW(BestVar) = (LONG)dv_ptr0;
    return 1;
  }
  else {
    exception = illegal_arguments;
    return -1;
  }
}


c_deleteff_des(){ /* destructive */
  register TERM BestVar,Vars,Rest;
  TERM Var,CurBestVar;
  register int best_size;
  register LONG_PTR dv_ptr,best_ptr,best_prev_ptr,prev_ptr,ptr,first_ptr,top;

  BestVar = ARG(1,3);
  Vars = ARG(2,3);
  Rest = ARG(3,3);
  
  DEREF(Vars);DEREF(Rest);
  
  while (ISLIST(Vars)){
    ptr = (LONG_PTR)UNTAGGED(Vars);
    Var = FOLLOW(ptr);
    Vars = FOLLOW(ptr+1);
    DEREF(Var);DEREF(Vars);
    if (IS_SUSP_VAR(Var)) goto cont;
  }
  return 0;
 cont:
  CurBestVar = Var;
  dv_ptr = (LONG_PTR)UNTAGGED(Var);
  best_size = DV_size(dv_ptr);
  first_ptr = best_ptr = ptr;
  best_prev_ptr = ptr;
  while (ISLIST(Vars)){
    prev_ptr = ptr;
    ptr = (LONG_PTR)UNTAGGED(Vars);
    Var = FOLLOW(ptr);
    Vars = FOLLOW(ptr+1);
    DEREF(Var);DEREF(Vars);
    if (IS_SUSP_VAR(Var)){
      dv_ptr = (LONG_PTR)UNTAGGED(Var);
      if (DV_size(dv_ptr)<best_size){
	best_ptr = ptr;
	best_prev_ptr = prev_ptr;
	best_size = DV_size(dv_ptr);
	CurBestVar = Var;
      }
    } else {
      top = prev_ptr+1;
      PUSHTRAILC(top,FOLLOW(top));  /* exclude this */
      FOLLOW(top) = FOLLOW(ptr+1);
    }
  }
  ASSIGN_f(BestVar,CurBestVar);
  if (best_ptr == best_prev_ptr){
    ASSIGN_f(Rest,FOLLOW(best_ptr+1));
  } else {
    top = best_prev_ptr+1;
    PUSHTRAILC(top,FOLLOW(top));  /* exclude best */
    FOLLOW(top) = FOLLOW(best_ptr+1);
    ASSIGN_f(Rest,ADDTAG(first_ptr,LST));
  }
  return 1;
}

b_CONSTRAINTS_NUMBER_cf(Var,N)
register LONG Var,N;
{
  LONG_PTR dv_ptr,dcs;
  int i;

  DEREF(Var);
  if (IS_SUSP_VAR(Var)){
    dv_ptr = (LONG_PTR)UNTAGGED(Var);
    i = 0;
    dcs = (LONG_PTR)DV_ins_cs(dv_ptr);
    i += count_cs_list(dcs);
/*
    if (!IS_UN_DOMAIN(dv_ptr)){
      dcs = (LONG_PTR)DV_min_cs(dv_ptr);
      i += count_cs_list(dcs);

      dcs = (LONG_PTR)DV_max_cs(dv_ptr);
      i += count_cs_list(dcs);

      dcs = (LONG_PTR)DV_dom_cs(dv_ptr);
      i += count_cs_list(dcs);
    }
*/
    PUSHTRAILB(A_DV_dcs_no(dv_ptr),DV_dcs_no(dv_ptr));
    DV_dcs_no(dv_ptr) = i;
    FOLLOW(N) = MAKEINT(i);
    return 1;
  } else {
    exception = illegal_arguments;
    return -1;
  }
}

int count_cs_list(list)
LONG_PTR list;
{
  int i = 0;

  while (ISLIST((LONG)list)){
    i ++;
    list = (LONG_PTR)UNTAGGED(list);
    list = (LONG_PTR)CS_NEXT(list);
  }
  return i;
}

print_constr(constr)
LONG_PTR constr;
{
  printf("constr(%x)\n",constr);
  printf("(status_word=%x,frame=%x\n",CS_STATUS(constr),CS_FRAME(constr));
}

print_susp_frame(fp,symptr)
LONG_PTR fp;
SYM_REC_PTR symptr;
{
  int arity;
  SYM_REC_PTR insert_cfunc();
  
  arity = GET_ARITY(symptr);
  printf("constr%d(",arity);
/*  printf("%s(\n");GET_NAME(symptr); */
  while (arity>0){
    write_term(FOLLOW(fp+arity));
    arity--;
    if (arity!=0) printf(",");
  }
  printf(")\n");
}
  
display_constraint(frame)
LONG_PTR frame;
{
  SYM_REC_PTR sym_ptr;
  int arity,i;

  sym_ptr = (SYM_REC_PTR)FOLLOW(((LONG_PTR)AR_REEP(frame)+2));
  arity = GET_ARITY(sym_ptr);
  writepname(stdout, GET_NAME(sym_ptr), GET_LENGTH(sym_ptr));
  printf("(");
  for (i = arity; i>0; i--){ 
    write_term(*(frame+i));
    if (i!=1) printf(",");
  }
  printf(")\n");
}

display_constraints(){

  LONG_PTR frame;

  frame = sfreg;
  while (frame<local_bottom-8){
    display_constraint(frame);
    frame = (LONG_PTR)AR_PREV(frame);
  }
  return 1;
}

display_susp_frame(frame)
LONG_PTR frame;
{
  SYM_REC_PTR sym_ptr;
  int arity,i;
  LONG op;

  op = *(frame+3);
  write_term(op); printf("\n");

  op = *(frame+2);
  write_term(op); printf("\n");

  op = *(frame+1);
  write_term(op); printf("\n");

  sym_ptr = (SYM_REC_PTR)FOLLOW(UNTAGGED(FOLLOW(frame-4))+2*sizeof(LONG));
  printf("sym_ptr = %x\n",sym_ptr);
  printf("%x\n",*frame);
  printf("%x\n",*(frame-1));
  printf("%x\n",*(frame-2));
  printf("%x\n",*(frame-3));
  printf("%x\n",*(frame-4));
  printf("%x\n",*(frame-5));
  return 1;
}


