/*
*   File   : domain.c
*   Author : Neng-Fa ZHOU Copyright (C) All rights reserved
*   Updated: Augest 1996
*   Purpose: Primitives for finite domains
*/
#include "state.h"
#include "event.h"

#define MASK_FF (unsigned long int)0xffffffff
#define MASK_80 (unsigned long int)0x80000000

#define WORD_PTR_OFFSET(dv_ptr,elm,w_ptr,offset){ \
   w_ptr = (LONG_PTR)DV_base_ptr(dv_ptr) + (elm-DV_low_val(dv_ptr))/32; \
   offset = ((elm-DV_low_val(dv_ptr)) % 32); }

#define WORD_OFFSET(dv_ptr,elm,w,w_ptr,offset){ \
   WORD_PTR_OFFSET(dv_ptr,elm,w_ptr,offset); \
   w = FOLLOW(w_ptr); }

#define NEXT_IN_WORD(from,w,w_ptr,offset,mask){ \
    mask = (MASK_FF << offset); \
    while ((w & mask) == 0){ \
      w_ptr++; \
      w = FOLLOW(w_ptr); \
      from = from + 32 - offset; \
      offset = 0; \
      mask = MASK_FF; \
    } \
  }

#define  NEXT_IN_ELM(from,w,offset,mask4,mask){ \
  mask = (0x1 << offset); \
  while ((w & mask) ==0){  \
    from++; \
    mask <<= 1; \
  } \
 }

#define PREV_IN_WORD(to,w,w_ptr,offset,mask){ \
    mask = (MASK_FF >> (31-offset)); \
    while ((w & mask) == 0){ \
      w_ptr--; \
      w = FOLLOW(w_ptr); \
      to = to - offset -1; \
      offset = 31; \
      mask = MASK_FF; \
     } \
   }


#define  PREV_IN_ELM(to,w,offset,mask4,mask){ \
  mask = (0x1 << offset); \
  while ((w & mask) ==0){  \
    to--; \
    mask >>= 1; \
 } \
}

#define BIT_WISE_AND(w1,w2,new_w,mask,count){ \
   new_w = w1; \
   mask = 0x1; \
   while (w2!=0){ \
     if ((w2 & mask) == mask){ \
        if ((new_w & mask) == mask) { \
           new_w &= (~mask); \
           count++; \
         } \
         w2 &= (~mask); \
     } \
     mask <<= 1; \
   } \
 }

int fd_region_low;
int fd_region_up;

/******** c_DM_CREATE *************/
c_DM_CREATE_DVAR(){
  register LONG Var,From,To;
  int res;
  
  Var = ARG(1,3);
  From = ARG(2,3);
  To  = ARG(3,3);
  
  DEREF(Var);
  DEREF(From);DEREF(To);
  From = INTVAL(From); To = INTVAL(To);

  res=create_domain_var_bv(Var,From,To);
  if (res!=1) return res;
  return 1;
}

c_DM_CREATE_DVARS(){
  register LONG Vars,From,To,Var;
  register LONG_PTR sreg;
  int res;

  Vars = ARG(1,3);
  From = ARG(2,3);
  To  = ARG(3,3);
  
  DEREF(Vars);
  DEREF(From);DEREF(To);
  From = INTVAL(From); To = INTVAL(To);

  while (!ISNIL(Vars)) {
    sreg = (LONG_PTR)UNTAGGED(Vars);
    Var = *sreg;
    DEREF(Var);
    Vars = *(sreg+1);
    DEREF(Vars);
    res=create_domain_var_bv(Var,From,To);
    if (res!=1) return res;
  }
  return 1;
}

int create_domain_var_bv(Var,from,to)
register LONG Var;
long int from,to;
{
  register LONG_PTR old_dv_ptr,dv_ptr;

  dv_ptr = hreg;
  DV_var(dv_ptr) =  ((LONG)hreg | SUSP);
  DV_ins_cs(dv_ptr) = nil_sym;
  DV_min_cs(dv_ptr) = nil_sym;
  DV_max_cs(dv_ptr) = nil_sym;
  DV_dom_cs(dv_ptr) = nil_sym;
  DV_dcs_no(dv_ptr) = 0;
  DV_delta(dv_ptr) = nil_sym;
  if (ISVAR(Var)){
    FOLLOW(Var) = (LONG)hreg;
    PUSHTRAIL(Var);
  } else if (IS_SUSP_VAR(Var)){
    old_dv_ptr = (LONG_PTR)UNTAGGED(Var);
    if (IS_UN_DOMAIN(old_dv_ptr)){
      PUSHTRAILB(old_dv_ptr,FOLLOW(old_dv_ptr));
      FOLLOW(old_dv_ptr) = (LONG)hreg;
      DV_ins_cs(dv_ptr) = DV_ins_cs(old_dv_ptr);
      DV_min_cs(dv_ptr) = DV_min_cs(old_dv_ptr);
      DV_max_cs(dv_ptr) = DV_max_cs(old_dv_ptr);
      DV_dom_cs(dv_ptr) = DV_dom_cs(old_dv_ptr);
    } else {
      return domain_region_aux(old_dv_ptr,from,to);
    }
  }
  else {
    exception = illegal_arguments;
    return -1;
  }
  DV_type(dv_ptr) = IT_DOMAIN;
  DV_first(dv_ptr) = from;
  DV_last(dv_ptr) = to;
  DV_size(dv_ptr) = to-from+1;
  hreg += SIZE_OF_DV;
  return 1;
}

fd_vector_min_max(){
  LONG low,up;

  low = ARG(1,2);
  up = ARG(2,2);
  DEREF(low); DEREF(up);
  if (ISINT(low))
    fd_region_low = INTVAL(low);
  if (ISINT(up))
    fd_region_up = INTVAL(up);
  if (!ISINT(low) || !ISINT(up))
    return unify(low,MAKEINT(fd_region_low)) && unify(up,MAKEINT(fd_region_up));
  else return 1;
}

it_to_bv(dv_ptr)
register LONG_PTR dv_ptr;
{
  long int from,to,i;
  LONG_PTR top;

  from = DV_first(dv_ptr);
  if (from<fd_region_low){ 
    from = fd_region_low;
    UPDATE_FIRST(dv_ptr,from);
  }
  to = DV_last(dv_ptr);
  if (to>fd_region_up){
    to = fd_region_up;
    UPDATE_LAST(dv_ptr,to);
  }
	
  i = to-from+1;
  if (DV_size(dv_ptr)!=i) UPDATE_SIZE(dv_ptr,i);

  from = (from>=0) ? (from/32)*32 : (from-32)/32*32;
  to = (to>=0) ? (to/32)*32 : (to-32)/32*32;
  
  DV_low_val(dv_ptr) = from;
  DV_up_val(dv_ptr) = to;
  DV_base_ptr(dv_ptr) = (LONG)hreg;

  for (i=from; i<=to; i += 32)
    *hreg++ = MASK_FF;

  return 1;
}

/******** domain_include  *************/
c_DM_INCLUDE(){
  LONG Var1,Var2;
  LONG_PTR dv_ptr1,dv_ptr2;

  Var1 = ARG(1,2); DEREF(Var1);
  Var2 = ARG(2,2); DEREF(Var2);

  if (IS_SUSP_VAR(Var1) && IS_SUSP_VAR(Var2)){
    dv_ptr1 = (LONG_PTR)UNTAGGED(Var1);
    dv_ptr2 = (LONG_PTR)UNTAGGED(Var2);
    return domain_include(dv_ptr1,dv_ptr2);
  } 
  if (IS_SUSP_VAR(Var1) && ISINT(Var2)){
    dv_ptr1 = (LONG_PTR)UNTAGGED(Var1);
    return dm_true(dv_ptr1,INTVAL(Var2));
  }
  if (ISINT(Var1) && ISINT(Var2))
    return Var1==Var2;
  return 0;
}

int domain_include(dv_ptr1,dv_ptr2)
register LONG_PTR dv_ptr1,dv_ptr2;
{
  register unsigned long int w1,w2,first_offset,last_offset,elm,mask;
  register LONG_PTR w_ptr1,w_ptr2,last_w_ptr;

  if (DV_first(dv_ptr1)>DV_first(dv_ptr2) || DV_last(dv_ptr1)<DV_last(dv_ptr2))
    return 0;
  if (DV_size(dv_ptr1)<DV_size(dv_ptr2)) return 0; 

  if (IS_IT_DOMAIN(dv_ptr1))
    return (DV_first(dv_ptr1)<=DV_first(dv_ptr2) && DV_last(dv_ptr1)>=DV_last(dv_ptr2));
  if (IS_IT_DOMAIN(dv_ptr2)){
    it_to_bv(dv_ptr2);
    UPDATE_TYPE(dv_ptr2,BV_DOMAIN);
  }
  /* bv_bv */

  elm = DV_first(dv_ptr2);
  WORD_PTR_OFFSET(dv_ptr1,elm,w_ptr1,first_offset);
  WORD_PTR_OFFSET(dv_ptr2,elm,w_ptr2,first_offset);
  elm = DV_last(dv_ptr2);
  WORD_PTR_OFFSET(dv_ptr2,elm,last_w_ptr,last_offset);
  mask = (MASK_FF<<first_offset); 
  w1 = FOLLOW(w_ptr1);
  
  w1 &= mask;
  w2 = FOLLOW(w_ptr2);
  w2 &= mask;
  while (w_ptr2!=last_w_ptr){
    if ((w1 & w2) != w2) return 0;
    w_ptr1++;
    w_ptr2++;
    w1 = FOLLOW(w_ptr1);
    w2 = FOLLOW(w_ptr2);
  }
  mask = (~(MASK_FF<<last_offset) | (1<<last_offset));
  w1 &= mask;
  w2 &= mask;
  if ((w1 & w2) != w2) return 0;
  return 1;
}

test_include(dv_ptr1,dv_ptr2)
LONG_PTR dv_ptr1,dv_ptr2;
{
  int i;
  i = DV_first(dv_ptr2);
  while (i < DV_last(dv_ptr2)){
    if (!dm_true(dv_ptr1,i)) { printf("strange\n"); exit(0);}
    if (IS_BV_DOMAIN(dv_ptr2))
      i = domain_next_bv(dv_ptr2,i+1);
    else
      i++;
  }
  if (!dm_true(dv_ptr1,i)) { printf("strange\n"); exit(0);}
  return 1;
}

domain_include_bv_it(dv_ptr1,dv_ptr2)
register LONG_PTR dv_ptr1,dv_ptr2;
{
  register LONG from,to,i;

  from = DV_first(dv_ptr2);
  to = DV_last(dv_ptr2);
  for (i=from;i<=to;i++)
    if (domain_true_bv(dv_ptr1,i)==0) return 0;
  return 1;
}

/******** set_false *************/
domain_set_false_aux(dv_ptr,elm)
register LONG_PTR dv_ptr;
long int elm;
{
  long int first,last;
  
  first = DV_first(dv_ptr);
  last = DV_last(dv_ptr);
  if (DV_size(dv_ptr) == 2){ 
    if (elm == first) 
      elm = last; 
    else if (elm == last) 
      elm = first; 
    else return 1;
    ASSIGN_DVAR(dv_ptr,MAKEINT(elm));
    return 1;
  } 
  if (elm < first || elm > last) return 1;
  if (IS_IT_DOMAIN(dv_ptr))
    return domain_set_false_it(dv_ptr,elm);
  else 
    return domain_set_false_bv(dv_ptr,elm);
}
  
domain_set_false_it(dv_ptr,elm)
register LONG_PTR dv_ptr;
register int elm;
{
  long int count;

  if (elm==DV_first(dv_ptr)){
    UPDATE_FIRST(dv_ptr,elm+1);
    count = DV_size(dv_ptr);
    UPDATE_SIZE(dv_ptr,count-1);
    return 1;
  } else if (elm==DV_last(dv_ptr)){
    UPDATE_LAST(dv_ptr,elm-1);
    count = DV_size(dv_ptr);
    UPDATE_SIZE(dv_ptr,count-1);
    return 1;
  } else {
    it_to_bv(dv_ptr);
    UPDATE_TYPE(dv_ptr,BV_DOMAIN);
    domain_set_false_bv(dv_ptr,elm);
    return 1;
  }
}

domain_set_false_bv(dv_ptr,elm) 
register LONG_PTR dv_ptr;
register int elm;
{
  register LONG_PTR w_ptr,top;
  register unsigned long int w,mask,offset,mask4;
  register int count;

/*  printf("   set_false %x %d size=%d\n",dv_ptr,elm,DV_size(dv_ptr));  */
  if (elm == DV_first(dv_ptr)){  /* first elm */
    elm++;
    WORD_OFFSET(dv_ptr,elm,w,w_ptr,offset);
    NEXT_IN_WORD(elm,w,w_ptr,offset,mask);
    NEXT_IN_ELM(elm,w,offset,mask4,mask);

    UPDATE_FIRST(dv_ptr,elm);
    count = DV_size(dv_ptr);
    UPDATE_SIZE(dv_ptr,count-1);
  } else  if (elm == DV_last(dv_ptr)){ /* last elm */
    elm--;
    WORD_OFFSET(dv_ptr,elm,w,w_ptr,offset);
    PREV_IN_WORD(elm,w,w_ptr,offset,mask);
    PREV_IN_ELM(elm,w,offset,mask4,mask);
      
    UPDATE_LAST(dv_ptr,elm);
    count = DV_size(dv_ptr);
    UPDATE_SIZE(dv_ptr,count-1);
  } else {
    WORD_OFFSET(dv_ptr,elm,w,w_ptr,offset);
    mask = (0x1 << offset);
    if ((w & mask) == 0x0) return 1; /* already out */
    w = (w & (~mask));
    PUSHTRAILB(w_ptr,FOLLOW(w_ptr));
    FOLLOW(w_ptr) = w;
    INSERT_TRIGGER_dom(dv_ptr);
    UPDATE_DELTA(dv_ptr,elm);
    count = DV_size(dv_ptr);
    UPDATE_SIZE(dv_ptr,count-1);
  }
  return 1;
}

int exclude_one_elm_bv(dv_ptr,elm) 
register LONG_PTR dv_ptr;
register int elm;
{
  register LONG_PTR w_ptr;
  register unsigned long int w,new_w,mask,offset;

  WORD_OFFSET(dv_ptr,elm,w,w_ptr,offset);
  mask = (0x1 << offset);
  if ((w & mask) == 0x0) return 0; /* already out */

  new_w = (w & (~mask));
  PUSHTRAILB(w_ptr,w);
  FOLLOW(w_ptr) = new_w;
  INSERT_TRIGGER_dom(dv_ptr);
  UPDATE_DELTA(dv_ptr,elm);
  return 1;
}


/******** next *************/
b_DM_NEXT_ccf(DVar,E,NextE)
register LONG DVar,E,NextE;
{
  register LONG_PTR dv_ptr;
  long int elm;

  DEREF(DVar);DEREF(E);
  /*  if (!IS_SUSP_VAR(DVar)) {exception = illegal_arguments; return -1;}; */
  if (!IS_SUSP_VAR(DVar)) return 0;
  dv_ptr = (LONG_PTR)UNTAGGED(DVar);
  elm = INTVAL(E);
  if (elm >= DV_last(dv_ptr)) return 0;
  if (IS_IT_DOMAIN(dv_ptr))
    elm += 1;
  else 
    elm = domain_next_bv(dv_ptr,elm+1);
  ASSIGN_f(NextE,MAKEINT(elm));
  return 1;
}

domain_next_bv(dv_ptr,elm)
register LONG_PTR dv_ptr;
register long int elm;
{
  register LONG_PTR w_ptr;
  register unsigned long int w,mask,mask4,offset;
  
/*  printf("%x %d\n",dv_ptr,elm); */

  WORD_OFFSET(dv_ptr,elm,w,w_ptr,offset);
  NEXT_IN_WORD(elm,w,w_ptr,offset,mask);
  NEXT_IN_ELM(elm,w,offset,mask4,mask);    

  return elm;
}

/******** prev *************/
b_DM_PREV_ccf(DVar,E,PrevE)
register LONG DVar,E,PrevE;
{
  register LONG_PTR dv_ptr;
  long int elm;

  DEREF(DVar);DEREF(E);
  if (!IS_SUSP_VAR(DVar)) {exception = illegal_arguments; return -1;};
  dv_ptr = (LONG_PTR)UNTAGGED(DVar);
  elm = INTVAL(E);
  if (elm <= DV_first(dv_ptr)) return 0;
  if (IS_IT_DOMAIN(dv_ptr))
    elm -= 1;
  else 
    elm = domain_prev_bv(dv_ptr,elm-1);
  ASSIGN_f(PrevE,MAKEINT(elm));
  return 1;
}

domain_prev_bv(dv_ptr,elm)
register LONG_PTR dv_ptr;
register long int elm;
{
  register LONG_PTR w_ptr;
  register unsigned long int w,mask,mask4,offset;

  WORD_OFFSET(dv_ptr,elm,w,w_ptr,offset);
  PREV_IN_WORD(elm,w,w_ptr,offset,mask);
  PREV_IN_ELM(elm,w,offset,mask4,mask);    

  return elm;
}

/******** true *************/
b_DM_TRUE_cc(Var,E)
register LONG Var,E;
{
  LONG_PTR dv_ptr;

  DEREF(Var); 
  DEREF(E);
  if (!IS_SUSP_VAR(Var)){
    return 0;
  }
  dv_ptr = (LONG_PTR)UNTAGGED(Var);
  return dm_true(dv_ptr,INTVAL(E));
}

dm_true(dv_ptr,elm)
register LONG_PTR dv_ptr;
register long int elm;
{
  if (elm<DV_first(dv_ptr) || elm>DV_last(dv_ptr)) return 0;
  if (IS_IT_DOMAIN(dv_ptr))
    return 1;
  else if (IS_BV_DOMAIN(dv_ptr))
    return domain_true_bv(dv_ptr,elm);
  return 0;
}

domain_true_bv(dv_ptr,elm)
register LONG_PTR dv_ptr;
register long int elm;
{
  register LONG_PTR w_ptr;
  register unsigned long int w,mask,offset;
  
  WORD_OFFSET(dv_ptr,elm,w,w_ptr,offset);
  mask = (0x1 << offset);
  return ((w & mask) !=0);
}
 
/******** intersect *************/
int domain_intersect(dv_ptr1,dv_ptr2)
register LONG_PTR dv_ptr1,dv_ptr2;
{
  register long int first,last;
  LONG_PTR top;

  first = (DV_first(dv_ptr1)>=DV_first(dv_ptr2)) ? DV_first(dv_ptr1) : DV_first(dv_ptr2);
  last = (DV_last(dv_ptr1)<=DV_last(dv_ptr2)) ? DV_last(dv_ptr1) : DV_last(dv_ptr2);

  if (last<first) return 0; 
  if (last == first){
    if (dm_true(dv_ptr1,first) && dm_true(dv_ptr2,first)) {
      ASSIGN_DVAR(dv_ptr1,MAKEINT(first));
      ASSIGN_DVAR(dv_ptr2,MAKEINT(first));
      return 1;
    }
    return 0;
  }
  if (IS_IT_DOMAIN(dv_ptr1))
    if (IS_IT_DOMAIN(dv_ptr2))
      return domain_intersect_it(dv_ptr1,dv_ptr2,first,last);
    else {
      it_to_bv(dv_ptr1);UPDATE_TYPE(dv_ptr1,BV_DOMAIN);
    }
  else if (IS_IT_DOMAIN(dv_ptr2)){
    it_to_bv(dv_ptr2);
    UPDATE_TYPE(dv_ptr2,BV_DOMAIN);
  }
  return domain_intersect_bv(dv_ptr1,dv_ptr2,first,last);
}
  

domain_intersect_it(dv_ptr1,dv_ptr2,first,last)
register LONG_PTR dv_ptr1,dv_ptr2;
register LONG first,last;
{
  long int count;

  if (first != DV_first(dv_ptr1)){
    UPDATE_FIRST(dv_ptr1,first);
  }
  if (first!=DV_first(dv_ptr2)){
    UPDATE_FIRST(dv_ptr2,first);
  }

  if (last != DV_last(dv_ptr1)){
    UPDATE_LAST(dv_ptr1,last);
  } 
  if (last!=DV_last(dv_ptr2)){
    UPDATE_LAST(dv_ptr2,last);
  }
  count = last-first+1;
  if (count!=DV_size(dv_ptr1))
    UPDATE_SIZE(dv_ptr1,count);
  if (count!=DV_size(dv_ptr2))
    UPDATE_SIZE(dv_ptr2,count);
  return 1;
}
 
domain_intersect_bv(dv_ptr1,dv_ptr2,first,last)
register LONG_PTR dv_ptr1,dv_ptr2;
register LONG first,last;
{
  long int elm,count;
  register LONG_PTR w_ptr1,w_ptr2,top;
  register unsigned long int w1,w2,mask,mask4,offset;
  
  elm = (first/32)*32;
  while (elm<=last){
    WORD_OFFSET(dv_ptr1,elm,w1,w_ptr1,offset);
    WORD_OFFSET(dv_ptr2,elm,w2,w_ptr2,offset);
    PUSHTRAILB(w_ptr1,w1);
    PUSHTRAILB(w_ptr2,w2);
    FOLLOW(w_ptr1) = (w1 & w2);
    FOLLOW(w_ptr2) = (w1 & w2);
    elm += 32;
  }
  WORD_OFFSET(dv_ptr1,first,w1,w_ptr1,offset);
  NEXT_IN_WORD(first,w1,w_ptr1,offset,mask);
  NEXT_IN_ELM(first,w1,offset,mask4,mask);

  WORD_OFFSET(dv_ptr1,last,w1,w_ptr1,offset);
  PREV_IN_WORD(last,w1,w_ptr1,offset,mask);
  PREV_IN_ELM(last,w1,offset,mask4,mask);

  if (first > last) return 0;
  if (first == last){
    ASSIGN_DVAR(dv_ptr1,MAKEINT(first));
    ASSIGN_DVAR(dv_ptr2,MAKEINT(first));
    return 1;
  }
  if (first!=DV_first(dv_ptr1))
    UPDATE_FIRST(dv_ptr1,first);
  if (first!=DV_first(dv_ptr2))
    UPDATE_FIRST(dv_ptr2,first);
  if (last!=DV_last(dv_ptr1))
    UPDATE_LAST(dv_ptr1,last);
  if (last!=DV_last(dv_ptr2))
    UPDATE_LAST(dv_ptr2,last);
  count = count_domain_elms(dv_ptr1,first,last);
  if (count!=DV_size(dv_ptr1))
    UPDATE_SIZE(dv_ptr1,count);
  if (count!=DV_size(dv_ptr2))
    UPDATE_SIZE(dv_ptr2,count);

  return 1;
}
    
/******** region *************/
domain_region_aux(dv_ptr,from,to)
register LONG_PTR dv_ptr;
register int from,to;
{

  if (from <= DV_first(dv_ptr) && to >= DV_last(dv_ptr)) return 1;
  from = from>DV_first(dv_ptr) ? from : DV_first(dv_ptr);
  to = to<DV_last(dv_ptr) ? to : DV_last(dv_ptr);
  if (from > to) return 0;
  if (from == to) {
    if (dm_true(dv_ptr,from)){
      ASSIGN_DVAR(dv_ptr, MAKEINT(from));
      return 1;
    } else return 0;
  }
  if (IS_BV_DOMAIN(dv_ptr))
    return domain_region_bv(dv_ptr,from,to);
  else 
    return domain_region_it(dv_ptr,from,to); 
}

domain_region_bv(dv_ptr,from,to)
register LONG_PTR dv_ptr;
register int from,to;
{
  register unsigned long int w,count,mask,mask4,offset;
  register LONG_PTR w_ptr,top;

  WORD_OFFSET(dv_ptr,from,w,w_ptr,offset);
  NEXT_IN_WORD(from,w,w_ptr,offset,mask);
  NEXT_IN_ELM(from,w,offset,mask4,mask);

  WORD_OFFSET(dv_ptr,to,w,w_ptr,offset);
  PREV_IN_WORD(to,w,w_ptr,offset,mask);
  PREV_IN_ELM(to,w,offset,mask4,mask);
  
  if (from>to) return 0;
  if (from == to){
    ASSIGN_DVAR(dv_ptr,MAKEINT(from));
    return 1;
  }

  if (from!=DV_first(dv_ptr)){
    UPDATE_FIRST(dv_ptr,from);
  }
  if (to != DV_last(dv_ptr)){
    UPDATE_LAST(dv_ptr,to);
  }
  count = count_domain_elms(dv_ptr,DV_first(dv_ptr),DV_last(dv_ptr));
  if (count == 0) return 0;
  if (count!=DV_size(dv_ptr)){
    UPDATE_SIZE(dv_ptr,count);
  }
  return 1;
}

domain_region_it(dv_ptr,from,to)
register LONG_PTR dv_ptr;
register int from,to;
{
  LONG_PTR top;

  if (from != DV_first(dv_ptr)){
    UPDATE_FIRST(dv_ptr,from);
  }
  if (to != DV_last(dv_ptr)){
    UPDATE_LAST(dv_ptr,to);
  }
  UPDATE_SIZE(dv_ptr,to-from+1);
  return 1;
}

/******** min_max *************/
b_DM_MIN_MAX_cff(Var,Min,Max)
register LONG Var,Min,Max;
{
  register LONG_PTR dv_ptr;

  DEREF(Var);  
  if (IS_SUSP_VAR(Var)){
    dv_ptr = (LONG_PTR)UNTAGGED(Var);
    if (!IS_UN_DOMAIN(dv_ptr)){
      ASSIGN_f(Min,MAKEINT(DV_first(dv_ptr)));
      ASSIGN_f(Max,MAKEINT(DV_last(dv_ptr)));
      return 1;
    } else {
      exception = illegal_arguments;
      return -1;
    }
  } else if (ISINT(Var)) {
    ASSIGN_f(Min,Var);
    ASSIGN_f(Max,Var);
    return 1;
  }
  write_term(Var);
  exception = illegal_arguments;
  return -1;
}

/******** min *************/
b_DM_MIN_cf(Var,Min)
register LONG Var,Min;
{
  register LONG_PTR dv_ptr;

  DEREF(Var);  
  if (IS_SUSP_VAR(Var)){
    dv_ptr = (LONG_PTR)UNTAGGED(Var);
    if (!IS_UN_DOMAIN(dv_ptr)){
      ASSIGN_f(Min,MAKEINT(DV_first(dv_ptr)));
      return 1;
    } else {
      exception = illegal_arguments;
      return -1;
    }
  } else if (ISINT(Var)) {
    ASSIGN_f(Min,Var);
    return 1;
  }
  write_term(Var);
  exception = illegal_arguments;
  return -1;
}

/******** max *************/
b_DM_MAX_cf(Var,Max)
register LONG Var,Max;
{
  register LONG_PTR dv_ptr;

  DEREF(Var);  
  if (IS_SUSP_VAR(Var)){
    dv_ptr = (LONG_PTR)UNTAGGED(Var);
    if (!IS_UN_DOMAIN(dv_ptr)){
      ASSIGN_f(Max,MAKEINT(DV_last(dv_ptr)));
      return 1;
    } else {
      exception = illegal_arguments;
      return -1;
    }
  } else if (ISINT(Var)) {
    ASSIGN_f(Max,Var);
    return 1;
  }
  write_term(Var);
  exception = illegal_arguments;
  return -1;
}

/******** count *************/
b_DM_COUNT_cf(Var,Count)
register LONG Var,Count;
{
  LONG_PTR dv_ptr;

  DEREF(Var); 
  if (IS_SUSP_VAR(Var)){
    dv_ptr = (LONG_PTR)UNTAGGED(Var);
    if (!IS_UN_DOMAIN(dv_ptr)){
      ASSIGN_f(Count,MAKEINT(DV_size(dv_ptr)));
      return 1;
    }
  }
  exception = illegal_arguments;
  return -11;
}

c_FD_DELTA()
{
  LONG var,delta,deltaTerm;
  LONG_PTR dv_ptr;

  var = ARG(1,2);
  delta = ARG(2,2);

  DEREF(var); 
  DEREF(delta);
  if (IS_SUSP_VAR(var)){
    dv_ptr = (LONG_PTR)UNTAGGED(var);
    if (!IS_UN_DOMAIN(dv_ptr)){
      if (DV_delta(dv_ptr)==nil_sym){ 
	PUSHTRAILB(A_DV_delta(dv_ptr),nil_sym); 
	DV_delta(dv_ptr) = (LONG)hreg;
        NEW_HEAP_FREE;
      }
      deltaTerm = ADDTAG(hreg,STR);
      FOLLOW(hreg++) = (LONG)domain_delta;
      FOLLOW(hreg++) = UNTAGGED(var);
      FOLLOW(hreg++) = DV_delta(dv_ptr);
      FOLLOW(delta) = deltaTerm;
      return 1;
    }
  }
  FOLLOW(delta) = nil_sym;
  return 1;
  /*
  exception = illegal_arguments;
  return -1;
  */
}

c_FD_DELTA_ELMS()
{
  LONG varDelta,elms;
  LONG var,delta,excludedElms;
  LONG_PTR ptr,ptr2,dv_ptr;

  varDelta = ARG(1,2);
  elms = ARG(2,2);

  DEREF(varDelta); /* domain_delta(var,delta) */
  DEREF(elms);
  if (ISATOM(varDelta)) return unify(elms,nil_sym);
  ptr = (LONG_PTR)UNTAGGED(varDelta);
  var = FOLLOW(ptr+1);
  DEREF(var);
  if (IS_SUSP_VAR(var)){
    delta = FOLLOW(FOLLOW(ptr+2));
    excludedElms = copyList(delta);
    FOLLOW(elms) = excludedElms;
    if (excludedElms!=nil_sym){
      ptr2 = ptr+2;
      PUSHTRAILB(ptr2,FOLLOW(ptr2)); 
      dv_ptr = (LONG_PTR)UNTAGGED(var);
      FOLLOW(ptr2) = DV_delta(dv_ptr);
    }
  } else FOLLOW(elms)=nil_sym;
  return 1;
}

copyList(list)
LONG list;
{
  LONG_PTR ptr;
  
  if (ISLIST(list)){
    UNTAG(list);
    ptr = hreg; hreg += 2;
    FOLLOW(ptr) = FOLLOW(list);
    FOLLOW(ptr+1) = copyList(FOLLOW((LONG_PTR)list+1));
    return ADDTAG(ptr,LST);
  } else return nil_sym;
}

c_FD_DELTA_INTERVAL()
{
  LONG varDelta,start,end;
  LONG var,delta;
  LONG_PTR ptr,ptr2,dv_ptr;

  varDelta = ARG(1,3);
  start = ARG(2,3);
  end = ARG(3,3);

  DEREF(varDelta); /* domain_delta(var,delta) */
  ptr = (LONG_PTR)UNTAGGED(varDelta);
  var = FOLLOW(ptr+1);
  DEREF(var);
  if (IS_SUSP_VAR(var)){
    ptr2 = ptr+2; /* address of delta */
    unify(start,ADDTAG(FOLLOW(ptr2),INT_TAG));
    dv_ptr = (LONG_PTR)UNTAGGED(var);
    unify(end,ADDTAG(DV_delta(dv_ptr),INT_TAG));
    PUSHTRAILB(ptr2,FOLLOW(ptr2)); 
    FOLLOW(ptr2) = DV_delta(dv_ptr);
    return 1;
  }
  unify(start,MAKEINT(0));
  unify(end,MAKEINT(0));
  return 1;
}

int count_domain_elms(dv_ptr,from,to)
register LONG_PTR dv_ptr;
register LONG from,to;
{
  register LONG_PTR w_ptr;
  register unsigned long int w,mask,offset,count;
  
  count = 0;
  WORD_OFFSET(dv_ptr,from,w,w_ptr,offset);
  mask = (0x1 << offset);
  while (from<=to){
    if ((w & mask) != 0) count++;
    mask <<= 1;
    if (mask == 0){
      mask = 0x1; w_ptr++; w = FOLLOW(w_ptr);
    }
    from++;
  }
  return count;
}
/******** misc *************/
print_domain_var(){
  LONG Var;

  LONG_PTR dv_ptr;
  Var = ARG(1,1);
  DEREF(Var);
  
  if (!IS_SUSP_VAR(Var)) return 1;
  dv_ptr = (LONG_PTR)UNTAGGED(Var);
  return print_dvar(dv_ptr);
}


print_dvar(dv_ptr)
LONG_PTR dv_ptr;
{
  LONG_PTR ptr;

  printf("%x:%d..%d:%d",DV_var(dv_ptr),DV_first(dv_ptr),DV_last(dv_ptr),DV_size(dv_ptr));
}

print_domain(dv_ptr)
LONG_PTR dv_ptr;
{
  int i,first,last;
 
  first = DV_first(dv_ptr);
  last = DV_last(dv_ptr);
  
  printf("elms: ");
  for (i=first;i<=last;i++){
    if (dm_true(dv_ptr,i)) printf("%i ",i);
  }
  printf("\n");
  return 1;
}

show_list(ptr)
LONG_PTR ptr;
{
  LONG_PTR  ptr1,ptr2;
  int i,arity;
  SYM_REC_PTR sym_ptr;
  
  ptr1 = ptr;
  while (ptr1 != NULL){
    ptr2 = (LONG_PTR)FOLLOW(ptr1);
    SHOW_SUSP(ptr2); 
    ptr1 = (LONG_PTR)FOLLOW(ptr1+1);
  }
}

print_bv(dv_ptr)
LONG_PTR dv_ptr;
{
  LONG from,to;
  register LONG_PTR w_ptr;
  unsigned long int w,mask,offset,count;

  from = DV_first(dv_ptr); to = DV_last(dv_ptr);
  printf("from=%d to=%d\n",from,to);
  while (from<=to){
    WORD_OFFSET(dv_ptr,from,w,w_ptr,offset);
    printf("%x\n",w);
    from +=32;
  }
}

Cboot_domain(){
  SYM_REC_PTR insert_cpred();
  extern int c_notin();
  extern int exclude_elm_dvars();
  extern c_DM_CREATE();
  extern int display_constraints();
  extern c_DM_INCLUDE();
  extern c_deleteff_des();

  fd_region_low = -320;
  fd_region_up = 320;
  
  insert_cpred("c_DM_CREATE_DVARS",3,c_DM_CREATE_DVARS);
  insert_cpred("c_DM_CREATE_DVAR",3,c_DM_CREATE_DVAR);
  insert_cpred("print_domain_var",1,print_domain_var);
  insert_cpred("exclude_elm_dvars",3,exclude_elm_dvars);
  insert_cpred("c_notin",3,c_notin);
  insert_cpred("display_constraints",0,display_constraints);
  insert_cpred("c_DM_INCLUDE",2,c_DM_INCLUDE);
  insert_cpred("fd_vector_min_max",2,fd_vector_min_max);
  insert_cpred("c_deleteff_des",3,c_deleteff_des);
  insert_cpred("fd_delta",2,c_FD_DELTA);
  insert_cpred("fd_delta_interval",3,c_FD_DELTA_INTERVAL);
  insert_cpred("fd_delta_elms",2,c_FD_DELTA_ELMS);
}
