/*
*   File   : float.c
*   Author : Neng-Fa ZHOU
*   Date   : Completed November 1993
*   Purpose: functions of float points
*/
#include "basic.h"
#include "term.h"
#include "macro.h"
#include <math.h>

#define Bit20          0x100000
#define Bits15to19     0x0f8000

#define MIN(x, y)      (x > y ? y : x)

#define EXP_SIGN       0x20000000
#define MANT_SIGN      0x40000000
#define EXP_MAGN(op)   (((unsigned)(op & 0xf8)) >> 3)
#define MANT_MAGN(op)  (((unsigned)(op & 0x1fffff00)) >> 8)

extern FILE *curr_out;

double frexp(), ldexp();   

int eval_arith();

/* converts floats from WAM representation to 
 *  the machine representation 
 */
double floatval(op)
register LONG op;
{
  double f;
  unsigned int w1,w2,w3,w,*pw;
  
  UNTAG(op);
/*  if (FOLLOW(op) != (LONG)float_psc) {
    printf("number required\n");
    return -1; 
  }
*/
  w1 = INTVAL(*((LONG_PTR)op+1));
  w2 = INTVAL(*((LONG_PTR)op+2));
  w3 = INTVAL(*((LONG_PTR)op+3));

  
  pw = (unsigned int *)&f;
  w1 = (w1 << 22) | (w2 >> 5);
  *pw = w1;
  w1 = ((w2  << 27) | w3);
  *(pw+1) = w1;
  return f;
}

/* "encodefloat" converts floats from the machine representation
 *             to the WAM representation.
 *  '$float'(int1,int2,int2);
*/
LONG encodefloat(op)
double op;
{
  register LONG temp;
  unsigned int *pw,w1,w2;

  temp = ADDTAG(hreg,STR);
  NEW_HEAP_NODE((LONG)float_psc);  /* '$float'(int1,int2,int3) */
  
  pw = (unsigned int *)&op;
  w1 = *pw; 
  w2 = *(pw+1);
  NEW_HEAP_NODE(MAKEINT(w1>>22));                              /* int1 */
  NEW_HEAP_NODE(MAKEINT(((w1 & 0x3fffff) << 5) | (w2>>27)));   /* int2 */
  NEW_HEAP_NODE(MAKEINT(w2 & 0x7ffffff));                      /* int3 */

  if (arreg-1000 < hreg) { 
    quit("control stack overflow\n");
  }
  return temp;
}

equal_to(op1,op2)
register LONG op1,op2;
{
  LONG top1;
  double floatval();

  DEREF(op1);DEREF(op2);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (!ISNUM(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }
  
  if (ISINT(op1))
    if (ISINT(op2))
      return (op1 == op2);
    else    /* op1 is int, op2 is not int */
      return prettymuch_equal((double)INTVAL(op1),floatval(op2));
  else     /* op1 is not int */
    if (ISINT(op2))
      return prettymuch_equal(floatval(op1),(double)INTVAL(op2));      
    else   /* op1 and op2 are both floats */
      return prettymuch_equal(floatval(op1),floatval(op2));
}

greater_than(op1,op2)
register LONG op1,op2;
{

  double floatval();
  LONG top1;

  DEREF(op1);
  DEREF(op2);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (!ISNUM(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }

  if (ISINT(op1)) 
    if (ISINT(op2))
      return (INTVAL(op1) > INTVAL(op2));
    else    /* op1 is int, op2 is not int */
      return ((double)INTVAL(op1) > floatval(op2));
  else     /* op1 is not int */
    if (ISINT(op2))
      return (floatval(op1) > (double)INTVAL(op2));
    else    /* op1 and op2 are both floats */
      return (floatval(op1)>floatval(op2));
}

greater_equal(op1,op2)
register LONG op1,op2;
{
  LONG top1;
  double floatval();

  DEREF(op1);
  DEREF(op2);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (!ISNUM(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }

  if (ISINT(op1))
    if (ISINT(op2))
      return (INTVAL(op1) >= INTVAL(op2));
    else    /* op1 is int, op2 is not int */
      return ((double)INTVAL(op1) >= floatval(op2));
  else     /* op1 is not int */
    if (ISINT(op2))
      return (floatval(op1) >= (double)INTVAL(op2));
    else    /* op1 and op2 are both floats */
      return (floatval(op1)>=floatval(op2));
}

b_FLOAT_MINUS_cf(op1,op2)
register LONG op1,op2;
{
  double floatval();

  DEREF(op1);DEREF(op2);
  ASSIGN_f(op2,encodefloat(0.0-floatval(op1)));
  return 1;
}
 
b_FLOAT_LOG_cf(op1,op2)
register LONG op1,op2;
{
  double floatval();
  LONG top1;
  
  DEREF(op1);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }

  if (ISINT(op1)) 
    ASSIGN_f(op2,encodefloat(log((double)INTVAL(op1))));
  else 
    ASSIGN_f(op2,encodefloat(log(floatval(op1))));
  return 1;
}

b_FLOAT_POW_ccf(op1,op2,op3)
register LONG op1,op2,op3;
{
  double f1,f2;
  double floatval();
  LONG top1;
  
  DEREF(op1);DEREF(op2);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (!ISNUM(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }

  if (ISINT(op1))
    f1 = (double)INTVAL(op1);
  else 
    f1 = floatval(op1);
  if (ISINT(op2))
    f2 = (double)INTVAL(op2);
  else 
    f2 = floatval(op2);
  ASSIGN_f(op3,encodefloat(pow(f1,f2)));
  return 1;
}

b_FLOAT_SQRT_cf(op1,op2)
register LONG op1,op2;
{
  register double f1;
  double floatval();
  LONG top1;

  DEREF(op1);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) {
      write_term(op1);printf(" is less than zero in sqrt");return -1;}
    op1 = top1;
  }

  if (ISINT(op1))
    f1 = (double)INTVAL(op1);
  else 
    f1 = floatval(op1);
  ASSIGN_f(op2,encodefloat(sqrt(f1)));
  return 1;
}

b_FLOAT_ABS_cf(op1,op2)
register LONG op1,op2;
{
  register double f1;
  double floatval();
  LONG top1;
  
  DEREF(op1);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (ISINT(op1))
    ASSIGN_f(op2,MAKEINT(abs(INTVAL(op1))));
  else 
    ASSIGN_f(op2,encodefloat(fabs(floatval(op1))));
  return 1;
}

b_FLOAT_EXP_cf(op1,op2)
register LONG op1,op2;
{
  register double f1;
  double floatval();
  LONG top1;
  
  DEREF(op1);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }

  if (ISINT(op1))
    f1 = (double)INTVAL(op1);
  else 
    f1 = floatval(op1);
  ASSIGN_f(op2,encodefloat(exp(f1)));
  return 1;
}

b_FLOAT_SIN_cf(op1,op2)
register LONG op1,op2;
{
  register double f1;
  double floatval();
  LONG top1;
  
  DEREF(op1);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }

  if (ISINT(op1))
    f1 = (double)INTVAL(op1);
  else 
    f1 = floatval(op1);
  ASSIGN_f(op2,encodefloat(sin(f1)));
  return 1;
}


b_FLOAT_COS_cf(op1,op2)
register LONG op1,op2;
{
  register double f1;
  double floatval();
  LONG top1;

  DEREF(op1);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }

  if (ISINT(op1))
    f1 = (double)INTVAL(op1);
  else 
    f1 = floatval(op1);
  ASSIGN_f(op2,encodefloat(cos(f1)));
  return 1;
}


b_FLOAT_TAN_cf(op1,op2)
register LONG op1,op2;
{
  register double f1;
  double floatval();
  LONG top1;

  DEREF(op1);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }

  if (ISINT(op1))
    f1 = (double)INTVAL(op1);
  else 
    f1 = floatval(op1);
  ASSIGN_f(op2,encodefloat(tan(f1)));
  return 1;
}

b_FLOAT_ATAN_cf(op1,op2)
register LONG op1,op2;
{
  register double f1;
  double floatval();
  LONG top1;

  DEREF(op1);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (ISINT(op1))
    f1 = (double)INTVAL(op1);
  else 
    f1 = floatval(op1);
  ASSIGN_f(op2,encodefloat(atan(f1)));
  return 1;
}

b_FLOAT_ASIN_cf(op1,op2)
register LONG op1,op2;
{
  register double f1;
  double floatval();
  LONG top1;

  DEREF(op1);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (ISINT(op1))
    f1 = (double)INTVAL(op1);
  else 
    f1 = floatval(op1);
  ASSIGN_f(op2,encodefloat(asin(f1)));
  return 1;
}

b_FLOAT_ACOS_cf(op1,op2)
register LONG op1,op2;
{
  register double f1;
  double floatval();
  LONG top1;

  DEREF(op1);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (ISINT(op1))
    f1 = (double)INTVAL(op1);
  else 
    f1 = floatval(op1);
  ASSIGN_f(op2,encodefloat(acos(f1)));
  return 1;
}

b_FLOAT_WRITE_c(op)
register LONG op;
{
  double floatval();
  
  DEREF(op);
  fprintf(curr_out,"%g",floatval(op)); 
  return 1;
}

b_FLOAT_FLOOR_cf(op1,op2)
register LONG op1,op2;
{
  double floatval();
  double temp;
  int tempi;
  LONG top1;

  DEREF(op1);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (ISINT(op1)) {
    ASSIGN_f(op2,op1);
  }
  else {
    temp = floatval(op1);
    tempi = (int)temp;
    if (temp>=tempi)
      ASSIGN_f(op2,MAKEINT(tempi));
    else 
      ASSIGN_f(op2,MAKEINT(tempi-1));
  }
  return 1;
}

b_FLOAT_FLOAT_cf(op1,op2)
register LONG op1,op2;
{
  LONG top1;

  DEREF(op1);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (ISINT(op1)) 
    ASSIGN_f(op2,encodefloat(INTVAL(op1)*1.0));
  else 
    ASSIGN_f(op2,op1);
  return 1;
}

b_FLOAT_ROUND_cf(op1,op2)
register LONG op1,op2;
{
  double floatval();
  LONG top1;

  DEREF(op1);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (ISINT(op1)) 
    ASSIGN_f(op2,op1);
  else 
    ASSIGN_f(op2,MAKEINT((int)(floatval(op1)+0.5)));
  return 1;
}

b_FLOAT_TRUNCATE_cf(op1,op2)
register LONG op1,op2;
{  
  double floatval();
  double f;
  LONG top1;

  DEREF(op1);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (ISINT(op1)) 
    ASSIGN_f(op2,op1);
  else {
    f = floatval(op1);
    if (f>=0.0) 
      ASSIGN_f(op2,MAKEINT((int)f));
    else
      ASSIGN_f(op2,MAKEINT(-(int)f));
  }
  return 1;
}

b_FLOAT_SIGN_cf(op1,op2)
register LONG op1,op2;
{
  extern SIGN();

  return SIGN(op1,op2);
}

b_FLOAT_CEILING_cf(op1,op2)
register LONG op1,op2;
{
  double floatval();
  double temp;
  int tempi;
  LONG top1;

  DEREF(op1);
  if (!ISNUM(op1)) {
    NEW_VAR(top1);
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  if (ISINT(op1)) 
    ASSIGN_f(op2,op1);
  else {
    temp = floatval(op1);
    tempi = (int)temp;
    if (temp>tempi)
      ASSIGN_f(op2,MAKEINT(tempi+1));
    else 
      ASSIGN_f(op2,MAKEINT(tempi));
  }
  return 1;
}

b_XOR_ccf(op1,op2,op3)
register LONG op1,op2,op3;
{
  LONG top1;

  DEREF(op1);
  if (!ISNUM(op1)) {
    if (eval_arith(op1,&top1)<0) return -1;
    op1 = top1;
  }
  DEREF(op2);
  if (!ISNUM(op2)){
    if (eval_arith(op2,&top1)<0) return -1;
    op2 = top1;
  }
  
  if (ISINT(op1) && ISINT(op2)) {
    ASSIGN_f(op3,MAKEINT(INTVAL(op1)^INTVAL(op2)));
    return 1;
  }
  else {
    exception = integer_expected;
    return -1;
  }
}
    
prettymuch_equal(op1, op2)
double op1, op2;
{
   double min, diff;

   if ((op1 < 0.0 && op2 > 0.0) || (op2 < 0.0 && op1 > 0.0))
      return 0;
   if (op1 < 0)
      op1 = -op1;
   if (op2 < 0)
      op2 = -op2;
   diff = op1 - op2;
   if (diff < 0)
      diff = -diff;
   min = MIN(op1, op2);
   if (min == 0.0)
      return (op1 == op2);
   else return ((diff/min) < EPSILON);
}







