/*
*   File   : global.c
*   Author : Neng-Fa ZHOU
*   Date   : Completed November 1993
*   Purpose: Management of the global database
*   Copyright (C) All rights reserved
*/
#include "cpreds.h"
#include "dynamic.h"

#define GET_GLOBAL_SYM(name,arity,sym_ptr) \
  DEREF(name);DEREF(arity);   \
  sym_ptr = GET_ATM_SYM_REC(name); \
  sym_ptr = insert(GET_NAME(sym_ptr),GET_LENGTH(sym_ptr),INTVAL(arity));

b_IS_CONSULTED_c(goal)
register LONG goal;
{
  register LONG_PTR top;

 is_consulted:
  if (ISVAR(goal)){
    NDEREF(goal,is_consulted);
    return 0;
  }
  if (ISSTRUCT(goal))
    return (GET_ETYPE(GET_STR_SYM_REC(goal))==T_DYNA);
  else if (ISATOM(goal))
    return (GET_ETYPE(GET_ATM_SYM_REC(goal))==T_DYNA);
  return 0;
}

b_SET_DYNAMIC_cc(name,arity)
register LONG name,arity;
{
  register LONG_PTR top;
  register SYM_REC_PTR sym_ptr;
  
  GET_GLOBAL_SYM(name,arity,sym_ptr);
  GET_ETYPE(sym_ptr)=T_DYNA;
  return 1;
}

b_GET_SYM_TYPE_ccf(name,arity,type)
register LONG name,arity,type;
{
  register LONG_PTR top;
  register SYM_REC_PTR sym_ptr;

  GET_GLOBAL_SYM(name,arity,sym_ptr);
  ASSIGN_f(type,MAKEINT(GET_ETYPE(sym_ptr)));
  return 1;
}

b_IS_DYNAMIC_cc(name,arity)
register LONG name,arity;
{
  register LONG_PTR top;
  register SYM_REC_PTR sym_ptr;
  
  GET_GLOBAL_SYM(name,arity,sym_ptr);
  return (GET_ETYPE(sym_ptr)==T_DYNA) ? 1 : 0;
}

b_IS_ORDINARY_cc(name,arity)
register LONG name,arity;
{
  register LONG_PTR top;
  register SYM_REC_PTR sym_ptr;

  GET_GLOBAL_SYM(name,arity,sym_ptr);
  return (GET_ETYPE(sym_ptr)==T_ORDI) ? 1 : 0;
}

b_GLOBAL_CREATE_cccc(name,arity,value,part) 
register LONG name,arity,value,part;
{
  return b_GLOBAL_SET_cccc(name,arity,value,part);
}

b_GLOBAL_SET_cccc(name,arity,value,part)
register LONG name,arity,value,part;
 {
   LONG_PTR top;
   SYM_REC_PTR sym_ptr;
   LONG oldValue;
  
   GET_GLOBAL_SYM(name,arity,sym_ptr);
   DEREF(part); part = INTVAL(part);
   if (GET_ETYPE(sym_ptr)==T_DYNA)
     oldValue = (LONG)GET_EP(sym_ptr);
   else {
     GET_ETYPE(sym_ptr) = T_DYNA;
     oldValue = nil_sym;
   }
   PAREA_TOP(part);
   DEREF(value);
   if (TAG(value) == ATM) 
     GET_EP(sym_ptr) = (int (*)())value;
   else {
     GET_EP(sym_ptr) = (int (*)())heap_to_parea_reuse(value,oldValue);
     UPDATE_CURR_FENCE(part);
   }
   return 1;
}

b_GLOBAL_SET_ccc(name,arity,value)
register LONG name,arity,value;
 {
   LONG_PTR top;
   SYM_REC_PTR sym_ptr;
   LONG oldValue;
  
   GET_GLOBAL_SYM(name,arity,sym_ptr);
   if (GET_ETYPE(sym_ptr)==T_DYNA)
     oldValue = (LONG)GET_EP(sym_ptr);
   else {
     GET_ETYPE(sym_ptr) = T_DYNA;
     oldValue = nil_sym;
   }
   PAREA_TOP(1);
   DEREF(value);
   if (TAG(value) == ATM) 
     GET_EP(sym_ptr) = (int (*)())value;
   else {
     GET_EP(sym_ptr) = (int (*)())heap_to_parea_reuse(value,oldValue);
     UPDATE_CURR_FENCE(1);
   }
   return 1;
 }

c_OLD_GLOBAL_SET4(){
  LONG name,arity,value,part;
  SYM_REC_PTR sym_ptr;
  name = ARG(1,4);
  arity = ARG(2,4);
  value = ARG(3,4);
  part = ARG(4,4);DEREF(part);part=INTVAL(part);
  GET_GLOBAL_SYM(name,arity,sym_ptr);
  GET_ETYPE(sym_ptr) = T_DYNA;
  PAREA_TOP(part);
  GET_EP(sym_ptr) = (int (*)())heap_to_parea_reuse(value,nil_sym);
  UPDATE_CURR_FENCE(part);
  return 1;
}

c_OLD_GLOBAL_SET3(){
  LONG name,arity,value;
  SYM_REC_PTR sym_ptr;
  name = ARG(1,3);
  arity = ARG(2,3);
  value = ARG(3,3);
  GET_GLOBAL_SYM(name,arity,sym_ptr);
  GET_ETYPE(sym_ptr) = T_DYNA;
  PAREA_TOP(1);
  GET_EP(sym_ptr) = (int (*)())heap_to_parea_reuse(value,nil_sym);
  UPDATE_CURR_FENCE(1);
  return 1;
}
  
c_OLD_GLOBAL_GET(){
  LONG name,arity,value;
  SYM_REC_PTR sym_ptr;
  name = ARG(1,3);
  arity = ARG(2,3);
  value = ARG(3,3);
  GET_GLOBAL_SYM(name,arity,sym_ptr);
  if (GET_ETYPE(sym_ptr)!=T_DYNA) {
    return 0;
  } else {
    return unify(value,GET_EP(sym_ptr));
  }
}
  
b_GLOBAL_GET_ccf(name,arity,value)
register LONG name,arity,value;
{
  LONG_PTR top,varVector;
  SYM_REC_PTR sym_ptr;
  LONG term,maxVarNo=0;

  GET_GLOBAL_SYM(name,arity,sym_ptr);
  if (GET_ETYPE(sym_ptr)!=T_DYNA) {
    return 0;
  }
  term = (LONG)GET_EP(sym_ptr);
  if (TAG(term)==ATM)
    ASSIGN_f(value,term);
  else {
    varVector = arreg-1000; /* safe? */
    ASSIGN_f(value,unnumberVarTerm(term,varVector,&maxVarNo));    
    QUIT_IF_OVERFLOW_HEAP(1000+maxVarNo);
  }
  return 1;
}

b_ISGLOBAL_cc(name,arity)
register LONG name,arity;
{
  register LONG_PTR top;
  register SYM_REC_PTR sym_ptr;
    
  GET_GLOBAL_SYM(name,arity,sym_ptr);
  return (GET_ETYPE(sym_ptr)!=T_DYNA) ? 0 : 1;
}

b_GLOBAL_DEL_cc(name,arity)
register LONG name,arity;
 {
  register LONG_PTR top;
  register SYM_REC_PTR sym_ptr;
  
  GET_GLOBAL_SYM(name,arity,sym_ptr);
  GET_ETYPE(sym_ptr)= T_ORDI;
  return 1;
}

b_GLOBAL_INSERT_HEAD_cccc(name,arity,value,part)
register LONG name,arity,value,part;
{
  register LONG_PTR top;
  register SYM_REC_PTR sym_ptr;
  register LONG temp1;

  GET_GLOBAL_SYM(name,arity,sym_ptr);
  DEREF(part); part = INTVAL(part);
  PAREA_TOP(part);
  temp1 = heap_to_parea_reuse(value,nil_sym);
  GET_EP(sym_ptr) = (int (*)())make_cons(temp1,(LONG)GET_EP(sym_ptr));
  UPDATE_CURR_FENCE(part);
  return 1;
}

b_GLOBAL_INSERT_TAIL_ccc(name,arity,value)
register LONG name,arity,value;
 {
   register LONG_PTR top;
   register SYM_REC_PTR sym_ptr;
   register LONG temp1;

   GET_GLOBAL_SYM(name,arity,sym_ptr);
   PAREA_TOP(1);
   temp1 = heap_to_parea_reuse(value,nil_sym);
   assert_tail(sym_ptr,make_cons(temp1,nil_sym));
   UPDATE_CURR_FENCE(1);
   return 1;
 }

assert_tail(sym_ptr,term)
SYM_REC_PTR sym_ptr;
LONG term;
{
  register LONG_PTR top,ptr;
  register LONG op;
  
  if (ISNIL((LONG)GET_EP(sym_ptr)))
    GET_EP(sym_ptr) = (int (*)())term;
  else {
    op = (LONG)GET_EP(sym_ptr);
    DEREF(op);
    while (ISLIST(op)){
      ptr = (LONG_PTR)UNTAGGED(op)+1;
      op = FOLLOW(ptr);
      DEREF(op);
    }
    FOLLOW(ptr) = term;
  }
  return 1;
}
  
b_GLOBAL_REMOVE_CLAUSE_ccc(name,arity,no)
register LONG name,arity,no;
{
  register LONG_PTR top,prev;
  SYM_REC_PTR ptr;
  BYTE found;
  register LONG temp,clause;

  GET_GLOBAL_SYM(name,arity,ptr);
  DEREF(no);
  temp = (LONG)GET_EP(ptr); 
  prev = (LONG_PTR)&GET_EP(ptr);
  if (ISSTRUCT(temp)){  /* hashtab */
    prev = (LONG_PTR)UNTAGGED(temp)+GET_SYM_ARITY(temp);
    temp = FOLLOW(prev);
  }
  while (ISLIST(temp)) {
    UNTAG(temp);
    clause = FOLLOW(temp);  /* CAR */
    DEREF(clause);
    UNTAG(clause);
    if (*((LONG_PTR)clause+1)==no) {
      *((LONG_PTR)clause+2)=MAKEINT(0);    /* 0 means deed*/
      FOLLOW(prev) = FOLLOW((LONG_PTR)temp+1);
      return 1;
    }
    prev = (LONG_PTR)temp+1;
    temp = FOLLOW(prev);
  }
  return 1;
}
  
make_cons(car,cdr)
LONG car,cdr;
{
  LONG temp;
  
  temp = (LONG)ADDTAG(parea_top,LST);    
  *parea_top++ = car;
  *parea_top++ = cdr;
  return temp;
}  

heap_to_parea_reuse(value,oldValue) 
LONG value,oldValue;
{
  LONG_PTR treg0;
  LONG temp,tmp_op,varno=1;

  treg0 = treg;
  temp = numberVarCopyToParea(value,oldValue,&varno);
  UNDO_TRAILING;
  return temp;
}
  
numberVarCopyToParea(term,oldTerm,varno)
LONG term,oldTerm;
int *varno;
{
  LONG_PTR ptr;

  DEREF(term);
  if (TAG(term)==ATM || IsNumberedVar(term))
    return term;
  else if (ISVAR(term)){
    ASSIGN_TRAIL_VALUE(term,NumberVar(*varno));
    *varno = *varno+1;
    return FOLLOW(term);
  } else if (ISLIST(term)){
    UNTAG(term);
    if (ISLIST(oldTerm)){
      ptr = (LONG_PTR)UNTAGGED(oldTerm);
      FOLLOW(ptr) = numberVarCopyToParea(FOLLOW(term),FOLLOW(ptr),varno);
      FOLLOW(ptr+1) = numberVarCopyToParea(FOLLOW((LONG_PTR)term+1),FOLLOW(ptr+1),varno);
    } else {
      ALLOCATE_FROM_PAREA(2,ptr);
      FOLLOW(ptr) = numberVarCopyToParea(FOLLOW(term),nil_sym,varno);
      FOLLOW(ptr+1) = numberVarCopyToParea(FOLLOW((LONG_PTR)term+1),nil_sym,varno);
    }
    return ADDTAG(ptr,LST);
  } else if (ISSTRUCT(term)){
    int i,arity;
    UNTAG(term);
    arity = GET_ARITY((SYM_REC_PTR)FOLLOW(term));
    ptr = (LONG_PTR)UNTAGGED(oldTerm);
    if (ISSTRUCT(oldTerm) && GET_ARITY((SYM_REC_PTR)FOLLOW(ptr))>arity){
      FOLLOW(ptr) = FOLLOW(term);
      for (i=1;i<=arity;i++){
	FOLLOW(ptr+i) = numberVarCopyToParea(FOLLOW((LONG_PTR)term+i),FOLLOW(ptr+i),varno);
      }
    } else {
      ALLOCATE_FROM_PAREA(arity+1,ptr);
      FOLLOW(ptr) = FOLLOW(term);
      for (i=1;i<=arity;i++){
	FOLLOW(ptr+i) = numberVarCopyToParea(FOLLOW((LONG_PTR)term+i),nil_sym,varno);
      }
    }
    return ADDTAG(ptr,STR);
  } else quit("Dynamic values cannot contain suspension variables\n");
}

c_UNNUMBER_VARS(){
  TERM term,unnumberedTerm,temp;
  LONG_PTR varVector=arreg-7;
  int maxVarNo=0;

  term = ARG(1,2);DEREF(term);
  unnumberedTerm = ARG(2,2); 
  temp = unnumberVarTerm(term,varVector,&maxVarNo);
  return unify(temp,unnumberedTerm);
}
  
  
