/*
*   File   : kapi.c
*   Author : Neng-Fa ZHOU Copyright (C) All rights reserved
*   Updated  : November 1998
*   Purpose: External language interface
*/
#include "cpreds.h"
#include "kapi.h"
#include <stdlib.h>
#include "frame.h"

SYM_REC_PTR  objectRef;
extern char *string_in;
extern char lastc;

/* Test Prolog terms in C */
/*
int PisVar(CELL term){
  DEREF(term);
  return ISVAR(term);
}

int PisAtom(CELL term){
  DEREF(term);
  return ISATOM(term);
}

int PisInt(CELL term){
  DEREF(term);
  return ISINT(term);
}

int PisLong(CELL term){
  DEREF(term);
  return 0;
}

int PisReal(CELL term){
  DEREF(term);
  return ISFLOAT(term);
}

int PisAddr(CELL term){
  DEREF(term);
  return (ISSTRUCT(term) && FOLLOW(UNTAGGED(term))==(LONG)objectRef);
}

int PisCompound(CELL term){
  DEREF(term);
  return (ISLIST(term) || ISSTRUCT(term));
}
*/
/* Prolog to C */
/*
int PvalueOfInt(CELL term){
  DEREF(term);
  return INTVAL(term);
}
*/
double PvalueOfReal(term)
CELL term;
{
  double floatval();
  DEREF(term);
  return floatval(term);
}


int PvalueOfAddr(term)
CELL term;
{
  register unsigned int w1,w2;
  DEREF(term);
  UNTAG(term);
  w1 = INTVAL(*((LONG_PTR)term+1));
  w2 = INTVAL(*((LONG_PTR)term+2));
  return (w2<<16 | w1);
}

char* PnumberToStr(term)
CELL term;
{
  printf("PnumberToStr used\n");
  return " ";
}

/* Manipulate compound terms */

int ParityOfCompound(term)
CELL term;
{
  DEREF(term);
  return GET_ARITY(GET_SYM_REC(term));
}

CELL PargOfCompound(term, n)
CELL term;
int n;
{
  return get_ARG(n+1,term);
}

/* Return functors as strings*/
char *PascOfFunc(term)
CELL term;
{
  return get_NAME(term);
}

char *PascOfAtom(term)
CELL term;
{
  return get_NAME(term);
}

char *PnameToAsc(term)
CELL term;
{
  printf("PnameToAsc used \n");
}

/* UNIFY */
int Punify(term1,term2)
CELL term1;
CELL term2;
{
  return unify(term1,term2);
}

int PuInt(term,i)
CELL term;
int i;
{
  TERM op;
  
  op = MAKEINT(i);
  return unify(term,op);
}

int PuAtom(term,str)
CELL term;
char *str;
{
  TERM op = build_ATOM(str);
  return unify(term,op);
}

int PuStr(term,str)
CELL term;
char *str;
{
  TERM op = build_ATOM(str);
  return b_NAME0_cf(op,term);
}

int PuAddr(term,a)
CELL term;
void *a;
{
  register LONG temp;

  temp = ADDTAG(hreg,STR);
  NEW_HEAP_NODE((LONG)objectRef);  /* '$addr'(int1,int2) */
  NEW_HEAP_NODE(MAKEINT(((unsigned long int)a<<16)>>16));     
  NEW_HEAP_NODE(MAKEINT(((unsigned long int)a >>16)));
  return unify(term,temp);
}

int atomToTerm(){
  TERM op1,op2;
  op1 = ARG(1,2);
  op2 = ARG(2,2);
  DEREF(op1);
  DEREF(op2);
  if (!ISATOM(op1)){exception=illegal_arguments; return  -1;};
  PnameToTerm(get_NAME(op1),op2);
  return 1;
}

int PnameToTerm(str, term)
char *str;
CELL term;
{
  int i;
  TERM Read;

  string_in = str;
  Read=build_STRUCTURE("read_vars",2);
  build_ARG(term);
  build_VAR();
  PcallF(Read);
  string_in = NULL;
   return 1;
}

int PuReal(term,d)
CELL term;
double d;
{
  return unify(term,encodefloat(d));
}

/* Create terms */
CELL PnewVar(){
  return build_VAR();
}

CELL Patom(str)
char *str;
{
  return build_ATOM(str);
}

CELL create_FUNCTOR(name, arity)
     char * name;
     int arity;
{
  TERM op;
  int i;

  op = build_STRUCTURE(name,arity);
  for (i=1;i<=arity;i++){
    build_VAR();
  }
  return op;
}

CELL Pfunctorn(name, arity)
CELL name;
int arity;
{
  TERM op;
  char *str;
  int i;

  str = get_NAME(name);
  op = build_STRUCTURE(str,arity);
  for (i=1;i<=arity;i++){
    build_VAR();
  }
  return op;
}

/* Manipulate lists */
/*
int PisList(CELL term){
  DEREF(term);
  return ISLIST(term);
}

int PisEndList(CELL term){
  DEREF(term);
  return ISNIL(term);
}
*/
CELL get_CAR(term)
     CELL term;
{
  DEREF(term);
  return *((LONG_PTR)UNTAGGED(term));
}

CELL PlistCar(term)
CELL term;
{
  DEREF(term);
  return *((LONG_PTR)UNTAGGED(term));
}

CELL get_CDR(term)
     CELL term;
{
  DEREF(term);
  return *((LONG_PTR)UNTAGGED(term)+1);
}

CELL PlistCdr(term)
CELL term;
{
  DEREF(term);
  return *((LONG_PTR)UNTAGGED(term)+1);
}

int PlistLength(term)
CELL term;
{
  register int len=0;
  DEREF(term);
  while (!PisEndList(term)){
    len++;
    term = PlistCdr(term);
    DEREF(term);
  }
  return len;
}

CELL Plistn(length)
int length;
{
  TERM lst,tail;
  
  if (length<=0) return build_NIL();
  else {
    length--;
    lst = build_LIST();
    build_VAR();
    tail = build_VAR();
    while (length>0){
      FOLLOW(tail) = build_LIST();
      build_VAR();
      tail = build_VAR();
      length--;
    }
    FOLLOW(tail) = build_NIL();
  }
  return lst;
}

CELL create_LIST(length)
int length;
{
  return Plistn(length);
}
    
/* Invoke a predicate */
TERM kapi_copy_term(t)
TERM t;
{
  return copy_term(t);
}

int PcallF(term)
CELL term;
{
  return call_bprolog_term(term);
}

int call_bprolog_term(term)
CELL term;
{
  int res;
/*
  printf("call_bprolog_term ");write_term(term);printf("\n");
*/
  DEREF(term);
  
  toam0.entry++;
  toam0.hreg = hreg;
  local_top = arreg-255;

  breg  = local_top;
  arreg = breg;
  local_top = arreg-9;
  AR_AR(arreg) = (LONG)arreg;
  AR_CPS(arreg) = (LONG)addr_halt;    /* CPS : return 1 on final success */
  AR_TOP(arreg) = (LONG)local_top;  /* TOP */
  AR_BTM(arreg) = ADDTAG((LONG)arreg,AR_IS_NONDET); /* BTM */
  AR_B(arreg) = (LONG)breg;          /* B    */
  AR_CPF(arreg)  = (LONG)addr_halt0;   /* CPF : return 0 on final failure */
  AR_H(arreg) = (LONG)hreg;          /* H    */
  AR_T(arreg) = (LONG)treg;          /* T    */
  AR_SF(arreg) = (LONG)sfreg;          /* T    */
  hbreg = hreg;

  *local_top-- = term;
  *local_top = (LONG)arreg;
  arreg = local_top;
  AR_CPS(arreg) = (LONG)addr_halt;
  AR_TOP(arreg) = (LONG)local_top - 3;
  AR_BTM(arreg) = ADDTAG((LONG)arreg,AR_IS_NONDET); /* BTM */

  inst_begin = (LONG_PTR) GET_EP(enter_dyn_call);
  init_signals_internal(); 
  res = toam(inst_begin,local_top,arreg,breg,hbreg,hreg,treg,sfreg);

  if (res==1)  compactHeap(term); 

  toam0.entry--;
  if (toam0.entry<1){ /* only one thread remaining, reclaim the stack areas */
    breg = toam0.breg;
    arreg = toam0.arreg;
    treg = toam0.treg;
    sfreg = toam0.sfreg;
    local_top = toam0.local_top;
    tareg = toam0.tareg;
  }
  return res;
}

int PcallX(command)
TERM command;
{
  return PcallF(command);
}

int call_bprolog_string(cmd)
     char *cmd;
{
  TERM Pcmd;
  Pcmd = build_VAR();
  PnameToTerm(cmd,Pcmd);
  return PcallF(Pcmd);
}

int PexecP(cmd)
char *cmd;
{
  return call_bprolog_string(cmd);
}

int Pexecute(cmd)
char *cmd;
{
  return PexecP(cmd);
}

/* Initialize B-Prolog */
PinitP(argc, argv)
int argc;
char *argv[];
{
  initialize_bprolog(argc,argv);
}

void jni_interface(){
}

void plc_sup(){
}

/* Mis */
char *PkanjiOfAtom(str)
char *str;
{
  return NULL;
}

compactHeap(term)
TERM term;
{
  LONG termCp,varno=1,maxVarNo=0;
  LONG_PTR varVector=toam0.local_top;
 
  parea_top = (LONG_PTR)UNTAGGED(curr_fence+4000);
  termCp = numberVarCopyToParea(term,nil_sym,&varno);
  hreg = toam0.hreg;
  copyBackToHeap(term,termCp,varVector,&maxVarNo);
  toam0.hreg = hreg;
}

copyBackToHeap(term,termCp,varVector,maxVarNo)
TERM term,termCp;
LONG_PTR varVector;
int *maxVarNo;
{
  LONG_PTR ptr1,ptr2;
  DEREF(term);
  if (ISLIST(term)){
    ptr1 = (LONG_PTR)UNTAGGED(term);
    ptr2 = (LONG_PTR)UNTAGGED(termCp);
    if (ptr1<toam0.hreg){
      FOLLOW(ptr1) = copyBackToHeap(FOLLOW(ptr1),FOLLOW(ptr2),varVector,maxVarNo);
      FOLLOW(ptr1+1) = copyBackToHeap(FOLLOW(ptr1+1),FOLLOW(ptr2+1),varVector,maxVarNo);
    } else {
      ptr1 = hreg; hreg += 2;
      FOLLOW(ptr1) = unnumberVarTerm(FOLLOW(ptr2),varVector,maxVarNo);
      FOLLOW(ptr1+1) = unnumberVarTerm(FOLLOW(ptr2+1),varVector,maxVarNo);
      term = ADDTAG(ptr1,LST);
    }
    return term;
  } else if (ISSTRUCT(term)){
    int arity,i;
    ptr1 = (LONG_PTR)UNTAGGED(term);
    ptr2 = (LONG_PTR)UNTAGGED(termCp);
    arity = GET_ARITY((SYM_REC_PTR)FOLLOW(ptr1));
    if (ptr1<=toam0.hreg){
      for (i=1;i<=arity;i++){
	FOLLOW(ptr1+i) = copyBackToHeap(FOLLOW(ptr1+i),FOLLOW(ptr2+i),varVector,maxVarNo);
      }
    } else {
      ptr1 = hreg; hreg += arity+1;
      FOLLOW(ptr1) = FOLLOW(ptr2);
      for (i=1;i<=arity;i++){
	FOLLOW(ptr1+i) = unnumberVarTerm(FOLLOW(ptr2+i),varVector,maxVarNo);
      }
      term = ADDTAG(ptr1,STR);
    }
    return term;
  }
  return unnumberVarTerm(termCp,varVector,maxVarNo);
}
  

  





