/* loader.c */

#include "basic.h"
#include "inst.h"
#include "term.h"
#include "macro.h"

#define MAXSYMS 5023

#define READ_DATA(x,y)  (y - fread(x, sizeof(*x), y, fp))
#define RELOC_ADDR(offset) ((LONG_PTR)curr_fence + offset)
#define BUILTIN 1

/* macro for creating hash chains */
#define GEN_COND_JUMP(opcode,arg2,arg3,ep) \
   *ep++ = opcode;                        \
   *ep++ = arg2;                          \
   *ep++ = (LONG)arg3; 

#define GEN_COND_JUMP2(opcode,arg2,ep) \
   *ep++ = opcode;                        \
   *ep++ = (LONG)arg2;

#define GEN_JUMP(opcode,arg1,ep) \
  *ep++ = opcode;                \
  *ep++ = (LONG)arg1;
      
#define CHECK_PCODE(ptr,size) \
   if (ptr + 1000 + size >= max_fence){ \
     printf("code_area_overflow\n");   \
     exit(1); \
   }                                     

#define BB4(bptr) (LONG)((((((*(BYTE_PTR)bptr << 8) | *((BYTE_PTR)bptr+1)) << 8) | *((BYTE_PTR)bptr+2)) << 8) | *((BYTE_PTR)bptr+3))

#define X_L { \
  temp = BB4(inst_addr); \
  *inst_addr = (LONG)(rreg + temp); \
  inst_addr++; \
}

#define Y_L {   \
  temp = BB4(inst_addr); \
  *inst_addr = temp; \
  inst_addr++; \
}

/* 
#define VX 0
#define UX 1
*/
#define Z4_L {Z_L; Z_L; Z_L; Z_L;}
#define Z5_L {Z_L; Z_L; Z_L; Z_L;Z_L;}

#define Z_L {   \
  temp = BB4(inst_addr); \
  switch (temp & 0x7) { \
    case 0: *inst_addr = (LONG)(rreg+(temp>>3)); break; \
    case 1: *inst_addr = ((LONG)(rreg+(temp>>3)) | 0x1); break; \
    case 6: \
    case 2: *inst_addr = temp | INT_TAG; break; \
    default: *inst_addr = temp; \
    } \
    inst_addr++; \
}

#define N_L {   \
  temp = BB4(inst_addr); \
  *inst_addr = temp; \
  inst_addr++; \
}

#define A_L { \
    temp = BB4(inst_addr); \
    *(LONG_PTR)inst_addr = (LONG)RELOC_ADDR(temp); \
    inst_addr++; \
    }

#define I_L { \
    temp = BB4(inst_addr); \
    *(LONG_PTR)inst_addr = MAKEINT(temp); \
    inst_addr++; \
    }

#define S_L { \
    temp = BB4(inst_addr); \
   *inst_addr = (LONG)reloc_table[temp]; \
    inst_addr++; \
  }

#define C_L { \
    temp = BB4(inst_addr); \
    *inst_addr = ADDTAG(reloc_table[temp],ATM); \
    inst_addr++; \
  }

static LONG_PTR inst_addr;
static LONG_PTR hptr;
static SYM_REC_PTR reloc_table[MAXSYMS];
static LONG_PTR last_text;

LONG   eof_flag;
LONG   psc_bytes, text_bytes, index_bytes;
BYTE   magic;

struct hrec {
  LONG     l;
  LONG_PTR link;
};

#define INDEXTABSIZE 1024

struct hrec indextab[INDEXTABSIZE];

int    (*load_routine[80])();

int dynload = 0;
LONG temp;

static FILE *fp;

/****************************************************************************/
/*                                                                          */
/* fix_bb4: fixes the byte-backwards problem.  It is passed a pointer to a  */
/* sequence of 4 bytes read in from a file as bytes.  It then converts      */
/* those bytes to represent a number.  This code works for any machine, and */
/* makes the byte-code machine independent.                                 */
/*                                                                          */
/****************************************************************************/
fix_bb4(lptr)
BYTE_PTR lptr;
{
   LONG_PTR numptr = (LONG_PTR)lptr;

   *numptr = (((((*lptr << 8) | *(lptr+1)) << 8) | *(lptr+2)) << 8) | *(lptr+3);
}

/****************************************************************************/
/*                                                                          */
/*    Load the file into permanent space, starting from "curr_fence".       */
/* Data segment first (mixed psc entries and name strings), then text       */
/* segment, ended with a virtual instruction "endfile <pointer>" where the  */
/* pointer is a pointer to the next text segment (of another byte code      */
/* file).                                                                   */
/*                                                                          */
/****************************************************************************/
loader(file,file_type)
CHAR_PTR file;
int file_type;
{
  LONG     err_msg;

#ifdef WIN32
   fp = fopen(file, "rb");
#else
   fp = fopen(file, "r");
#endif
   if (fp == NULL)
      return 10;
/*  printf("\n     ...... loading file %s\n", file); */

   while ((eof_flag = READ_DATA(&magic, 1)) == 0) {
     exception = illegal_file;
     if (magic!=17) return -1;
     
     if (eof_flag = READ_DATA(&magic, 1))
       return 1;
     if (magic!=18) return -1;

     if (eof_flag = READ_DATA(&magic, 1))
       return 1;
     if (magic!=19) return -1;

     if (eof_flag = READ_DATA(&magic, 1))
       return 1;
     if (magic!=3) return -1;
     
     if (eof_flag = READ_DATA(&psc_bytes, 1))
       return 1;
     fix_bb4(&psc_bytes);
     if (eof_flag = READ_DATA(&text_bytes, 1))
       return 1;
     fix_bb4(&text_bytes);
     if (eof_flag = READ_DATA(&index_bytes, 1))
       return 1;
     fix_bb4(&index_bytes);
     
     err_msg = load_syms(file_type);
     if (err_msg != 0) {
       close(fp);
       return -1;  /* eventually upper level routines will determine */
     } 
     
      err_msg = load_text();
      if (err_msg != 0) {
	printf("error %d loading file %s: bad text segment\n", err_msg, file);
	return -1;  /* eventually upper level routines will determine */
      }

      err_msg = load_index();
      if (err_msg != 0) {
	printf("error %d in (index) loading file %s: bad index segment\n", err_msg, file);
	return -1;  /*eventually upper level routines will determine */
      }
      if (eof_flag = READ_DATA(inst_addr, 2))
	return eof_flag;
     fix_bb4(inst_addr);
     if (*inst_addr != endfile)
       *inst_addr = endfile;
     inst_addr++;                 /* skip opcode */
     *inst_addr = 0;              /* force 0 address (LONG) */
     last_text  = (LONG_PTR)inst_addr;
     inst_addr++;
     curr_fence = (CHAR_PTR)inst_addr;
   }
  fclose(fp);
  return 0;
}  /* end of loader */

/*************************************************************************
*                                                                        *
* Load_syms is a function which loads a symbol table given in a byte     *
* code file into an appropriate format in the pcs table.  As part of     *
* its function it resolves entry points for byte code intructions (call  *
* to relloc_addr), and maintains a tableau so that instructions          *
* with indices into the pcs table may have those indices resloved before *
* loading them in the intruction array (byte code program space).  The   *
* intructions are loaded by a separate function.                         *
* The function returns a short integer which is an error code.  Relevant *
* error codes are listed below.                                          *
*                                                                        *
*      1: in load_syms: incomplete or missing psc count                  *
*      2: in load_syms: incomplete or missing ep in record dec           *
*      3: in load_syms: incomplete or missing arity in record dec        *
*      4: in load_syms: incomplete or missing length in record dec       *
*      5: in load_syms: incomplete or missing name in record dec         *
*      8: can not open file                                              *
*                                                                        *
*************************************************************************/
load_syms(file_type)
int file_type;
{
   CHAR     name[256];
   LONG     ep_offset;
   LONG     i = 0, j, count = 0;
   CHAR     temp_name[256];
   BYTE     temp_len;
   BYTE     temp_arity;
   SYM_REC_PTR insert();

/*   printf("=> load_syms \n"); */
   while (count < psc_bytes && eof_flag == 0) {
      if (eof_flag = READ_DATA(&ep_offset, 1))
	 return 2;
      fix_bb4(&ep_offset);
      if (eof_flag = READ_DATA(&temp_arity, 1))
	 return 3;
      if (eof_flag = READ_DATA(&temp_len, 1))
	 return 4;
      if (eof_flag = READ_DATA(name, temp_len))
	 return 5;
/*
      if (dynload == 1 ) {
	strncpy(temp_name,name,temp_len);
	printf("load (%s,%i)\n",temp_name,temp_len); 
      }
*/
      reloc_table[i] = insert(name, temp_len, temp_arity);
      if (set_temp_ep(reloc_table[i], ep_offset*4,file_type)!=0)
	return -1;
      count += temp_len + 6;
      i++;
    }
   
   if (count!=psc_bytes)
     printf("error in load_syms count=%i,psc_bytes=%i\n",count,psc_bytes);

   ALIGN(CHAR_PTR,curr_fence);
   
   for (j = 0; j < i; j++)
      set_real_ep(reloc_table[j], (LONG_PTR)curr_fence);
   return 0;
}  /* end of load_syms */

/************************************************************************
*                                                                       *
* Load_text loads the byte code instruction from a byte code file to    *
* the byte code program space.  References to indices to the pcs table  *
* are resolved with the use of the macro st_index.  New index relies    *
* on the symbol table array which is assigned values by load_syms.      *
* The routine assumes the current length (8/18/84) of byte code         *
* intructions when reading from the byte code file.                     *
* The function returns a short integer which is a error code.  Relevant *
* error codes are listed below.                                         *
*                                                                       *
*      6: in load_text: byte code Operands are non-existent             *
*      7: in load_text: Illegal instruction from PIL file               *
*      8: can not open file                                             *
*                                                                       *
************************************************************************/
load_text()
{
   LONG current_opcode = 0;
   LONG count = 0;
   LONG opval;
   
/*   printf("==> load_text \n");*/

   /* set text segments chain */
   inst_addr = (LONG_PTR)curr_fence;

   if (inst_begin == 0) 
      inst_begin = (LONG_PTR)inst_addr;
   else 
      *last_text = (LONG)inst_addr;

   CHECK_PCODE(curr_fence,text_bytes);

   while (count < text_bytes &&
	  (eof_flag = READ_DATA(&current_opcode, 1)) == 0) {
     fix_bb4(&current_opcode);
     *inst_addr++ = current_opcode;
/*
     printf("opcode=%d %s len=%d\n",current_opcode,xxopcode[current_opcode].name,xxopcode[current_opcode].size); 
*/
      (*load_routine[xxopcode[current_opcode].type])();
      count += xxopcode[current_opcode].size;
    }
/*      printf("count=%i,text_bytes=%i\n",count,text_bytes);*/
   if (count != text_bytes)  {
     return 9;               /* missing instructions */
   }
   return 0;
}  /* end of load_text */


load_index()
{
   LONG     hash_inst_addr,hash_reg,alt, clause_no, temp_len;
   LONG     count = 0;
   LONG_PTR gen_index();

   CHECK_PCODE((CHAR_PTR)inst_addr,index_bytes);
/*   printf("==> load_index \n"); */
   while (count < index_bytes && eof_flag == 0) {
      if (eof_flag = READ_DATA(&hash_inst_addr, 1))
	 return 10;
      fix_bb4(&hash_inst_addr);
      hash_inst_addr = (LONG)RELOC_ADDR(hash_inst_addr);
      if (eof_flag = READ_DATA(&hash_reg, 1))
	 return 10;
      fix_bb4(&hash_reg);
      if (eof_flag = READ_DATA(&clause_no, 1))
	 return 1;
      fix_bb4(&clause_no);
      if (eof_flag = READ_DATA(&alt, 1))
	 return 1;
      fix_bb4(&alt);
      alt = (LONG) RELOC_ADDR(alt);
      if (eof_flag = get_index_tab(clause_no, &temp_len))
	 return eof_flag;
      inst_addr = gen_index(hash_inst_addr,hash_reg,clause_no,alt);
      count += (16 + temp_len);
   }
   return 0;
}


get_index_tab(clause_no,lenptr)
LONG     clause_no;
LONG_PTR lenptr;
{
   LONG     hashval, size, j;
   LONG     count = 0;
   BYTE     type;
   LONG     val,ttype;
   LONG_PTR label;

   hptr = hreg;
   size = hsize(clause_no);
   for (j = 0; j < size; j++) {
      indextab[j].l = 0;
      indextab[j].link = (LONG_PTR)(&(indextab[j].link));
   }
   for (j = 0; j < clause_no; j++) {
      if (eof_flag = READ_DATA(&type, 1))
	 return 11;
      switch (type) {
        case 'i': if (eof_flag = READ_DATA(&val, 1))
	  return 12;
	  fix_bb4(&val);
	  val = MAKEINT(val);
	  count += 9;
	  ttype=0;
	  break;
        case 's': if (eof_flag = READ_DATA(&val, 1))
	  return 12;
	  fix_bb4(&val);
	  count += 9;
	  val = (LONG)reloc_table[val];
	  if (val==(LONG)list_psc) {
	    ttype=1;
	  }
	  else {
	    ttype=2;
	  }
	  break;
        case 'c': if (eof_flag = READ_DATA(&val, 1))
	  return 12;
	  fix_bb4(&val);
	  count += 9;
	  val = (LONG)reloc_table[val];
	  if (val==UNTAGGED(nil_sym))
	    ttype=3;
	  else
	    ttype=4;
	  val = ADDTAG(val,ATM);
	  break;
      }
      if (eof_flag = READ_DATA(&label, 1))
	 return 13;
      fix_bb4(&label);
      label = RELOC_ADDR((LONG)label);
      hashval = IHASH(val, size);
      inserth(ttype,val,label, &indextab[hashval]);
   }
   *lenptr = count;
   return 0;
}  /* end of get_index_tab */


LONG_PTR gen_index(hash_inst_addr,hash_reg,clause_no,alt)
LONG        hash_inst_addr,hash_reg,clause_no,alt;
{
   LONG_PTR ep1, ep2;
   LONG     j, size;
   LONG  ttype, val;
   LONG_PTR label,jumpaddr,temp;

   size = hsize(clause_no);
   ep1  = inst_addr;
   *ep1++ = tabsize;
   *ep1++ = size;
   ep2 = inst_addr + 2 + size;
   temp = (LONG_PTR)(hash_inst_addr) + 2;  /* fill the slot in the hash inst*/
   *temp++ = size;
   *temp = (LONG)ep1;
   for (j = 0; j < size; j++) {
     if (indextab[j].l == 0) {
       *ep1++ = alt;
     } 
     else  
       if (indextab[j].l == 1) {
         *ep1++ = (LONG)*((LONG_PTR)(indextab[j].link) + 2);
       } 
     else 
       {            /* create conditional jump insts */
	 *ep1++ = (LONG)ep2;
	 temp = (LONG_PTR)(indextab[j].link);
	 while (*temp != (LONG)temp) 
	   {
	     ttype = *temp++;
	     val   = *temp++;
	     label = (LONG_PTR)*temp++;
	     switch (ttype) 
	       {
	       case 0 :
		 jumpaddr = ep2+5;
		 GEN_COND_JUMP(hash_jmpn_int,val,jumpaddr,ep2);
		 GEN_JUMP(jmp,label,ep2);
		 break;
	       case 1 :
		 jumpaddr = ep2+4;
		 GEN_COND_JUMP2(hash_jmpn_list,jumpaddr,ep2);
		 GEN_JUMP(jmp,label,ep2);
		 break;
	       case 2 :
		 jumpaddr = ep2+5;
		 GEN_COND_JUMP(hash_jmpn_struct,val,jumpaddr,ep2);
		 GEN_JUMP(jmp,label,ep2);
		 break;
	       case 3 :
		 jumpaddr = ep2+4;
		 GEN_COND_JUMP2(hash_jmpn_nil,jumpaddr,ep2);
		 GEN_JUMP(jmp,label,ep2);
		 break;
	       case 4 :
		 jumpaddr = ep2+5;
		 GEN_COND_JUMP(hash_jmpn_atom,val,jumpaddr,ep2);
		 GEN_JUMP(jmp,label,ep2);
		 break;
	       }
	     temp = (LONG_PTR)*temp;
	     CHECK_PCODE((CHAR_PTR)ep2,0);
	   }
	 GEN_JUMP(jmp,alt,ep2);
       }
   }
   return ep2;
 }

inserth(ttype,val,label, bucket)
LONG ttype, val;
LONG_PTR label;
struct   hrec *bucket;
{
  register LONG_PTR temp;

   bucket->l++;
   temp = (LONG_PTR)&(bucket->link);
   if (bucket->l > 1) {
      temp = (LONG_PTR)*temp;
      while ((LONG_PTR)*temp != temp) 
	{ 
	  
/*	  printf("type = %i \n", (LONG)*temp++);
	  printf("val  = %i \n", (LONG)*temp++);
	  printf("label= %x \n", (LONG)*temp++);*/
	  temp = (LONG_PTR)*(temp+3);
	  }
    }
   *temp = (LONG)hptr;
   *hptr++ = ttype;
   *hptr++ = val;
   *hptr++ = (LONG)label;
   *hptr++ = *temp + 12;     /* *hptr++ = (LONG)hptr; */
 }


hsize(numentry)
LONG numentry;
{
  register LONG i, j, temp;

   temp = numentry + 1;

hashsod:
   j = temp / 2 + 1;
   for (i = 2; i <= j; i++) {
      if (i != temp && (temp % i) == 0) {
         temp++;
         goto hashsod;
      }
   }
   return temp;
}


l_A(){
  register LONG temp;
  
  if (READ_DATA(inst_addr, 1))
    quit("incomplete instruction\n"); 
  else {
    A_L;
  }
}

l_AA(){
  register LONG temp;
  
  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    A_L;
    A_L;
  }
}

l_C(){
  register LONG temp;

  if (READ_DATA(inst_addr, 1))
    quit("incomplete instruction\n"); 
  else {
    C_L;
  }
}
  
l_CA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    C_L;
    A_L;
  }
}

l_E(){
}

l_I(){
  register LONG temp;

  if (READ_DATA(inst_addr, 1))
    quit("incomplete instruction\n"); 
  else {
    I_L;
  }
}
  
l_IA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    I_L;
    A_L;
  }
}
  
l_III(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    I_L;
    I_L;
    I_L;
  }
}

l_N(){
  register LONG temp;

  if (READ_DATA(inst_addr, 1))
    quit("incomplete instruction\n"); 
  else {
    N_L;
  }
}

l_NS(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    N_L;
    S_L;
  }
}

l_NAAAS(){
  register LONG temp;

  if (READ_DATA(inst_addr, 5))
    quit("incomplete instruction\n"); 
  else {
    N_L;
    A_L;
    A_L;
    A_L;
    S_L;
  }
}

l_NA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    N_L;
    A_L;
  }
}

l_NNA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    N_L;
    N_L;
    A_L;
  }
}

l_NNSA(){
  register LONG temp;

  use_tabling = 1; /* loading table_allocate */
  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    N_L;
    N_L;
    S_L;
    A_L;
  }
}

l_NN(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    N_L;
    N_L;
  }
}

l_NZ(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    N_L;
    Z_L;
  }
}

l_NZZ(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    N_L;
    Z_L;
    Z_L;
  }
}
  
l_NAZ(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    N_L;
    A_L;
    Z_L;
  }
}

l_NAZZ(){
  register LONG temp;

  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    N_L;
    A_L;
    Z_L;
    Z_L;
  }
}

l_NAZZZ(){
  register LONG temp;

  if (READ_DATA(inst_addr, 5))
    quit("incomplete instruction\n"); 
  else {
    N_L;
    A_L;
    Z_L;
    Z_L;
    Z_L;
  }
}

l_NAZZZZ(){
  register LONG temp;

  if (READ_DATA(inst_addr, 6))
    quit("incomplete instruction\n"); 
  else {
    N_L;
    A_L;
    Z_L;
    Z_L;
    Z_L;
    Z_L;
  }
}

l_Z6(){
  register LONG temp;

  if (READ_DATA(inst_addr, 6))
    quit("incomplete instruction\n"); 
  else {
    Z_L;
    Z_L;
    Z_L;
    Z_L;
    Z_L;
    Z_L;
  }
}
  

l_S(){
  register LONG temp;

  if (READ_DATA(inst_addr, 1))
    quit("incomplete instruction\n"); 
  else {
    S_L;
  }
}

l_SA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    S_L;
    A_L;
  }
}

l_X(){
  register LONG temp;

  if (READ_DATA(inst_addr, 1))
    quit("incomplete instruction\n"); 
  else {
    X_L;
  }
}

l_XA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    A_L;
  }
}

l_XAA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    A_L;
    A_L;
  }
}

l_XAAA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    A_L;
    A_L;
    A_L;
  }
}

 
l_XC(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    C_L;
  }
}

l_XCAA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    C_L;
    A_L;
    A_L;
  }
}

l_XI(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    I_L;
  }
}
  
l_XIAA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    I_L;
    A_L;
    A_L;
  }
}

l_XNA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    N_L;
    A_L;
  }
}
  
l_XNAA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    N_L;
    A_L;
    A_L;
  }
}

l_XS(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    S_L;
  }
}

l_XSAA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    S_L;
    A_L;
    A_L;
  }
}

l_XX(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    X_L;
  }
}

l_XXX(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    X_L;
    X_L;
  }
}

l_XXY(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    X_L;
    Y_L;
  }
}

l_XY(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    Y_L;
  }
}

l_XYA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    Y_L;
    A_L;
  }
}
l_XYX(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    Y_L;
    X_L;
  }
}

l_XYY(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    X_L;
    Y_L;
    Y_L;
  }
}

l_Y(){
  register LONG temp;

  if (READ_DATA(inst_addr, 1))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
  }
}

l_YA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    A_L;
  }
}

l_YAA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    A_L;
    A_L;
  }
}

l_AAS(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    A_L;
    A_L;
    S_L;
  }
}

l_AASA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    A_L;
    A_L;
    S_L;
    A_L;
  }
}

l_YAAS(){
  register LONG temp;

  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    A_L;
    A_L;
    S_L;
  }
}

l_YAAA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    A_L;
    A_L;
    A_L;
  }
}

l_YC(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    C_L;
  }
}

l_YCA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    C_L;
    A_L;
  }
}
l_YCAA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    C_L;
    A_L;
    A_L;
  }
}

l_YI(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    I_L;
  }
}

l_YIA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    I_L;
    A_L;
  }
}
l_YIAA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    I_L;
    A_L;
    A_L;
  }
}

l_YNA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    N_L;
    A_L;
  }
}

l_YNAA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    N_L;
    A_L;
    A_L;
  }
}

l_YS(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    S_L;
  }
}

l_YSA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    S_L;
    A_L;
  }
}

l_YSAA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    S_L;
    A_L;
    A_L;
  }
}

l_YX(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    X_L;
  }
}

l_YXX(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    X_L;
    X_L;
  }
}

l_YXY(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    X_L;
    Y_L;
  }
}

l_YY(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    Y_L;
  }
}

l_YYA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    Y_L;
    A_L;
  }
}

l_YYX(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    Y_L;
    X_L;
  }
}

l_YYY(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    Y_L;
    Y_L;
  }
}

l_Z(){
  register LONG temp;

  if (READ_DATA(inst_addr, 1))
    quit("incomplete instruction\n"); 
  else {
    Z_L;
  }
}

l_ZZ(){
  register LONG temp;

  if (READ_DATA(inst_addr, 2))
    quit("incomplete instruction\n"); 
  else {
    Z_L;
    Z_L;
  }
}

l_Z4(){
  register LONG temp;

  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    Z4_L;
  }
}

l_Z5(){
  register LONG temp;

  if (READ_DATA(inst_addr, 5))
    quit("incomplete instruction\n"); 
  else {
    Z5_L;
  }
}

l_ZZA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    Z_L;
    Z_L;
    A_L;
  }
}
  
l_ZZZ(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    Z_L;
    Z_L;
    Z_L;
  }
}

l_AAA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    A_L;
    A_L;
    A_L;
  }
}

l_NAA(){
  register LONG temp;

  if (READ_DATA(inst_addr, 3))
    quit("incomplete instruction\n"); 
  else {
    N_L;
    A_L;
    A_L;
  }
}

l_YYYY(){
  register LONG temp;

  if (READ_DATA(inst_addr, 4))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    Y_L;
    Y_L;
    Y_L;
  }
}

l_YYYYYY(){
  register LONG temp;

  if (READ_DATA(inst_addr, 6))
    quit("incomplete instruction\n"); 
  else {
    Y_L;
    Y_L;
    Y_L;
    Y_L;
    Y_L;
    Y_L;
  }
}
  
init_load_routine()
{
   load_routine[ A ]= l_A;
   load_routine[ AAA ] = l_AAA;
   load_routine[ C ]= l_C;
   load_routine[ CA ]= l_CA;
   load_routine[ E ]= l_E;
   load_routine[ I ]= l_I;
   load_routine[ IA ]= l_IA;
   load_routine[ III ]= l_III;
   load_routine[ M ]= l_Y;
   load_routine[ MC ]= l_YC;
   load_routine[ N ]= l_N;
   load_routine[ NA ] = l_NA;
   load_routine[ NAA ] = l_NAA;
   load_routine[ NAZ ] = l_NAZ;
   load_routine[ NAZZ ] = l_NAZZ;
   load_routine[ NAZZZ ] = l_NAZZZ;
   load_routine[ NAZZZZ ] = l_NAZZZZ;
   load_routine[ NN ] = l_NN;
   load_routine[ NS ] = l_NS;
   load_routine[ NZZ ]= l_NZZ;
   load_routine[ R ]= l_X;
   load_routine[ RC ]= l_XC;
   load_routine[ S ]= l_S;
   load_routine[ SA ]= l_SA;
   load_routine[ XA ]= l_XA;
   load_routine[ XAA ]= l_XAA;
   load_routine[ XAAA ]= l_XAAA;
   load_routine[ XCAA ]= l_XCAA;
   load_routine[ XI ]= l_XI;
   load_routine[ XIAA ]= l_XIAA;
   load_routine[ XNA ]= l_XNA;
   load_routine[ XNAA ]= l_XNAA;
   load_routine[ XS ]= l_XS;
   load_routine[ XSAA ]= l_XSAA;
   load_routine[ XX ]= l_XX;
   load_routine[ XXX ]= l_XXX;
   load_routine[ XXY ]= l_XXY;
   load_routine[ XY ]= l_XY;
   load_routine[ XYA ]= l_XYA;
   load_routine[ XYX ]= l_XYX;
   load_routine[ XYY ]= l_XYY;
   load_routine[ YA ]= l_YA;
   load_routine[ YAA ]= l_YAA;
   load_routine[ YAAA ]= l_YAAA;
   load_routine[ YCA ]= l_YCA;
   load_routine[ YCAA ]= l_YCAA;
   load_routine[ YI ]= l_YI;
   load_routine[ YIA ]= l_YIA;
   load_routine[ YIAA ]= l_YIAA;
   load_routine[ YNA ]= l_YNA;
   load_routine[ YNAA ]= l_YNAA;
   load_routine[ YS ]= l_YS;
   load_routine[ YSA ]= l_YSA;
   load_routine[ YSAA ]= l_YSAA;
   load_routine[ YX ]= l_YX;
   load_routine[ YXX ]= l_YXX;
   load_routine[ YXY ]= l_YXY;
   load_routine[ YY ]= l_YY;
   load_routine[ YYA ]= l_YYA;
   load_routine[ YYX ]= l_YYX;
   load_routine[ YYY ]= l_YYY;
   load_routine[ YYYY ] = l_YYYY;
   load_routine[ YYYYYY ] = l_YYYYYY;
   load_routine[ Z ]= l_Z;
   load_routine[ ZZ ]= l_ZZ;
   load_routine[ ZZA ]= l_ZZA;
   load_routine[ ZZZ ]= l_ZZZ;
   load_routine[ AA ] = l_AA;
   load_routine[ NAAAS ] = l_NAAAS;
   load_routine[ YAAS ] = l_YAAS;
   load_routine[ AAS ] = l_AAS;
   load_routine[ NNA] = l_NNA;
   load_routine[ Z4 ] = l_Z4;
   load_routine[ Z5 ] = l_Z5;
   load_routine[ AASA] = l_AASA;
   load_routine[ NNSA] = l_NNSA;
   load_routine[ Z6 ] = l_Z6;
 }

dyn_loader(sym_ptr,file_type)
SYM_REC_PTR sym_ptr;
int file_type;
{
   extern   CHAR_PTR getenv();
   CHAR     s[256], s1[256], s3[256];
   CHAR_PTR s2;
   LONG     i, returnval;
   
   dynload = 1;
   namestring(sym_ptr, s1);
   if (*s1 == '/' || *s1 == '.') 
     return loader(s1,file_type);
   else if (file_type != BUILTIN){
     if (*s1 == '~'){
       s2 = getenv("HOME");
       if (s2==NULL){
	 fputs("the environment variable HOME is not set\n",stderr);
       }
       if (*(s1+1) == '/'){
	 scat(s2,&s1[1],s3);
	 return loader(s3,file_type);
       }
       else {
	 i = strlen(s2);
	 i--;
	 while  (*(s2 + i) != '/'){
	   *(s2+i) = '\0';
	   i--;
	 }
	 scat(s2,&s1[1],s3); 
	 return loader(s3,file_type);
       }
     }
     return loader(s1,file_type);
   }
   else {
     s2 = getenv("SIMPATH");
     if (s2==NULL){
       fputs("the environment variable SIMPATH is not set\n",stderr);
     }
     while (1) {
       while (*s2 == ':' || *s2 == ' ')
	 s2++;
       i = 0;
       if (*s2 == '\0')    /* file not found */
	 return 1;
       while (*s2 && *s2 != ' ' && *s2 != ':') 
	 s[i++] = *s2++;
       s[i++] = '/';
       s[i] = '\0';
       scat(s, s1, s3);
       returnval = loader(s3,file_type);

       if (returnval == 0)
	 return 0;
     }
   }
 }

/************************************************************************
*                                                                       *
*  The hash function uses the arity and character string associated     *
*  with a predicate, constant, or structure to find the proper bucket   *
*  (a bucket is a linked list within the pcs table) to insert or locate *
*  pcs entries.                                                         *
*                                                                       *
************************************************************************/

int hash(name, length, arity)  /* hashing function on name,returning a */
register CHAR_PTR name;        /* bucket number in the hash table      */
register LONG     length;
BYTE     arity;
{
  register int bucknum;

   bucknum = arity + 1;
   if (length > 0) {              /* first */
      bucknum = bucknum + *name;
      if (length > 1) {           /* last */
         bucknum = (bucknum << 2) + *(name + length - 1);
         if (length > 2) {        /* middle */
            bucknum = (bucknum << 2) + *(name + length / 2);
            if (length > 3)
               bucknum = (bucknum << 2) + *(name+(length / 2) - 1);
         }
      }
   }
   return abs(bucknum % BUCKET_CHAIN);

}  /* end of hash */

/****************************************************************************/

SYM_REC_PTR search(name, length, arity, sym_ptr)
CHAR_PTR name;
BYTE     arity;
LONG     length;
register SYM_REC_PTR sym_ptr;
{
  register unsigned short i,res;
  CHAR_PTR nameptr;

  while (sym_ptr!=NULL) {
    if (arity  == GET_ARITY(sym_ptr) && length == GET_LENGTH(sym_ptr)) {
      res = TRUE;
      nameptr = GET_NAME(sym_ptr);
      for (i = 0; i < length && res == TRUE; i++)
	if (*(name + i) != *(nameptr + i))
	  res = FALSE;
      if (res == TRUE) return sym_ptr;
    }
    sym_ptr = GET_NEXT(sym_ptr);
  }
  return sym_ptr;
}  /* end of search */

/****************************************************************************/
SYM_REC_PTR insert(name, length, arity)
CHAR_PTR name;
LONG     length;
BYTE     arity;
{
   int      bucket_no;
   BYTE     found;
   register int i;
   register SYM_REC_PTR sym_ptr;

   bucket_no = hash(name, length, arity);
   sym_ptr = search(name, length, arity, hash_table[bucket_no]);

   if (sym_ptr != NULL) {   /* found sym record */
     return sym_ptr; 
   }

   ALIGN(CHAR_PTR, curr_fence);  /* insert a sym record */
   sym_ptr = (SYM_REC_PTR)curr_fence;   
   curr_fence += sizeof(struct sym_rec);  
   CHECK_PCODE(curr_fence,length);

   GET_ETYPE(sym_ptr)  = T_ORDI;
   GET_ARITY(sym_ptr)  = arity;
   GET_LENGTH(sym_ptr) = length;
   GET_NEXT(sym_ptr) = hash_table[bucket_no];
   GET_NAME(sym_ptr)   = curr_fence;

   hash_table[bucket_no] = sym_ptr;
   
   for (i = 0; i < length; i++)
      *curr_fence++ = *name++;
   *curr_fence++ = '\0';

   ALIGN(CHAR_PTR, curr_fence);
   return sym_ptr;
 }  /* end of insert */

/*********************************************/
SYM_REC_PTR insert_cpred(name,arity,func)
CHAR_PTR name;
BYTE arity;
int (*func)();
{
  int      bucket_no,length;
  SYM_REC_PTR sym_ptr;

  length = strlen(name);
  sym_ptr = insert(name,length,arity);
  GET_ETYPE(sym_ptr) = C_PRED;
  GET_EP(sym_ptr) = func;
  return sym_ptr;
}

SYM_REC_PTR insert_cfunc(name,arity)
CHAR_PTR name;
BYTE arity;
{
  return insert(name,strlen(name),arity);
}

set_temp_ep(sym_ptr, ep,file_type)
SYM_REC_PTR sym_ptr;
LONG ep;
int file_type;
{
  int type;
  
  if (ep < 0) 
    return 0;
  if (file_type==BUILTIN){ /* builtin */
    if (sym_ptr!=damon_load) {
/*      insert_builtin(sym_ptr); */
    }
    GET_ETYPE(sym_ptr) = T_TEMP_PRED;
    GET_EP(sym_ptr) = (int (*)())ep;
    return 0;
  }
  /*
    type = GET_ETYPE(sym_ptr);
    if (sym_ptr!=damon_load && (type==T_PRED || type==C_PRED || type == T_DYNA)) {
    fprintf(stderr,"** warning %s/%i is redefined\n",GET_NAME(sym_ptr),GET_ARITY(sym_ptr)); 
    }
  insert_compiled(sym_ptr);
  */
  GET_ETYPE(sym_ptr) = T_TEMP_PRED;
  GET_EP(sym_ptr) = (int (*)())ep;
  return 0;
}

/*********************************************/
set_real_ep(sym_ptr, base)
SYM_REC_PTR sym_ptr;
CHAR_PTR    base;
{
   if (GET_ETYPE(sym_ptr) == T_TEMP_PRED) {
     GET_EP(sym_ptr) = (int (*)())(base + (int)GET_EP(sym_ptr));  /*???*/
      GET_ETYPE(sym_ptr) = T_PRED;
   }
}


/************************************?????????????????????*********/

/*
unset_file_ptr(sym_ptr)
SYM_REC_PTR sym_ptr;
{
   GET_ETYPE(sym_ptr) = T_ORDI;
   (LONG)GET_EP(sym_ptr) = 0;
}
*/

/************************************?????????????????????*********/

namestring(sym_ptr, s)
SYM_REC_PTR sym_ptr;
CHAR_PTR    s;
{
   LONG     i, len;
   CHAR_PTR st;

   len = GET_LENGTH(sym_ptr);
   st  = GET_NAME(sym_ptr);

#ifdef DEBUG_NAMESTRING
   printf("namestring: len = %d    string = %s\n", len, st);
#endif
   for (i = 0; i < len; i++)
      *s++ = *st++;
   *s = '\0';
}

c_current_predicates(){
  LONG list,temp0,temp1,cell;
  int i;
  SYM_REC_PTR sym_ptr;

  list = ARG(1,1);
  
  temp1 = nil_sym;
  for (i = 0; i < BUCKET_CHAIN; ++i) {
    sym_ptr = hash_table[i];
    while (sym_ptr!=NULL){
      switch (GET_ETYPE(sym_ptr)){
      case T_DYNA: 
      case T_PRED: 
      case C_PRED: 
	cell = ADDTAG(insert(GET_NAME(sym_ptr),GET_LENGTH(sym_ptr),0),ATM);
	cell = make_struct2("/",cell,MAKEINT(GET_ARITY(sym_ptr)));
	temp0 = ADDTAG((LONG)hreg,LST);
	NEW_HEAP_NODE(cell);
	NEW_HEAP_NODE(temp1);
	temp1 = temp0;
	break;
      default: break;
      }
      sym_ptr = GET_NEXT(sym_ptr);
    }
  }
  return unify(list,temp1);
}

malloc_from_pcode(size)
int size;
{
  LONG_PTR ptr;

  ALIGN(CHAR_PTR, curr_fence);
  ptr = (LONG_PTR)curr_fence;
  curr_fence += size;
  return (LONG)ptr;
}
