/* mic.c */

#include <sys/types.h>
#ifdef WIN32
#include <time.h>
#else
#include <sys/times.h>
#endif

#ifndef CLK_TCK
#if defined HZ
#define CLK_TCK HZ
#else
#define CLK_TCK 60
#endif
#endif

#include "basic.h"
#include "macro.h"
#include "term.h"
#include "frame.h"

float floatval();
extern SYM_REC_PTR insert(),insert_cfunc();
LONG greater_than_sym,less_than_sym,equal_sym;

Cboot_mic()
{
  extern c_functor(),c_arg();
  extern int c_get_main_args();
  extern int c_chdir();
  extern int c_UNDERSCORE_NAME();

  greater_than_sym = ADDTAG(insert_cfunc(">",0),ATM);
  less_than_sym = ADDTAG(insert_cfunc("<",0),ATM);
  equal_sym = ADDTAG(insert_cfunc("=",0),ATM);
  insert_cpred("c_get_main_args",1,c_get_main_args);
  insert_cpred("chdir",1,c_chdir);
  insert_cpred("c_UNDERSCORE_NAME",1,c_UNDERSCORE_NAME);
}


c_UNDERSCORE_NAME(){
  LONG atom;
  SYM_REC_PTR sym_ptr;
  char *name;

  atom = ARG(1,1);
  DEREF(atom);
  sym_ptr = (SYM_REC_PTR)UNTAGGED(atom);
  name = GET_NAME(sym_ptr);
  return *name == '_';
}

c_chdir(){
  LONG dir = ARG(1,1);
  SYM_REC_PTR sym_ptr;

  DEREF(dir);
  if(!ISATOM(dir)){
    exception=atom_expected;
    return -1;
  } 
  sym_ptr = (SYM_REC_PTR)UNTAGGED(dir);
  chdir(GET_NAME(sym_ptr));
  return 1;
}

show_ar_chain(ar)
LONG_PTR ar;
{
  while ((LONG)ar!=FOLLOW(ar)){
    show_frame(ar);
    ar = (LONG_PTR)AR_AR(ar);
  }
}

show_sf_chain(sf)
LONG_PTR sf;
{
  if ((LONG)sf!=FOLLOW(sfreg)){
    show_frame(sf);
    sf = (LONG_PTR)AR_PREV(sf);
  }
}


show_frame(ar)
LONG_PTR ar;
{
  LONG_PTR sp;
  LONG type,no,i;
  
  NO_RESERVED_SLOTS(ar,type,no);
  printf("AR_BTM(ar)=%x\n",AR_BTM(ar));
  printf("*******%x*******(%d,%d)\n",ar,(LONG_PTR)(UNTAGGED(AR_BTM(ar)))-ar,no);
  printf("   AR  = %x\n",AR_AR(ar));
  printf("   CPS = %x\n",AR_CPS(ar));
  if (no>=9)   printf("   H  = %x\n",AR_H(ar));
  sp = (LONG_PTR)UNTAGGED(AR_BTM(ar));
  while (sp>ar){
    printf("%x %x\n",sp,FOLLOW(sp));
    printf("      ");write_term(FOLLOW(sp));printf("\n");
    sp--;
  }
  printf("**************\n");
  printf("-----------------\n",ar); 
  sp = ar-no;
  while ((LONG)sp>AR_TOP(ar)){
    printf("%x %x\n",sp,FOLLOW(sp));
    printf("      ");write_term(FOLLOW(sp));printf("\n");
    sp--;
  }
  printf("-----------------\n",ar); 
}

#ifdef WIN32
LONG cputime()
{
  return (LONG)(1000*((float)clock() / (float)CLOCKS_PER_SEC));
}
#else
LONG cputime()
{ struct tms t;
  (void)times(&t);
  return (1000*t.tms_utime/CLK_TCK); /* _user_ time in msecs */
}
#endif

b_COMPARE_fcc(op1,op2,op3)
register LONG op1,op2,op3;
{
   /* op2, op3: two terms to be compared; 
      op1 = '<' if op2<op3;
      op1 = '>' if op2>op3;
      op1 = '=' if op2=op3;
      */
  register int i;

  i = compare(op2, op3);
  if (i==0) 
    ASSIGN_f(op1,equal_sym);
  else if (i>0)
    ASSIGN_f(op1,greater_than_sym);
  else 
    ASSIGN_f(op1,less_than_sym);
  return 1;
}

compare(val1, val2)
register LONG val1, val2;
{
   register LONG_PTR top;
   int              a, b;
   LONG              c;
   SYM_REC_PTR       psc_ptr1, psc_ptr2;
   float             fpt_val;

   DEREF(val2);
   
   SWITCH_OP(val1,lcompare1,
	     {if (ISVAR(val2)) return (val1 - val2); else return -1;},
	     
	     {SWITCH_OP(val2,lcompare2,
			{return 1;}, 
			{goto compare_atom_atom;}, 
			{return -1;}, 
			{return -1;},
			{return 1;});},

	     {if (!ISLIST(val2)) return 1;
	     else {
	       UNTAG(val1);
	       UNTAG(val2);
	       c = compare(FOLLOW(val1), FOLLOW(val2));
	       if (c)
		 return c;
	       else
		 return compare(FOLLOW(((LONG_PTR)val1)+1),
				FOLLOW(((LONG_PTR)val2)+1));
	     }},

	     {SWITCH_OP(val2,lcompare3,
			{return 1;},
			{if (ISATOM(val2)) goto compare_symbol; else return 1;},
			{return -1;},
			{goto compare_symbol;},
			{return 1;});},

	     {SWITCH_OP(val2,lcompare4,
			{return (val1-val2);},
			{return -1;},
			{return -1;},
			{return -1;},
			{return (val1-val2);});});

 compare_atom_atom:
   if (ISINT(val1)) 
     if (ISINT(val2)) 
       return (INTVAL(val1) - INTVAL(val2));
     else 
       return -1; /* int atom */
   else
     if (ISATOM(val2)) 
       goto compare_symbol;
     else
       return 1; /* atom int */

 compare_symbol:
   psc_ptr1 = GET_SYM_REC(val1);
   psc_ptr2 = GET_SYM_REC(val2);
   a = GET_ARITY(psc_ptr1);
   b = GET_ARITY(psc_ptr2);
   if (a != b)
     return (a - b);
   c = comalpha(psc_ptr1, psc_ptr2);
   if (c || a == 0)
     return c;
   UNTAG(val1);
   UNTAG(val2);
   for (b = 1; b <= a; b++) {
     c = compare(FOLLOW(((LONG_PTR)val1)+b),
		 FOLLOW(((LONG_PTR)val2)+b));
     if (c)
       break;
   }
   return c;
 }

comalpha(psc_ptr1, psc_ptr2)
SYM_REC_PTR psc_ptr1, psc_ptr2;
{
   CHAR name1[256], name2[256];

   if (psc_ptr1 == psc_ptr2)
      return 0;
   namestring(psc_ptr1, name1);
   namestring(psc_ptr2, name2);
   return strcmp(name1, name2);
}


c_INCREMENTARG(){
  register LONG     op1,op2;
  LONG_PTR top;

  op1 = ARG(1,2);
  op2 = ARG(2,2);

  DEREF(op1);
  DEREF(op2);
  if (!ISINT(op1)) {
    exception = integer_expected;
    return -1;
  }
  op1 = INTVAL(op1);
  if (ISSTRUCT(op2) && op1>0 && op1 <= GET_STR_SYM_ARITY(op2)) {
    top = (LONG_PTR)UNTAGGED(op2)+op1;
    op2 = *top;
    PUSHTRAILB(top,op2);
    if (!ISINT(op2)) {
      exception = integer_expected;
      return -1;
    }
    *top = MAKEINT(INTVAL(op2)+1);
    return 1;
  }
  else if (ISLIST(op2) && op1>0 && op1 <= 2) {
    top = (LONG_PTR)UNTAGGED(op2)+op1-1;
    op2 = *top;
    PUSHTRAILB(top,op2);
    if (!ISINT(op2)) {
      exception = integer_expected;
      return -1;
    }
    *top = MAKEINT(INTVAL(op2)+1);
    return 1;
  }
  exception = structure_expected;
  return -1;
}

c_DECREMENTARG(){
  register LONG     op1,op2;
  LONG_PTR top;
  
  op1 = ARG(1,2);
  op2 = ARG(2,2);
  
  DEREF(op1);DEREF(op2);
  if (!ISINT(op1)) {
    exception = integer_expected;
    return -1;
  }
  op1 = INTVAL(op1);
  if (ISSTRUCT(op2) && op1>0 && op1 <= GET_STR_SYM_ARITY(op2)) {
    top = (LONG_PTR)UNTAGGED(op2)+op1;
    op2 = *top;
    PUSHTRAILB(top,op2);
    if (!ISINT(op2)) {
      exception = integer_expected;
      return -1;
    }
    *top = MAKEINT(INTVAL(op2)-1); 
    return 1;
  }
  else if (ISLIST(op2) && op1 > 0 && op1 <= 2) {
    top = (LONG_PTR)UNTAGGED(op2)+op1-1;
    op2 = *top;
    PUSHTRAILB(top,op2);
    if (!ISINT(op2)) {
      exception = integer_expected;
      return -1;
    }
    *top = MAKEINT(INTVAL(op2)-1);
    return 1;
  }
  exception = structure_expected;
  return -1;
}

b_DESTRUCTIVE_SET_ARG_ccc(op1,op2,op3)
register LONG op1,op2,op3;
{
  LONG_PTR top;
  
  DEREF(op1);DEREF(op2);DEREF(op3);
  if (IS_SUSP_VAR(op3)) op3 = UNTAGGED(op3);
  if (!ISINT(op1)) {
    exception = integer_expected;
    return -1;
  }
  op1 = INTVAL(op1);
  if (ISSTRUCT(op2) && op1>0 && op1 <= GET_STR_SYM_ARITY(op2)) {
    top = (LONG_PTR)UNTAGGED(op2)+op1;
    FOLLOW(top) = op3;
    return 1;
  }
  else if (ISLIST(op2) && op1>0 && op1 <= 2) {
    top = (LONG_PTR)UNTAGGED(op2)+op1-1;
    FOLLOW(top) = op3;
    return 1;
  }
  exception = structure_expected;
  return -1;
}

int b_NAME0_cf(op1,op2)  /* op2 is made to be the string of the name of op1*/ 
register LONG     op1,op2;
{
  SYM_REC_PTR       psc_ptr;
  CHAR_PTR name;
  LONG     i, len, nlist;
  register LONG_PTR top;
  LONG_PTR stack_top;

  DEREF(op1);
  if (!ISATOM(op1)) {
    exception = atom_expected;
    return -1;
  }
  psc_ptr = GET_ATM_SYM_REC(op1);
  name    = GET_NAME(psc_ptr);
  len     = GET_LENGTH(psc_ptr);
  if (len == 0) {
    ASSIGN_f(op2,nil_sym);
    return 1;
  }
  nlist   = ADDTAG(hreg,LST);
  for (i = 0; i < len; i++) {
    FOLLOW(hreg++) = MAKEINT(*name++);
    top = hreg++;
    FOLLOW(top) = ADDTAG(hreg,LST);
  }
  FOLLOW(top) = nil_sym;
  ASSIGN_f(op2,nlist);
  return 1;
}  /* b_NAME0 */


b_BLDATOM_fc(op1,op2)
register LONG     op1,op2;
{            /* make op1 to be an atom with name string op2       */

  int     a, n;
  CHAR_PTR name,s;
  register LONG  op3;
  register LONG_PTR top;
  LONG ptr;
  
  name = s = (char *)hreg;
  n = 0;
  
  do {
    DEREF(op2);
    if (ISNIL(op2))
      break;
    if (ISLIST(op2)) {
      UNTAG(op2);
      op3 = FOLLOW(op2);  
      DEREF(op3);
      if (!ISINT(op3)) {
	printf("Error: bldatom, non integer\n");
	return 0;
      }
      a = INTVAL(op3);
      if (a < 0 ) {
	exception = integer_expected;
	return -1;
      }
      *s++ = a;
      n++;
      op2 = FOLLOW((op2+4));
    } else {
      exception = list_expected;
      return -1;
    }
  } while (1);
  ASSIGN_f(op1,ADDTAG(insert(name, n, 0),ATM));
  return 1;
}  /* b_BLDATOM */


b_SYSTEM0_cf(op1,op2)  /* op1: a list of int (string) for CShell commands */
register LONG     op1,op2;
{
   register LONG_PTR top;
   SYM_REC_PTR ptr;
   CHAR     s[256];

   DEREF(op1);
   if (!ISATOM(op1)){
     exception = atom_expected;
     return -1;
   }
   ptr = GET_ATM_SYM_REC(op1);
   namestring(ptr, s);
   ASSIGN_f(op2,MAKEINT(system(s)));
   return 1;
}

b_LOAD_cfc(op1,op2,op3)
register LONG     op1,op2,op3;
{  /* arg1: the byte code file to be loaded
    * arg2: the return code, 0 => success
    * arg3: file type
    */
  register LONG_PTR top;
  SYM_REC_PTR	     psc_ptr;

  DEREF(op1);DEREF(op3);
  psc_ptr = GET_ATM_SYM_REC(op1);
  op1 = dyn_loader(psc_ptr,INTVAL(op3));
  ASSIGN_f(op2, MAKEINT(op1));
  return 1;
}

c_STATISTICS()
{
  LONG_PTR lstktop;

  printf("\nControl stack + Heap: %d bytes\n",maxmem*sizeof(LONG));
  printf("Control stack in use: %d bytes\n",(local_bottom-arreg)*sizeof(LONG));
  printf("Heap stack in use:    %d bytes\n\n", (hreg-heap_bottom)*sizeof(LONG));
  printf("Program area:         %d bytes\n", maxpspace*sizeof(LONG));
  printf("Program area in use:  %d bytes\n\n", (LONG)curr_fence - (LONG)pspace);
  printf("Trail stack:          %d bytes\n", maxtrail*sizeof(LONG));
  printf("Trail stack in use:   %d bytes\n\n", (trail_bottom-treg)*sizeof(LONG));

  printf("Table area:           %d bytes\n", tableAreaSize*sizeof(LONG));
  printf("Table area in use:    %d bytes\n", (tareg-tableArea)*sizeof(LONG));
  return 1;
}

c_STATISTICS0()
{
   if (!unify(ARG(1,8), MAKEINT(maxpspace*sizeof(LONG))))
     return 0;     /* max program area */
   if (!unify(ARG(2,8), MAKEINT((((LONG)curr_fence-(LONG)pspace)))))
      return 0;     /* program area in use */
   if (!unify(ARG(3,8), MAKEINT((hreg-heap_bottom)*sizeof(LONG))))
      return 0;     /* heap in use */
   if (!unify(ARG(4,8), MAKEINT((arreg-hreg)*sizeof(LONG))))
      return 0;     /* stack area (local, global) free */
   if (!unify(ARG(5,8), MAKEINT((local_bottom-arreg)*sizeof(LONG))))
      return 0;     /* control stack in use */
   if (!unify(ARG(6,8), MAKEINT(maxmem*sizeof(LONG))))
      return 0;     /* total stack area size */
   if (!unify(ARG(7,8), MAKEINT(maxtrail*sizeof(LONG))))
      return 0;     /* total trail size */
   if (!unify(ARG(8,8), MAKEINT((trail_bottom-treg)*sizeof(LONG))))
      return 0;
   return 1;
}

b_HASHVAL_cf(op1,op2)  /* op1 a term, op2 the hash value of op1 */
register LONG     op1, op2;
{
   register LONG_PTR top;
   int type;

   ASSIGN_f(op2,MAKEINT(hashval(op1)));
   return 1;
 }

hashval(op)
LONG op;
{
  int  arity,sum;
  SYM_REC_PTR	     psc_ptr;
  LONG_PTR top;
  
 nhashval:
  switch(TAG(op)) {
  case REF:
    NDEREF(op,nhashval);
    return 0;
    break;
  case ATM : 
    return  ((op & 0xffff));
    break;
  case LST: 
    if (ISLIST(op)){
      UNTAG(op);
      return hashval(*(LONG_PTR)op)+hashval(*((LONG_PTR)op+1));
    } else return 0;
    break;
  case STR :
    if (ISSTRUCT(op)){
      psc_ptr = GET_STR_SYM_REC(op);
      arity = GET_ARITY(psc_ptr);
      UNTAG(op);
      sum = (((LONG)psc_ptr & 0xffff));
      while (arity>0) {
	sum += hashval(*((LONG_PTR)op+arity));
	arity--;
      }
      return sum;
    } else return 0;
    break;
  }
}

b_HASHVAL1_cf(op1,op2)  /* op1 a term, op2 the hash value of the main functor of op1*/
register LONG     op1, op2;
{
   register LONG_PTR top;
   int type;
   
   DEREF(op1);
   ASSIGN_f(op2,MAKEINT(hashval1(op1)));
   return 1;
 }

hashval1(op)
register LONG op;
{
  int  arity,sum;
  SYM_REC_PTR	     psc_ptr;

  switch(TAG(op)) {
  case REF:
    printf("1st arg of b_HASHVAL1 cannot be a variable\n");
    return 0;
    break;
  case ATM : 
    return (op & 0x00ffffff)>>2;
    break;
  case LST: 
    return ((LONG)list_psc &  0x00ffffff) >>2;
    break;

  case STR :
    psc_ptr = GET_STR_SYM_REC(op);
    return ((LONG)psc_ptr &  0x00ffffff)>>2;
    break;
  }
}

LONG make_struct1(f,op1)
char *f;
LONG op1;
{
  SYM_REC_PTR psc_ptr;
  LONG return_value;

  psc_ptr = insert_cfunc(f,1);
  return_value = ADDTAG(hreg,STR);
  NEW_HEAP_NODE((LONG)psc_ptr);
  BUILD_VALUE(op1,lab1);
  return return_value;
}

LONG make_struct2(f,op1,op2)
char *f;
LONG op1,op2;
{
  SYM_REC_PTR psc_ptr;
  LONG return_value;

  psc_ptr = insert_cfunc(f,2);
  return_value = ADDTAG(hreg,STR);
  NEW_HEAP_NODE((LONG)psc_ptr);
  BUILD_VALUE(op1,lab1);
  BUILD_VALUE(op2,lab2);
  return return_value;
}

LONG make_struct3(f,op1,op2,op3)
char *f;
LONG op1,op2,op3;
{
  SYM_REC_PTR psc_ptr;
  LONG return_value;

  psc_ptr = insert_cfunc(f,3);
  return_value = ADDTAG(hreg,STR);
  NEW_HEAP_NODE((LONG)psc_ptr);
  BUILD_VALUE(op1,lab1);
  BUILD_VALUE(op2,lab2);
  BUILD_VALUE(op3,lab3);
  return return_value;
}

LONG make_struct4(f,op1,op2,op3,op4)
char *f;
LONG op1,op2,op3,op4;
{
  SYM_REC_PTR psc_ptr;
  LONG return_value;

  psc_ptr = insert_cfunc(f,4);
  return_value = ADDTAG(hreg,STR);
  NEW_HEAP_NODE((LONG)psc_ptr);
  BUILD_VALUE(op1,lab1);
  BUILD_VALUE(op2,lab2);
  BUILD_VALUE(op3,lab3);
  BUILD_VALUE(op4,lab4);
  return return_value;
}

LONG make_struct5(f,op1,op2,op3,op4,op5)
char *f;
LONG op1,op2,op3,op4,op5;
{
  SYM_REC_PTR psc_ptr;
  LONG return_value;

  psc_ptr = insert_cfunc(f,5);
  return_value = ADDTAG(hreg,STR);
  NEW_HEAP_NODE((LONG)psc_ptr);
  BUILD_VALUE(op1,lab1);
  BUILD_VALUE(op2,lab2);
  BUILD_VALUE(op3,lab3);
  BUILD_VALUE(op4,lab4);
  BUILD_VALUE(op5,lab5);
  return return_value;
}

LONG make_struct_dummy(psc_ptr)
SYM_REC_PTR psc_ptr;
{
  int arity,i;
  LONG return_value;

  arity = GET_ARITY(psc_ptr);
  return_value = ADDTAG(hreg,STR);
  NEW_HEAP_NODE((LONG)psc_ptr);
  for (i=1;i<=arity;i++){
    NEW_HEAP_FREE;
  }
  return return_value;
}

LONG make_cons_dummy(){
  LONG return_value;
  
  return_value = ADDTAG(hreg,LST);
  NEW_HEAP_FREE;
  NEW_HEAP_FREE;
  return return_value;
}

LONG make_struct_with_args(fp,psc_ptr)
LONG_PTR fp;
SYM_REC_PTR psc_ptr;
{
  int arity,i;
  LONG return_value;

  arity = GET_ARITY(psc_ptr);
  if (arity == 0) 
    return ADDTAG(insert(GET_NAME(psc_ptr),GET_LENGTH(psc_ptr),0),ATM);
  else {
    return_value = ADDTAG(hreg,STR);
    NEW_HEAP_NODE((LONG)psc_ptr);
    for (i=1;i<=arity;i++){
      BUILD_VALUE(ARG(i,arity),lab1);
    }
    return return_value;
  }
}

b_GET_LENGTH_cf(op1,op2)
register LONG op1,op2;
{
  DEREF(op1);
  ASSIGN_f(op2,MAKEINT(GET_SYM_LENGTH(op1)));
  return 1;
}
  
c_SHOW_NONDET_FRAME() {
  LONG op1,i;
  LONG_PTR tempreg;

  op1 = ARG(1,1);
  op1 = INTVAL(op1);
  tempreg = (LONG_PTR)*arreg;
  printf("AR=   = %x\n",tempreg);
  printf("(AR)  = %x\n", *tempreg); 
  printf("CPS   = %x\n", *(tempreg-1)); 
  printf("TOP   = %x\n", *(tempreg-2)); 
  printf("B     = %x\n", *(tempreg-3)); 
  printf("CPF   = %x\n", *(tempreg-4)); 
  printf("H     = %x\n", *(tempreg-5)); 
  printf("T     = %x\n", *(tempreg-6)); 
  for (i=1;i<= op1 ;i++) {
    printf("%x",*(tempreg-6-i));
    if (*(tempreg-6-i) == (LONG)(tempreg-6-i))
      printf(" ***\n");
    else
      if ((*(tempreg-6-i) > (LONG)hreg) && (*(tempreg-6-i) < (LONG)(tempreg-6-op1)))
      printf(" ???\n");
    else 
      printf("\n");
  }
  return 1;
}

c_SHOW_SUSP_FRAME() {
  LONG arity,i,op1,op2;
  LONG_PTR tempreg;
  
  arity = ARG(1,1);DEREF(arity);
  arity = INTVAL(arity);
  tempreg = (LONG_PTR)*arreg;

  op1 = *(tempreg+2);DEREF(op1);
  if (INTVAL(op1)!=1) return 1;
  
  for (i = arity; i>0; i--){ 
    printf("Y%i    = %x  ", i,*(tempreg+i)); write_term(*(tempreg+i));printf("\n"); 
  } 
  printf("AR=   = %x\n",tempreg);
  printf("(AR)  = %x\n", *tempreg); 
  printf("CPS   = %x\n", *(tempreg-1)); 
  printf("TOP   = %x\n", *(tempreg-2)); 
  printf("SYM   = %x\n", *(tempreg-3)); 
  printf("PREV  = %x\n", *(tempreg-4)); 
  printf("NEXT  = %x\n", *(tempreg-5)); 
  return 1;
}

c_SHOWHREG()
{
  printf("hreg=%x\n",hreg);
}

b_CPUTIME_f(op1)
register LONG op1;
{
  extern LONG cputime();
  LONG msec;

  
  msec = cputime();
  ASSIGN_f(op1,MAKEINT(msec));
  return 1;
}
  
membchk() 
{
  LONG op1,op2;
  op1 = ARG(1,2);  DEREF(op1);
  op2 = ARG(2,2);
  return membchk2(op1,op2);
}

membchk2(op1,op2)
LONG op1,op2;
{
  LONG op3,op4;
  DEREF(op2);
  if (!ISLIST(op2))
    return 0;
  GET_CAR(op3,op2);
  GET_CDR(op4,op2);
  if (!identical(op1,op3))
    goto l1;
  return 1;
 l1:
  return membchk2(op1,op4);
}
    

quit(s)
CHAR_PTR s;
{
   printf(s);
   exit(0);
}

/* concatinate s1 and s2 into s3 */

scat(s1, s2, s3)
CHAR_PTR s1, s2, s3;
{
   while (*s1)             /* copy s1 into s3, without the EOS */
      *s3++ = *s1++;
   while (*s3++ = *s2++)   /* add s2 onto s3, including the EOS */
      ;
}


/*********** functor(T,F,N) *********/
c_functor(){
  TERM op1,op2,op3;

  op1 = ARG(1,3);
  op2 = ARG(2,3);
  op3 = ARG(3,3);
  return cfunctor1(op1,op2,op3);
}

int cfunctor1(op1,op2,op3)
     LONG op1,op2,op3;
{
  int i;
  LONG top1,top2;
  SYM_REC_PTR psc_ptr;

  DEREF(op1);
  SWITCH_OP(op1,n_functor1,
	    {goto functor_vdd;},
	    { top1 = MAKEINT(0);
	      if (unify(op2,op1) && unify(op3,top1))
		return 1;
	      else
		return 0;},
	    {top1 = (LONG)insert(".",1,0);
	     top2 = MAKEINT(2);
	     goto functor_compound_d_d;},
	    
	    {psc_ptr = GET_SYM_REC(op1);
	     top1 = (LONG)insert(GET_NAME(psc_ptr),GET_LENGTH(psc_ptr),0);
	     top2 = MAKEINT(GET_ARITY(psc_ptr));
	     goto functor_compound_d_d;},
	    
	    {goto functor_vdd;});

 functor_compound_d_d:
  top1 = ADDTAG(top1,ATM);
  return (unify(op2,top1) && unify(op3,top2)) ?  1 : 0;

 functor_vdd:
  DEREF(op2);
  if (ISINT(op2)) {
    if (unify(op1,op2) && unify(op3,MAKEINT(0)))
      return 1;
    else
      return 0;
  }
  if (!ISATOM(op2)) {
    exception = atom_expected;
    return -1;
  }
  DEREF(op3);
  if (!ISINT(op3)) {
    exception = integer_expected;
    return -1;
  }
  op3 = INTVAL(op3);
  if (op3<0) {
    printf("type error: arity must be greater than 0");
  }
  psc_ptr = GET_SYM_REC(op2);
  if (GET_ARITY(psc_ptr)>0) {
    printf("type error: main functor of functor/3 must be constant\n");
  }
  if (arreg-100<hreg) {
    quit("Heap overflow\n");
  }
  if (op3==0) {
    unify(op1,ADDTAG(insert(GET_NAME(psc_ptr),GET_LENGTH(psc_ptr),op3),ATM));
    return 1;
  };
  if (op3==2 && GET_NAME(psc_ptr)[0]=='.' && GET_LENGTH(psc_ptr)==1) {
    unify(op1,ADDTAG(hreg,LST));
  } else {
    top1 = (LONG)insert(GET_NAME(psc_ptr),GET_LENGTH(psc_ptr),op3);
    unify(op1,ADDTAG(hreg,STR));
    *hreg++ = top1;
  }
  for (i=0;i<op3;hreg++,i++)
    MAKE_FREE(LONG,*hreg);
  return 1;
}

/************* arg(N,T,A) *************/
c_arg(){
  TERM op1,op2,op3;
  
  op1 = ARG(1,3);
  op2 = ARG(2,3);
  op3 = ARG(3,3);
  
  return carg1(op1,op2,op3);
}

int carg1(op1,op2,op3)
register LONG op1,op2,op3;
{
  DEREF(op1);
  op1 = INTVAL(op1);
  if (op1 <= 0) {
/*    printf("type error: index for arg must be > 0");*/
    exception = out_of_range;
    return -1;
  }
  DEREF(op2);
  if (ISSTRUCT(op2) && op1 <= GET_SYM_ARITY(op2))
    if (unify(*((LONG_PTR) UNTAGGED(op2) + op1), op3))
      return 1;
    else 
      return 0;
  if (ISLIST(op2) && op1 <= 2)
    if (unify(*((LONG_PTR) UNTAGGED(op2) + op1 - 1), op3))
      return 1;
    else
      return 0;
  else {
/*    printf("instantiation error: second argument of arg must be instantiated\n");*/
    exception = structure_expected;
    return -1;
  }
}

c_get_main_args(){
  TERM list;
  
  list = ARG(1,1);

  return unify(list,main_args);
}

c_SAVE_AR(){
  TERM AR;
  
  AR = ARG(1,1);
  return unify(AR,ADDTAG(FOLLOW(arreg),ATM));
}

c_WRITE_AR(){
  TERM AR;
  
  AR = ARG(1,1);
  DEREF(AR);
  printf("%x",UNTAGGED(AR));
  return 1;
}

/*
check_strange_dvar(arreg,hreg)
LONG_PTR arreg,hreg;
{
  LONG_PTR sp;
  LONG op;

  
  for (sp = arreg+1; sp<=arreg+5; sp++){
    op = FOLLOW(sp);
    while (op<hreg && op>heap_bottom && ISREF(op) && !ISFREE(op)){
      if (IS_SUSP_VAR(FOLLOW(op)) && op!=UNTAGGED(FOLLOW(op)))
	printf("************STRANGE**********%x %x\n",op,FOLLOW(op));
      op = FOLLOW(op);
    }
  }
}
*/


