/*
*   File   : booltable.c
*   Author : Neng-Fa ZHOU Copyright (C) All rights reserved
*   Updated: November 1993
*   Purpose: Primitives for Boolean table
*/
#include "state.h"
#define ISTUPLE(op) if (!ISSTRUCT(op)){exception = structure_expected; return -1;} 

/* Primitives for Boolean tables */
c_BT_FIRST(){
  register LONG op1;

  op1 = ARG(1,1);

  DEREF(op1);
  ISTUPLE(op1);
  return bt_first1(op1);
}

c_BT_LAST(){
  register LONG op1;

  op1 = ARG(1,1);
  DEREF(op1);
  ISTUPLE(op1);
  return bt_last1(op1);
}
    
c_BT_NEXT(){
  register LONG op1,op2,op3;

  op1 = ARG(1,3); op2 = ARG(2,3); op3 = ARG(3,3);
  DEREF(op1);DEREF(op2);DEREF(op3);
  ISTUPLE(op1);
  if (!ISVAR(op2)) {exception = variable_expected; return -1;};
  ISTUPLE(op3);
  return bt_next3(op1,op2,op3);
}
  
c_BT_PREV(){
  register LONG op1,op2,op3;

  op1 = ARG(1,3); op2 = ARG(2,3); op3 = ARG(3,3);
  DEREF(op1);DEREF(op2);DEREF(op3);
  ISTUPLE(op1);
  if (!ISVAR(op2)) {exception = variable_expected; return -1;};
  ISTUPLE(op3);
  return bt_prev3(op1,op2,op3);
}

c_BT_TRUE(){
  register LONG op1;

  op1 = ARG(1,1);
  DEREF(op1);
  ISTUPLE(op1);
  return  bt_true1(op1);
}

c_BT_FALSE(){
  register LONG op1;

  op1 = ARG(1,1);
  DEREF(op1);
  ISTUPLE(op1);
  return bt_false1(op1);
}

c_BT_SET_TRUE(){
  register LONG op1;
  
  op1 = ARG(1,1);
  DEREF(op1);
  ISTUPLE(op1);
  return  bt_set_true1(op1);
}

c_BT_SET_FALSE(){
  register LONG op1;

  op1 = ARG(1,1);
  DEREF(op1);
  ISTUPLE(op1);
  return bt_set_false_1(op1);
}

c_BT_SET_FALSE1(){
  register LONG op1,op2;
  
  op1 = ARG(1,2); op2 = ARG(2,2);
  DEREF(op1);DEREF(op2);
  ISTUPLE(op1);
  return bt_set_false_2(op1,op2);
}

c_BT_COUNT(){
  register LONG op1,op2;
  
  op1 = ARG(1,2); op2 = ARG(2,2);
  DEREF(op1);
  ISTUPLE(op1);
  return bt_count2(op1,op2);  
}

/***************************************************
  bt(T,S)
  T is in the form of p(X1,X2,...,Xn) where Xi's are lists
  S is 0 (means false state) or 1 (means true state)
  ***************************************************/
c_BT(){
  register LONG T,S;

  SYM_REC_PTR psc_ptr;
  NODE_PTR root;
  register int arity,i,size;
  NODE_PTR new_node();
  
  T = ARG(1,2); S = ARG(2,2);
  DEREF(T);DEREF(S);
  ISTUPLE(T);

  S = INTVAL(S);
  psc_ptr = GET_STR_SYM_REC(T);

  if (!IS_ORDI(psc_ptr) && GET_ETYPE(psc_ptr)!=T_BOOT){
    exception = illegal_arguments;
    return -1;
  }
  
  GET_ETYPE(psc_ptr) = T_BOOT;
  root = (NODE_PTR)new_node(MAKEINT(0));
  GET_EP(psc_ptr) = (int (*)())root;
  
  arity = psc_ptr->arity;
  for (i=1;i<=arity;i++) {
    hashtabs[i] = NULL;
    tabsizes[i] = 0;
  }
  build_tree(root,1,arity,UNTAGGED(T),S);
  return 1;
}

int build_tree(parent,arg_no,arity,T,S) /* return count */
NODE_PTR parent;
register int arg_no,arity;
register LONG T,S;
{
  register NODE_PTR node;
  register LONG arg;
  LONG elm,tag;
  register int total_count;
  NODE_PTR lookup_children();
  LONG_PTR malloc_from_heap();

  arg = *((LONG_PTR)T+arg_no);
  DEREF(arg);

/*  write_term(arg);printf("\n"); */

  if (parent->hashtab==NULL) {
    if (hashtabs[arg_no]!=NULL) {
      parent->hashtab = hashtabs[arg_no];
      parent->size = tabsizes[arg_no];
    }
    else {
      parent->hashtab = (LONG_PTR)malloc_from_heap(MAX_SIZE);
      initialize_block(parent->hashtab,MAX_SIZE);
      hashtabs[arg_no] = parent->hashtab;
    }
  }
  total_count = 0;
  if (arg_no==arity) {
    while (!ISNIL(arg)) {
      elm = *(LONG_PTR)UNTAGGED(arg);
      DEREF(elm);
      arg = *((LONG_PTR)UNTAGGED(arg)+1);
      DEREF(arg);
      node = (NODE_PTR)lookup_children(parent,S,elm,&tag);
      if (S==1) {
	node->count=1;
      total_count += tag;
      }
    }
  }
  else {
    while (!ISNIL(arg)) {
      elm = *(LONG_PTR)UNTAGGED(arg);
      DEREF(elm);
      arg = *((LONG_PTR)UNTAGGED(arg)+1);
      DEREF(arg);
      node = (NODE_PTR)lookup_children(parent,S,elm,&tag);
      total_count += build_tree(node,arg_no+1,arity,T,S);

    }
  }
  parent->count += total_count;
  tabsizes[arg_no] = parent->size;
  return total_count;
}

initialize_block(addr,size)
register LONG_PTR addr;
register int size;
{
  while (size>0) {
    *addr++ = (LONG)NULL;
    size--;
  }
}

NODE_PTR lookup_children(parent,S,elm,tag)
NODE_PTR parent;
LONG S,elm;
LONG_PTR tag;
{
  BYTE elm_type;
  LONG hashval,index,temp;
  NODE_PTR node,prev;
  LONG_PTR children1,children2;
  register HASHREC_PTR pt1,pt2;
  LONG_PTR malloc_from_heap();
  NODE_PTR new_node();
  prev = parent->last;
  hashval = ((unsigned long int)elm>>2)%MAX_SIZE;
  pt1 = (HASHREC_PTR)(parent->hashtab + hashval);
  pt2 = (HASHREC_PTR)*(LONG_PTR)pt1;
  while (pt2!=NULL && pt2->value != elm) {
    pt1 = pt2;
    pt2 = pt2->next;
  }
  if (pt2==NULL) {  /* not found */
    pt2 = (HASHREC_PTR)malloc_from_heap(sizeof(HASHREC)/4);
    pt1->next = pt2;
    pt2->next = NULL;
    pt2->value = elm;
    index = parent->size;
    pt2->index = index;
    parent->size += 1;
  } 
  else 
    index = pt2->index;
  
  if (parent->children == NULL) {
    parent->children = (LONG_PTR)malloc_from_heap(MAX_SIZE);
    initialize_block(parent->children,MAX_SIZE);
  }
    
  children1 = parent->children;
  children2 = (LONG_PTR)*(children1+127);
  while (index>=127) {
    if (children2==NULL) {
      children2 = (LONG_PTR)malloc_from_heap(MAX_SIZE);
      initialize_block(children2,MAX_SIZE);
      *(children1+127) = (LONG)children2;
    }
    children1 = children2;
    children2 = (LONG_PTR)*(children1+127);
    index -= 127;
  }
  node = (NODE_PTR)*(children1+index);
  if (node==NULL) {
    *tag = 1;
    node = (NODE_PTR)new_node(elm);
    *(children1+index) = (LONG)node;
    if (S==1) {
      if (parent->first==NULL) 
	parent->first = node;
      node->prev = prev;
      if (prev!=NULL)
	prev->succ = node;
      parent->last = node;
    }
  }
  else *tag = 0;
  return node;
}

NODE_PTR locate_node(parent,elm)
NODE_PTR parent;
LONG elm;
{
  BYTE elm_type;
  LONG hashval,index,temp;
  LONG_PTR children1,children2;
  register HASHREC_PTR pt1,pt2;
  
  hashval = ((unsigned long int)elm>>2)%MAX_SIZE;
  pt1 = (HASHREC_PTR)(parent->hashtab + hashval);
  pt2 = (HASHREC_PTR)*(LONG_PTR)pt1;

  while (pt2!=NULL && pt2->value != elm) {
    pt1 = pt2;
    pt2 = pt2->next;
  }
  if (pt2==NULL) {  /* not found */
    return 0;
  } 
  index = pt2->index;
  children1 = parent->children;
  children2 = (LONG_PTR)*(children1+127);
  while (index>=127) {
    if (children2==NULL) 
      return 0;
    children1 = children2;
    children2 = (LONG_PTR)*(children1+127);
    index -= 127;
  }
  return (NODE_PTR)*(children1+index);
}
 
LONG_PTR malloc_from_heap(size)
int size;
{
  LONG_PTR ptr;

/*  printf("==> malloc_from_heap hreg= %x,size=%i\n",hreg,size);*/
  ptr = hreg;
  hreg += size;

  if (arreg-1000 < hreg) { 
    printf("control_stack_overflow\n");           
    exit(1);
  }
/*  printf("<== malloc_from_heap hreg=%x\n",hreg);*/
  return ptr;
}

NODE_PTR new_node(elm)
LONG elm;
{
  register NODE_PTR node;
  
/*  printf("sizeof(NODE)/4=%i\n",sizeof(NODE)/4);*/
  node = (NODE_PTR)malloc_from_heap(sizeof(NODE)/4);
  node->value = elm;
  node->count = 0;
  node->size =0;
  node->hashtab = NULL;
  node->prev = NULL;
  node->succ = NULL;
  node->first = NULL;
  node->last = NULL;
  node->children = NULL;
  return node;
}

/***********************************************************
bt_first1(T)
the first tuple matching the pattern T
***********************************************************/
int bt_first1(T)
LONG T;
{
  SYM_REC_PTR psc_ptr;
  NODE_PTR node;
  int arg_no,arity;
  register LONG arg;

  psc_ptr = GET_STR_SYM_REC(T);
  node = (NODE_PTR)GET_EP(psc_ptr);
  if (GET_ETYPE(psc_ptr)!=T_BOOT) {
    /*    printf("%s is not a Boolean table\n",GET_NAME(psc_ptr));*/
    exception = illegal_predicate;
    return -1;
  }
  if (node==NULL || node->count==0)
    return 0;
  arity = GET_ARITY(psc_ptr);
  UNTAG(T);
  arg_no = 1;
  arg = *((LONG_PTR)T+arg_no);
  DEREF(arg);
  while (ISNONVAR(arg) && arg_no<arity) {
    node = locate_node(node,arg);
    if (node==NULL || node->count==0) return 0;
    arg_no++;
    arg = *((LONG_PTR)T+arg_no);
    DEREF(arg);
  }
  if (ISINT(arg) || ISATOM(arg)) {
    node = locate_node(node,arg);
    if (node==NULL || node->count==0) return 0;
    return 1;
  }
  else return first_node(node,arg_no,arity,T);
}

int first_node(node,arg_no,arity,T)
NODE_PTR node;
int arg_no,arity;
LONG T;
{
  register LONG arg;
  
  if (arg_no>arity) 
    return 1;
  arg = *((LONG_PTR)T+arg_no);
  DEREF(arg);
  node = node->first;
  if (node==NULL) return 0;
  if (first_node(node,arg_no+1,arity,T)) {
    ASSIGN_W_TEST(arg,node->value);
    PUSHTRAIL(arg);
    return 1;
  }
  return 0;
}

/***********************************************************
bt_last1(T)
the last tuple matching the pattern T
***********************************************************/
int bt_last1(T)
LONG T;
{
  SYM_REC_PTR psc_ptr;
  NODE_PTR node;
  int arg_no,arity;
  register LONG arg;

  psc_ptr = GET_STR_SYM_REC(T);
  node = (NODE_PTR)GET_EP(psc_ptr);
  if (GET_ETYPE(psc_ptr)!=T_BOOT) {
/*    printf("%s is not a Boolean table\n",GET_NAME(psc_ptr));*/
    exception = illegal_predicate;
    return -1;
  }
  if (node==NULL || node->count==0)
    return 0;
  arity = GET_ARITY(psc_ptr);
  UNTAG(T);
  arg_no = 1;
  arg = *((LONG_PTR)T+arg_no);
  DEREF(arg);
  while ((ISINT(arg) || ISATOM(arg)) && arg_no<arity) {
    node = locate_node(node,arg);
    if (node==NULL || node->count==0) return 0;
    arg_no++;
    arg = *((LONG_PTR)T+arg_no);
    DEREF(arg);
  }
  if (ISINT(arg) || ISATOM(arg)) {
    node = locate_node(node,arg);
    if (node==NULL || node->count==0) return 0;
    return 1;
  }
  else return last_node(node,arg_no,arity,T);
}

last_node(node,arg_no,arity,T)
NODE_PTR node;
int arg_no,arity;
LONG T;
{
  register LONG arg;
  
  if (arg_no>arity)
    return 1;
  arg = *((LONG_PTR)T+arg_no);
  DEREF(arg);
  node = node->last;
  if (node==NULL) return 0;
  if (last_node(node,arg_no+1,arity,T)) {
    ASSIGN_W_TEST(arg,node->value);
    PUSHTRAIL(arg);
    return 1;
  }
  return 0;
}

/***********************************************************
bt_next3(T1,T2,T3)
T2 is the next tuple of T1 that matches T3
***********************************************************/
int bt_next3(T1,T2,T3)
LONG T1,T2,T3;
{
  SYM_REC_PTR psc_ptr;
  register NODE_PTR node,node1;
  register LONG arg1,arg2;
  register int arg_no,arity,i;
  LONG temp_hreg;

  psc_ptr = GET_STR_SYM_REC(T3);
  if (GET_ETYPE(psc_ptr)!=T_BOOT) {
    /*    printf("%s is not a Boolean table\n",GET_NAME(psc_ptr)); */
    exception = illegal_predicate;
    return -1;
  }
  node = (NODE_PTR)GET_EP(psc_ptr);
  if (node==NULL || node->count==0)
    return 0;
  arity = GET_ARITY(psc_ptr);
  if (ISNONVAR(T2)) {
    printf("*** error : the second argument of next is nonvar\n");
    return 0;
  }
  FOLLOW(T2) = ADDTAG(hreg,STR);
  PUSHTRAIL(T2);
  temp_hreg = (LONG)hreg;
  NEW_HEAP_NODE((LONG)psc_ptr);
  for (i=1;i<=arity;i++,hreg++) {
    MAKE_FREE(LONG,*hreg);
  }

  arg_no = 1;
  UNTAG(T1);
  UNTAG(T3);
  arg1 = *((LONG_PTR)T1+arg_no);
  arg2 = *((LONG_PTR)T3+arg_no);
  DEREF(arg1);
  DEREF(arg2);
  while (ISNONVAR(arg1) && arg1== arg2) {
    node = locate_node(node,arg1);
    if (node==NULL || node->count==0)
      return 0;
    ASSIGN_W_TEST(*((LONG_PTR)temp_hreg+arg_no),arg1);
    arg_no++;
    arg1 = *((LONG_PTR)T1+arg_no);
    arg2 = *((LONG_PTR)T3+arg_no);
    DEREF(arg1);
    DEREF(arg2);
  }
  if (ISVAR(arg2)) {
    if (next_node(node,arg_no,arity,T1,temp_hreg))
      return 1;
    else 
      return 0;
  }
  else if (ISVAR(arg1)) {
    if (first_node(node,arg_no,arity,temp_hreg)) 
      return 1;
    else 
      return 0;
  }
  else {
    node1 = locate_node(node,arg1);
    node = locate_node(node,arg2);
    if (next_to(node1,node)) {
      if (first_node(node,arg_no+1,arity,temp_hreg)) {
	ASSIGN_W_TEST(*((LONG_PTR)temp_hreg+arg_no),node->value);
	return 1;
      }
      else 
	return 0;
    }
    else 
      return 0;
  }
}

/***********************************************************
bt_prev3(T1,T2,T3)
T2 is the previous tuple of T1 that matches T3
***********************************************************/
int bt_prev3(T1,T2,T3)
LONG T1,T2,T3;
{
  SYM_REC_PTR psc_ptr;
  register NODE_PTR node,node1;
  register LONG arg1,arg2;
  register int arg_no,arity,i;
  LONG temp_hreg;

  psc_ptr = GET_STR_SYM_REC(T3);
  if (GET_ETYPE(psc_ptr)!=T_BOOT) {
    /*printf("%s is not a Boolean table\n",GET_NAME(psc_ptr));*/
    exception = illegal_predicate;
    return -1;
  }
  node = (NODE_PTR)GET_EP(psc_ptr);
  if (node==NULL || node->count==0)
    return 0;
  arity = GET_ARITY(psc_ptr);
  if (ISNONVAR(T2)) {
    printf("*** error : the second argument of next is nonvar\n");
    return 0;
  }
  FOLLOW(T2) = ADDTAG(hreg,STR);
  PUSHTRAIL(T2);
  temp_hreg = (LONG)hreg;
  NEW_HEAP_NODE((LONG)psc_ptr);
  for (i=1;i<=arity;i++,hreg++) {
    MAKE_FREE(LONG,*hreg);
  }

  arg_no = 1;
  UNTAG(T1);
  UNTAG(T3);
  arg1 = *((LONG_PTR)T1+arg_no);
  arg2 = *((LONG_PTR)T3+arg_no);
  DEREF(arg1);
  DEREF(arg2);
  while (ISNONVAR(arg1) && arg1== arg2) {
    node = locate_node(node,arg1);
  if (node==NULL || node->count==0)
    return 0;
    ASSIGN_W_TEST(*((LONG_PTR)temp_hreg+arg_no),arg1);
    arg_no++;
    arg1 = *((LONG_PTR)T1+arg_no);
    arg2 = *((LONG_PTR)T3+arg_no);
    DEREF(arg1);
    DEREF(arg2);
  }
  if (ISVAR(arg2)) {
    if (prev_node(node,arg_no,arity,T1,temp_hreg))
      return 1;
    else 
      return 0;
  }
  else if (ISVAR(arg1)) {
    if (last_node(node,arg_no,arity,temp_hreg)) 
      return 1;
    else 
      return 0;
  }
  else {
    node1 = locate_node(node,arg1);
    node = locate_node(node,arg2);
    if (next_to(node,node1)) {
      if (last_node(node,arg_no+1,arity,temp_hreg)) {
	ASSIGN_W_TEST(*((LONG_PTR)temp_hreg+arg_no),node->value);
	return 1;
      }
      else 
	return 0;
    }
    else 
      return 0;
  }
}

next_node(parent,arg_no,arity,T1,T2)
NODE_PTR parent;
int arg_no,arity;
LONG T1,T2;
{
  LONG arg;
  register NODE_PTR node;

  arg = *((LONG_PTR)T1+arg_no);
  DEREF(arg);
  if (ISVAR(arg)) 
    return 0;

  node = locate_node(parent,arg);
  if (node==NULL || node->count==0)
      return 0;
  if (arg_no==arity) {
    node = node->succ;
    if (node==NULL)
      return 0;
    ASSIGN_W_TEST(*((LONG_PTR)T2+arg_no),node->value);
    return 1;
  }
  
  if (next_node(node,arg_no+1,arity,T1,T2)) {
    ASSIGN_W_TEST(*((LONG_PTR)T2+arg_no),node->value);
    return 1;
  }
  else {
    node = node->succ;
    while (node!=NULL) {
      if (first_node(node,arg_no+1,arity,T2)) {
	ASSIGN_W_TEST(*((LONG_PTR)T2+arg_no),node->value);
	return 1;
      }
      node = node->succ;
    }
    return 0;
  }
}

int next_to(node1,node2)
NODE_PTR node1,node2;
{
  while (node1 != NULL && node1 != node2) {
    node1 = node1->succ;
  }
  if (node1==node2)
    return 1;
  else
    return 0;
}

prev_node(parent,arg_no,arity,T1,T2)
NODE_PTR parent;
int arg_no,arity;
LONG T1,T2;
{
  LONG arg;
  register NODE_PTR node;

  arg = *((LONG_PTR)T1+arg_no);
  DEREF(arg);
  if (ISVAR(arg)) 
    return 0;

  node = locate_node(parent,arg);
  if (node==NULL || node->count==0)
      return 0;
  if (arg_no==arity) {
    node = node->prev;
    if (node==NULL)
      return 0;
    ASSIGN_W_TEST(*((LONG_PTR)T2+arg_no),node->value);
    return 1;
  }
  
  if (prev_node(node,arg_no+1,arity,T1,T2)) {
    ASSIGN_W_TEST(*((LONG_PTR)T2+arg_no),node->value);
    return 1;
  }
  else {
    node = node->prev;
    while (node!=NULL) {
      if (last_node(node,arg_no+1,arity,T2)) {
	ASSIGN_W_TEST(*((LONG_PTR)T2+arg_no),node->value);
	return 1;
      }
      node = node->prev;
    }
    return 0;
  }
}

/***********************************************************
bt_set_true1(T)
***********************************************************/
int bt_set_true1(T)
LONG T;
{
  SYM_REC_PTR psc_ptr;
  NODE_PTR node;
  int arg_no,arity,count;

  psc_ptr = GET_STR_SYM_REC(T);
  if (GET_ETYPE(psc_ptr)!=T_BOOT) {
    /* printf("%s is not a Boolean table\n",GET_NAME(psc_ptr)); */
    exception = illegal_predicate;
    return -1;
  }
  node = (NODE_PTR)GET_EP(psc_ptr);
  if (node==NULL)
    return 0;
  arity = GET_ARITY(psc_ptr);
  UNTAG(T);
  count = set_true1(node,1,arity,T);
  if (count!=0) {
    PUSHTRAILB(&(node->count),node->count);
    node->count += count;
  }
  return 1;
}

int set_true1(node,arg_no,arity,T) 
NODE_PTR node;
LONG T;
int arg_no,arity;
{
  int count;
  LONG arg;
  NODE_PTR chd,temp;

  arg = *((LONG_PTR)T+arg_no);
  DEREF(arg);


  if (ISATOM(arg) || ISINT(arg)) {
    chd = locate_node(node,arg);
    if (chd==NULL)
      return 0;
    if (arg_no==arity) {
      if (chd->count == 1)
	return 0;
      PUSHTRAILB(&(chd->count),chd->count);
      chd->count = 1;
      link_node(chd,node);
      return 1;
    }
    else {
      count = set_true1(chd,arg_no+1,arity,T);
      if (count==0) return 0;
      PUSHTRAILB(&(chd->count),chd->count);
      chd->count += count;
      if (chd->count==count) 
	link_node(chd,node);
      return count;
    }
  }
  else {   /* ISVAR(arg) */
    count = set_all_true(node,node->children);
    if (count!=0) {
      PUSHTRAILB(&(node->count),node->count);
      node->count = count;
    }
    return count;
  }
}

link_node(chd,node)
register NODE_PTR chd,node;
{
  register NODE_PTR temp;

  temp = node->last;
  if (temp==NULL) {
    PUSHTRAILB(&(node->last),(LONG)temp);
    node->last = chd;
  }

  temp = node->first;
  PUSHTRAILB(&(node->first),(LONG)temp);
  node->first = chd;

  PUSHTRAILB(&(chd->succ),(LONG)chd->succ);
  chd->succ = temp;

  PUSHTRAILB(&(chd->prev),(LONG)chd->prev);
  chd->prev = NULL;

  if (temp!=NULL) {
    PUSHTRAILB(&(temp->prev),(LONG)temp->prev);
    temp->prev = chd;
  }
}

int set_all_true(parent,children) 
NODE_PTR parent;
LONG_PTR children;
{
  register int i;
  int total_count,count;
  NODE_PTR node,prev;
  
  total_count = 0;
  while (children!=NULL) {
    for (i=0;i<=126;i++) {
      node = (NODE_PTR)*(children+i);
      if (node!=NULL) {
	if (node->count==0) {
	  if (node->children!=NULL) { /* node leaf */
	    count = set_all_true(node,node->children);
	    PUSHTRAILB(&(node->count),node->count);
	    node->count = count;
	    total_count += count;
	  }
	  else {   /* leaf */
	    PUSHTRAILB(&(node->count),node->count);
	    node->count = 1;
	    total_count += 1;
	  }
	  if (parent->first==NULL) {
	    PUSHTRAILB(&(parent->first),(LONG)NULL);
	    parent->first = node;
	  } 
	  prev = parent->last;
	  PUSHTRAILB(&(parent->last),(LONG)prev);
	  parent->last = node;
	  if (prev!=NULL) {
	    PUSHTRAILB(&(prev->succ),(LONG)prev->succ);
	    prev->succ = node;
	  }
	  PUSHTRAILB(&(node->prev),(LONG)node->prev);
	  node->prev = prev;
	}
      }
    }
    children = (LONG_PTR)*(children+127);
  }
  return total_count;
}
    
/***********************************************************
bt_set_false_1(T)
***********************************************************/
int bt_set_false_1(T) 
LONG T;
{
  SYM_REC_PTR psc_ptr;
  register NODE_PTR node;
  int arg_no,arity,count;
  LONG arg;

  psc_ptr = GET_STR_SYM_REC(T);
  if (GET_ETYPE(psc_ptr)!=T_BOOT) {
    /* printf("%s is not a Boolean table\n",GET_NAME(psc_ptr));*/
    exception = illegal_predicate;
    return -1;
  }
  node = (NODE_PTR)GET_EP(psc_ptr);
  if (node==NULL)
    return 0;
  if (node->count==0)
    return 1;
  arity = GET_ARITY(psc_ptr);
  UNTAG(T);
  arg = *((LONG_PTR)T+1);
  DEREF(arg);
  if (ISATOM(arg) || ISINT(arg)) {
    count = set_false1(node,arg,1,arity,T);
    if (count!=0) {
      PUSHTRAILB(&(node->count),node->count);
      node->count -= count;
    }
  }
  else {
    PUSHTRAILB(&(node->count),node->count);
    node->count = 0;
  }
  return 1;
}

/***********************************************************
bt_set_false_2(T,Tag)
***********************************************************/
int bt_set_false_2(T,Tag)
LONG T,Tag;
{
  SYM_REC_PTR psc_ptr;
  register NODE_PTR node;
  int arg_no,arity,count;
  LONG arg;

  psc_ptr = GET_STR_SYM_REC(T);
  if (GET_ETYPE(psc_ptr)!=T_BOOT) {
    /*printf("%s is not a Boolean table\n",GET_NAME(psc_ptr));*/
    exception = illegal_predicate;
    return -1;
  }
  node = (NODE_PTR)GET_EP(psc_ptr);
  if (node==NULL)
    return 0;
  if (node->count==0)
    return 1;
  arity = GET_ARITY(psc_ptr);
  UNTAG(T);
  arg = *((LONG_PTR)T+1);
  DEREF(arg);
  if (ISATOM(arg) || ISINT(arg)) {
    count = set_false1(node,arg,1,arity,T);
    if (count!=0) {
      PUSHTRAILB(&(node->count),node->count);
      node->count -= count;
      unify(Tag,MAKEINT(1));
    }
  }
  else {
    PUSHTRAILB(&(node->count),node->count);
    node->count = 0;
    unify(Tag,MAKEINT(1));
  }
  return 1;
}

int set_false1(node,arg,arg_no,arity,T) 
NODE_PTR node;
LONG arg,T;
int arg_no,arity;
{
  int count;
  NODE_PTR chd,temp;

  chd = locate_node(node,arg);
  if (chd==NULL)
    return 0;
  if (arg_no==arity) {
    if (chd->count == 0)
      return 0;
    PUSHTRAILB(&(chd->count),chd->count);
    chd->count = 0;
    unlink_node(chd,node);
    return 1;
  }
  arg_no += 1;
  arg = *((LONG_PTR)T+arg_no);
  DEREF(arg);
  if (ISVAR(arg)) {
    count = chd->count;
    if (count!=0) {
      PUSHTRAILB(&(chd->count),count);
      chd->count = 0;
      unlink_node(chd,node);
      return count;
    }
    else return 0;
  }
  count = set_false1(chd,arg,arg_no,arity,T);
  if (count==0) return 0;
  PUSHTRAILB(&(chd->count),(LONG)chd->count);
  chd->count -= count;
  if (chd->count==0) 
    unlink_node(chd,node);
  return count;
}

unlink_node(chd,node)
NODE_PTR chd,node;
{
  NODE_PTR temp;
  
  if (node->first==chd) {
    PUSHTRAILB(&(node->first),(LONG)node->first);
    node->first = chd->succ;
  }

  if (node->last==chd) {
    PUSHTRAILB(&(node->last),(LONG)node->last);
    node->last = chd->prev;
  }

  temp = chd->prev;
  if (temp!=NULL) {
    PUSHTRAILB(&(temp->succ),(LONG)temp->succ);
    temp->succ = chd->succ;
  }

  temp = chd->succ;
  if (temp!=NULL) {
    PUSHTRAILB(&(temp->prev),(LONG)temp->prev);
    temp->prev = chd->prev;
  }
   
} 

/***********************************************************
bt_true1(T)
***********************************************************/
int bt_true1(T1)
LONG T1;
{
  SYM_REC_PTR psc_ptr;
  register NODE_PTR node;
  int arg_no,arity;
  LONG arg;

  psc_ptr = GET_STR_SYM_REC(T1);
  if (GET_ETYPE(psc_ptr)!=T_BOOT) {
    /*printf("%s is not a Boolean table\n",GET_NAME(psc_ptr));*/
    exception = illegal_predicate;
    return -1;
  }
  node = (NODE_PTR)GET_EP(psc_ptr);
  if (node==NULL || node->count==0)
    return 0;
  arity = GET_ARITY(psc_ptr);

  arg_no = 1;
  UNTAG(T1);
  arg = *((LONG_PTR)T1+arg_no);
  DEREF(arg);
  while (ISNONVAR(arg) && arg_no<arity) {
    node = locate_node(node,arg);
    if (node==NULL || node->count==0)
      return 0;
    arg_no++;
    arg = *((LONG_PTR)T1+arg_no);
    DEREF(arg);
  }
  if (ISNONVAR(arg)) {
    node = locate_node(node,arg);
    if (node==NULL) return 0;
  }
  if (node->count==0)
    return 0;
  else
    return 1;
}

/***********************************************************
bt_false1(T)
***********************************************************/
int bt_false1(T1) 
LONG T1;
{
  SYM_REC_PTR psc_ptr;
  NODE_PTR node;
  int arg_no,arity;
  LONG arg;

  psc_ptr = GET_STR_SYM_REC(T1);
  if (GET_ETYPE(psc_ptr)!=T_BOOT) {
    /*printf("%s is not a Boolean table\n",GET_NAME(psc_ptr));*/
    exception = illegal_predicate;
    return -1;
  }
  node = (NODE_PTR)GET_EP(psc_ptr);
  if (node==NULL || node->count==0)
    return 1;
  arity = GET_ARITY(psc_ptr);

  arg_no = 1;
  UNTAG(T1);
  arg = *((LONG_PTR)T1+arg_no);
  DEREF(arg);
  while (ISNONVAR(arg) && arg_no<arity) {
    node = locate_node(node,arg);
    if (node==NULL || node->count==0)
      return 1;
    arg_no++;
    arg = *((LONG_PTR)T1+arg_no);
    DEREF(arg);
  }
  if (ISNONVAR(arg)) {
    node = locate_node(node,arg);
    if (node==NULL) return 1;
  }
  if (node->count==0)
    return 1;
  else
    return 0;
}

/***********************************************************
  bt_count2(T1,N)
***********************************************************/
int bt_count2(T1,T2) 
LONG T1,T2;
{
  SYM_REC_PTR psc_ptr;
  NODE_PTR node;
  int arg_no,arity,i;
  LONG arg;

  psc_ptr = GET_STR_SYM_REC(T1);
  if (GET_ETYPE(psc_ptr)!=T_BOOT) {
    /*printf("%s is not a Boolean table\n",GET_NAME(psc_ptr));*/
    exception = illegal_predicate;
    return -1;
  }
  node = (NODE_PTR)GET_EP(psc_ptr);
  if (node==NULL || node->count==0) {
    unify(T2,MAKEINT(0));
    return 1;
  }
  arity = GET_ARITY(psc_ptr);

  arg_no = 1;
  UNTAG(T1);
  arg = *((LONG_PTR)T1+arg_no);
  DEREF(arg);
  while (ISNONVAR(arg)) {
    node = locate_node(node,arg);
    if (node==NULL)
      return 1;
    if (node->count==0) {
      unify(T2,MAKEINT(0));
      return 1;
    }
    arg_no++;
    if (arg_no>arity) {
      return unify(T2,MAKEINT(node->count));  
    }
    arg = *((LONG_PTR)T1+arg_no);
    DEREF(arg);
  }
  for (i=arg_no+1;i<=arity;i++){
    arg = *((LONG_PTR)T1+i);
    DEREF(arg);
    if (!ISVAR(arg)){
      /*      printf("Wrong tuple pattern: bt_count\n"); */
      exception = illegal_predicate;
      return -1;                             
    }
  }
  return unify(T2,MAKEINT(node->count));  
}

print_bt_tree(){
  LONG T;
  SYM_REC_PTR psc_ptr;
  int arity;
  NODE_PTR root;
  
  T = ARG(1,1);
  DEREF(T);
  psc_ptr = GET_STR_SYM_REC(T);
  arity = psc_ptr->arity;
  root = (NODE_PTR)GET_EP(psc_ptr);
  if (root==NULL) return 0;
  print_bt_node(root);
}

print_bt_node(node)
NODE_PTR node;
{
  int chd_num,i;
  LONG_PTR children;
  NODE_PTR chd;
  
  print_node(node);
  children = node->children;
  if (children==NULL) return 1;
  
  while (*(children+127)!=(LONG)NULL) {
    for (i=0;i<=126;i++) {
      if (*(children+i)!=(LONG)NULL) 
	print_bt_node((NODE_PTR)*(children+i));
    }
    children = (LONG_PTR)*(children+127);
  }
  for (i=0;i<=126;i++) {
    if (*(children+i)!=(LONG)NULL) 
      print_bt_node((NODE_PTR)*(children+i));
  }
}

print_node(node)
NODE_PTR node;
{
  printf("==>print_node\n");
  printf("node->value=");write_term(node->value);printf("\n");
  printf("node->hashtab=%x\n",node->hashtab);
  printf("node->size=%x\n",node->size);
  printf("node->count=%x\n",node->count);
  printf("node->prev=%x\n",node->prev);
  printf("node->succ=%x\n",node->succ);
  printf("node->first=%x\n",node->first);
  printf("node->last=%x\n",node->last);
  printf("node->children=%x\n\n\n",node->children);
}

