/*
*   File   : table.c
*   Author : Neng-Fa ZHOU Copyright (C) All rights reserved
*   Updated  : April 1999
*   Purpose: Primitives on table area
*/
#include "basic.h"
#include "term.h"
#include "macro.h"
#include "frame.h"

extern char *malloc();
extern FILE *curr_out;
/****** memmory manager for the table area *****/
allocateTableArea(){
  tableArea = (LONG_PTR)malloc(tableAreaSize*sizeof(LONG));
  if (!tableArea) quit("Not enough core!\n");
  tableAreaBottom = tableArea+tableAreaSize;
  initializeTableArea();
}

initializeTableArea()
{
  tareg = tableArea;
}

  
/************ Primitives on numbered terms *********/
numberVarTerm(term,num)
TERM term;
int num;
{
  DEREF(term);
  if (TAG(term)==ATM || IsNumberedVar(term)) 
    return num;
  else if (ISVAR(term)){
    ASSIGN_TRAIL_VALUE(term,NumberVar(num));  /* Borrow the tag LST */
    return num+1;
  } else if (ISLIST(term)) {
    UNTAG(term);
    num = numberVarTerm(*(LONG_PTR)term,num);
    num = numberVarTerm(*((LONG_PTR)term+1),num);
    return num;
  } else if (ISSTRUCT(term)){
    int i,arity;
    UNTAG(term);
    arity = GET_ARITY((SYM_REC_PTR)FOLLOW(term));
    for (i = 1; i <= arity; i++){
      num = numberVarTerm(*((LONG_PTR)term + i),num);
    }
    return num;
  } else quit("Suspension variables cannot occur in tabled calls!\n");
}

numberVarTermTestGround(term,num,ground)
TERM term;
int num,*ground;

{
  DEREF(term);
  if (TAG(term)==ATM) return num;
  else if (IsNumberedVar(term)){*ground=0;return num;}
  else if (ISVAR(term)){
    ASSIGN_TRAIL_VALUE(term,NumberVar(num));  /* Borrow the tag LST */
    *ground=0;
    return num+1;
  } else if (ISLIST(term)) {
    UNTAG(term);
    num = numberVarTermTestGround(*(LONG_PTR)term,num,ground);
    num = numberVarTermTestGround(*((LONG_PTR)term+1),num,ground);
    return num;
  } else if (ISSTRUCT(term)){
    int i,arity;
    UNTAG(term);
    arity = GET_ARITY((SYM_REC_PTR)FOLLOW(term));
    for (i = 1; i <= arity; i++){
      num = numberVarTermTestGround(*((LONG_PTR)term + i),num,ground);
    }
    return num;
  } else quit("Suspension variables cannot occur in tabled calls!\n");
}

TERM unnumberVarTerm(term,varVector,maxVarNo)
TERM term;
LONG_PTR varVector;
LONG_PTR maxVarNo;
{
  LONG_PTR ptr;
  int varNo;

  DEREF_NONVAR(term);
  if (TAG(term)==ATM)
    return term;
  else if (IsNumberedVar(term)) {
    varNo = INTVAL(term);
    if (varNo>FOLLOW(maxVarNo)){
      FOLLOW(maxVarNo) = varNo;
      FOLLOW(varVector-varNo) = (LONG)hreg;
      NEW_HEAP_FREE;
      return FOLLOW(varVector-varNo);
    } else {
      ptr = (LONG_PTR)FOLLOW(varVector-varNo);
      if (ptr>hreg){ /* var on the local stack */
	FOLLOW(varVector-varNo) = (LONG)hreg;
	FOLLOW(ptr) = (LONG)hreg;
	NEW_HEAP_FREE;
      } 
      return FOLLOW(ptr);
    }
  }  else if (ISLIST(term)) {
    UNTAG(term);
    ptr = hreg; hreg += 2;
    QUIT_IF_OVERFLOW_HEAP(1000);
    FOLLOW(ptr) = unnumberVarTerm(*(LONG_PTR)term,varVector,maxVarNo);
    FOLLOW(ptr+1) = unnumberVarTerm(*((LONG_PTR)term+1),varVector,maxVarNo);
    return ADDTAG(ptr,LST);
  } else { /* is structure */
    int i,arity;
    SYM_REC_PTR sym_ptr;
    UNTAG(term);
    sym_ptr = (SYM_REC_PTR)FOLLOW(term);
    arity = GET_ARITY(sym_ptr);
    ptr = hreg; hreg += arity+1;
    QUIT_IF_OVERFLOW_HEAP(1000);
    FOLLOW(ptr) = (LONG)sym_ptr;
    for (i = 1; i <= arity; i++){
      FOLLOW(ptr+i) = unnumberVarTerm(*((LONG_PTR) term + i),varVector,maxVarNo);
    }
    return ADDTAG(ptr,STR);
  }
}

TERM copyNumberedTerm(term)
TERM term;
{
  LONG_PTR ptr;
  
  DEREF_NONVAR(term);
  if (TAG(term)==ATM || IsNumberedVar(term)) 
    return term;
  else if (ISLIST(term)) {
    UNTAG(term);
    ptr = hreg; hreg += 2;
    QUIT_IF_OVERFLOW_HEAP(1000);
    FOLLOW(ptr) = copyNumberedTerm(*(LONG_PTR)term);
    FOLLOW(ptr+1) = copyNumberedTerm(*((LONG_PTR)term+1));
    return ADDTAG(ptr,LST);
  } else { /* is structure */
    int i,arity;
    SYM_REC_PTR sym_ptr;
    UNTAG(term);
    sym_ptr = (SYM_REC_PTR)FOLLOW(term);
    arity = GET_ARITY(sym_ptr);
    ptr = hreg; hreg += arity+1;
    QUIT_IF_OVERFLOW_HEAP(1000);
    FOLLOW(ptr) = (LONG)sym_ptr;
    for (i = 1; i <= arity; i++){
      FOLLOW(ptr+i) = copyNumberedTerm(*((LONG_PTR) term + i));
    }
    return ADDTAG(ptr,STR);
  }
}

TERM copyNumberedTermToTable(term)
TERM term;
{
  LONG_PTR ptr;

  DEREF_NONVAR(term);
  if (TAG(term)==ATM || IsNumberedVar(term)) return term;
  if (ISLIST(term)){
    UNTAG(term);
    AllocateFromTableArea(ptr,2);
    FOLLOW(ptr) = copyNumberedTermToTable(FOLLOW(term));
    FOLLOW(ptr+1) = copyNumberedTermToTable(FOLLOW((LONG_PTR)term+1));
    return ADDTAG(ptr,LST);  
  }
  { /* struct */
    SYM_REC_PTR sym_ptr;
    int arity,i;

    UNTAG(term);
    sym_ptr = (SYM_REC_PTR)FOLLOW(term);
    arity = GET_ARITY(sym_ptr);
    AllocateFromTableArea(ptr,arity+1);
    FOLLOW(ptr) = (LONG)sym_ptr;
    for (i=1;i<=arity;i++){
      FOLLOW(ptr+i) = copyNumberedTermToTable(FOLLOW((LONG_PTR)term+i));
    }
    return ADDTAG(ptr,STR);
  }
}


unifyNumberedTerms(t1,t2)
TERM t1,t2;
{
  DEREF_NONVAR(t1);DEREF_NONVAR(t2);
  if (TAG(t1)==ATM || IsNumberedVar(t1) || IsNumberedVar(t2)) 
    return t1==t2;
  else if (ISLIST(t1)) {
    if (ISLIST(t2)){
      UNTAG(t1); UNTAG(t2);
      return (unifyNumberedTerms(*(LONG_PTR)t1,*(LONG_PTR)t2) &&
	      unifyNumberedTerms(*((LONG_PTR)t1+1),*((LONG_PTR)t2+1)));
    } else return 0;
  } else { /* t1 is STRUCT */ 
    if (ISSTRUCT(t2)){
      int i,arity;
      UNTAG(t1); UNTAG(t2);
      if (FOLLOW(t1)!=FOLLOW(t2)) return 0;
      arity = GET_ARITY((SYM_REC_PTR)FOLLOW(t1));
      for (i = 1; i <= arity; i++){
	if (!unifyNumberedTerms(*((LONG_PTR)t1+i),*((LONG_PTR)t2+i))) return 0;
      }
      return 1;
    } else return 0;
  }
}
    
/************************ Primitives on frames *****************/
copyTabledCallArgs(src,des1,des2,arity,varVector)
LONG_PTR src,des1,des2,varVector;
int arity;
{
  int i, num = 1,maxVarNo=0,ground;
  LONG arg,top1;
  LONG_PTR treg0;

  treg0 = treg;
  
  for (i=0;i<arity;i++){  
    arg = FOLLOW(src-i);
    DEREF(arg);
    if (TAG(arg)==ATM){
      FOLLOW(des1+i) = arg;
      FOLLOW(des2-i) = arg;
    } else if (ISVAR(arg)){
      top1 = NumberVar(num); 
      ASSIGN_TRAIL_VALUE(arg,top1);

      /* build a new var */
      maxVarNo = num;
      FOLLOW(varVector-num) = (LONG)(des2-i);
      FOLLOW(des2-i) = (LONG)(des2-i);
      
      FOLLOW(des1+i) = top1;
      num++;
    } else {
      ground=1;
      num = numberVarTermTestGround(arg,num,&ground);
      FOLLOW(des1+i) = copyNumberedTermToTable(arg);
      if (ground) FOLLOW(des2-i) = arg;
      else FOLLOW(des2-i) = unnumberVarTerm(arg,varVector,&maxVarNo);
    }
  }
  /* undo numberVar*/
  UNDO_TRAILING;
}

tableHashValue(ptr,arity) /* ptr points to stack */
     LONG_PTR ptr;
     int arity;
{
  int i;
  SYM_REC_PTR	     psc_ptr;

  LONG term;
  int hvalue=0;
  for (i=0;i<arity;i++){
    term = FOLLOW(ptr-i);
    DEREF(term);
    if (TAG(term)==ATM) 
      hvalue += (term & 0x00ffffff)>>2;
    else if (IsNumberedVar(term))
      hvalue += INTVAL(term);
    else if (ISLIST(term)){
      UNTAG(term);
      hvalue += hashval(FOLLOW(term))+hashval(FOLLOW((LONG_PTR)term+1));
    }
    else {
      UNTAG(term);
      psc_ptr = (SYM_REC_PTR)FOLLOW(term);
      arity = GET_ARITY(psc_ptr);
      hvalue += ((LONG)psc_ptr &  0x00ffffff)>>2;
      for (i=1;i<=arity;i++){
	hvalue += hashval(FOLLOW((LONG_PTR)term+i));
      }
    }
  }
  return hvalue;
}

/******************* SUBGOAL TABLE ********************/
allocateSubgoalTable(){
  LONG i;

  AllocateFromTableArea(subgoalTable,subgoalTableBucketSize);
  for (i=0;i<subgoalTableBucketSize;i++){
    FOLLOW(subgoalTable+i) = (LONG)NULL;
  }
}

LONG_PTR registerAndLookupSubgoalTable(subgoalTableEntryPtr,sym,arity)
LONG_PTR subgoalTableEntryPtr;
SYM_REC_PTR sym;
int arity;
{
  LONG_PTR entryPtrPtr0,entryPtr,ptr1,ptr2;
  LONG t1,t2;
  int i,hvalue;
  
  ptr1 = GT_ARG_ADDR(subgoalTableEntryPtr);
  hvalue = (tableHashValue(ptr1+arity-1,arity)+(LONG)sym) % subgoalTableBucketSize;
  entryPtrPtr0  = subgoalTable+hvalue;
  entryPtr = (LONG_PTR)FOLLOW(entryPtrPtr0);

  while (entryPtr!=NULL){ /* lookup */
    if (GT_SYM(entryPtr)!=(LONG)sym) goto cont;
    
    ptr2 = GT_ARG_ADDR(entryPtr);
    for (i=0;i<arity;i++){
      t1 = FOLLOW(ptr1+i);  /* the same */
      t2 = FOLLOW(ptr2+i);
      if (t1==t2); 
      else if (ISCOMPOUND(t1) && unifyNumberedTerms(t1,t2));
      else goto cont;
    }
    return entryPtr;
  cont:
    entryPtr = (LONG_PTR)GT_NEXT(entryPtr);
  }
  GT_NEXT(subgoalTableEntryPtr) = FOLLOW(entryPtrPtr0); /* not found, then register */
  FOLLOW(entryPtrPtr0) = (LONG)subgoalTableEntryPtr;
  return NULL;
}

/********************** ANSWER TABLE *********************/   
LONG_PTR allocateAnswerTable(){
  LONG i,size;
  LONG_PTR tab;
  
  size = answerTableBucketSize+3;
  AllocateFromTableArea(tab,size);
  for (i=0;i<answerTableBucketSize;i++){
    FOLLOW(tab+i) = (LONG)NULL;
  }
  ANSWERTABLE_FIRST(tab) = (LONG)ANSWERTABLE_DUMMY_ADDR(tab);  
  ANSWERTABLE_LAST(tab) = AddRevisedBit(ANSWERTABLE_DUMMY_ADDR(tab)); 
  ANSWER_NEXT_IN_TABLE(ANSWERTABLE_DUMMY_ADDR(tab)) = (LONG)NULL;  
  return tab;
}

addAnswerAnswerTable(ar,arity)
LONG_PTR ar;
LONG arity;
{
  int i,hvalue,num=1;
  LONG_PTR tab,entry,answer,last_answer,ptr1,ptr2,treg0;
  LONG top1,t1,t2;
  
  tab = (LONG_PTR)AR_TABLE(ar); /* subgoal table */
  tab = (LONG_PTR)GT_ANSWERTABLE(tab); /* answer table */
  ptr1 = ar+arity;
  treg0 = treg;

/*  printf("arity=%d\n",arity); */
  for (i=0;i<arity;i++){  /* numberVar */
    t1 = FOLLOW(ptr1-i);DEREF(t1);
/*    printf("i=%d   ",i);myWriteTerm(t1);printf("\n"); */
    if (TAG(t1)==ATM || IsNumberedVar(t1));
    else if (ISVAR(t1)){
      ASSIGN_TRAIL_VALUE(t1,NumberVar(num));
      num++;
    } else num = numberVarTerm(t1,num);
  }
  
  hvalue = tableHashValue(ptr1,arity) % answerTableBucketSize;
  entry = tab+hvalue;
  answer = (LONG_PTR)FOLLOW(entry);
  while (answer!=NULL){
    ptr2 = ANSWER_ARG_ADDR(answer);
    for (i=0;i<arity;i++){
      t1 = FOLLOW(ptr1-i); DEREF_NONVAR(t1);
      t2 = FOLLOW(ptr2+i);
      if (t1==t2); 
      else if (ISCOMPOUND(t1) && unifyNumberedTerms(t1,t2));
      else goto cont;
    }
    UNDO_TRAILING; /* variants */
    return;
  cont:
    answer = (LONG_PTR)ANSWER_NEXT_IN_CHAIN(answer);
  }
  /* not found, then add it into the answer table */
  AllocateFromTableArea(answer,arity+2);
  ptr2 = ANSWER_ARG_ADDR(answer);
  for (i=0;i<arity;i++){
    t1 = FOLLOW(ptr1-i);DEREF(t1);
    if (TAG(t1)==ATM || IsNumberedVar(t1))
      FOLLOW(ptr2+i) = t1;
    else 
      FOLLOW(ptr2+i) = copyNumberedTermToTable(t1);
  }
  UNDO_TRAILING;
/*
  printAnswer(ANSWER_ARG_ADDR(answer),arity);
  */
  ANSWER_NEXT_IN_CHAIN(answer) = FOLLOW(entry);
  FOLLOW(entry) = (LONG)answer;
  ANSWER_NEXT_IN_TABLE(answer) = (LONG)NULL;
  
  last_answer = (LONG_PTR)RemoveRevisedBit(ANSWERTABLE_LAST(tab));
  ANSWER_NEXT_IN_TABLE(last_answer) = (LONG)answer;
  ANSWERTABLE_LAST(tab) = AddRevisedBit(answer);
}

int useAnswerAnswerTable(ar,arity,varVector)
LONG_PTR ar,varVector;
LONG arity;
{
  LONG_PTR ptr1,ptr2,answer,top;
  TERM t1,t2;
  int i,varNo=0;
  
  answer = (LONG_PTR)AR_CURRA(ar);
  answer = (LONG_PTR)ANSWER_NEXT_IN_TABLE(answer);
  if (answer==NULL) return NULL;
  ptr1 = ar+2*arity;
  ptr2 = ANSWER_ARG_ADDR(answer); 
  for (i=0;i<arity;i++){
    t1 = FOLLOW(ptr1-i);DEREF(t1); 
    if (TAG(t1)!=ATM) {
      t2 = FOLLOW(ptr2+i);
      t2 = unnumberVarTerm(t2,varVector,&varNo);
      unify(t1,t2);
    }
  }
  AR_CURRA(ar) = (LONG)answer;
  return 1;
}

/***************************************************************/
int c_VARIANT(){
  int i,maxVarNo=0;
  TERM op1, op2;
  LONG_PTR treg0 = treg;
  LONG top1;
  LONG varTable[10];

  op1 = ARG(1,2);
  op2 = ARG(2,2);
  
  numberVarTerm(op1,1);
  numberVarTerm(op2,1);
  op1 =unnumberVarTerm(op1,varTable,&maxVarNo);
  numberVarTerm(op1,1);
  maxVarNo=0;
  op2 =unnumberVarTerm(op2,varTable,&maxVarNo);
  numberVarTerm(op2,1);

  i = unifyNumberedTerms(op1,op2);
  while (treg != treg0){
    POPTRAIL(treg);
  }
  return i!=0;
}
  
int c_INITIALIZE_TABLE(){

  initializeTableArea();
  allocateSubgoalTable();
  return 1;
}
  
int c_TABLE_BUCKET_SIZES(){
  TERM gt_bucket_size,at_bucket_size;

  gt_bucket_size = ARG(1,2);DEREF(gt_bucket_size);
  at_bucket_size = ARG(2,2);DEREF(at_bucket_size);

  if (ISINT(gt_bucket_size))
    subgoalTableBucketSize = INTVAL(gt_bucket_size);
  else  
    unify(gt_bucket_size,MAKEINT(subgoalTableBucketSize));

  if (ISINT(at_bucket_size))
    answerTableBucketSize = INTVAL(at_bucket_size);
  else  
    unify(at_bucket_size,MAKEINT(answerTableBucketSize));
  
  return 1;
}



/*****************************************************************/
printTable(){
  int i;
  LONG_PTR subgoalTableEntryPtr;
  
  for (i=0;i<subgoalTableBucketSize;i++){
    subgoalTableEntryPtr = (LONG_PTR)FOLLOW(subgoalTable+i);
    if (subgoalTableEntryPtr!=NULL) printSubgoalTableEntry(subgoalTableEntryPtr);
  }
}

printSubgoalTableEntry(ptr)
LONG_PTR ptr;
{
  int i,arity;
  LONG_PTR tab,answer;

  SYM_REC_PTR sym;
  sym = (SYM_REC_PTR)GT_SYM(ptr);
  arity = GET_ARITY(sym);
  
  printAnswer(GT_ARG_ADDR(ptr),arity);
  tab = (LONG_PTR)GT_ANSWERTABLE(ptr);
  
  for (i=0;i<answerTableBucketSize;i++){
    answer = (LONG_PTR)FOLLOW(tab+i);
    while (answer!=NULL){
      printf("     ");printAnswer(ANSWER_ARG_ADDR(answer),arity);
      answer = (LONG_PTR)ANSWER_NEXT_IN_CHAIN(answer);
    }
  }
}
      
printAnswer(ptr,arity)
LONG_PTR ptr;
int arity;
{
  int i;
  LONG term;

  for (i=0;i<arity;i++){
    term = FOLLOW(ptr+i);
    myWriteTerm(term);
  }
  printf("\n");
}

printAnswerStack(ptr,arity)
LONG_PTR ptr;
int arity;
{
  int i;
  LONG term;

  for (i=0;i<arity;i++){
    term = FOLLOW(ptr-i);
    myWriteTerm(term);
  }
  printf("\n");
}

myWriteTerm(term)
LONG term;
{
  DEREF(term);
  if (ISVAR(term)) printf("_%x",term); 
  else if (TAG(term)==ATM) write_term(term);
  else if (TAG(term)==LST) {
    if (IsNumberedVar(term)) printf("%x ",term);
    else {
      UNTAG(term);
      printf(".(");
      myWriteTerm(FOLLOW(term));
      printf(",");
      myWriteTerm(FOLLOW((LONG_PTR)term+1));
      printf(")");
    }
  } else {
    SYM_REC_PTR sym_ptr;
    int arity,i;
    UNTAG(term);
    sym_ptr = (SYM_REC_PTR)FOLLOW(term);
    writepname(curr_out,GET_NAME(sym_ptr),GET_LENGTH(sym_ptr));
    if (GET_ARITY(sym_ptr)>0) {
      fprintf(curr_out,"(");
      arity = GET_ARITY(sym_ptr);
      for (i=1;i<=arity;i++) {
	myWriteTerm(*((LONG_PTR)term+i));
	if (i<arity)
	  fprintf(curr_out,",");
      }
      fprintf(curr_out,")");
    }
  }
}

isAncestor(chd_ar,anc_ar)
LONG_PTR chd_ar,anc_ar;
{
  while (chd_ar < anc_ar){
    chd_ar = (LONG_PTR)AR_AR(chd_ar);
  }
  return chd_ar==anc_ar;
}

printGoalTableARCPF(){
  int i;
  LONG_PTR subgoalTableEntryPtr;
  
  for (i=0;i<subgoalTableBucketSize;i++){
    subgoalTableEntryPtr = (LONG_PTR)FOLLOW(subgoalTable+i);
    while (subgoalTableEntryPtr!=NULL){
      printSubgoalTableAntryARCPF(subgoalTableEntryPtr);
      subgoalTableEntryPtr = (LONG_PTR)GT_NEXT(subgoalTableEntryPtr);
    }
  }
}

printSubgoalTableAntryARCPF(subgoalTableEntryPtr)
LONG_PTR subgoalTableEntryPtr;
{
  LONG_PTR ptr;
  
  ptr = (LONG_PTR)GT_AR(subgoalTableEntryPtr);
  if (ptr!=NULL){
    printf("entry=%x entry_ar=%x entry_ar_cpf=%x\n",subgoalTableEntryPtr,ptr,AR_CPF(ptr));
  }
}

