/* term.h */

#define REF               0x0   /* 00 */
#define LST               0x3   /* 11 */
#define STR               0x1   /* 01 */
#define SUSP              0x80000001
#define ATM               0x2   /* 10 */
#define TAG_MASK          0x80000003
#define INT_TAG           0x80000002
#define VAL_MASK          0x7ffffffc
#define VAL_MASK1         0xfffffffc
#define EPSILON           0.00001
#define MASK2             0x2
#define FFFF              0xffffffff
/****************************************************************************/
#define FOLLOW(op)        (*(LONG_PTR)(op))     /* return what op points to */
#define TAG(op)           (((LONG)op) & 0x3)
#define ISVAR(op)         (!(TAG(op)))
#define ISREF(op)         (!(TAG(op)))
#define ISNUM(op)         (ISINT(op) || ISFLOAT(op))
#define ISLIST(op)        (((op) & TAG_MASK) == LST)
#define ISSTRUCT(op)      ((TAG_MASK & (op)) == STR)
#define IS_SUSP_VAR(op)   (((op) & TAG_MASK) == SUSP)
#define ISINT(op)         (((op) & TAG_MASK) == INT_TAG)
#define ISFLOAT(op)       (ISSTRUCT(op) && FOLLOW(UNTAGGED(op))==(LONG)float_psc)
#define ISNONVAR(op)      (TAG(op))
#define ISFREE(var)       ((LONG)var == FOLLOW(var)) /* must be a ref */
#define ISATOM(op)        (((op) & TAG_MASK) == ATM)
#define ISNIL(op)         ((op) == nil_sym)
#define ISNIL1(op)        ((op) == nil_sym1)
#define ISCOMPOUND(op)    (((op) & 0x1) == STR && (op)>0)
#define IS_VAR_OR_STRUCT(op) (((op) & MASK2) == 0)

#define MAKEINT(op)       (((LONG)(op) << 2) | INT_TAG)
#define INTVAL(op)        (((LONG)(op) <<1) >> 3)
#define NUMVAL(op)        (ISINT(op) ? INTVAL(op) : floatval(op))

#define ADDTAG(op,tag)    ((LONG)(op) | tag)

/*
#define POS_OVERFLOW(op)  ((LONG)(op) & 0xf8000000)
#define NEG_OVERFLOW(op)  (((LONG)(op) & 0xf8000000) != 0xf8000000)
#define POS_ANS(op1,op2)  (((LONG)(op1)&0x40000000) == ((LONG)(op2)&0x40000000))
*/
#define UNTAG(op)         ((op) &= 0x7ffffffc) /* remove the tag from op */
#define UNTAGGED(op)      (((LONG)op) & 0x7ffffffc) /* return val of op w/out tag */

#define GET_ATM_SYM_REC(op)    ((SYM_REC_PTR)UNTAGGED(op))
#define GET_STR_SYM_REC(op)    ((SYM_REC_PTR)(FOLLOW(UNTAGGED(op))))
#define GET_SYM_REC(op)    (ISATOM(op) ? ((SYM_REC_PTR)UNTAGGED(op)) : ((SYM_REC_PTR)(FOLLOW(UNTAGGED(op)))))

#define GET_SYM_ARITY(op)  GET_ARITY(GET_SYM_REC(op))
#define GET_STR_SYM_ARITY(op)  GET_ARITY(GET_STR_SYM_REC(op))

#define GET_SYM_LENGTH(op) GET_LENGTH(GET_SYM_REC(op))

#define DEREF(op)        while (ISVAR(op)) {              \
                            top =  (LONG_PTR)FOLLOW(op);  \
                            if ((LONG)top==op)            \
			       break;                     \
 	                    op = (LONG)top; \
		         }

#define DEREF_NONVAR(op) while (ISVAR(op)) \
                            op = FOLLOW(op)  \

#define NDEREF(op,labl)  top = (LONG_PTR)FOLLOW(op);    \
			 if ((LONG)top!=op) { \
                            op = (LONG)top;   \
			    goto labl;        \
			 }

/****************************************************************************/
/* The following are macros for setting heap values. */

#define MAKE_FREE(type,var)  (var) = (type)&(var)
/* must pass a simple pointer, not an expression */

#define NEW_HEAP_FREE *hreg = (LONG)hreg; hreg++

/* make a free variable on the top of the heap */

#define NEW_HEAP_NODE(x)          *hreg++ = x
/* make a new heap node with value x (one word type) */

/*****************************************************************************/
#define GET_ETYPE(ptr)  ((ptr)->etype)
#define GET_ARITY(ptr)  ((ptr)->arity)
#define GET_LENGTH(ptr) ((ptr)->length)
#define GET_EP(ptr)     ((ptr)->ep)
#define GET_NAME(ptr)   ((ptr)->nameptr)
#define GET_NEXT(ptr)   ((ptr)->next)

#define IS_PRED(sym)    (GET_ETYPE(sym) == T_PRED)
#define IS_CPRED(sym)   (GET_ETYPE(sym) == C_PRED)
#define IS_DYNA(sym)    (GET_ETYPE(sym) == T_DYNA)
#define IS_ORDI(sym)    (GET_ETYPE(sym) == T_ORDI)

#define IHASH(val,size)  (((unsigned long int)val >>2 ) % size)

#define ALIGN(type,ptr)  ptr = (type)(((LONG)ptr + 3) & 0xfffffffc)

#define ASSIGN_f(op,value) FOLLOW(op)=value

#define SWITCH_OP(op,nderef,VarCode,AtmCode,LstCode,StrCode,SuspCode) \
 nderef: switch (TAG(op)) { \
	    case REF: \
	      NDEREF(op,nderef); \
   	      VarCode; \
              break; \
            case ATM: \
	      AtmCode; \
              break; \
	    case LST: \
	      LstCode; \
              break; \
	    case STR: \
	      if (op>0) StrCode \
              else SuspCode;break;}

#define SWITCH_OP_STRUCT(op,nderef,VarCode,StrCode,SuspCode) \
    nderef:if (ISSTRUCT(op)) \
              StrCode \
           else if (ISVAR(op)){ \
 	      NDEREF(op,nderef); \
   	      VarCode;} \
           else if (op<0 && TAG(op) == STR) \
              SuspCode; \

#define SWITCH_OP_ATM(op,nderef,VarCode,AtmCode,SuspCode) \
 nderef: if (IS_VAR_OR_STRUCT(op)){ \
           if (ISVAR(op)){ \
	      NDEREF(op,nderef); \
   	      VarCode;} \
           else if (op<0) \
              SuspCode;} \
           AtmCode

#define SWITCH_OP_LST(op,nderef,VarCode,LstCode,SuspCode) {\
 nderef: if (ISVAR(op)){ \
            NDEREF(op,nderef); \
   	    VarCode;} \
         else if (ISLIST(op)) \
            LstCode \
         else if (IS_SUSP_VAR(op)) \
            SuspCode;}

#define SWITCH_OP_CONS(op,nderef,VarCode,LstCode,SuspCode) {\
nderef:   if (ISLIST(op)) \
            LstCode; \
          if (ISVAR(op)){ \
            NDEREF(op,nderef); \
   	    VarCode;} \
          else if (IS_SUSP_VAR(op)) \
            SuspCode;}

#define SWITCH_OP_NIL(op,nderef,VarCode,NilCode,SuspCode) \
 nderef: if (ISVAR(op)){ \
            NDEREF(op,nderef); \
   	    VarCode;} \
         else if (ISNIL1(op)) \
            NilCode \
         else if (IS_SUSP_VAR(op)) \
            SuspCode

#define SWITCH_OP_LST_NIL(op,nderef,VarCode,LstCode,NilCode,SuspCode) \
 nderef: if (ISLIST(op)) \
            LstCode \
         else if (ISNIL1(op)) \
            NilCode \
         else if (ISVAR(op)){ \
            NDEREF(op,nderef); \
   	    VarCode;} \
         else if (IS_SUSP_VAR(op)) \
            SuspCode

#define SWITCH_OP_VAR(op,nderef,VarCode,SuspCode,OtherCode) \
 nderef: if (IS_VAR_OR_STRUCT(op)){ \
             if (ISVAR(op)){ \
	        NDEREF(op,nderef); \
   	        VarCode;} \
             else if (op<0) \
              SuspCode \
             else OtherCode;} \
          else OtherCode


#define TRIGGER_ON 0x1
#define INTERRUPT 0x2
/*
#define SHIFT_TRIGGERS { \
  int i; \
  for (i=trigger_no;i>0;i--) { \
      triggers[i+1]=triggers[i]; \
      event_flag[i+1] = event_flag[i]; \
 }}
*/
#define INSERT_TRIGGER_var_ins(op) {\
  toam_signal_vec |= TRIGGER_ON; \
  triggers[++trigger_no] = (LONG_PTR)(op); \
  event_flag[trigger_no] = EVENT_VAR_INS; }

#define INSERT_TRIGGER_dvar_ins(op) {\
  toam_signal_vec |= TRIGGER_ON; \
  triggers[++trigger_no] = (LONG_PTR)(op); \
  event_flag[trigger_no] = EVENT_DVAR_INS; }

#define INSERT_TRIGGER_min(op) \
  if (DV_min_cs(op) != (LONG)NULL) { \
    toam_signal_vec |= TRIGGER_ON; \
    triggers[++trigger_no] = (LONG_PTR)(op); \
    event_flag[trigger_no] = EVENT_DVAR_MIN; }

#define INSERT_TRIGGER_max(op) \
  if (DV_max_cs(op) != (LONG)NULL) { \
    toam_signal_vec |= TRIGGER_ON; \
    triggers[++trigger_no] = (LONG_PTR)(op); \
    event_flag[trigger_no] = EVENT_DVAR_MAX; }

#define INSERT_TRIGGER_dom(op) \
  if (DV_dom_cs(op) != (LONG)NULL) { \
    toam_signal_vec |= TRIGGER_ON; \
    triggers[++trigger_no] = (LONG_PTR)(op); \
    event_flag[trigger_no] = EVENT_DVAR_DOM; }

/***************************************************************************/
/* Macros for manipulating domain variables */
#define A_DV_var(dv_ptr) dv_ptr
#define DV_var(dv_ptr) FOLLOW(dv_ptr)

#define A_DV_type(dv_ptr) (dv_ptr+1)
#define DV_type(dv_ptr) FOLLOW(dv_ptr+1)

#define A_DV_ins_cs(dv_ptr) (dv_ptr+2)
#define DV_ins_cs(dv_ptr) FOLLOW(dv_ptr+2)

#define A_DV_dcs_no(dv_ptr) (dv_ptr+3)
#define DV_dcs_no(dv_ptr) FOLLOW(dv_ptr+3)

#define A_DV_min_cs(dv_ptr) (dv_ptr+4)
#define DV_min_cs(dv_ptr) FOLLOW(dv_ptr+4)

#define A_DV_max_cs(dv_ptr) (dv_ptr+5)
#define DV_max_cs(dv_ptr) FOLLOW(dv_ptr+5)

#define A_DV_dom_cs(dv_ptr) (dv_ptr+6)
#define DV_dom_cs(dv_ptr) FOLLOW(dv_ptr+6)

#define A_DV_size(dv_ptr) (dv_ptr+7)
#define DV_size(dv_ptr) FOLLOW(dv_ptr+7)

#define A_DV_first(dv_ptr) (dv_ptr+8)
#define DV_first(dv_ptr) FOLLOW(dv_ptr+8)

#define A_DV_last(dv_ptr) (dv_ptr+9)
#define DV_last(dv_ptr) FOLLOW(dv_ptr+9)

#define A_DV_delta(dv_ptr) (dv_ptr+10)
#define DV_delta(dv_ptr) FOLLOW(dv_ptr+10)

#define A_DV_low_val(dv_ptr) (dv_ptr+11)
#define DV_low_val(dv_ptr) FOLLOW(dv_ptr+11)

#define A_DV_up_val(dv_ptr) (dv_ptr+12)
#define DV_up_val(dv_ptr) FOLLOW(dv_ptr+12)

#define A_DV_base_ptr(dv_ptr) (dv_ptr+13)
#define DV_base_ptr(dv_ptr) FOLLOW(dv_ptr+13)

#define SIZE_OF_DV 14

#define BV_DOMAIN  -1
#define IT_DOMAIN   0
#define UN_DOMAIN   1

#define IS_BV_DOMAIN(var)  (DV_type(var)<0)
#define IS_IT_DOMAIN(var)  (DV_type(var)==0)
#define IS_UN_DOMAIN(var)  (DV_type(var)==1)

#define SHOW_SUSP(ar) { \
		       sym_ptr = (SYM_REC_PTR)FOLLOW(UNTAGGED(FOLLOW(ar-4))+2*sizeof(LONG)); \
		       printf("%s\n",GET_NAME(sym_ptr)); \
                       for (i = GET_ARITY(sym_ptr); i>0; i--){ \
                       printf("Y%i    = %x  ", i,*(ar+i)); write_term(*(ar+i));printf("\n"); \
		       } \
                        printf("AR    = %x\n", ar); \
                        printf("(AR)  = %x\n", *ar); \
                        printf("CPS   = %x\n", *(ar-1)); \
                        printf("TOP   = %x\n", *(ar-2)); \
                        printf("BTM   = %x\n", *(ar-3)); \
                        printf("REEP   = %x\n",*(ar-4)); \
                        printf("Prev  = %x\n", *(ar-5)); \
		      }


#define SHOW_FRAME(ar) { \
   LONG_PTR new_top; \
   printf("------- %x-------\n",ar); \
   printf("AR_BTM=%x \n ",FOLLOW(ar-3)); \
   for (new_top = (LONG_PTR)FOLLOW(ar-3); new_top>ar; new_top--){ \
       printf("%x  ",FOLLOW(new_top)); printf("\n"); \
    } \
    printf("*");printf("%x",FOLLOW(ar)); printf("*\n"); \
   for (new_top = ar-1; new_top>(LONG_PTR)FOLLOW(ar-2);new_top--){ \
      printf("%x\n",FOLLOW(new_top));} \
   printf("**************\n"); }

#define SHOW_NAR(reg)   printf("AR    = %x\n", reg); \
                        printf("(AR)  = %x\n", *reg); \
                        printf("CPS   = %x\n", *(reg-1)); \
                        printf("B     = %x\n", *(reg-2)); \
                        printf("H     = %x\n", *(reg-3)); \
                        printf("T     = %x\n", *(reg-4)); \
                        printf("CPF   = %x\n", *(reg-5)); \
                        printf("TOP  = %x\n", *(reg-6)); \
                        for (top1=(LONG)(reg-7);top1>= *(reg-6);top1 -= 4) printf("%x\n",*(LONG_PTR)top1)

#define SHOW_HEAP       tempreg = heap_bottom; \
                        printf("heap\n");      \
                        while (tempreg<hreg) {  \
                            printf("%x == %x\n",tempreg,*tempreg); \
			    tempreg++; } 



