/*
*   File   : toam.c
*   Author : Neng-Fa ZHOU Copyright (C) All rights reserved
*   Updated  : December 1998
*   Purpose: Emulator of ATOAM
*/

#include "basicd.h"
#include "term.h"
#include "inst.h"
#include "frame.h"
#include "toam.h"
#include <signal.h>
#include <setjmp.h>

/*
#define TRACE_TOAM
#define DEBUG_DELAY
#define DEBUG_CALL
#define DEBUG_INST
*/

extern int (*builtins[])();
extern double   floatval();
extern LONG     encodefloat();
extern int lookup_dm_elm(),exclude_one_elm_bv();
extern LONG_PTR registerAndLookupSubgoalTable();
extern LONG_PTR allocateAnswerTable();

int bprolog_initialized = 0;
jmp_buf restart_env;

void exception_handler(signo)
int signo;
{
  switch (signo) {
  case SIGINT:
    signal(SIGINT,exception_handler);
/*    longjmp(restart_env,-1); */
    toam_signal_vec |= INTERRUPT;
    break;
  default:
    printf("signal not handled: signo=%d\n",signo);
    exit(2);
    break;
  }
}

void exception_handler_halt(signo)
int signo;
{
  switch (signo) {
  case SIGINT:
    exit(0);
    break;
  default:
    printf("signal not handled: signo=%d\n",signo);
    exit(2);
    break;
  }
}

is_near_ancestor(ar1,ar2)
register LONG_PTR ar1,ar2;
{
  if (ar1==ar2) return 1;
  ar2 = (LONG_PTR)AR_AR(ar2);
  if (ar1==ar2) return 1;
  ar2 = (LONG_PTR)AR_AR(ar2);
  if (ar1==ar2) return 1;
  return 0;
}

void init_signals_internal(){
  toam_signal_vec = 0;
#ifdef WIN32
  signal(SIGINT,exception_handler_halt);
#else
  if (signal(SIGINT,exception_handler_halt)==SIG_ERR)
    printf("can't catch SIGINT\n");
#endif
}

void init_signals(){
  toam_signal_vec = 0;
#ifdef WIN32
  signal(SIGINT,exception_handler);
#else
  if (signal(SIGINT,exception_handler)==SIG_ERR)
    printf("can't catch SIGINT\n");
#endif
}

initialize_bprolog(argc,argv)
int argc;
char *argv[];
{
  LONG_PTR my_inst_begin;
  if (bprolog_initialized==1) return;
  /*  print_argv(argc,argv); */
  init_toam(argc, argv);	/* set up memory according to arguments */
  init_load_routine();		/* set up array of pointers to functions */
  file_init();		/* set up array of pointers to functions */
  init_loading(argc, argv);	
/*  printf("after init_loading\n");*/
  init_builtins();
  init_signals(); 
  Cboot();
  bprolog_initialized = 1;
}

print_argv(argc,argv)
int argc;
char *argv[];
{
  int i;
  for (i=0;i<argc;i++){
    printf("%s ",argv[i]);
  }
  printf("\n");
}

int start_toam(){
/*  if (setjmp(restart_env)==0){ */
  toam0.entry = 0;
  return toam(inst_begin,local_top,arreg,breg,hbreg,hreg,treg,sfreg);
/*  } else { 
    breg = (LONG_PTR)UNTAGGED((LONG)GET_EP(interrupt_psc));
    arreg = breg;
    local_top = (LONG_PTR)*(arreg-2);
    hbreg = (LONG_PTR)*(arreg-6);
    inst_begin = (LONG_PTR)*(arreg-5);
    treg = trail_bottom - 1;
    sfreg = (LONG_PTR)*(arreg-8);
    hreg = heap_bottom;
    printf("\n");
    return start_toam();
  }
*/
}

is_exit_point(opcode)
int opcode;
{
  if ((opcode >= call && opcode <=executev) ||
      (opcode >= allocate_flat && opcode <= allocate_nondet) ||
      opcode == allocate_susp ||
      (opcode >= move_struct_x  && opcode <= move_yw_yy ) ||
      (opcode >= get_ar_cps && opcode <= jmpn_det_get_ar_cps) ||
      (opcode >= unify_arg_nil && opcode <= unify_arg_iii))
    return 1;
  else 
    return 0;
}

int toam(P,LOCAL_TOP,AR,B,HB,H,T,SF)
register LONG_PTR P,LOCAL_TOP,AR,B,HB,H,T,SF;
{
  register LONG op1;
  register LONG_PTR top,sreg; /* sreg also used to save AR(AR) */
  register LONG     
    op2;                 
  register LONG  tmp_op,op3; /* op3 also used to save CPS(AR) */        
  register LONG nil_sym1;
  register SYM_REC_PTR     sym_ptr;
  register LONG   i;
  register int   arity;
  register LONG_PTR dv_ptr,trigger_var;
  LONG op4;
  LONG goal;
  SYM_REC_PTR insert(),insert_cfunc();
  LONG_PTR status_ptr;

  nil_sym1 = nil_sym;
  
#ifdef ToamProfile
Cboot_profile();
#endif

  /**************************************************************************/
  goto contcase_read;
 contcase_write:

  UPDATE_SPACE_COUNTERS(LOCAL_TOP,H,T);
#ifdef DEBUG_INST
    printf("%x\t", P); printf(" T=%x  H=%x HB=%x ",T,H,HB);
    printf("%s\n",xxopcode[*P].name);    
#endif
#ifdef TRACE_TOAM
  if (trace_toam) {
      printf("%x\t", P); 
      printf("%s\n",xxopcode[*P].name);    
  }
#endif
  switch (*P++) {		/* switch on current opcode */
    
    /* Conditional jump */
  CASE(noop)
    CONTCASE_R;
    
  CASE(jmp)
    P = (LONG_PTR)*P;
    OVERFLOW_C;
    CATCH_WAKE_EVENT;
    CONTCASE_R;
    
  CASE(djmp)
    P = (LONG_PTR)*P;
    CONTCASE_R;
    
  CASE(jmpn_eq_struct_x)
    goto r_jmpn_eq_struct_x;

  CASE(jmpn_eq_struct_y)
    goto r_jmpn_eq_struct_y;

  CASE(jmpn_eq_atom_x)
    goto r_jmpn_eq_atom_x;

  CASE(jmpn_eq_atom_y)
    goto r_jmpn_eq_atom_y;
    
  CASE(jmpn_nil_x) 
    goto r_jmpn_nil_x;

  CASE(jmpn_nil_y) 
    goto r_jmpn_nil_y;
    
  CASE(switch_list_x) 
    goto r_switch_list_x;

  CASE(switch_list_y) 
    goto r_switch_list_y;

  CASE(switch_list_yxx)
    goto r_switch_list_yxx;

  CASE(switch_list_yxy)
    goto r_switch_list_yxy;

  CASE(switch_list_yyx)
    goto r_switch_list_yyx;

  CASE(switch_list_yyy)
    goto r_switch_list_yyy;

  CASE(jmpn_eq_int_x) 
    goto r_jmpn_eq_int_x;

  CASE(jmpn_eq_int_y) 
    goto r_jmpn_eq_int_y;

  CASE(jmpn_eql) 
    goto r_jmpn_eql;
    
  CASE(jmp_eql) 
    goto r_jmp_eql;

  CASE(jmp_eql_yy) 
    goto r_jmp_eql_yy;

  CASE(jmpn_gt) 
    goto r_jmpn_gt;

  CASE(jmpn_gt_yy) 
    goto r_jmpn_gt_yy;

  CASE(jmpn_ge) 
    goto r_jmpn_ge;

  CASE(jmpn_ge_yy) 
    goto r_jmpn_ge_yy;

  CASE(jmpn_id) 
    goto r_jmpn_id;
    
  CASE(jmp_id) 
    goto r_jmp_id;

  CASE(jmpn_var_x) 
    goto r_jmpn_var_x;

  CASE(jmpn_var_y) 
    goto r_jmpn_var_y;

  CASE(jmp_var_x) 
    goto r_jmp_var_x;

  CASE(jmp_var_y) 
    goto r_jmp_var_y;

  CASE(jmpn_atom_x) 
    goto r_jmpn_atom_x;

  CASE(jmpn_atom_y) 
    goto r_jmpn_atom_y;

  CASE(jmpn_atomic_x) 
    goto r_jmpn_atomic_x;

  CASE(jmpn_atomic_y) 
    goto r_jmpn_atomic_y;
    
  CASE(jmpn_num_x) 
    goto r_jmpn_num_x;

  CASE(jmpn_num_y) 
    goto r_jmpn_num_y;

  CASE(jmpn_float_x) 
    goto r_jmpn_float_x;

  CASE(jmpn_float_y) 
    goto r_jmpn_float_y;
    
  CASE(jmpn_int_x) 
    goto r_jmpn_int_x;

  CASE(jmpn_int_y) 
    goto r_jmpn_int_y;

  CASE(hash_jmpn_nil) 
    goto r_hash_jmpn_nil;
  
  CASE(hash_jmpn_list)
    goto r_hash_jmpn_list;
    
  CASE(hash_jmpn_int) 
    goto r_hash_jmpn_int;
    
  CASE(hash_jmpn_atom)
    goto r_hash_jmpn_atom;

  CASE(hash_jmpn_struct)
    goto r_hash_jmpn_struct;

  CASE(hash_jmpn_struct_x) 
    goto r_hash_jmpn_struct_x;

  CASE(hash_jmpn_struct_y)
    goto r_hash_jmpn_struct_y;
  
  CASE(hash_jmpn_struct_xx)
    goto r_hash_jmpn_struct_xx;

  CASE(hash_jmpn_struct_xy)
    goto r_hash_jmpn_struct_xy;

  CASE(hash_jmpn_struct_yx)
    goto r_hash_jmpn_struct_yx;

  CASE(hash_jmpn_struct_yy)
    goto r_hash_jmpn_struct_yy;

    /* Unify */
  CASE(unify_struct_x) 
    goto r_unify_struct_x;
    
  CASE(unify_struct_y) 
    OPERAND_YW;
    UNIFY_STRUCT_Y(op1,op2,w_unify_struct_y,{toam_PUSHTRAIL(op1);},{},{goto backtrack;});
    
  CASE(unify_list_x)   
    goto r_unify_list_x;
    
  CASE(unify_list_y)   
    OPERAND_Y;
    UNIFY_LIST_Y(op1,w_unify_list_y,{toam_PUSHTRAIL(op1);},{},{goto backtrack;});
    
  CASE(unify_nil_x) 
    goto r_unify_nil_x;

  CASE(unify_nil_y) 
    goto r_unify_nil_y;

  CASE(unify_atom_x) 
  CASE(unify_int_x) 
    goto r_unify_atom_x;
    
  CASE(unify_atom_y) 
  CASE(unify_int_y) 
    OPERAND_YW;
    UNIFY_ATOMIC12_Y(op1,op2,w_unify_atom_y,{toam_PUSHTRAIL(op1);},{},{goto backtrack;});
    
  CASE(unify_ux_ux) 
    goto r_unify_ux_ux;
    
  CASE(unify_ux_uy) 
    goto r_unify_ux_uy;
    
  CASE(unify_uy_uy) 
    goto r_unify_uy_uy;

  CASE(unify_cons_x) 
    goto r_unify_cons_x;

  CASE(unify_cons_y) 
    goto r_unify_cons_y;

  CASE(fork_unify_struct_y)
    goto r_fork_unify_struct_y;

  CASE(fork_unify_list_y)
    goto r_fork_unify_list_y;
  
  CASE(fork_unify_nil_y)
    goto r_fork_unify_nil_y;

  CASE(fork_unify_atom_y)
  CASE(fork_unify_int_y)
    goto r_fork_unify_atom_y;

  CASE(fork_unify_uy_uy)
    goto r_fork_unify_uy_uy;

  /* Fork_unicut */
  CASE(fork_unicut_struct_y)
    goto r_fork_unicut_struct_y;

  CASE(fork_unicut_list_y)
    goto r_fork_unicut_list_y;
  
  CASE(fork_unicut_nil_y)
    goto r_fork_unicut_nil_y;

  CASE(fork_unicut_atom_y)
  CASE(fork_unicut_int_y)
    goto r_fork_unicut_atom_y;

  CASE(fork_unicut_uy_uy)
    goto r_fork_unicut_uy_uy;

    /* Unify0 */
  CASE(unify0_struct_y) 
    OPERAND_YW;
    UNIFY_STRUCT_Y(op1,op2,w_unify0_struct_y,{toam_PUSHTRAIL_UC_0(op1);},{},{goto backtrack0;});
    
  CASE(unify0_list_y)   
    OPERAND_Y;
    UNIFY_LIST_Y(op1,w_unify0_list_y,{toam_PUSHTRAIL_UC_0(op1);},{},{goto backtrack0;});

  CASE(unify0_nil_y) 
    goto r_unify0_nil_y;

  CASE(unify0_atom_y) 
  CASE(unify0_int_y) 
    OPERAND_YW;
    UNIFY_ATOMIC12_Y(op1,op2,w_unify0_atom_y,{toam_PUSHTRAIL_UC_0(op1);},{},{goto backtrack0;});
    
  CASE(unify0_uy_uy) 
    OPERAND_YY;
    SWITCH_OP(op1,w_unify0_uy_uy,
	      {goto unify_v_d;},
	      {UNIFY_ATOMIC21_Y(op2,op1,w_unify0_uy_uy_1,{toam_PUSHTRAIL_UC_0(op2);},{},{goto backtrack0;});},
	      {goto unify0_lst_d;},
	      {goto unify0_str_d;},
	      {goto unify_susp_d;});

    /* Unicut */
  CASE(unicut_struct_y)  
    goto r_unicut_struct_y;

  CASE(unicut_list_y)    
    goto r_unicut_list_y;

  CASE(unicut_nil_y) 
    goto r_unicut_nil_y;

  CASE(unicut_atom_y) 
  CASE(unicut_int_y) 
    goto r_unicut_atom_y;

  CASE(unicut_uy_uy)
    goto r_unicut_uy_uy;

  CASE(unicut)
    CONTCASE_R;
      
/* Unify argument */
  CASE(unify_arg_nil)  
    toam_NEW_HEAP_NODE(nil_sym1);
    CONTCASE_W;

  CASE(unify_arg_atom)  
  CASE(unify_arg_int)   
    OPERAND_W;
    toam_NEW_HEAP_NODE(op2);
    CONTCASE_W;
    
  CASE(unify_arg_ux_ux)    
    BUX;
    BUX;
    P += 4;  /* skip unify_arg_ux unify_arg_ux */
    CONTCASE_W;	
    
  CASE(unify_arg_ux)    
    BUX;
    CONTCASE_W;	

  CASE(unify_arg_ux_vx)
    BUX;
    BVX;
    P += 4;  /* skip unify_arg_ux unify_arg_vy */
    CONTCASE_W;	

  CASE(unify_arg_ux_vy)
    BUX;
    BVY;
    P += 4;  /* skip unify_arg_ux unify_arg_vy */
    CONTCASE_W;	

  CASE(unify_arg_uy_uy)    
    BUY(w_unify_arg_uy_uy_1);
    BUY(w_unify_arg_uy_uy_2);
    P += 4;  /* skip unify_arg_uy unify_arg_uy */
    CONTCASE_W;	
   
  CASE(unify_arg_uy)    
    BUY(w_unify_arg_uy);
    CONTCASE_W;

  CASE(unify_arg_wy)
    BWY;
    CONTCASE_W;	

  CASE(unify_arg_vx_vx)
    BVX;

  CASE(unify_arg_vx)    
    BVX;
    CONTCASE_W;

  CASE(unify_arg_vy_vy)
    BVY;

  CASE(unify_arg_vy)    
    BVY;
    CONTCASE_W;

  CASE(unify_arg_list)  
    BL;
    CONTCASE_W;

  CASE(unify_arg_struct)
    BS;
    CONTCASE_W;
    
  CASE(unify_arg_void_one) 
    toam_NEW_HEAP_FREE;
    CONTCASE_W;

  CASE(unify_arg_void)     
    op1 = *P++;
    while (op1>0) {
      toam_NEW_HEAP_FREE;
      op1--;
    }
    CONTCASE_W;

  CASE(unify_arg_vx_vy)
    BVX;
    BVY;
    CONTCASE_W;

  CASE(unify_arg_vx_ux)
    BVX;
    BUX;
    CONTCASE_W;

  CASE(unify_arg_vx_uy)
    BVX;
    BUY(w_unify_arg_vx_uy);
    CONTCASE_W;

  CASE(unify_arg_vy_vx)
    BVY;
    BVX;
    CONTCASE_W;

  CASE(unify_arg_vy_ux)
    BVY;
    BUX;
    CONTCASE_W;

  CASE(unify_arg_vy_uy)
    BVY;
    BUY(w_unify_arg_vy_uy);
    CONTCASE_W;

  CASE(unify_arg_iii)
    toam_NEW_HEAP_NODE(*P++);
    toam_NEW_HEAP_NODE(*P++);
    toam_NEW_HEAP_NODE(*P++);
    CONTCASE_W;
    
    /* Move */
  CASE(move_struct_x) 
    MOVE_STRUCT_X;
    CONTCASE_W;
    
  CASE(move_struct_y) 
    goto r_move_struct_y;

  CASE(move_list_x) 
    MOVE_LIST_X;
    CONTCASE_W;
    
  CASE(move_list_y) 
    goto r_move_list_y;

  CASE(move_nil_x) 
    goto r_move_nil_x;
    
  CASE(move_nil_y) 
    goto r_move_nil_y;

  CASE(move_atom_x) 
  CASE(move_int_x) 
    goto r_move_atom_x;
    
  CASE(move_atom_y) 
  CASE(move_int_y) 
    MOVE_ATOMIC_Y;
    CONTCASE_R;

  CASE(move_x_ux)
    MOVE_X_UX;
    CONTCASE_R;

  CASE(move_x_uy) 
    goto r_move_x_uy;
    
  CASE(move_y_ux) 
    MOVE_Y_UX;
    CONTCASE_R;

  CASE(move_yy_yy_yy)
    MOVE_Y_UY(w_move_yy_yy_yy);

  CASE(move_yy_yy)
    MOVE_Y_UY(w_move_yy_yy1);

  CASE(move_y_uy) 
    MOVE_Y_UY(w_move_y_uy);
    CONTCASE_R;

  CASE(move_vx) 
    goto r_move_vx;
    
  CASE(move_vy) 
    goto r_move_vy;

  CASE(move_x_wy)
    goto r_move_x_wy;

  CASE(move_yy_yw)
    MOVE_Y_UY(w_move_yy_yw);
    
  CASE(move_y_wy)
    MOVE_Y_WY;
    CONTCASE_R;

  CASE(move_yw_yy)
    MOVE_Y_WY;
    MOVE_Y_UY(w_move_yw_yy);
    CONTCASE_R;

    /* Numeric instructions */
  CASE(and)
    goto r_and;

  CASE(or)
    goto r_or;

  CASE(lshiftl)
    goto r_lshiftl;

  CASE(lshiftr)
    goto r_lshiftr;
  
  CASE(complement)
    goto r_complement;

  CASE(add_xxx)
    goto r_add_xxx;

  CASE(add_xyx)
    goto r_add_xyx;

  CASE(add_yyx)
    goto r_add_yyx;

  CASE(add)
    goto r_add;

  CASE(add1_y)
    goto r_add1_y;

  CASE(sub_xxx) 
    goto r_sub_xxx;

  CASE(sub_xyx) 
    goto r_sub_xyx;

  CASE(sub_yxx) 
    goto r_sub_yxx;

  CASE(sub_yyx) 
    goto r_sub_yyx;

  CASE(sub) 
    goto r_sub;

  CASE(sub1_y) 
    goto r_sub1_y;

  CASE(mul) 
    goto r_mul;

  CASE(div_) 
    goto r_div;

  CASE(idiv) 
    goto r_idiv;

  CASE(divge) 
    goto r_divge;

  CASE(divle) 
    goto r_divle;

  CASE(mod)
    goto r_mod;

/* Parameter passing */
  CASE(para_struct)
    PS;
    CONTCASE_W;
    
  CASE(para_list)
    PL;
    CONTCASE_W;

  CASE(para_nil)
    PN;
    CONTCASE_R;

  CASE(para_atom)
    PW;
    CONTCASE_R;

  CASE(para_int)
    PW;
    CONTCASE_R;

  CASE(para_ux_ux_ux)
    PUX;
    
  CASE(para_ux_ux)
    PUX;

  CASE(para_ux)
    PUX;
    CONTCASE_R;

  CASE(para_uy_uy_uy_uy)
    PUY;
    
  CASE(para_uy_uy_uy)
    PUY;

  CASE(para_uy_uy)
    PUY;

  CASE(para_uy)
    PUY;
    CONTCASE_R;

  CASE(para_vx)
    PVX;
    CONTCASE_R;

  CASE(para_vy_vy)
    PVY;

  CASE(para_vy)
    PVY;
    CONTCASE_R;

  CASE(para_void_one)
    op1 = (LONG)LOCAL_TOP;
    *LOCAL_TOP-- = op1;
    CONTCASE_R;

  CASE(para_void)
    op1 = *P++;
    while (op1>0) {
      op2 = (LONG)LOCAL_TOP;
      *LOCAL_TOP-- = op2;
      op1--;
    }
    CONTCASE_R;

  CASE(para_vy_ux)
    goto r_para_vy_ux;
    
  CASE(para_vy_uy)
    PVY;
    PUY;
    CONTCASE_R;

  CASE(para_ux_vy)
    PUX;
    PVY;
    CONTCASE_R;
  
  CASE(para_ux_ux_uy)
    PUX;

  CASE(para_ux_uy)
    PUX;
    PUY;
    CONTCASE_R;

  CASE(para_ux_uy_uy)
    PUX;
    PUY;
    PUY;
    CONTCASE_R;

  CASE(para_uy_vy)
    PUY;
    PVY;
    CONTCASE_R;

  CASE(para_ux_uy_ux)
    PUX;

  CASE(para_uy_ux)
    PUY;
    PUX;
    CONTCASE_R;

  CASE(para_uy_ux_ux)
    goto r_para_uy_ux_ux;
    
  CASE(para_uy_ux_uy)
    goto r_para_uy_ux_uy;

  CASE(para_uy_uy_ux)
    PUY;
    PUY;
    PUX;
    CONTCASE_R;
    
/* Procedural instructions */
  CASE(call)
    goto r_call;

  CASE(call_d)
    goto r_call_d;

  CASE(callv)
    goto r_callv;
    
  CASE(execute)
    goto r_execute;
    
  CASE(executev) 
    goto r_executev;
    
  CASE(return_a) 
    RETURN_A;
    CATCH_WAKE_EVENT;
    CONTCASE_R;
        
  CASE(return_b)
    RETURN_B;
    CATCH_WAKE_EVENT;
    CONTCASE_R;
    
  CASE(return_d)
    goto r_return_d;

  CASE(jmpn_det)
    JMPN_DET;
    
  CASE(save_ht_jmp) 
    SAVE_HT_JMP;
    CONTCASE_R;

    /* Allocate Instructions */
  CASE(allocate_flat) 
    goto r_allocate_flat;

  CASE(allocate_flat_b) 
    goto r_allocate_flat_b;

  CASE(allocate_nondet) 
    goto r_allocate_nondet;

    /* Backtracking */
  CASE(fail) 
    BACKTRACK;

  CASE(fail0)
    goto backtrack0;
    
  CASE(fork)
    goto r_fork;
/*
  CASE(commit)
    goto r_commit;
*/
  CASE(cut)
    CUT;
    CONTCASE_R;

  CASE(cut_fail)
    goto r_cut_fail;

  CASE(cut_return)
    CUT;
    RETURN_A;
    CATCH_WAKE_EVENT;
    CONTCASE_R;
    
  CASE(save_b) 
    goto r_save_b;

  CASE(getbreg)
    goto r_getbreg;

  CASE(putbreg)
    goto r_putbreg;

  CASE(getpbreg)
    goto r_getpbreg;

/* Hashing */
  CASE(hash_x) 
    goto r_hash_x;

  CASE(hash_y) 
    goto r_hash_y;

    /* Builtin */
  CASE(arg)
    goto r_arg;

  CASE(arg0)
    op1 = *P++;
    goto arg_idd;
    
  CASE(setarg) 
    goto setarg1;

  CASE(setarg0)
    op1 = *P++;
    goto setarg_idd;
    
  CASE(functor) 
   goto functor1;
    
  CASE(func_arity)
    goto arity1;

  CASE(get_ar_cps)
    goto r_get_ar_cps;
    
  CASE(put_ar_cps)
    goto r_put_ar_cps;
    
  CASE(move_ar_cps)
    goto r_move_ar_cps;
    
  CASE(jmpn_det_get_ar_cps)
    goto r_jmpn_det_get_ar_cps;

  CASE(builtin0)
    goto r_builtin0;
    
  CASE(builtin1)
    goto r_builtin1;

  CASE(builtin2)
    goto r_builtin2;
    
  CASE(builtin3)
    goto r_builtin3;

  CASE(builtin4)
    goto r_builtin4;

  CASE(allocate_susp)
    goto r_allocate_susp;
    
  CASE(delay)
    goto r_delay;

  CASE(end_delay)
    goto r_end_delay;
    
  CASE(nondet)
    goto r_nondet;

  CASE(jmpn_dvar_y)
    goto r_jmpn_dvar_y;

  CASE(jmpn_susp_var_y)
    goto r_jmpn_susp_var_y;
/*
  CASE(jmp_susp)
    goto r_jmp_susp;
*/  
  CASE(domain_set_false_yx)
    goto r_domain_set_false_yx;

  CASE(domain_set_false_yy)
    goto r_domain_set_false_yy;

  CASE(domain_set_false)
    goto r_domain_set_false;

  CASE(no_vars_gt)
    goto r_no_vars_gt;

  CASE(domain_min_max)
    goto r_domain_min_max;

  CASE(domain_min)
    goto r_domain_min;

  CASE(domain_max)
    goto r_domain_max;

  CASE(domain_min_max_u)
    goto r_domain_min_max_u;

  CASE(domain_min_u)
    goto r_domain_min_u;

  CASE(domain_max_u)
    goto r_domain_max_u;

  CASE(domain_region)
    goto r_domain_region;

  CASE(domain_region_min)
    goto r_domain_region_min;

  CASE(domain_region_max)
    goto r_domain_region_max;
/*

  CASE(first_indomain)
    goto r_first_indomain;

  CASE(rest_indomain)
    goto r_rest_indomain;
*/
  
/* Miscellaneous instructions */
  CASE(halt) 
    goto r_halt;

  CASE(halt0) 
    goto r_halt0;

  CASE(endfile) 
    goto r_endfile;
    
  CASE(table_allocate)
    goto r_table_allocate;

  CASE(table_add_use_answer)
    goto r_table_add_use_answer;

  CASE(table_use_answer_fork)
    goto r_table_use_answer_fork;

  CASE(table_check_completion)
    goto r_table_check_completion;
  
  CASE(table_cut) 
    goto r_table_cut;

  CASE(table_end) 
    goto r_table_end;

  CASE(interval_consistent_eq)
    goto r_interval_consistent_eq;

  CASE(interval_consistent_ge)
    goto r_interval_consistent_ge;

  CASE(nary_to_binary)
    goto r_nary_to_binary;

  CASE(trigger_ins_min_max)
    goto r_trigger_ins_min_max;
  
  CASE(interval_consistent_eq_nocoe)
    goto r_interval_consistent_eq_nocoe;

  default:
    printf("\nIllegal opcode hex %x at %x\n", *--P, P);
    exit(1);
  }  /* end switch */
  
 contcase_read:			/* LOCAL_TOP OF EXECUTION LOOP : Read Mode */

  UPDATE_SPACE_COUNTERS(LOCAL_TOP,H,T);
#ifdef DEBUG_INST 
    printf("%x\t", P); printf(" T=%x  H=%x HB=%x ",T,H,HB);
    printf("%s\n",xxopcode[*P].name);    
#endif 
#ifdef TRACE_TOAM
 if (trace_toam) {
      printf("%x\t", P); 
      printf("%s\n",xxopcode[*P].name);    
    }
#endif
  switch (*P++) {		/* switch on current opcode */
    
    /* Conditional jump */
  CASE(noop) 
    CONTCASE_R;
    
  CASE(jmp) 
    P = (LONG_PTR)*P;
    OVERFLOW_C;
    CATCH_WAKE_EVENT;
    CONTCASE_R;
    
  CASE(djmp)
    P = (LONG_PTR)*P;
    CONTCASE_R;
    
  CASE(jmpn_eq_struct_x) 
  r_jmpn_eq_struct_x:
    OPERAND_XW;
    goto jmpn_eq_struct;

  CASE(jmpn_eq_struct_y)
  r_jmpn_eq_struct_y:
/*    printf("jmpn_eq_struct_y Y(%d) ",*P);  */
    OPERAND_YW;
/*
    DEREF(op1);
    printf("deref(op1)=%x str=%s/%d",op1,GET_NAME((SYM_REC_PTR)op2),GET_ARITY((SYM_REC_PTR)op2)); 
*/
   SWITCH_OP_STRUCT(op1,jmpn_eq_struct,
	    {P = (LONG_PTR)*(P+1);CONTCASE_R;},
	    {tmp_op = UNTAGGED(op1);
	     if ((*(LONG_PTR)tmp_op) == op2) {
	       P += 2;
	       sreg = (LONG_PTR)tmp_op + 1;
/*
	       printf(" arg1=%x arg2=%x\n",FOLLOW(sreg),FOLLOW(sreg+1));
*/
	       CONTCASE_R;}},
	    {P = (LONG_PTR)*(P+1);CONTCASE_R;});
  P = (LONG_PTR)*P;
  CONTCASE_R;
    
  CASE(jmpn_eq_atom_x) 
  r_jmpn_eq_atom_x:
    OPERAND_XW;
    goto jmpn_eq_atom;

  CASE(jmpn_eq_atom_y) 
  r_jmpn_eq_atom_y:
    OPERAND_YW;
    SWITCH_OP_ATM(op1,jmpn_eq_atom,
	    {P = (LONG_PTR)*(P+1);CONTCASE_R;},
	    {if (op1==op2) {P += 2;CONTCASE_R;}},
	    {P = (LONG_PTR)*(P+1);CONTCASE_R;});
  P = (LONG_PTR)*P;
  CONTCASE_R;
    
  CASE(jmpn_nil_x) 
  r_jmpn_nil_x:
    OPERAND_X;
    goto jmpn_nil1;

  CASE(jmpn_nil_y) 
  r_jmpn_nil_y:
    OPERAND_Y;
  SWITCH_OP_NIL(op1,jmpn_nil1,
	    {P = (LONG_PTR)*(P+1);CONTCASE_R;},
	    {P += 2;CONTCASE_R;},
	    {P = (LONG_PTR)*(P+1);CONTCASE_R;});
  P = (LONG_PTR)*P;
  CONTCASE_R;
    
  CASE(switch_list_x) 
  r_switch_list_x:
    OPERAND_X;
    goto switch_list;

  CASE(switch_list_y)  
  r_switch_list_y:
    OPERAND_Y;
    SWITCH_OP_LST_NIL(op1,switch_list,
	    {P = (LONG_PTR)*(P+1); CONTCASE_R;},
	    {P += 3; sreg = (LONG_PTR)UNTAGGED(op1); CONTCASE_R;},
	    {P=(LONG_PTR)*P; CONTCASE_R;},
            {P = (LONG_PTR)*(P+1); CONTCASE_R;});
  P = (LONG_PTR)*(P+2);
  CONTCASE_R;


  CASE(switch_list_yxx)
  r_switch_list_yxx:
    OPERAND_Y;
    SWITCH_OP_LST_NIL(op1,nswitch_list_yxx,
		       {P = (LONG_PTR)*(P+1); 
			goto contcase_read;},
		       {P += 4;
			UNTAG(op1);
			XC(*P++) = *(LONG_PTR)op1; 
			XC(*P++) = *((LONG_PTR)op1+1); 
			goto contcase_read;},
		       {P = (LONG_PTR)*P; 
			goto contcase_read;},
		       {P = (LONG_PTR)*(P+1); 
			goto contcase_read;});
    P = (LONG_PTR)*(P+2);
    goto contcase_read; 

  CASE(switch_list_yxy)
  r_switch_list_yxy:
    OPERAND_Y;
    SWITCH_OP_LST_NIL(op1,nswitch_list_yxy,
		       {P = (LONG_PTR)*(P+1); 
			goto contcase_read;},
		       {P += 4;
			UNTAG(op1);
			XC(*P++) = *(LONG_PTR)op1; 
			YC(*P++) = *((LONG_PTR)op1+1); 
			goto contcase_read;},
		       {P = (LONG_PTR)*P; 
			goto contcase_read;},
		       {P = (LONG_PTR)*(P+1); 
			goto contcase_read;});
    P = (LONG_PTR)*(P+2);
    goto contcase_read; 

  CASE(switch_list_yyx)
    r_switch_list_yyx:
    OPERAND_Y;
    SWITCH_OP_LST_NIL(op1,nswitch_list_yyx,
		       {P = (LONG_PTR)*(P+1); 
			goto contcase_read;},
		       {P += 4;
			UNTAG(op1);
			YC(*P++) = *(LONG_PTR)op1; 
			XC(*P++) = *((LONG_PTR)op1+1); 
			goto contcase_read;},
		       {P = (LONG_PTR)*P; 
			goto contcase_read;},
		       {P = (LONG_PTR)*(P+1); 
			goto contcase_read;});
    P = (LONG_PTR)*(P+2);
    goto contcase_read; 

  CASE(switch_list_yyy)
  r_switch_list_yyy:
    OPERAND_Y;
    SWITCH_OP_LST_NIL(op1,nswitch_list_yyy,
		       {P = (LONG_PTR)*(P+1); 
			goto contcase_read;},
		       {P += 4;
			UNTAG(op1);
			YC(*P++) = *(LONG_PTR)op1; 
			YC(*P++) = *((LONG_PTR)op1+1); 
			goto contcase_read;},
		       {P = (LONG_PTR)*P; 
			goto contcase_read;},
		       {P = (LONG_PTR)*(P+1); 
			goto contcase_read;});
    P = (LONG_PTR)*(P+2);
    goto contcase_read; 
   
  CASE(jmpn_eq_int_x)   
  r_jmpn_eq_int_x:
    OPERAND_XW;
    goto jmpn_eq_int;

  CASE(jmpn_eq_int_y) 
  r_jmpn_eq_int_y:
    OPERAND_YW;
    SWITCH_OP_ATM(op1,jmpn_eq_int,
	    {P = (LONG_PTR)*(P+1); CONTCASE_R;},
	    {if (op1==op2) {P += 2;CONTCASE_R;}},
	    {P = (LONG_PTR)*(P+1); CONTCASE_R;});
  P = (LONG_PTR)*P;
  CONTCASE_R;

  CASE(jmpn_eql)   
  r_jmpn_eql:
    profile_COUNT_REFERS(2);
    op1 = *P++;
    op2 = *P++;
    OP_CONT_DEREF(op1);
    OP_CONT_DEREF(op2);
    if (ISINT(op1) && ISINT(op2))
      if (op1 == op2) {
	P++;
	CONTCASE_R;
      }
      else {
	P = (LONG_PTR)*P;
	CONTCASE_R;
      }
    SAVE_H;
    op3 = equal_to(op1,op2);
    if (op3==1) {
      P++;
      CONTCASE_R;
    }
    else if (op3==0) {
      P = (LONG_PTR)*P;
      CONTCASE_R;
    }
    else {
      RAISE_EXCEPTION(number_expected,make_struct2("=:=",op1,op2));
    }
    
  CASE(jmp_eql)  
  r_jmp_eql:
    profile_COUNT_REFERS(2);
    op1 = *P++;
    op2 = *P++;
    OP_CONT_DEREF(op1);
    OP_CONT_DEREF(op2);
    goto jmp_eql1;

  CASE(jmp_eql_yy)  
  r_jmp_eql_yy:
    op1 = YC(*P++);DEREF(op1);
    op2 = YC(*P++);DEREF(op2);
 jmp_eql1:
    if (ISINT(op1) && ISINT(op2))
      if (op1 == op2) {
	P = (LONG_PTR)*P;
	CONTCASE_R;
      }
      else {
	P++;
	CONTCASE_R;
      }
    SAVE_H;
    op3 = equal_to(op1,op2);
    if (op3==1) {
      P = (LONG_PTR)*P;
      CONTCASE_R;
    }
    else if (op3==0) {
      P++;
      CONTCASE_R;
    }
    else {
      RAISE_EXCEPTION(number_expected,make_struct2("=\\=",op1,op2));
    }

  CASE(jmpn_gt)  
  r_jmpn_gt:
    profile_COUNT_REFERS(2);
    op1 = *P++;
    op2 = *P++;
    OP_CONT_DEREF(op1);
    OP_CONT_DEREF(op2);
    goto jmpn_gt1;

  CASE(jmpn_gt_yy)  
  r_jmpn_gt_yy:
    op1 = YC(*P++); DEREF(op1);
    op2 = YC(*P++); DEREF(op2);
 jmpn_gt1:
    if (ISINT(op1) && ISINT(op2)){
      if (INTVAL(op1)>INTVAL(op2)){
	P++;
	CONTCASE_R;
      }
      else {
	P = (LONG_PTR)*P;
	CONTCASE_R;
      }
    }
    SAVE_H;
    op3 = greater_than(op1,op2);
    if (op3==1) {
      P++;
      CONTCASE_R;
    }
    else if (op3==0){
      P = (LONG_PTR)*P;
      CONTCASE_R;
    }
    else {
      RAISE_EXCEPTION(number_expected,make_struct2(">",op1,op2));
    }

  CASE(jmpn_ge)   
  r_jmpn_ge:
    profile_COUNT_REFERS(2);
    op1 = *P++;
    op2 = *P++;
    OP_CONT_DEREF(op1);
    OP_CONT_DEREF(op2);
    goto jmpn_ge1;

  CASE(jmpn_ge_yy)   
  r_jmpn_ge_yy:
    op1 = YC(*P++);DEREF(op1);
    op2 = YC(*P++);DEREF(op2);
 jmpn_ge1:
    if (ISINT(op1) && ISINT(op2)){
      if (INTVAL(op1)>=INTVAL(op2)){
	P++;
	CONTCASE_R;
      }
      else {
	P = (LONG_PTR)*P;
	CONTCASE_R;
      }
    }
    SAVE_H;
    op3 = greater_equal(op1,op2);
    if (op3==1) {
      P++;
      CONTCASE_R;
    }
    else if (op3==0){
      P = (LONG_PTR)*P;
      CONTCASE_R;
    }
    else {
      RAISE_EXCEPTION(number_expected,make_struct2(">=",op1,op2));
    }

  CASE(jmpn_id)  
  r_jmpn_id:
    profile_COUNT_REFERS(2);
    op1 = *P++;
    OP_CONT_DEREF(op1);
    op2 = *P++;
    OP_CONT_DEREF(op2);
    if (op1 == op2 || (ISCOMPOUND(op1) && identical(op1,op2))){
      P++;
      CONTCASE_R;
    } else {
      P = (LONG_PTR)*P;
      CONTCASE_R;
    }
    
  CASE(jmp_id) 
  r_jmp_id:
    profile_COUNT_REFERS(2);
    op1 = *P++;
    OP_CONT_DEREF(op1);
    op2 = *P++;
    OP_CONT_DEREF(op2);
    if (op1 == op2 || (ISCOMPOUND(op1) && identical(op1,op2))){
      P = (LONG_PTR)*P;
      CONTCASE_R;
    }
    else {
      P++;
      CONTCASE_R;
    }

  CASE(jmpn_var_x)   
  r_jmpn_var_x:
    OPERAND_X;
    goto jmpn_var;

  CASE(jmpn_var_y)  
  r_jmpn_var_y:
    OPERAND_Y;
    SWITCH_OP_VAR(op1,jmpn_var,
		  {P++;CONTCASE_R;},
		  {P++;CONTCASE_R;},
		  {P = (LONG_PTR)*P; CONTCASE_R;});

  CASE(jmp_var_x)   
  r_jmp_var_x:
    OPERAND_X;
    SWITCH_OP_VAR(op1,jmp_var,
		  {P = (LONG_PTR)*P;CONTCASE_R;},
		  {P = (LONG_PTR)*P;CONTCASE_R;},
		  {P++; CONTCASE_R;});

  CASE(jmp_var_y)  
  r_jmp_var_y:
    OPERAND_Y;
    goto jmp_var;

  CASE(jmpn_atom_x)  
  r_jmpn_atom_x:
    OPERAND_X;
    goto jmpn_atom;

  CASE(jmpn_atom_y)  
  r_jmpn_atom_y:
    OPERAND_Y;
  jmpn_atom:
    DEREF(op1);
    if (ISATOM(op1)){
      P++;
      CONTCASE_R;
    }
    P = (LONG_PTR)*P;
    CONTCASE_R;

  CASE(jmpn_atomic_x)   
  r_jmpn_atomic_x:
    OPERAND_X;
    goto jmpn_atomic;

  CASE(jmpn_atomic_y)  
  r_jmpn_atomic_y:
    OPERAND_Y;
  jmpn_atomic:
    DEREF(op1);
    if (TAG(op1)==ATM || ISFLOAT(op1)){
      P++;
      CONTCASE_R;
    }
    P = (LONG_PTR)*P; 
    CONTCASE_R;
    
  CASE(jmpn_num_x)  
  r_jmpn_num_x:
    OPERAND_X;
    goto jmpn_num;

  CASE(jmpn_num_y)  
  r_jmpn_num_y:
    OPERAND_Y;
  jmpn_num:
  DEREF(op1);
  if (ISNUM(op1)) {
    P++;
    CONTCASE_R;
  }
  P = (LONG_PTR)*P;
  CONTCASE_R;

  CASE(jmpn_float_x) 
  r_jmpn_float_x:
    OPERAND_X;
    goto jmpn_float;

  CASE(jmpn_float_y)  
  r_jmpn_float_y:
    OPERAND_Y;
  jmpn_float:
  DEREF(op1);
  if (ISFLOAT(op1)){
    P++;
    CONTCASE_R;
  }
  P = (LONG_PTR)*P;
  CONTCASE_R;
    
  CASE(jmpn_int_x)  
  r_jmpn_int_x:
    OPERAND_X;
    goto jmpn_int1;

  CASE(jmpn_int_y)  
  r_jmpn_int_y:
    OPERAND_Y;
  jmpn_int1:
  DEREF(op1);
  if (ISINT(op1)) {
    P++;
    CONTCASE_R;
  }
  P = (LONG_PTR)*P;
  CONTCASE_R;

  CASE(hash_jmpn_nil)  
  r_hash_jmpn_nil:
    if (ISNIL1(op1)) {
      P++;
      CONTCASE_R;
    }
    else {
      P = (LONG_PTR)*P;
      CONTCASE_R;
    }
  
  CASE(hash_jmpn_list)
  r_hash_jmpn_list:
    if (ISLIST(op1)) {
      P++;
      sreg = (LONG_PTR)UNTAGGED(op1);
      CONTCASE_R;
    }
    else {
      P = (LONG_PTR)*P;
      CONTCASE_R;
    }
    
  CASE(hash_jmpn_int)  
  r_hash_jmpn_int:
    OPERAND_W;
    if (op1==op2) {
      P++;
      CONTCASE_R;
    }
    else {
      P = (LONG_PTR)*P;
      CONTCASE_R;
    }
    
  CASE(hash_jmpn_atom)
  r_hash_jmpn_atom:
    OPERAND_W;
    if (op1==op2) {
      P++;
      CONTCASE_R;
    }
    else {
      P = (LONG_PTR)*P;
      CONTCASE_R;
    }

  CASE(hash_jmpn_struct)
  r_hash_jmpn_struct:
    OPERAND_W;
    if (ISSTRUCT(op1) && (LONG)sym_ptr == op2) {  /* tmp_op and sym_ptr got value at hash */
      P++;
      sreg = (LONG_PTR)tmp_op + 1;
      CONTCASE_R;
    }
    P = (LONG_PTR)*P;
    CONTCASE_R;

  CASE(hash_jmpn_struct_x)  /* combine jmpn_struct and fetch_var_x */
  r_hash_jmpn_struct_x:
    OPERAND_W;
    if (ISSTRUCT(op1)) {
      if ((LONG)sym_ptr == op2) {  /* tmp_op and sym_ptr got value at hash */
	P += 2;      /* skip fetch_vars_xx */
	sreg = (LONG_PTR)tmp_op + 1;
	XC(*P++) = *sreg++;
	CONTCASE_R;
      }
    }
    P = (LONG_PTR)*P;
    CONTCASE_R;

  CASE(hash_jmpn_struct_y)
  r_hash_jmpn_struct_y:
    OPERAND_W;
    if (ISSTRUCT(op1)) {
      if ((LONG)sym_ptr == op2) {  /* tmp_op and sym_ptr got value at hash */
	P += 2;      /* skip fetch_vars_xx */
	sreg = (LONG_PTR)tmp_op + 1;
	YC(*P++) = *sreg++;
	CONTCASE_R;
      }
    }
    P = (LONG_PTR)*P;
    CONTCASE_R;
  
  CASE(hash_jmpn_struct_xx)
  r_hash_jmpn_struct_xx:
    OPERAND_W;
    HASH_JMPN_STRUCT(XC(*P++),XC(*P++));

  CASE(hash_jmpn_struct_xy)
  r_hash_jmpn_struct_xy:
    OPERAND_W;
    HASH_JMPN_STRUCT(XC(*P++),YC(*P++));

  CASE(hash_jmpn_struct_yx)
  r_hash_jmpn_struct_yx:
    OPERAND_W;
    HASH_JMPN_STRUCT(YC(*P++),XC(*P++));


  CASE(hash_jmpn_struct_yy)
  r_hash_jmpn_struct_yy:
    OPERAND_W;
    HASH_JMPN_STRUCT(YC(*P++),YC(*P++));

    /* Unify */
  CASE(unify_struct_x)  
  r_unify_struct_x:
    OPERAND_XW;
    UNIFY_STRUCT_X(op1,op2,rr_unify_struct_x);
    
  CASE(unify_struct_y)  
  r_unify_struct_y:
    OPERAND_YW;
    UNIFY_STRUCT_Y(op1,op2,rr_unify_struct_y,{toam_PUSHTRAIL(op1);},{},{goto backtrack;});

  CASE(unify_list_x)    
  r_unify_list_x:
    OPERAND_X;
    UNIFY_LIST_X(op1,rr_unify_list_x);
    
  CASE(unify_list_y)    
  r_unify_list_y:
    OPERAND_Y;
    UNIFY_LIST_Y(op1,rr_unify_list_y,{toam_PUSHTRAIL(op1);},{},{goto backtrack;});

  CASE(unify_nil_x)     
  r_unify_nil_x:
    OPERAND_X;
    UNIFY_NIL_X(op1,rr_unify_nil_x);

  CASE(unify_nil_y)     
  r_unify_nil_y:
    OPERAND_Y;
    UNIFY_NIL_Y(op1,rr_unify_nil_y,{toam_PUSHTRAIL(op1);},{},{goto backtrack;});

  CASE(unify_atom_x)    
  CASE(unify_int_x)     
  r_unify_atom_x:
    OPERAND_XW;
    UNIFY_ATOMIC12_X(op1,op2,rr_unify_atom_x,{goto backtrack;});

  CASE(unify_atom_y)    
  CASE(unify_int_y)     
  r_unify_atom_y:
    OPERAND_YW;
    UNIFY_ATOMIC12_Y(op1,op2,rr_unify_atom_y,{toam_PUSHTRAIL(op1);},{},{goto backtrack;});
    
  CASE(unify_ux_ux)     
  r_unify_ux_ux:
    OPERAND_XX;
    SWITCH_OP(op1,rr_unify_ux_ux,
	      {goto unify_nsv_d;},
	      {UNIFY_ATOMIC21_X(op2,op1,rr_unify_ux_ux2,{goto backtrack;});},
	      {goto unify_lst_d;},
	      {goto unify_str_d;},
	      {goto unify_susp_d;});

  CASE(unify_ux_uy)     
  r_unify_ux_uy:
    OPERAND_XY;
    SWITCH_OP(op1,rr_unify_ux_uy,
	      {goto unify_nsv_d;},
	      {UNIFY_ATOMIC21_Y(op2,op1,rrr_unify_ux_uy,{toam_PUSHTRAIL(op2);},{},{goto backtrack;});},
	      {goto unify_lst_d;},
	      {goto unify_str_d;},
	      {goto unify_susp_d;});

/*  unify_nsv_d: */
    SWITCH_OP_VAR(op2,unify_nsv_d,
		  {if (op1 != op2) {
		    if (op1 < op2) {
		      ASSIGN_in_toam(op2,op1,{toam_PUSHTRAIL(op2);});
		    } else {		/* op1 points to op2 */
		      ASSIGN_in_toam(op1,op2,{toam_PUSHTRAIL(op1);});
		    }
		  }
		   CONTCASE_R;},
		  {goto unify_v_susp;},
		  {});
  ASSIGN_in_toam(op1,op2,{toam_PUSHTRAIL_h(op1);});
  CONTCASE_R;

  CASE(unify_uy_uy)     
  r_unify_uy_uy:
    OPERAND_YY;
    SWITCH_OP(op1,unify_d_d,
	      /* unify_v_d: */
	       {SWITCH_OP_VAR(op2,unify_v_d,
			  {goto unify_v_v;},
			  {goto unify_v_susp;},{});
		ASSIGN_in_toam(op1,op2,{toam_PUSHTRAIL(op1);});
		CONTCASE_R;},
	      {UNIFY_ATOMIC21_Y(op2,op1,rr_unify_uy_uy,{toam_PUSHTRAIL(op2);},{},{goto backtrack;});},
	      {goto unify_lst_d;},
	      {goto unify_str_d;},
	      {goto unify_susp_d;});
 unify_v_v:
    if (op1 != op2) {    /* op1 is free, op2 is free */
      if (op1 < op2) {
	if (op1 < (LONG) H) {	/* op1 not in loc stack */
	  ASSIGN_in_toam(op2,op1,toam_PUSHTRAIL(op2));
	} else {		/* op1 points to op2 */
	  ASSIGN_in_toam(op1,(LONG)H,{toam_PUSHTRAIL(op1);});
	  ASSIGN_in_toam(op2,(LONG)H,{toam_PUSHTRAIL(op2);});
	  toam_NEW_HEAP_FREE;
	}
      } else {		/* op1 > op2 */
	if (op2 < (LONG) H) {
	  ASSIGN_in_toam(op1,op2,{toam_PUSHTRAIL(op1);});
	} else {
	  ASSIGN_in_toam(op1,(LONG)H,{toam_PUSHTRAIL(op1);});
	  ASSIGN_in_toam(op2,(LONG)H,{toam_PUSHTRAIL(op2);});
	  toam_NEW_HEAP_FREE;
	}
      }
    }
    CONTCASE_R;
    
  CASE(unify_cons_x) 
  r_unify_cons_x:
    tmp_op = *P++;
    op1 = YC(tmp_op);
    op2 = XC(*P++);
    UNIFY_CONS(rr_unify_cons_x,rrr_unify_cons_x,
	       {toam_NEW_HEAP_NODE(op2);},
	       {toam_NEW_HEAP_NODE(op2);},
	       {UNIFY_ATOMIC21_X(op2,op1,rrrr_unify_cons_x,{goto backtrack;});});


  CASE(unify_cons_y) 
  r_unify_cons_y:
    tmp_op = *P++;
    op1 = YC(tmp_op);
    op2 = YC(*P++);
    UNIFY_CONS(rr_unify_cons_y, rrr_unify_cons_y,
	       {toam_BUILD_VALUE(op2,rrrr_unify_cons_y);},
	       {toam_BUILD_VALUE(op2,rrrrr_unify_cons_y);},
	       {UNIFY_ATOMIC21_Y(op2,op1,rrrrrr_unify_cons_y,{toam_PUSHTRAIL(op2);},{},{goto backtrack;});});


  CASE(fork_unify_struct_y)
  r_fork_unify_struct_y:
    OPERAND_YW;
    UNIFY_STRUCT_Y(op1,op2,rr_fork_unify_struct_y,
		   {FORK; toam_PUSHTRAIL_UC_0(op1);},
		   {FORK;},
		   {goto backtrack00;});

  CASE(fork_unify_list_y)
  r_fork_unify_list_y:
    OPERAND_Y;
    UNIFY_LIST_Y(op1,rr_fork_unify_list_y,
                 {FORK;toam_PUSHTRAIL_UC_0(op1);},
		 {FORK;},
		 {goto backtrack00;});

  CASE(fork_unify_nil_y)
  r_fork_unify_nil_y:
    OPERAND_Y;
    UNIFY_NIL_Y(op1,rr_fork_unify_nil_y,
                {FORK;toam_PUSHTRAIL_UC_0(op1);},
		{FORK;},
		{goto backtrack00;});

  CASE(fork_unify_atom_y)
  CASE(fork_unify_int_y)
  r_fork_unify_atom_y:
    OPERAND_YW;
    UNIFY_ATOMIC12_Y(op1,op2,rr_fork_unify_atom_y,
                   {FORK;toam_PUSHTRAIL_UC_0(op1);},
		   {FORK;},
		   {goto backtrack00;});

  CASE(fork_unify_uy_uy)
  r_fork_unify_uy_uy:
    OPERAND_YY;
    SWITCH_OP(op1,rr_fork_unify_uy_uy,
	       {FORK;goto unify_v_d;},
	       {UNIFY_ATOMIC21_Y(op2,op1,rrr_fork_unify_uy_uy,
			       {FORK;toam_PUSHTRAIL_UC_0(op2);},
			       {FORK;},
			       {goto backtrack00;});},
	       {FORK;goto unify_lst_d;},
	       {FORK;goto unify_str_d;},
	       {FORK;goto unify_susp_d;});
  
  /* Fork_unicut */
  CASE(fork_unicut_struct_y)
  r_fork_unicut_struct_y:
    OPERAND_YW;
    UNIFY_STRUCT_Y(op1,op2,rr_fork_unicut_struct_y,
		   {CUT; toam_PUSHTRAIL(op1);P++;},
		   {FORK;},
		   {goto backtrack00;});
    
  CASE(fork_unicut_list_y)
  r_fork_unicut_list_y:
    OPERAND_Y;
    UNIFY_LIST_Y(op1,rr_fork_unicut_list_y,
		 {CUT;toam_PUSHTRAIL(op1);P++;},
		 {FORK;},
		 {goto backtrack00;});

  CASE(fork_unicut_nil_y)
  r_fork_unicut_nil_y:
    OPERAND_Y;
    UNIFY_NIL_Y(op1,rr_fork_unicut_nil_y,
		{CUT;toam_PUSHTRAIL(op1);P++;},
		{CUT;P++;},
		{goto backtrack00;});

  CASE(fork_unicut_atom_y)
  CASE(fork_unicut_int_y)
  r_fork_unicut_atom_y:
    OPERAND_YW;
    UNIFY_ATOMIC12_Y(op1,op2,rr_fork_unicut_atom_y,
		   {CUT;toam_PUSHTRAIL(op1);P++;},
		   {CUT;P++;},
		   {goto backtrack00;});

  CASE(fork_unicut_uy_uy)
  r_fork_unicut_uy_uy:
    OPERAND_YY;
    SWITCH_OP(op1,rr_fork_unicut_uy_uy,
	       {CUT; P += 2; goto unify_v_d;}, /* skip cut */
	       {UNIFY_ATOMIC21_Y(op2,op1,rrr_fork_unicut_uy_uy,
			       {CUT; toam_PUSHTRAIL(op2);P += 2;},
			       {CUT; P += 2;},
			       {goto backtrack00;});},
	      {FORK; goto unify_lst_d;},      /* cut follows */
	      {FORK; goto unify_str_d;},
	      {CUT;goto unify_susp_d;});
  
  /* Unify0 */
  CASE(unify0_struct_y)   
  r_unify0_struct_y:
    OPERAND_YW;
    UNIFY_STRUCT_Y(op1,op2,rr_unify0_struct_y,{toam_PUSHTRAIL_UC_0(op1);},{},{goto backtrack0;});
    
  CASE(unify0_list_y)     
  r_unify0_list_y:
    OPERAND_Y;
    UNIFY_LIST_Y(op1,rr_unify0_list_y,{toam_PUSHTRAIL_UC_0(op1);},{},{goto backtrack0;});

  CASE(unify0_nil_y)      
  r_unify0_nil_y:
    OPERAND_Y;
    UNIFY_NIL_Y(op1,rr_unify0_nil_y,{toam_PUSHTRAIL_UC_0(op1);},{},{goto backtrack0;});

  CASE(unify0_atom_y)     
  CASE(unify0_int_y)      
  r_unify0_atom_y:
    OPERAND_YW;
    UNIFY_ATOMIC12_Y(op1,op2,rr_unify0_atom_y,{toam_PUSHTRAIL_UC_0(op1);},{},{goto backtrack0;});
    
  CASE(unify0_uy_uy)      
    OPERAND_YY;
    SWITCH_OP(op1,unify0_d_d,
	       {goto unify_v_d;},
	       {UNIFY_ATOMIC21_Y(op2,op1,unify0_d_d_1,{toam_PUSHTRAIL_UC_0(op2);},{},{goto backtrack0;});},
	       {goto unify0_lst_d;},
	       {goto unify0_str_d;},
	      {goto unify_susp_d;});

    /* Unicut */
  CASE(unicut_struct_y)  
  r_unicut_struct_y:    
    OPERAND_YW;
    UNIFY_STRUCT_Y(op1,op2,rr_unicut_struct_y,{CUT;toam_PUSHTRAIL(op1);},{},{goto backtrack0;});

  CASE(unicut_list_y)    
  r_unicut_list_y:
    OPERAND_Y;
    UNIFY_LIST_Y(op1,rr_unicut_list_y,{CUT;toam_PUSHTRAIL(op1);},{},{goto backtrack0;});

  CASE(unicut_nil_y)     
  r_unicut_nil_y:
    OPERAND_Y;
    UNIFY_NIL_Y(op1,rr_unicut_nil_y,{CUT;toam_PUSHTRAIL(op1);},{CUT;},{goto backtrack0;});
    
  CASE(unicut_atom_y)    
  CASE(unicut_int_y)     
  r_unicut_atom_y:
    OPERAND_YW;
    UNIFY_ATOMIC12_Y(op1,op2,rr_unicut_atom_y,{CUT;toam_PUSHTRAIL(op1);},{CUT;},{goto backtrack0;});
     
  CASE(unicut_uy_uy)
  r_unicut_uy_uy:
    OPERAND_YY;
    SWITCH_OP(op1,rr_unicut_uy_uy,
	       {CUT; P++; goto unify_v_d;}, /* skip cut */
	       {UNIFY_ATOMIC21_Y(op2,op1,rrr_unicut_uy_uy,
			       {CUT; toam_PUSHTRAIL(op2);P++;},
			       {CUT; P++;},
			       {goto backtrack0;});},
	       {goto unify_lst_d;},      /* cut follows */
	       {goto unify_str_d;},
	      {CUT;P++;goto unify_susp_d;});
   
 CASE(unicut)
    CUT;
    CONTCASE_R;
      
/* Unify argument */
  CASE(unify_arg_nil)  
    op1 = *sreg++;
    SWITCH_OP_VAR(op1,unify_arg_nil1,
		  {ASSIGN_in_toam(op1,nil_sym1,{toam_PUSHTRAIL_h(op1);});
		   CONTCASE_R;},

		  {op2 = nil_sym1; goto bind_susp_var_value;},

		  {if (ISNIL1(op1)) CONTCASE_R;
		  else  BACKTRACK;});

  CASE(unify_arg_atom) 
  CASE(unify_arg_int)  
    OPERAND_W;		
    op1 = *sreg++; 
    UNIFY_ATOMIC12_X(op1,op2,unify_arg_atomic,{goto backtrack;});

  CASE(unify_arg_ux_ux)  
    P += 3;

  CASE(unify_arg_ux)  
  r_unify_arg_ux:
    OPERAND_X;
    op2 = *sreg++;
    SWITCH_OP(op1,rr_unify_arg_ux,
	       {goto unify_nsv_d;},
	       {UNIFY_ATOMIC21_X(op2,op1,rrr_unify_arg_ux,{goto backtrack;});},
	       {goto unify_lst_d;},
	       {goto unify_str_d;},
	      {goto unify_susp_d;});

  CASE(unify_arg_ux_vx)  
    P += 3;
    goto r_unify_arg_ux;

  CASE(unify_arg_ux_vy)  
    P += 3;
    goto r_unify_arg_ux;
    
  CASE(unify_arg_uy_uy)    
    P += 3;

  CASE(unify_arg_uy) 
  CASE(unify_arg_wy)
  r_unify_arg_uy:
    op1 = *sreg++;
    op2 = YC(*P++);
    SWITCH_OP(op1,rr_unify_arg_uy,
	       {goto unify_nsv_d;},
	       {UNIFY_ATOMIC21_Y(op2,op1,rrr_unify_arg_uy,{toam_PUSHTRAIL(op2);},{},{goto backtrack;});},
	       {goto unify_lst_d;},
	       {goto unify_str_d;},
	      {goto unify_susp_d;});

  CASE(unify_arg_vx_vx)
    RVX;

  CASE(unify_arg_vx) 
    RVX;
    CONTCASE_R;

  CASE(unify_arg_vy_vy)
    RVY;

  CASE(unify_arg_vy)  
    RVY;
    CONTCASE_R;

  CASE(unify_arg_list)   
    op1 = *sreg++;
    UNIFY_LIST_X(op1,r_unify_arg_list);

  CASE(unify_arg_struct) 
    OPERAND_W;
    op1 = *sreg;
    UNIFY_STRUCT_X(op1,op2,r_unify_arg_struct);
    
  CASE(unify_arg_void_one)  
    sreg += 1;
    CONTCASE_R;      

  CASE(unify_arg_void)   
    OPERAND_W;
    sreg += op2;
    CONTCASE_R;      

  CASE(unify_arg_vx_vy)
    RVX;
    RVY;
    CONTCASE_R;      

  CASE(unify_arg_vx_ux)
    RVX;
    goto r_unify_arg_ux;

  CASE(unify_arg_vx_uy)
    RVX;
    goto r_unify_arg_uy;

  CASE(unify_arg_vy_vx)
    RVY;
    RVX;
    CONTCASE_R;      

  CASE(unify_arg_vy_ux)
    RVY;
    goto r_unify_arg_ux;

  CASE(unify_arg_vy_uy)
    RVY;
    goto r_unify_arg_uy;

  CASE(unify_arg_iii)
    RI(unify_arg_iii1);
    RI(unify_arg_iii2);
    RI(unify_arg_iii3);
    CONTCASE_R;
    
    /* Move */
  CASE(move_struct_x) 
    MOVE_STRUCT_X;
    CONTCASE_W;
    
  CASE(move_struct_y) 
  r_move_struct_y:
    MOVE_STRUCT_Y;
    CONTCASE_W;

  CASE(move_list_x)
    MOVE_LIST_X;
    CONTCASE_W;
    
  CASE(move_list_y)  
  r_move_list_y:
    MOVE_LIST_Y;
    CONTCASE_W;

  CASE(move_nil_x) 
  r_move_nil_x:
    MOVE_NIL_X;
    CONTCASE_R;
    
  CASE(move_nil_y)  
  r_move_nil_y:
    MOVE_NIL_Y;
    CONTCASE_R;

  CASE(move_atom_x) 
  CASE(move_int_x) 
  r_move_atom_x:
    MOVE_ATOMIC_X;
    CONTCASE_R;

  CASE(move_atom_y)  
  CASE(move_int_y) 
    MOVE_ATOMIC_Y;
    CONTCASE_R;

  CASE(move_x_ux)
    MOVE_X_UX;
    CONTCASE_R;

  CASE(move_x_uy)
  r_move_x_uy:
    MOVE_X_UY(w_move_x_uy);
    CONTCASE_R;
    
  CASE(move_y_ux)
    MOVE_Y_UX;
    CONTCASE_R;

  CASE(move_yy_yy_yy)
    MOVE_Y_UY(r_move_yy_yy_yy);

  CASE(move_yy_yy)
    MOVE_Y_UY(r_move_yy_yy1);

  CASE(move_y_uy)  
    MOVE_Y_UY(r_move_y_uy);
    CONTCASE_R;

  CASE(move_vx)   
  r_move_vx:
    MOVE_VX;
    CONTCASE_R;
    
  CASE(move_vy) 
  r_move_vy:
    MOVE_VY;
    CONTCASE_R;

  CASE(move_x_wy)
  r_move_x_wy:
    MOVE_X_WY;
    CONTCASE_R;

  CASE(move_yy_yw)
    MOVE_Y_UY(r_move_yy_yw);

  CASE(move_y_wy)
    MOVE_Y_WY;
    CONTCASE_R;
    
  CASE(move_yw_yy)
    MOVE_Y_WY;
    MOVE_Y_UY(r_move_yw_yy);
    CONTCASE_R;

    /* Numeric instructions */
  CASE(and)
  r_and:
    profile_COUNT_REFERS(3);
    op1 = *P++;OP_CONT_DEREF(op1);
    op2 = *P++;OP_CONT_DEREF(op2);
    op3 = *P++;OP_ADDR(op3);
    if (ISINT(op1) && ISINT(op2)) 
      FOLLOW(op3) = MAKEINT(INTVAL(op2) & INTVAL(op1));
    else {
      SAVE_H;
      if (AND(op1,op2,op3)<0){
	RAISE_EXCEPTION(exception,make_struct2("/\\",op1,op2));
      } 
      RESTORE_H;
    }
    CONTCASE_R;

  CASE(or)
  r_or:
    profile_COUNT_REFERS(3);
    op1 = *P++;OP_CONT_DEREF(op1);
    op2 = *P++;OP_CONT_DEREF(op2);
    op3 = *P++;OP_ADDR(op3);
    if (ISINT(op1) && ISINT(op2)) {
      FOLLOW(op3) = MAKEINT(INTVAL(op2) | INTVAL(op1));
    } else {
      SAVE_H;
      if (OR(op1,op2,op3)<0){
	RAISE_EXCEPTION(exception,make_struct2("\\/",op1,op2));
      }
      RESTORE_H;
    }
    CONTCASE_R;

  CASE(lshiftl)
  r_lshiftl:
    profile_COUNT_REFERS(3);
    op1 = *P++;OP_CONT_DEREF(op1);
    op2 = *P++;OP_CONT_DEREF(op2);
    op3 = *P++;OP_ADDR(op3);
    if (ISINT(op1) && ISINT(op2)) 
      FOLLOW(op3) = MAKEINT(INTVAL(op1) << INTVAL(op2));
    else {
      SAVE_H;
      if (SHIFTL(op1,op2,op3)<0) {
	RAISE_EXCEPTION(exception,make_struct2("<<",op1,op2));
      }
      RESTORE_H;
    }
    CONTCASE_R;

  CASE(lshiftr)
  r_lshiftr:
    profile_COUNT_REFERS(3);
    op1 = *P++;OP_CONT_DEREF(op1);
    op2 = *P++;OP_CONT_DEREF(op2);
    op3 = *P++;OP_ADDR(op3);
    if (ISINT(op1) && ISINT(op2)) 
      FOLLOW(op3) = MAKEINT(INTVAL(op1) >> INTVAL(op2));
    else {
      SAVE_H;
      if (SHIFTR(op1,op2,op3)<0){
	RAISE_EXCEPTION(exception,make_struct2(">>",op1,op2));
      }
      RESTORE_H;
    }
    CONTCASE_R;
  
  CASE(complement)
  r_complement:
  complement1:
    profile_COUNT_REFERS(2);
    op1 = *P++;OP_CONT_DEREF(op1);
    op2 = *P++;OP_ADDR(op2);
    if (ISINT(op1)) 
      FOLLOW(op2) = MAKEINT(~INTVAL(op1));
    else {
      SAVE_H;
      if (NEG(op1,op2)<0){
	RAISE_EXCEPTION(integer_expected,make_struct1("\\",op1));
      }
      RESTORE_H;
    }
    CONTCASE_R;

  CASE(add_xxx)
  r_add_xxx:
    {OPERAND_XX;DEREF(op1);DEREF(op2);
     op3 = *P++;
     goto add1;}

  CASE(add_xyx)
  r_add_xyx:
    {OPERAND_XY;DEREF(op1);DEREF(op2);
     op3 = *P++;
     goto add1;}
  
  CASE(add_yyx)
  r_add_yyx:
    {OPERAND_YY;DEREF(op1);DEREF(op2);
     op3 = *P++;
     goto add1;}

  CASE(add)
  r_add:
    profile_COUNT_REFERS(3);
    op1 = *P++;OP_CONT_DEREF(op1);
    op2 = *P++;OP_CONT_DEREF(op2);
    op3 = *P++;OP_ADDR(op3);
  /*------------------------------------------------------------------*/
 add1:
  if (ISINT(op1)) 
    if (ISINT(op2))
      FOLLOW(op3) = MAKEINT(INTVAL(op1) + INTVAL(op2));
    else if (ISFLOAT(op2)){
      toam_ENCODEFLOAT(op3,(double)INTVAL(op1) + floatval(op2));}
    else 
      goto external_add;
  else  if (ISFLOAT(op1))
    if (ISINT(op2)) {
      toam_ENCODEFLOAT(op3,floatval(op1) + (double)INTVAL(op2));}
    else if (ISFLOAT(op2)){
      toam_ENCODEFLOAT(op3,floatval(op1) + floatval(op2));}
    else
      goto external_add;
  else goto external_add;
  CONTCASE_R;
 external_add:
  SAVE_H;
  if (ADD(op1,op2,op3)<0){
    RAISE_EXCEPTION(exception,make_struct2("+",op1,op2));
  }
  RESTORE_H;
  CONTCASE_R;
  /*------------------------------------------------------------------*/

  CASE(add1_y)
  r_add1_y:
    op1 = (LONG)Y(*P++);
    op2 = FOLLOW(op1);
  add11:
    DEREF(op2);
    if (ISINT(op2)) {
      FOLLOW(op1)=MAKEINT(INTVAL(op2)+1);
      CONTCASE_R;
    }
    else if (ISFLOAT(op2)){
      toam_ENCODEFLOAT(op1,floatval(op2)+1.0);
      CONTCASE_R;
    }
    else {
      SAVE_H;
      if (eval_arith(op2,&op4)<0) {
	RAISE_EXCEPTION(number_expected,make_struct2("+",op2,MAKEINT(1)));
      }
      RESTORE_H;
      op2 = op4;
      goto add11;
    }

  CASE(sub_xxx)   
  r_sub_xxx:
    {OPERAND_XX;DEREF(op1);DEREF(op2);
     op3 = *P++;
     goto sub1;}

  CASE(sub_xyx)   
  r_sub_xyx:
    {OPERAND_XY;DEREF(op1);DEREF(op2);
     op3 = *P++;
     goto sub1;}

  CASE(sub_yxx)   
  r_sub_yxx:
    {OPERAND_YX;DEREF(op1);DEREF(op2);
     op3 = *P++;
     goto sub1;}

  CASE(sub_yyx)   
  r_sub_yyx:
    {OPERAND_YY;DEREF(op1);DEREF(op2);
     op3 = *P++;
     goto sub1;}

  CASE(sub)   
  r_sub:
    profile_COUNT_REFERS(3);
    op1 = *P++;OP_CONT_DEREF(op1);
    op2 = *P++;OP_CONT_DEREF(op2);
    op3 = *P++;OP_ADDR(op3);
  /*------------------------------------------------------------------*/
 sub1:
  if (ISINT(op1)) 
    if (ISINT(op2))
      FOLLOW(op3) = MAKEINT(INTVAL(op1) - INTVAL(op2));
    else if (ISFLOAT(op2)){
      toam_ENCODEFLOAT(op3,(double)INTVAL(op1) - floatval(op2));}
    else 
      goto external_sub;
  else  if (ISFLOAT(op1))
    if (ISINT(op2)) {
      toam_ENCODEFLOAT(op3,floatval(op1) - (double)INTVAL(op2));}
    else if (ISFLOAT(op2)){
      toam_ENCODEFLOAT(op3,floatval(op1) - floatval(op2));}
    else
      goto external_sub;
  else goto external_sub;
  CONTCASE_R;
 external_sub:
  SAVE_H;
  if (SUB(op1,op2,op3)<0){
    RAISE_EXCEPTION(exception,make_struct2("-",op1,op2));
  }
  RESTORE_H;
  CONTCASE_R;
  /*------------------------------------------------------------------*/
      
  CASE(sub1_y)  
  r_sub1_y:
    op1 = (LONG)Y(*P++);
    op2 = FOLLOW(op1);
  sub11:
    DEREF(op2);
    if (ISINT(op2)) {
      FOLLOW(op1)=MAKEINT(INTVAL(op2)-1);
      CONTCASE_R;
    }
    else if (ISFLOAT(op2)){
      toam_ENCODEFLOAT(op1,floatval(op2)-1.0);
      CONTCASE_R;
    }
    else {
      SAVE_H;
      if (eval_arith(op2,&op4)<0) {
	RAISE_EXCEPTION(number_expected,make_struct2("-",op2,MAKEINT(1)));
      }
      RESTORE_H;
      op2 = op4;
      goto sub11;
    }

  CASE(mul)     
  r_mul:
    profile_COUNT_REFERS(3);
    op1 = *P++;OP_CONT_DEREF(op1);
    op2 = *P++;OP_CONT_DEREF(op2);
    op3 = *P++;OP_ADDR(op3);
    goto mul1;

  CASE(div_)    
  r_div:
    profile_COUNT_REFERS(3);
    op1 = *P++;OP_CONT_DEREF(op1);
    op2 = *P++;OP_CONT_DEREF(op2);
    op3 = *P++;OP_ADDR(op3);
    goto div1;

  CASE(idiv)   
  r_idiv:
    profile_COUNT_REFERS(3);
    op1 = *P++;OP_CONT_DEREF(op1);
    op2 = *P++;OP_CONT_DEREF(op2);
    op3 = *P++;OP_ADDR(op3);
    goto idiv1;

  CASE(divge) /* op3 = op1 /> op2: op3*op2>=op1 */
  r_divge:{
    profile_COUNT_REFERS(3);
    op1 = *P++;OP_CONT_DEREF(op1);
    op2 = *P++;OP_CONT_DEREF(op2);
    op3 = *P++;OP_ADDR(op3);
    op1 = INTVAL(op1); op2 = INTVAL(op2);
    tmp_op = op1/op2;
    if (tmp_op*op2<op1) tmp_op++;
    FOLLOW(op3) = MAKEINT(tmp_op);
    CONTCASE_R;
  }

  CASE(divle) /* op3 = op1 /< op2: op3*op2<=op1 */
  r_divle:{
    profile_COUNT_REFERS(3);
    op1 = *P++;OP_CONT_DEREF(op1);
    op2 = *P++;OP_CONT_DEREF(op2);
    op3 = *P++;OP_ADDR(op3);
    op1 = INTVAL(op1); op2 = INTVAL(op2);
    tmp_op = op1/op2+1;
    if (tmp_op*op2>op1) tmp_op--;
    FOLLOW(op3) = MAKEINT(tmp_op);
    CONTCASE_R;
  }
    
  CASE(mod)
  r_mod:
    profile_COUNT_REFERS(3);
    op1 = *P++;OP_CONT_DEREF(op1);
    op2 = *P++;OP_CONT_DEREF(op2);
    op3 = *P++;OP_ADDR(op3);
    goto mod1;

/* Parameter passing */
  CASE(para_struct)
    PS;
    CONTCASE_W;
    
  CASE(para_list)
    PL;
    CONTCASE_W;

  CASE(para_nil)
    PN;
    CONTCASE_R;

  CASE(para_atom)
    PW;
    CONTCASE_R;

  CASE(para_int)
    PW;
    CONTCASE_R;

  CASE(para_ux_ux_ux)
    PUX;
    
  CASE(para_ux_ux)
    PUX;

  CASE(para_ux)
    PUX;
    CONTCASE_R;

  CASE(para_uy_uy_uy_uy)
    PUY;
    
  CASE(para_uy_uy_uy)
    PUY;

  CASE(para_uy_uy)
    PUY;

  CASE(para_uy)
    PUY;
    CONTCASE_R;

  CASE(para_vx)
    PVX;
    CONTCASE_R;

  CASE(para_vy_vy)
    PVY;

  CASE(para_vy)
    PVY;
    CONTCASE_R;

  CASE(para_void_one)
    op1 = (LONG)LOCAL_TOP;
    *LOCAL_TOP-- = op1;
    CONTCASE_R;

  CASE(para_void)
    op1 = *P++;
    while (op1>0) {
      op2 = (LONG)LOCAL_TOP;
      *LOCAL_TOP-- = op2;
      op1--;
    }
    CONTCASE_R;

  CASE(para_vy_ux)
  r_para_vy_ux:
    PVY;
    PUX;
    CONTCASE_R;
    
  CASE(para_vy_uy)
    PVY;
    PUY;
    CONTCASE_R;

  CASE(para_ux_vy)
    PUX;
    PVY;
    CONTCASE_R;
  
  CASE(para_ux_ux_uy)
    PUX;

  CASE(para_ux_uy)
    PUX;
    PUY;
    CONTCASE_R;

  CASE(para_ux_uy_uy)
    PUX;
    PUY;
    PUY;
    CONTCASE_R;

  CASE(para_uy_vy)
    PUY;
    PVY;
    CONTCASE_R;

  CASE(para_ux_uy_ux)
    PUX;

  CASE(para_uy_ux)
    PUY;
    PUX;
    CONTCASE_R;

  CASE(para_uy_ux_ux)
  r_para_uy_ux_ux:
    PUY;
    PUX;
    PUX;
    CONTCASE_R;
    
  CASE(para_uy_ux_uy)
  r_para_uy_ux_uy:
    PUY;
    PUX;
    PUY;
    CONTCASE_R;

  CASE(para_uy_uy_ux)
    PUY;
    PUY;
    PUX;
    CONTCASE_R;
    
/* Procedural instructions */
  CASE(call)
  r_call:
    *LOCAL_TOP = (LONG)AR;
    AR = LOCAL_TOP;
    AR_CPS(AR) = (LONG)(P+1);
    sym_ptr = (SYM_REC_PTR)*P;
    goto call_sub;

  CASE(call_d)
  r_call_d:
    CALL_D;
    CONTCASE_R;

  CASE(callv)
  r_callv:
    op2 = *P++; OP_CONT_DEREF(op2);
    CALLV;
    *LOCAL_TOP = (LONG)AR;
    AR = LOCAL_TOP;
    AR_CPS(AR) = (LONG)P;
    goto call_sub;

  CASE(execute)
  r_execute:
    sym_ptr = (SYM_REC_PTR)*P;
    goto call_sub;
    
  CASE(executev) 
  r_executev:{
    sreg = (LONG_PTR)AR_AR(AR);
    op2 = *P++; OP_CONT_DEREF(op2);
    CPS = AR_CPS(AR);       /* first discard the frame */
    LOCAL_TOP = (LONG_PTR)UNTAGGED(AR_BTM(AR));
    CALLV;                  /* then copy the arguments */
    AR = LOCAL_TOP;
    AR_AR(AR) = (LONG)sreg;
    AR_CPS(AR) = (LONG)CPS;
    goto call_sub;
  }
    
  CASE(return_a)   
    RETURN_A;
    CATCH_WAKE_EVENT;
    CONTCASE_R;
        
  CASE(return_b)
    RETURN_B;
    CATCH_WAKE_EVENT;
    CONTCASE_R;
    
  CASE(return_d)
  r_return_d:{
    sreg = AR;
    top = (LONG_PTR)AR_STATUS(AR);
    P = (LONG_PTR)AR_CPS(AR); 
    AR = (LONG_PTR)AR_AR(AR); 
    if (FOLLOW(top)==SUSP_REENT){
      FOLLOW(top) = SUSP_RUN;
      CONNECT_WOKEN_FRAME(sreg);
      P = (LONG_PTR)FOLLOW(AR_REEP(sreg)); 
    } else {
      FOLLOW(top) = SUSP_SLEEP;
    }
    CATCH_WAKE_EVENT; 
    CONTCASE_R;
  }
    
  CASE(jmpn_det)
    JMPN_DET;
    
  CASE(save_ht_jmp)  
    SAVE_HT_JMP;
    CONTCASE_R;

    /* Allocate Instructions */
  CASE(allocate_flat) 
  r_allocate_flat:{
    op1 = *P++; 
    AR_BTM(AR) = ADDTAG((AR+op1),AR_IS_FLAT);
    op1 = *P++;
    LOCAL_TOP =  AR-op1;

    INCR_REFER_Y("allocate_flat",op1-3);
    AR_TOP(AR) = (LONG)LOCAL_TOP;

#ifdef GC
    INITIALIZE_STACK_VARS(4);
    INVOKE_GC;
#else
    OVERFLOW_C;
#endif
    CATCH_WAKE_EVENT;
    CONTCASE_R;
  }

  CASE(allocate_flat_b)   
  r_allocate_flat_b:{
    op1 = *P++; 
    AR_B(AR) = (LONG)B;
    AR_BTM(AR) = ADDTAG((AR+op1),AR_IS_FLAT_B);
    op1 = *P++;
    LOCAL_TOP =  AR-op1;
    INCR_REFER_Y("allocate_flat",op1-3);
    AR_TOP(AR) = (LONG)LOCAL_TOP;

#ifdef GC
    INITIALIZE_STACK_VARS(5);
    INVOKE_GC;
#else
    OVERFLOW_C;
#endif
    CATCH_WAKE_EVENT;
    CONTCASE_R;
  }

  CASE(allocate_nondet) 
  r_allocate_nondet:{
    op1 = *P++;
    AR_BTM(AR) = ADDTAG((AR+op1),AR_IS_NONDET);
    op1 = *P++;
    LOCAL_TOP = AR - op1;
    INCR_REFER_Y("allocate_flat",op1-7);
    AR_TOP(AR) = (LONG)LOCAL_TOP;

#ifdef GC
    INITIALIZE_STACK_VARS(9);
    INVOKE_GC_NONDET;
#else
    OVERFLOW_C;
#endif

    CATCH_WAKE_EVENT;
    AR_B(AR) = (LONG)B;
    AR_TABLE(AR) = (LONG)AR_TABLE_ADDR(AR);
    AR_H(AR) = (LONG)H;
    AR_T(AR) = (LONG)T;
    AR_SF(AR) = (LONG)SF;
    B = AR;
    HB = H;
    P += 2; /* skip nondet */
    CONTCASE_R;
  }

    /* Backtracking */
  CASE(fail)     
  backtrack: {
    trigger_no = 0;
    toam_signal_vec = 0;
    INCR_REFER_Y("backtrack",3);
    AR = B;
    H = HB;
    SF = (LONG_PTR)AR_SF(AR);

    {LONG_PTR top;
      top = (LONG_PTR)AR_T(AR);
      while (T != top) {
	POPTRAIL(T);
      }
   }
/*
    top = (LONG_PTR)AR_T(AR);
    while (T != top) {
      POPTRAIL(top);
    }
    T = (LONG_PTR)AR_T(AR);
*/

    P = (LONG_PTR)AR_CPF(AR);
    LOCAL_TOP = (LONG_PTR)AR_TOP(AR);
    CONTCASE_R;
  }
  
  CASE(fail0)
  backtrack0:
    trigger_no = 0;
    toam_signal_vec = 0;
    INCR_REFER_Y("backtrack0",2);
    H = (LONG_PTR)HB;

    {LONG_PTR top;
    top = (LONG_PTR)AR_T(AR);
    while (T != top) {
      POPTRAIL(T);
    }
   }

/*
    top = (LONG_PTR)AR_T(AR);
    while (T != top) {
      POPTRAIL(top);
    }
    T = (LONG_PTR)AR_T(AR);
*/
    P = (LONG_PTR)AR_CPF(AR);
    CONTCASE_R;
  
  CASE(fork)
  r_fork:
    FORK;
    CONTCASE_R;
/*
  CASE(commit)
  r_commit:
    SB = (LONG_PTR)AR_B(AR);
    CONTCASE_R;
*/
  CASE(cut)
    CUT;
    CONTCASE_R;

  CASE(cut_fail)
  r_cut_fail:{
    TABLE_CUT(AR);
  rr_cut_fail:
    B = AR = (LONG_PTR)AR_B(AR); 
    LOCAL_TOP = (LONG_PTR)AR_TOP(AR);
    HB = (LONG_PTR)AR_H(AR);
    SF = (LONG_PTR)AR_SF(AR);
    goto backtrack0;
  }
    
  CASE(cut_return)
    CUT;
    RETURN_A;
    CATCH_WAKE_EVENT;
    CONTCASE_R;

  CASE(save_b)  
  r_save_b:
    AR_B(AR) = (LONG)B;
    CONTCASE_R;

  CASE(getbreg)
  r_getbreg:
    profile_COUNT_REFERS(1);
    op1 = *P++;
    OP_CONT_DEREF(op1);
    if (!ISVAR(op1)) {
      RAISE_EXCEPTION(variable_expected,make_struct1("getbreg",op1));
    }
    FOLLOW(op1) = ADDTAG(B,ATM);
    toam_PUSHTRAIL(op1);
    CONTCASE_R;

  CASE(putbreg)
  r_putbreg:{
    profile_COUNT_REFERS(1);
    op1 = *P++;
    OP_CONT_DEREF(op1); UNTAG(op1);

    while (B<(LONG_PTR)op1){ 
      sreg = (LONG_PTR)UNTAGGED(AR_TABLE(B)); 
      if (sreg>tableArea && sreg<tableAreaBottom) GT_AR(sreg) = NULL; 
      B = (LONG_PTR)AR_B(B);
    }

    B = (LONG_PTR)op1;
    HB = (LONG_PTR)AR_H(B);
    CONTCASE_R;
  }

  CASE(getpbreg)
  r_getpbreg:
    profile_COUNT_REFERS(1);
    op1 = *P++;
    OP_CONT_DEREF(op1);
    if (!ISVAR(op1)) {
      RAISE_EXCEPTION(variable_expected,make_struct1("_$savecp",op1));
    }
    if (AR==B)
      FOLLOW(op1) = ADDTAG(AR_B(AR),INT_TAG);
    else 
      FOLLOW(op1) = ADDTAG(B,INT_TAG);
    toam_PUSHTRAIL(op1);
    CONTCASE_R;

/* Hashing */
  CASE(hash_x)    
  r_hash_x:
    op1 = XC(*P);
    goto nhash;

  CASE(hash_y)   
  r_hash_y:
    op1 = YC(*P);
    SWITCH_OP(op1,nhash,
	      {P = (LONG_PTR)*(P+3); CONTCASE_R;},
	      {P = *(LONG_PTR *)(IHASH(op1,*(P+1))*4 + *(P+2));CONTCASE_R;},
	      {P = *(LONG_PTR *)(IHASH((LONG)list_psc, *(P+1))* 4 + *(P+2));
	       CONTCASE_R;},
	      {tmp_op = UNTAGGED(op1);
	       sym_ptr = (SYM_REC_PTR) FOLLOW(tmp_op);
	       P = *(LONG_PTR *)(IHASH((LONG)sym_ptr, *(P+1))*4 + *(P+2)); 
	       CONTCASE_R;},
	      {P = (LONG_PTR)*(P+3);CONTCASE_R;});
  
    /* Builtin */
  CASE(arg)
  r_arg:
      profile_COUNT_REFERS(3);
      op1 = *P++; OP_CONT_DEREF(op1);
    rr_arg:
      if (!ISINT(op1)){
	exception = integer_expected;
        printf("op1=%x *op1=%x\n",op1,FOLLOW(UNTAGGED(op1)));
	op2 = *P++;OP_CONT(op2);
	op3 = *P++; OP_CONT(op3);
	goto error_arg;
      }
      op1 = INTVAL(op1);
    arg_idd:
      op2 = *P++;OP_CONT(op2);
      op3 = *P++;
    rrr_arg:
      if (ISSTRUCT(op2)){
	UNTAG(op2);
	if (op1>0 && op1 <= GET_ARITY((SYM_REC_PTR)FOLLOW(op2))){
	  FAST_UNIFY(op3,*((LONG_PTR)op2 + op1));
	  CONTCASE_R;
	} else BACKTRACK;
      } else if (ISVAR(op2)){
	NDEREF(op2,rrr_arg);
	exception = structure_expected;
	OP_CONT(op3);op1 = MAKEINT(op1);
	goto error_arg;
      } else if (ISLIST(op2)){
	if (op1>0 && op1 <= 2){
	  FAST_UNIFY(op3,*((LONG_PTR)UNTAGGED(op2) + op1 - 1));
	  CONTCASE_R;
	} else BACKTRACK;
      }
      exception = structure_expected;
      op1 = MAKEINT(op1);OP_CONT(op3);
   error_arg:
      RAISE_EXCEPTION(exception,make_struct3("arg",op1,op2,op3));

  CASE(arg0)
    op1 = *P++;
    goto arg_idd;
    
  CASE(setarg) 
    goto setarg1;
    
  CASE(setarg0)
    op1 = *P++;
    goto setarg_idd;
    
  CASE(functor) 
  functor1:
    {
      profile_COUNT_REFERS(3);
      op1 = *P++; op2 = *P++; op3 = *P++;
      switch (op1 & 0x7) { 
      case XV:  
      case XV1: goto functor_vdd;
      case XU: 
      case XU1: op1=XC(((LONG)(op1) & VAL_MASK)); break;
      case YV: op1=(LONG)Y(op1>>3);goto functor_vdd;
      case YU: op1=YC(op1>>3);break;
      default: break;
      }
    SWITCH_OP(op1,functor_udd,
	      {toam_PUSHTRAIL(op1);
	       goto functor_vdd;},
	      
	      {i = MAKEINT(0);},

	      {i = MAKEINT(2);op1 = period_sym;},

	      {sym_ptr = GET_STR_SYM_REC(op1);
	       i = MAKEINT(GET_ARITY(sym_ptr));
	       op1 = ADDTAG(insert(GET_NAME(sym_ptr),GET_LENGTH(sym_ptr),0),ATM);},

	      {if (cfunctor1(op1,op2,op3)<0){
		 RAISE_EXCEPTION(illegal_arguments,make_struct3("functor",op1,op2,op3));} else CONTCASE_R;});
      FAST_UNIFY(op2,op1);
      FAST_UNIFY(op3,i);
      CONTCASE_R;

    functor_vdd:
      OP_CONT_DEREF(op2);
      OP_CONT_DEREF(op3);
      if (ISATOM(op2)) {
	if (ISINT(op3)){
	  sym_ptr = GET_ATM_SYM_REC(op2);
	  op3 = INTVAL(op3);
	  switch (op3) {
	  case 0:
	    FOLLOW(op1) = ADDTAG(insert(GET_NAME(sym_ptr),GET_LENGTH(sym_ptr),op3),ATM);
	    CONTCASE_R;
	  case 2:
	    if (op2 == period_sym){
	      FOLLOW(op1)=ADDTAG(H,LST);
	      break;
	    }
	  case 1:
	  case 3:
	  case 4:
	    FOLLOW(op1) = ADDTAG(H,STR);
	    *H++ = (LONG)insert(GET_NAME(sym_ptr),GET_LENGTH(sym_ptr),op3);
	    break;
	  default:
	    if (op3<0 || op3>255) {
	      RAISE_EXCEPTION(out_of_range,make_struct3("functor",op1,op2,MAKEINT(op3)));
	    }
	    FOLLOW(op1) = ADDTAG(H,STR);
	    *H++ = (LONG)insert(GET_NAME(sym_ptr),GET_LENGTH(sym_ptr),op3);
	    break;
	  }
	  OVERFLOW_C;
	  for (i=0;i<op3;H++,i++)
	    MAKE_FREE(LONG,*H);
	  CONTCASE_R;
	}
	else { /* op3 is not an integer */
	  RAISE_EXCEPTION(integer_expected,make_struct3("functor",op1,op2,op3));
	}
      }
      else if (ISNUM(op2)) {
	FOLLOW(op1) = op2;
	UNIFY_INT_Y(op3,MAKEINT(0),functor_num,{},{},{goto backtrack;});
      }
      else {
	RAISE_EXCEPTION(atom_expected,make_struct3("functor",op1,op2,op3));
      }
    }    
  
  CASE(func_arity)
    goto arity1;

  CASE(get_ar_cps)
  r_get_ar_cps:
    sreg = (LONG_PTR)AR_AR(AR);
    CPS = AR_CPS(AR);
    CONTCASE_R;
    
  CASE(put_ar_cps)
  r_put_ar_cps:
    op1 = *P++;
    AR = AR+op1;
    AR_AR(AR) = (LONG)sreg;
    AR_CPS(AR) = CPS;
    CONTCASE_R;
    
  CASE(move_ar_cps)
  r_move_ar_cps:
    op1 = *P++;
    *(AR+op1) = AR_AR(AR);
    *(AR+op1-1) = AR_CPS(AR);
    AR = AR+op1;
    CONTCASE_R;
    
  CASE(jmpn_det_get_ar_cps)
  r_jmpn_det_get_ar_cps:
    if (AR_IS_TOP && AR<B) {   /* det */
      sreg = (LONG_PTR)AR_AR(AR);
      CPS = AR_CPS(AR);
      P++;
      CONTCASE_R;
    }
    else {                 /* not det */
      P = (LONG_PTR)*P;
      CONTCASE_R;
    }

  CASE(builtin0)
  r_builtin0:
    i = (WORD)*P++;
    SAVE_H_T;
    tmp_op = (*builtins[i])();
    RESTORE_H;
    if (tmp_op==1){
      P++;
      CONTCASE_R;
    }
    else if (tmp_op==0) {
      P = (LONG_PTR)*P;
      CONTCASE_R;
    }
    else {
      RAISE_EXCEPTION(exception,make_struct1("builtin",MAKEINT(i)));
    }

  CASE(builtin1)
  r_builtin1:
    i = (WORD)*P;
    P += 2;
    profile_COUNT_REFERS(1);
    OPERAND_BA1;
    SAVE_H_T;
    tmp_op = (*builtins[i])(op1);
    RESTORE_H;
    if (tmp_op==1){
      CONTCASE_R;
    }
    else if (tmp_op==0) {
      P = (LONG_PTR)*(P-2);
      CONTCASE_R;
    }
    else {
      RAISE_EXCEPTION(exception,make_struct2("builtin",MAKEINT(i),op1));
    }

  CASE(builtin2)
  r_builtin2:
    i = (WORD)*P;
    P += 2;
    profile_COUNT_REFERS(2);
    OPERAND_BA2;
    SAVE_H_T;
    tmp_op = (*builtins[i])(op1,op2);
    RESTORE_H;
    if (tmp_op==1){
      CONTCASE_R;
    }
    else if (tmp_op==0) {
      P = (LONG_PTR)*(P-3);
      CONTCASE_R;
    }
    else {
      RAISE_EXCEPTION(exception,make_struct3("builtin",MAKEINT(i),op1,op2)); 
    }

    
  CASE(builtin3)
  r_builtin3:
    i = (WORD)*P;
    P += 2;
    profile_COUNT_REFERS(3);
    OPERAND_BA3;
    SAVE_H_T;
    tmp_op = (*builtins[i])(op1,op2,op3);
    RESTORE_H;
    if (tmp_op==1){
      CONTCASE_R;
    }
    else if (tmp_op==0) {
      P = (LONG_PTR)*(P-4);
      CONTCASE_R;
    }
    else {
      RAISE_EXCEPTION(exception,make_struct4("builtin",MAKEINT(i),op1,op2,op3));
    }

  CASE(builtin4)
  r_builtin4:
    i = (WORD)*P;
    P += 2;
    profile_COUNT_REFERS(4);
    OPERAND_BA4;
    SAVE_H_T;
    tmp_op = (*builtins[i])(op1,op2,op3,op4);
    RESTORE_H;
    if (tmp_op==1){
      CONTCASE_R;
    }
    else if (tmp_op==0) {
      P = (LONG_PTR)*(P-5);
      CONTCASE_R;
    }
    else {
      RAISE_EXCEPTION(exception,make_struct5("builtin",MAKEINT(i),op1,op2,op3,op4)); 
    }
    

 CASE(allocate_susp)
  r_allocate_susp:{
    op1 = *P++;
    AR_BTM(AR) = ADDTAG((AR+op1),AR_IS_SUSP);
    LOCAL_TOP = AR-(*P++);
    AR_TOP(AR) = (LONG)LOCAL_TOP;
#ifdef GC
    INITIALIZE_STACK_VARS(7);
    INVOKE_GC;
#else
    OVERFLOW_C;
#endif
    top = H++;
    AR_STATUS(AR) = (LONG)top; FOLLOW(top) = SUSP_START;
    CATCH_WAKE_EVENT;
    CONTCASE_R;
  }

 CASE(trigger_var_ins)
 r_trigger_var_ins:
   {op2 = *P++; OP_CONT(op2); DEREF(op2);
    SWITCH_OP(op2,rr_trigger_var_ins,
	      {CREATE_SUSP_VAR_ins(op2);},
	      {},
	      {SAVE_REGS;trigger_vars_ins(op2);RESTORE_REGS;},
	      {SAVE_REGS;trigger_vars_ins(op2);RESTORE_REGS;},
	      {INSERT_SUSP_CALL_ins(op2);});
    CONTCASE_R;
  }

 CASE(trigger_var_min)
 r_trigger_var_min:
   {op2 = *P++; OP_CONT(op2); DEREF(op2);
    SWITCH_OP(op2,rr_trigger_var_min,
	      {CREATE_SUSP_VAR_min(op2);},
	      {},
	      {SAVE_REGS;trigger_vars_min(op2);RESTORE_REGS;},
	      {SAVE_REGS;trigger_vars_min(op2);RESTORE_REGS;},
	      {INSERT_SUSP_CALL_min(op2);});
    CONTCASE_R;
  }

 CASE(trigger_var_max)
 r_trigger_var_max:
   {op2 = *P++; OP_CONT(op2); DEREF(op2);
    SWITCH_OP(op2,rr_trigger_var_max,
	      {CREATE_SUSP_VAR_max(op2);},
	      {},
	      {SAVE_REGS;trigger_vars_max(op2);RESTORE_REGS;},
	      {SAVE_REGS;trigger_vars_max(op2);RESTORE_REGS;},
	      {INSERT_SUSP_CALL_max(op2);});
    CONTCASE_R;
  }

 CASE(trigger_var_dom)
 r_trigger_var_dom:
   {op2 = *P++; OP_CONT(op2); DEREF(op2);
    SWITCH_OP(op2,rr_trigger_var_dom,
	      {CREATE_SUSP_VAR_dom(op2);},
	      {},
	      {SAVE_REGS;trigger_vars_dom(op2);RESTORE_REGS;},
	      {SAVE_REGS;trigger_vars_dom(op2);RESTORE_REGS;},
	      {INSERT_SUSP_CALL_dom(op2);});
    CONTCASE_R;
  }

  CASE(delay)  /* AASA */ /* sreg used */
  r_delay: {
    if (FRAME_IS_START(AR)){
      CONNECT_SUSP_FRAME;
      AR_REEP(AR) = (LONG)P;
      P += 4;
      CONTCASE_R;
    } if (AR_REEP(AR) == (LONG)P){ /* re-entrance already set */
      P = (LONG_PTR)FOLLOW((P+3));
      CONTCASE_R;
    } else {
      RESET_AR_REEP(P);
      P += 4;
      CONTCASE_R;
    }  
  }

 CASE(end_delay)
 r_end_delay:
   if (!FRAME_IS_START(AR)){
     DISCONNECT_SUSP_FRAME;
   } else {
     FOLLOW(AR_STATUS(AR)) = SUSP_EXIT;
   }
  CONTCASE_R;
    
 CASE(nondet)
 r_nondet:
   if ((LONG)LOCAL_TOP == AR_TOP(AR)){ /* turn this frame into a nondet frame */
      AR_B(AR) = (LONG)B;
      AR_H(AR) = (LONG)H;
      AR_T(AR) = (LONG)T;
      AR_SF(AR) = (LONG)SF;
      AR_TABLE(AR) = (LONG)AR_TABLE_ADDR(AR);
      B = AR;
      HB = H;
      P++;
      CONTCASE_R;
    }
    else {
      op1 = *(P-2); /* size in allocate_flat*/
      op2 = *P++; /* arity */
      for (i=op2;i>0;i--)  /* copy arguments */
	*LOCAL_TOP-- = FOLLOW(AR+i);
      *LOCAL_TOP-- = FOLLOW(AR);
      *LOCAL_TOP = AR_CPS(AR);
      AR = (LOCAL_TOP+1);
      LOCAL_TOP = AR - op1; 
      AR_BTM(AR) = ADDTAG((AR+op2),AR_IS_NONDET);
      AR_TOP(AR) = (LONG)LOCAL_TOP;
      AR_B(AR) = (LONG)B;
      AR_H(AR) = (LONG)H;
      AR_T(AR) = (LONG)T;
      AR_SF(AR) = (LONG)SF;
      AR_TABLE(AR) = (LONG)AR_TABLE_ADDR(AR);
#ifdef GC
    INITIALIZE_STACK_VARS(9);
#endif
      B = AR;
      HB = H;
/*      SHOW_NAR(AR,op2); */
      CONTCASE_R;
    }
    
  CASE(jmpn_dvar_y)
  r_jmpn_dvar_y:{
    OPERAND_Y;
    DEREF(op1);
    if (IS_SUSP_VAR(op1)){
      dv_ptr = (LONG_PTR)UNTAGGED(op1);
      if (!IS_UN_DOMAIN(dv_ptr)){
	P++;
	CONTCASE_R;
      }
    }
    P = *(LONG_PTR *)P;
    CONTCASE_R;
  }
  
  CASE(jmpn_susp_var_y)
  r_jmpn_susp_var_y:{
    OPERAND_Y;
    DEREF(op1);
    if (IS_SUSP_VAR(op1)){
      P++;
      CONTCASE_R;
    } else {
      P = *(LONG_PTR *)P;
      CONTCASE_R;
    }
  }

/*
 CASE(jmp_susp)
  r_jmp_susp:
    P = *(LONG_PTR *)P;
    AR_PREV(AR) = (LONG)NULL;
    OVERFLOW_C;
    CONTCASE_R;
*/

  CASE(domain_set_false_yx)
  r_domain_set_false_yx:
    {OPERAND_YX;
     goto domain_set_false1;}

  CASE(domain_set_false_yy)
  r_domain_set_false_yy:
    {OPERAND_YY;
     goto domain_set_false1;}

  CASE(domain_set_false)
  r_domain_set_false:
    {op1 = *P++;
     OP_CONT(op1);
     op2 = *P++;
     OP_CONT(op2);
     goto domain_set_false1;}

  CASE(no_vars_gt) /* NNA */
  r_no_vars_gt:{
    arity = *P++;
    op1 = *P++; /* limit */
    op3 = 0;  /* total no of vars */
    while (arity>0){
      op2 = YC(arity);
      SWITCH_OP(op2,rr_no_vars_gt,
		{op3++;},
		
		{},

		{op3 += count_vars(op2);},

		{op3 += count_vars(op2);},

		{op3++;}
		);
      if (op3>op1) {P++; CONTCASE_R;};
      arity--;
    }
    P = (LONG_PTR)*P;
    CONTCASE_R;
  }
     
  CASE(domain_min_max)
  r_domain_min_max:
    {op1 = *P++; OP_CONT(op1);
     DEREF_NONVAR(op1);
     if (ISINT(op1)){
       op2 = *P++;
       FOLLOW(op2) = op1;
       op2 = *P++;
       FOLLOW(op2) = op1;
       CONTCASE_R;
     } else {
       dv_ptr = (LONG_PTR)UNTAGGED(op1); 
       op1 = MAKEINT(DV_first(dv_ptr)); 
       op2 = *P++;
       FOLLOW(op2) = op1;
       op1 = MAKEINT(DV_last(dv_ptr)); 
       op2 = *P++;
       FOLLOW(op2) = op1;
       CONTCASE_R;
     } 
   }

  CASE(domain_min)
  r_domain_min:
    {op1 = *P++; OP_CONT(op1);
     DEREF_NONVAR(op1);
     if (ISINT(op1)){
       op2 = *P++;
       FOLLOW(op2) = op1;
       CONTCASE_R;
     } else {
       dv_ptr = (LONG_PTR)UNTAGGED(op1); 
       op1 = MAKEINT(DV_first(dv_ptr)); 
       op2 = *P++;
       FOLLOW(op2) = op1;
       CONTCASE_R;
     } 
   }

  CASE(domain_max)
  r_domain_max:
    {op1 = *P++; OP_CONT(op1);
     DEREF_NONVAR(op1);
     if (ISINT(op1)){
       op2 = *P++;
       FOLLOW(op2) = op1;
       CONTCASE_R;
     } else {
       dv_ptr = (LONG_PTR)UNTAGGED(op1); 
       op1 = MAKEINT(DV_last(dv_ptr)); 
       op2 = *P++;
       FOLLOW(op2) = op1;
       CONTCASE_R;
     } 
   }

  CASE(domain_region)
  r_domain_region:
    {OPERAND_BA3;
     DEREF_NONVAR(op1); DEREF_NONVAR(op2); DEREF_NONVAR(op3);
     op2 = INTVAL(op2); op3 = INTVAL(op3);
     if (IS_SUSP_VAR(op1))
       goto domain_region_1;
     else if (ISINT(op1)){
       op1 = INTVAL(op1);
       if (op1 >= op2 && op1 <= op3)
	 CONTCASE_R;
       else
	 BACKTRACK;
     } else {
       dv_ptr = (LONG_PTR)UNTAGGED(op1);
       if (op2 > DV_first(dv_ptr)) DV_first(dv_ptr) = op2;
       if (op3 < DV_last(dv_ptr)) DV_last(dv_ptr) = op3;
       if (DV_first(dv_ptr) > DV_last(dv_ptr)) BACKTRACK;
       CONTCASE_R;
     }
   }

  CASE(domain_region_min)
  r_domain_region_min:
    {OPERAND_BA2;
     DEREF_NONVAR(op1); DEREF_NONVAR(op2); 
     op2 = INTVAL(op2); 
/*     printf("region_min %x %d\n",op1,op2); */
     if (IS_SUSP_VAR(op1))
       goto domain_region_min_1;
     else if (ISINT(op1)){
       op1 = INTVAL(op1);
       if (op1 >= op2)
	 CONTCASE_R;
       else
	 BACKTRACK;
     } else {
       dv_ptr = (LONG_PTR)UNTAGGED(op1);
       if (op2 > DV_first(dv_ptr)) {
	 if (op2 > DV_last(dv_ptr)) BACKTRACK;
	 DV_first(dv_ptr) = op2;
       }
       CONTCASE_R;
     }
   } 

  CASE(domain_region_max)
  r_domain_region_max:
    {op1 = *P++; OP_CONT(op1); DEREF_NONVAR(op1); 
     op3 = *P++; OP_CONT(op3); DEREF_NONVAR(op3); 
     op3 = INTVAL(op3); 
/*     printf("region_max %x %d\n",op1,op3); */
     if (IS_SUSP_VAR(op1))
       goto domain_region_max_1;
     else if (ISINT(op1)){
       op1 = INTVAL(op1);
       if (op1 <= op3)
	 CONTCASE_R;
       else
	 BACKTRACK;
     } else {
       dv_ptr = (LONG_PTR)UNTAGGED(op1);
       if (op3 < DV_last(dv_ptr)) {
	 if (op3 < DV_first(dv_ptr)) BACKTRACK;
	 DV_last(dv_ptr) = op3;
       }
       CONTCASE_R;
     } 
   } 
/*
  CASE(first_indomain)
  r_first_indomain:
    {OPERAND_BA1;
     SWITCH_OP_VAR(op1,rr_first_indomain,
		   {CONTCASE_R;},
		   {},
		   {CONTCASE_R;});
     dv_ptr = (LONG_PTR)UNTAGGED(op1); 
     *LOCAL_TOP-- = (LONG)dv_ptr;
     op2 = DV_first(dv_ptr);
     *LOCAL_TOP-- = op2;

     *(LOCAL_TOP-4) = (LONG)B;  
     B = LOCAL_TOP;
     AR_AR(B) = (LONG)AR;
     AR_CPS(B) = (LONG)P;
     AR_BTM(B) = ADDTAG((AR+2),AR_IS_NONDET);
     LOCAL_TOP -= 9;
     OVERFLOW_C;
     AR_TOP(B) = (LONG)LOCAL_TOP;
     AR_CPF(B) = (LONG)addr_rest_indomain;
     AR_H(B) = (LONG)H;
     AR_T(B) = (LONG)T;
     AR_SF(B) = (LONG)SF;
     HB = H;

     toam_PUSHTRAILC(dv_ptr,op1);
     FOLLOW(dv_ptr) = MAKEINT(op2);
     INSERT_TRIGGER_dvar_ins(dv_ptr);
     goto trigger_on_handler;
   }

  CASE(rest_indomain)
  r_rest_indomain:
    {dv_ptr = (LONG_PTR)YC(2);
     op1 = YC(1);
     if (op1 >= DV_last(dv_ptr)) goto r_cut_fail;
     if (IS_BV_DOMAIN(dv_ptr))
       op1 = domain_next_bv(dv_ptr,op1+1);
     else op1++;
     YC(1) = op1;
     toam_PUSHTRAILC(dv_ptr,FOLLOW(dv_ptr));
     FOLLOW(dv_ptr) = MAKEINT(op1);
     INSERT_TRIGGER_dvar_ins(dv_ptr);
     RETURN_B;
     goto trigger_on_handler;
   } 
*/

  CASE(domain_min_max_u)
  r_domain_min_max_u:{
    op1 = *P++; OP_CONT_DEREF(op1); op1 = INTVAL(op1);  /* A */
    op2 = *P++; /* V */
    OP_Y_CONT(op2); DEREF_NONVAR(op2);
    if (IS_SUSP_VAR(op2)){ 
      dv_ptr = (LONG_PTR)UNTAGGED(op2); 
      op3 = *P++;
      FOLLOW(op3) = MAKEINT(op1*DV_first(dv_ptr)); 
      op3 = *P++;
      FOLLOW(op3) = MAKEINT(op1*DV_last(dv_ptr)); 
      CONTCASE_R;
    } else { 
      op2 = MAKEINT(op1*INTVAL(op2));
      op3 = *P++;
      FOLLOW(op3) = op2;
      op3 = *P++;
      FOLLOW(op3) = op2;
      CONTCASE_R;
    } 
  }

  CASE(domain_min_u)
  r_domain_min_u:{
    op1 = *P++; OP_CONT(op1);DEREF_NONVAR(op1); op1 = INTVAL(op1);  /* A */
    op2 = *P++; /* V */
    OP_Y_CONT(op2); DEREF_NONVAR(op2);
    if (IS_SUSP_VAR(op2)){ 
      dv_ptr = (LONG_PTR)UNTAGGED(op2); 
      op3 = *P++;
      FOLLOW(op3) = MAKEINT(op1*DV_first(dv_ptr)); 
      CONTCASE_R;
    } else { 
      op2 = MAKEINT(op1*INTVAL(op2));
      op3 = *P++;
      FOLLOW(op3) = op2;
      CONTCASE_R;
    } 
  }

  CASE(domain_max_u)
  r_domain_max_u:{
    op1 = *P++; OP_CONT(op1);DEREF_NONVAR(op1); op1 = INTVAL(op1);  /* A */
    op2 = *P++; /* V */
    OP_Y_CONT(op2); DEREF_NONVAR(op2);
    if (IS_SUSP_VAR(op2)){ 
      dv_ptr = (LONG_PTR)UNTAGGED(op2); 
      op3 = *P++;
      FOLLOW(op3) = MAKEINT(op1*DV_last(dv_ptr)); 
      CONTCASE_R;
    } else { 
      op2 = MAKEINT(op1*INTVAL(op2));
      op3 = *P++;
      FOLLOW(op3) = op2;
      CONTCASE_R;
    } 
  }

/* Miscellaneous instructions */
  CASE(halt)   
  r_halt:
    SAVE_H;
    return 1;

  CASE(halt0)   
  r_halt0:
    SAVE_H;
    return 0;

  CASE(endfile)  
  r_endfile:
    OPERAND_W;
    CONTCASE_R;

  CASE(table_allocate) /* N,N,S,A*/
  r_table_allocate:{
    LONG_PTR subgoalTableEntryPtr,pioneerTableEntryPtr;
    op1 = *P++; /* Arity */
    op2 = *P++; /* Frame size */
    op3 = *P++; /* Predicate symbol */
    AR -= op1;
    FOLLOW(AR-1) = FOLLOW(AR+op1-1);
    FOLLOW(AR) = FOLLOW(AR+op1);
/*
    printf("table_allocate\n");
*/
    AllocateSubgoalTableEntry(subgoalTableEntryPtr,op1);
    GT_SYM(subgoalTableEntryPtr) = op3;
    SAVE_AR_H_T;
    copyTabledCallArgs(AR+2*op1,GT_ARG_ADDR(subgoalTableEntryPtr),AR+op1,op1,AR-2);
    RESTORE_H;
    
    pioneerTableEntryPtr  = registerAndLookupSubgoalTable(subgoalTableEntryPtr,op3,op1);  
    if ((LONG)pioneerTableEntryPtr==NULL){ /* pioneer */
/*
      printf("%s ",GET_NAME((SYM_REC_PTR)op3));
      for (i=op1;i>0;i--){
	write_term(FOLLOW(AR+i)); printf("  ");
      }
      printf("\n");
*/
      GT_AR(subgoalTableEntryPtr) = (LONG)AR;
      top = allocateAnswerTable();
      GT_ANSWERTABLE(subgoalTableEntryPtr) = (LONG)top;
      AR_TABLE(AR) = (LONG)subgoalTableEntryPtr;
      AR_CURRA(AR) = ANSWERTABLE_FIRST(top);
      AR_PIONEER(AR) = (LONG)AR;
      P++; /* skip address of the last clause */
    } else {  /* follower */
      ReleaseSubgoalTableEntry(subgoalTableEntryPtr); 
      AR_CURRA(AR) = ANSWERTABLE_FIRST((LONG_PTR)GT_ANSWERTABLE(pioneerTableEntryPtr));
      sreg = (LONG_PTR)GT_AR(pioneerTableEntryPtr); /* pioneer's frame */
      if ((LONG)sreg==TABLE_COMPLETE){
	P = (LONG_PTR)*P;
	AR_TABLE(AR) = (LONG)pioneerTableEntryPtr;
      } else if ((LONG)sreg!=NULL){ /* steal the choice point */
	top = (LONG_PTR)AR_CPF(sreg);
	AR_CPF(sreg) = *P;
	P = top;
	AR_TABLE(AR) = (LONG)pioneerTableEntryPtr;
	AR_PIONEER(AR) = AR_PIONEER(sreg);
	GT_AR(pioneerTableEntryPtr) = (LONG)AR; 
      } else { 
	AR_TABLE(AR) = (LONG)pioneerTableEntryPtr;
	AR_PIONEER(AR) = (LONG)AR;
	P++;
	GT_AR(pioneerTableEntryPtr) = (LONG)AR; 
      }
    }
    AR_CPF(AR) = (LONG)P;
    
    AR_BTM(AR) = ADDTAG((AR+2*op1),AR_IS_TABLE);
    LOCAL_TOP = AR - op2;
    AR_TOP(AR) = (LONG)LOCAL_TOP;
    OVERFLOW_C;
#ifdef GC
    INITIALIZE_STACK_VARS(11);
#endif
    AR_B(AR) = (LONG)B;
    AR_H(AR) = (LONG)H;
    AR_T(AR) = (LONG)T;
    AR_SF(AR) = (LONG)SF;
    B = AR;
    HB = H;
    CONTCASE_R;
  }
    
  CASE(table_add_use_answer) /* N */
  r_table_add_use_answer:{
/*
    printf("table_add_use H=%d\n",H);
    printTable();
*/
    SAVE_REGS;
    op1 = *P++;
    addAnswerAnswerTable(AR,op1);
    if (useAnswerAnswerTable(AR,op1,LOCAL_TOP)!=0){
      RESTORE_H_T;
      RETURN_B;
      CONTCASE_R;
    } else goto backtrack;
  }

  CASE(table_use_answer_fork) /* NA */
  r_table_use_answer_fork:{

    SAVE_REGS;
    op1 = *P++;
    if (useAnswerAnswerTable(AR,op1,LOCAL_TOP)!=0){
      RESTORE_H_T;
      RETURN_B;
      CONTCASE_R;
    } else {
      AR_CPF(AR) = *P++;
      CONTCASE_R;
    }
  } 

  CASE(table_check_completion) /* A */
  r_table_check_completion:{

/*
    printf("table_check_completion\n");
    printTable();
*/
    sreg = (LONG_PTR)AR_TABLE(AR); /* goal table */
  rr_table_check_completion:
    if (!IS_PIONEER(AR)){
      top = (LONG_PTR)GT_ANSWERTABLE(sreg); /*answer table */
      if (RevisedAnswerTable(top)){
/*	printf("%x recompute %s\n",sreg,GET_NAME((SYM_REC_PTR)GT_SYM(sreg))); */
	ResetRevisedAnswerTable(top);
	GT_AR(sreg) = (LONG)AR;
	P = (LONG_PTR)*P;
	AR_CPF(AR) = (LONG)P;
	CONTCASE_R;
      } 
    }
    GT_AR(sreg) = TABLE_COMPLETE;
    goto rr_cut_fail;
  }
  
  CASE(table_cut) /* A */
  r_table_cut:{
    op1 = *P++;
    AR_CPF(AR) = (LONG)op1;
    while (B<AR){ /* not until equal */
      sreg = (LONG_PTR)AR_TABLE(B); 
      if (sreg>tableArea && sreg<tableAreaBottom) GT_AR(sreg) = NULL; 
      B = (LONG_PTR)AR_B(B);
    }
    B = AR;
    CONTCASE_R;
  }

  CASE(table_end) /* A */
  r_table_end:{
    sreg = (LONG_PTR)AR_TABLE(AR); 
    if (IS_PIONEER(AR) || AR_PIONEER(AR)==AR_AR(AR)){
      GT_AR(sreg) = TABLE_COMPLETE;
      goto rr_cut_fail;
    } 
    goto rr_table_check_completion;
  }

  CASE(interval_consistent_eq) /* N */
  r_interval_consistent_eq:{
    op1 = *P++; 
    SAVE_REGS;
    if (nary_interval_consistent(op1,0)==0){
      RESTORE_T;
      BACKTRACK;
    } else {
      RESTORE_H_T;
      CONTCASE_R;
    }
  }  

  CASE(interval_consistent_ge) /* N */
  r_interval_consistent_ge:{
    op1 = *P++; 
    SAVE_REGS;
    if (nary_interval_consistent(op1,9999999)==0){
      RESTORE_T;
      BACKTRACK;
    } else {
      RESTORE_H_T;
      CONTCASE_R;
    }
  }  

  CASE(nary_to_binary)
  r_nary_to_binary:{
    LONG op5,op6;
    OPERAND_BA4;
    op5 = *P++; OP_CONT(op5);
    op6 = *P++; OP_CONT(op6);
    SAVE_AR;
    inst_nary_to_binary(op1,op2,op3,op4,op5,op6);
    CONTCASE_R;
  }

  CASE(trigger_ins_min_max)
  r_trigger_ins_min_max:{
    op1 = *P++; 

    sreg = AR+2*op1+1; /* constant */
    op2 = FOLLOW(sreg);DEREF_NONVAR(op2);FOLLOW(sreg)=op2;
    for (i=1;i<=op1;i++){
      sreg = AR+op1+i; /* coe */
      op2 = FOLLOW(sreg);DEREF_NONVAR(op2);FOLLOW(sreg)=op2;
    }
    
    for (i=1;i<=op1;i++){
      op2 = FOLLOW(AR+i);DEREF_NONVAR(op2);
      if (IS_SUSP_VAR(op2)){
	INSERT_SUSP_CALL_ins(op2);
	INSERT_SUSP_CALL_max(op2);
	INSERT_SUSP_CALL_min(op2);
      }
    }
    CONTCASE_R;
  }
  
  CASE(interval_consistent_eq_nocoe) /* N */
  r_interval_consistent_eq_nocoe:{
    op1 = *P++; 
    SAVE_REGS;
    if (nary_interval_consistent_nocoe(op1,0)==0){
      RESTORE_T;
      BACKTRACK;
    } else {
      RESTORE_H_T;
      CONTCASE_R;
    }
  }  

 default:
    printf("\nIllegal opcode hex %x at %x\n", *--P, P);
    exit(1);
}  /* end switch: read */

  /***********************************************************************/

 /* unify_lst_d: */
  SWITCH_OP(op2,unify_lst_d,
	    {FOLLOW(op2) = op1;
	     toam_PUSHTRAIL(op2);
	     CONTCASE_R;},
	    {BACKTRACK;},
	    {if (op1 != op2) {
	      UNTAG(op1);
	      UNTAG(op2);
	      op3 = *(LONG_PTR)op1;
	      tmp_op = *(LONG_PTR)op2;
	      toam_CALL_UNIFY(op3,tmp_op,backtrack);
	      op1 = *((LONG_PTR)op1 + 1);
	      op2 = *((LONG_PTR)op2 + 1);
	      SWITCH_OP(op1,unify_lst_d_1,
			{goto unify_nsv_d;},
			{UNIFY_ATOMIC21_X(op2,op1,unify_lst_d_2,{goto backtrack;});},
			{goto unify_lst_d;},
			{goto unify_str_d;},
			{goto unify_susp_d;});
	    }
	     CONTCASE_R;},
	    {BACKTRACK;},
	    {goto bind_value_susp_var;});

/* unify_str_d: */
  SWITCH_OP_STRUCT(op2,unify_str_d,
	    {FOLLOW(op2) = op1;
	     toam_PUSHTRAIL(op2);
	     CONTCASE_R;},
	    {if (op1 != op2) {
	      UNTAG(op1);
	      UNTAG(op2);
	      sym_ptr = (SYM_REC_PTR)FOLLOW(op1);
	      if ((LONG)sym_ptr != FOLLOW(op2))
		BACKTRACK;
	      else {
		register long i;
		arity = GET_ARITY(sym_ptr)-1;
		for (i = 1; i <= arity; i++){
		  op3 = *((LONG_PTR)op1+i);
		  tmp_op = *((LONG_PTR)op2+i);
		  toam_CALL_UNIFY(op3,tmp_op,backtrack);
		}
		op1 = *((LONG_PTR)op1+i);
		op2 = *((LONG_PTR)op2+i);
		SWITCH_OP(op1, unify_str_d_1,
			  {goto unify_nsv_d;},
			  {UNIFY_ATOMIC21_X(op2,op1,unify_str_d_2,{goto backtrack;});},
			  {goto unify_lst_d;},
			  {goto unify_str_d;},
			  {goto unify_susp_d;});
	      }
	    }
	     CONTCASE_R;},
	    {goto bind_value_susp_var;});
  BACKTRACK;


/* unify0_lst_d: */
  SWITCH_OP(op2,unify0_lst_d,
	    {FOLLOW(op2) = op1;
	     toam_PUSHTRAIL_UC_0(op2);
	     CONTCASE_R;},
	    {BACKTRACK0;},
	    {if (op1 != op2) {
	      UNTAG(op1);
	      UNTAG(op2);
	      op3 = *(LONG_PTR)op1;
	      tmp_op = *(LONG_PTR)op2;
	      toam_CALL_UNIFY(op3,tmp_op,backtrack0);
	      op1 = *((LONG_PTR)op1 + 1);
	      op2 = *((LONG_PTR)op2 + 1);
	      SWITCH_OP(op1,unify0_lst_d_1,
			{goto unify_nsv_d;},
			{UNIFY_ATOMIC21_X(op2,op1,unify0_lst_d_2,{goto backtrack0;});},
			{goto unify0_lst_d;},
			{goto unify0_str_d;},
			{goto unify_susp_d;});
	    }
	     CONTCASE_R;},
	    {BACKTRACK0;},
	    {goto bind_value_susp_var;});
  
/* unify0_str_d: */
  SWITCH_OP_STRUCT(op2,unify0_str_d,
	    {NDEREF(op2,unify0_str_d);
	     FOLLOW(op2) = op1;
	     toam_PUSHTRAIL_UC_0(op2);
	     CONTCASE_R;},
	    {if (op1 != op2) {
	      UNTAG(op1);
	      UNTAG(op2);
	      sym_ptr = (SYM_REC_PTR)FOLLOW(op1);
	      if ((LONG)sym_ptr != FOLLOW(op2))
		BACKTRACK;
	      else {
		register long i;
		arity = GET_ARITY(sym_ptr)-1;
		for (i = 1; i <= arity; i++){
		  op3 = *((LONG_PTR)op1+i);
		  tmp_op = *((LONG_PTR)op2+i);
		  toam_CALL_UNIFY(op3,tmp_op,backtrack0);
		}
		op1 = *((LONG_PTR)op1+i);
		op2 = *((LONG_PTR)op2+i);
		SWITCH_OP(op1, unify0_str_d_1,
			  {goto unify_nsv_d;},
			  {UNIFY_ATOMIC21_X(op2,op1,unify0_str_d_2,{goto backtrack0;});},
			  {goto unify0_lst_d;},
			  {goto unify0_str_d;},
			  {goto unify_susp_d;});
	      }
	    }
	     CONTCASE_R;},
	    {goto bind_value_susp_var;});
  BACKTRACK0;


 unify_v_susp:
  FOLLOW(op1) = UNTAGGED(op2);
  toam_PUSHTRAIL(op1);
  CONTCASE_R;
  
  /* unify_susp_d: */
  SWITCH_OP_VAR(op2,unify_susp_d,
		{FOLLOW(op2) = UNTAGGED(op1); 
		 toam_PUSHTRAIL(op2); 
		 CONTCASE_R;},
		{goto unify_susp_vars;},
		{goto bind_susp_var_value;});

 bind_susp_var_value:
  BIND_SUSP_VAR_VALUE(op1,op2);

 bind_value_susp_var:
  BIND_SUSP_VAR_VALUE(op2,op1);
  
 bind_susp_var_str:
  BIND_SUSP_VAR_STR(op1,op2);
  
 bind_susp_var_list:
  BIND_SUSP_VAR_LIST(op1);
  
  /*------------------------------------------------------------------*/
 unify_susp_vars: /* both are suspending vars */
  SAVE_REGS;
  if (!unify_suspvar_suspvar(op1,op2)){
    RESTORE_REGS;
    BACKTRACK;
  } 
  RESTORE_REGS;
  CONTCASE_R;
  
  /*------------------------------------------------------------------*/
 interrupt_handler:
  B = (LONG_PTR)UNTAGGED((LONG)GET_EP(interrupt_psc));
  HB = (LONG_PTR)*(B-6);
  toam_signal_vec &= ~INTERRUPT;
  goto backtrack;

  /*------------------------------------------------------------------*/
 trigger_on_handler:
  arity = trigger_no;
  while (arity>0){
    trigger_var = triggers[arity];
    i = event_flag[arity--];
/*
    printf("handler %x trigger_no=%d trigger=%x\n",i,arity,trigger_var);  
*/
    switch (i) {
    case EVENT_VAR_INS:
      sreg = (LONG_PTR)DV_ins_cs(trigger_var);
      while (ISLIST((LONG)sreg)){
	sreg = (LONG_PTR)UNTAGGED(sreg);
	CONNECT_WOKEN_FRAME_ins(sreg);
	sreg = (LONG_PTR)CS_NEXT(sreg);
      }
      break;
    case EVENT_DVAR_INS:
      sreg = (LONG_PTR)DV_ins_cs(trigger_var);
      while (ISLIST((LONG)sreg)){
	sreg = (LONG_PTR)UNTAGGED(sreg);
	CONNECT_WOKEN_FRAME_ins(sreg);
	sreg = (LONG_PTR)CS_NEXT(sreg);
      }
      break;
    case EVENT_DVAR_MAX:
      sreg = (LONG_PTR)DV_max_cs(trigger_var); 
      while (ISLIST((LONG)sreg)){
	sreg = (LONG_PTR)UNTAGGED(sreg);
	CONNECT_WOKEN_FRAME_min_max(sreg);
	sreg = (LONG_PTR)CS_NEXT(sreg);
      }
      break;
    case EVENT_DVAR_MIN:
      sreg = (LONG_PTR)DV_min_cs(trigger_var); 
      while (ISLIST((LONG)sreg)){
	sreg = (LONG_PTR)UNTAGGED(sreg);
	CONNECT_WOKEN_FRAME_min_max(sreg);
	sreg = (LONG_PTR)CS_NEXT(sreg);
      }
      break;
    case EVENT_DVAR_DOM:
      sreg = (LONG_PTR)DV_dom_cs(trigger_var); 
      while (ISLIST((LONG)sreg)){
	sreg = (LONG_PTR)UNTAGGED(sreg);
	CONNECT_WOKEN_FRAME_min_max(sreg);
	sreg = (LONG_PTR)CS_NEXT(sreg);
      }
      break;
    }
  }
  trigger_no = 0;
  toam_signal_vec &= ~TRIGGER_ON;
  CONTCASE_R;
  
 
/*------------------------------------------------------------------*/
 call_sub:			/* call a struct (psc) */
  {
#ifdef DEBUG_CALL 
    printf("==> call_sub");
    printf("(name : %s/%d, type : %x)\n",GET_NAME(sym_ptr),GET_ARITY(sym_ptr),GET_ETYPE(sym_ptr)); 
#endif 
    switch (GET_ETYPE(sym_ptr)) {
    case T_ORDI :
      RAISE_EXCEPTION(undefined_predicate,make_struct2("/",ADDTAG(insert_cfunc(GET_NAME(sym_ptr),0),ATM),MAKEINT(GET_ARITY(sym_ptr))));
      
      break;
    case T_DYNA :
/*      printf("\n*** call dynamic pred %s/%i***\n",GET_NAME(sym_ptr),GET_ARITY(sym_ptr)); */
      arity = GET_ARITY(sym_ptr);
      if (arity==0) {  /* rearrange the frame */
	INCR_REFER_Y("call_sub",5);
	*(AR-2) = AR_CPS(AR);
	*(AR-1) = AR_AR(AR);
	*AR = ADDTAG(sym_ptr,ATM);
	AR--;
      }
      else {
	tmp_op = ADDTAG(H,STR);
	toam_NEW_HEAP_NODE((LONG)sym_ptr);
	INCR_REFER_Y("call_sub",arity);
	for (i=1;i<=arity;i++) {
	  toam_BUILD_VALUE(GRA(i,arity),call_sub_dyn);
	}
	*(AR+1) = tmp_op;
      }
      P = (LONG_PTR) GET_EP(enter_dyn_call);
      CONTCASE_R;	   
      break;
    case T_PRED :
      P = (LONG_PTR) GET_EP(sym_ptr);
      CONTCASE_R;	   
      break;
    case C_PRED :
      SAVE_REGS; sfreg=SF; local_top = LOCAL_TOP-2;
      op1 = ((*GET_EP(sym_ptr))());
      RESTORE_REGS;
      if (op1==1) {
	P = (LONG_PTR)AR_CPS(AR);
	AR = (LONG_PTR)AR_AR(AR);
	CONTCASE_R;	   
      }
      else if (op1==0) {
        BACKTRACK;
      }
      else {
	RAISE_EXCEPTION(exception,make_struct_with_args(AR,sym_ptr));
      }
      break;
    }
  }

 backtrack00:
  INCR_REFER_Y("backtrack00",1);
  H = (LONG_PTR)HB;

  {LONG_PTR top;
   top = (LONG_PTR)AR_T(AR);
   while (T != top) {
     POPTRAIL(T);
   }
 }
/*
  top = (LONG_PTR)AR_T(AR);
  while (T != top) {
    POPTRAIL(top);
  }
  T = (LONG_PTR)AR_T(AR);
*/
  P = (LONG_PTR)*P;
  CONTCASE_R;

  /*------------------------------------------------------------------*/
 catch_exception:{
   LONG_PTR parent_ar = (LONG_PTR)AR_AR(AR);

   LOCAL_TOP -= 256;  /* assume that the limit of frame size is 256 */
   *LOCAL_TOP-- = exception; 
                    /* the computation cannot restart from the frame because
                       registers are not saved */
   *LOCAL_TOP-- = goal;
   *LOCAL_TOP = (LONG)parent_ar;
   *(LOCAL_TOP-1) = AR_CPS(parent_ar);
   AR = LOCAL_TOP;

   if (GET_ETYPE(handle_exception_psc)==T_ORDI) {
     printf("Exception :");
     write_term(exception);
     printf(":");
     write_term(goal);
     printf("\n");
     B = (LONG_PTR)UNTAGGED((LONG)GET_EP(interrupt_psc)); 
     goto backtrack;
   }
   else {
     P = (LONG_PTR)GET_EP(handle_exception_psc);
     CONTCASE_R;
   }
 }  
  /*------------------------------------------------------------------*/
 setarg1:
  {
    profile_COUNT_REFERS(3);
    op1 = *P++;OP_CONT_DEREF(op1);
    if (!ISINT(op1)){
      RAISE_EXCEPTION(integer_expected,make_struct3("setarg",op1,op2,op3));
    }
    else op1 = INTVAL(op1);
  setarg_idd:
    op2 = *P++;OP_CONT_DEREF(op2);
    op3 = *P++;OP_CONT_DEREF(op3);
    if (IS_SUSP_VAR(op3)) op3 = UNTAGGED(op3);
    if (ISSTRUCT(op2))
      if (op1 <= 0 || op1 > GET_STR_SYM_ARITY(op2)) {
	RAISE_EXCEPTION(out_of_range,make_struct3("setarg",MAKEINT(op1),op2,op3));
      }
      else {
	tmp_op = (LONG)((LONG_PTR)(UNTAGGED(op2))+op1);
	toam_PUSHTRAILB((LONG_PTR)tmp_op,*(LONG_PTR)tmp_op);
	*(LONG_PTR)tmp_op = op3;
	CONTCASE_R;
      }
    else if (ISLIST(op2))
      if (op1<=0 || op1 > 2) {
	RAISE_EXCEPTION(out_of_range,make_struct3("setarg",MAKEINT(op1),op2,op3));
      }
      else {
	tmp_op = (LONG)((LONG_PTR)UNTAGGED(op2)+op1-1);
	toam_PUSHTRAILB((LONG_PTR)tmp_op,*(LONG_PTR)tmp_op);
	*(LONG_PTR)tmp_op = op3;
	CONTCASE_R;
      }
    else {
      RAISE_EXCEPTION(structure_expected,make_struct3("setarg",MAKEINT(op1),op2,op3));
    }
  }

  /*------------------------------------------------------------------*/
 arity1:
  {
    profile_COUNT_REFERS(2);
    op1 = *P++;
    op2 = *P++;
    switch (op1 & 0x7) { 
    case XV:  
    case XV1: 
    case YV: 
      toam_NEW_VAR(tmp_op);
      RAISE_EXCEPTION(structure_expected,make_struct3("functor",op1,tmp_op,op2));
      break;
    case XU: 
    case XU1: op1=XC(((LONG)(op1) & VAL_MASK)); goto arity_ud;
    case YU: op1=YC(op1>>3);goto arity_ud;
    default: 
      break;
    }
  arity_ud:
    switch (TAG(op1)) {
    case REF:
      NDEREF(op1,arity_ud);
      toam_NEW_VAR(tmp_op);
      RAISE_EXCEPTION(structure_expected,make_struct3("functor",op1,tmp_op,op2));
      break;
    case ATM:
      FAST_UNIFY(op2,MAKEINT(0));
      break;
    case LST:
      FAST_UNIFY(op2,MAKEINT(2));
      break;
    case STR:
      sym_ptr = GET_STR_SYM_REC(op1);
      FAST_UNIFY(op2,MAKEINT(GET_ARITY(sym_ptr)));
      break;
    }
    CONTCASE_R;
  }

  /*------------------------------------------------------------------*/
  mul1:
  if (ISINT(op1)) 
    if (ISINT(op2))
      FOLLOW(op3) = MAKEINT(INTVAL(op1) * INTVAL(op2));
    else if (ISFLOAT(op2)){
      toam_ENCODEFLOAT(op3,(double)INTVAL(op1) * floatval(op2));}
    else 
      goto external_mul;
  else  if (ISFLOAT(op1))
    if (ISINT(op2)) {
      toam_ENCODEFLOAT(op3,floatval(op1) * (double)INTVAL(op2));}
    else if (ISFLOAT(op2)){
      toam_ENCODEFLOAT(op3,floatval(op1) * floatval(op2));}
    else
      goto external_mul;
  else goto external_mul;
  CONTCASE_R;
 external_mul:
  SAVE_H;
  if (MUL(op1,op2,op3)<0){
    RAISE_EXCEPTION(exception,make_struct2("*",op1,op2));
  }
  RESTORE_H;
  CONTCASE_R;
  
  /*------------------------------------------------------------------*/
 div1:
  if (ISINT(op1))
    if (ISINT(op2)){
      op2 = INTVAL(op2);
      if (op2==0){
	RAISE_EXCEPTION(divide_by_zero,make_struct2("/",op1,MAKEINT(op2)));
      }
      toam_ENCODEFLOAT(op3,(double)INTVAL(op1) / (double)op2);
    }
    else if (ISFLOAT(op2)){
      if (floatval(op2)==0.0){
	RAISE_EXCEPTION(divide_by_zero,make_struct2("/",op1,op2));
      }
      toam_ENCODEFLOAT(op3,(double)INTVAL(op1) / floatval(op2));
    } 
    else goto external_div;
  else if (ISFLOAT(op1)) 
    if (ISINT(op2)){
      op2 = INTVAL(op2);
      if (op2==0){
	RAISE_EXCEPTION(divide_by_zero,make_struct2("/",op1,MAKEINT(op2)));
      }
      toam_ENCODEFLOAT(op3,floatval(op1) / (double)op2);
    }
    else if (ISFLOAT(op2)){
      if (floatval(op2)==0.0){
	RAISE_EXCEPTION(divide_by_zero,make_struct2("/",op1,op2));
      }
      toam_ENCODEFLOAT(op3,floatval(op1) / floatval(op2));
    }
    else goto external_div;
  else 
    goto external_div;
  CONTCASE_R;
 external_div:
  SAVE_H;
  if (DIV(op1,op2,op3)<0){
    RAISE_EXCEPTION(exception,make_struct2("/",op1,op2));
  }
  RESTORE_H;
  CONTCASE_R;
  
  /*------------------------------------------------------------------*/
  idiv1:
  if (ISINT(op1) && ISINT(op2)){
    op2 = INTVAL(op2);
    if (op2==0){
      RAISE_EXCEPTION(divide_by_zero,make_struct2("//",op1,MAKEINT(op2)));
    }
    FOLLOW(op3) = MAKEINT((int)(INTVAL(op1) / op2));
  }
  else {
    SAVE_H;
    if (IDIV(op1,op2,op3)<0) {
      RAISE_EXCEPTION(exception,make_struct2("//",op1,op2));
    }
    RESTORE_H;
  }
  CONTCASE_R;

  /*------------------------------------------------------------------*/
  mod1:
  if (ISINT(op1) && ISINT(op2)){
    op2 = INTVAL(op2);
    if (op2==0){
      RAISE_EXCEPTION(divide_by_zero,make_struct2("mod",op1,MAKEINT(op2)));
    }
    FOLLOW(op3) = MAKEINT((int)(INTVAL(op1) % op2));
  }
  else {
    SAVE_H;
    if (MOD(op1,op2,op3)<0){
      RAISE_EXCEPTION(exception,make_struct2("mod",op1,op2));
    }
    RESTORE_H;
  }
  CONTCASE_R;

  /*------------------------------------------------------------------*/
 domain_set_false1:{
   DEREF_NONVAR(op2);
   DEREF_NONVAR(op1);
/*
   {  printf("set_false "); write_term(op1); printf(" : %d  %d  ",DV_first(dv_ptr),DV_last(dv_ptr)); printf("  ");write_term(op2); printf("\n"); }
*/
   if (!IS_SUSP_VAR(op1)){  /* be sure to be a domain var */
     if (op1 == op2){
       BACKTRACK;
     }
     CONTCASE_R;
   }
   dv_ptr = (LONG_PTR)UNTAGGED(op1);  
   op1 = DV_size(dv_ptr);
   op2 = INTVAL(op2); 
   if (op1 == 2){ 
     if (op2 == DV_first(dv_ptr)) 
       op2 = DV_last(dv_ptr); 
     else if (op2 == DV_last(dv_ptr)) 
       op2 = DV_first(dv_ptr); 
     else CONTCASE_R;
     BIND_DVAR_VALUE(dv_ptr,MAKEINT(op2)); 
     CONTCASE_R;
   } 
   if (op2 < DV_first(dv_ptr) || op2 > DV_last(dv_ptr)) CONTCASE_R;
   i = 1; /* no of excluded elms */
   if (IS_IT_DOMAIN(dv_ptr)){ 
     if (op2==DV_first(dv_ptr)){
       toam_UPDATE_FIRST(dv_ptr,op2+1);
     } else if (op2==DV_last(dv_ptr)){
       toam_UPDATE_LAST(dv_ptr,op2-1);
     } else {
       toam_UPDATE_TYPE(dv_ptr,BV_DOMAIN);
       SAVE_H; SAVE_HB,SAVE_T; 
       it_to_bv(dv_ptr);
       exclude_one_elm_bv(dv_ptr,op2);
       RESTORE_H_T; 
     }
   } else {
     if (op2==DV_first(dv_ptr)){
       op2 = domain_next_bv(dv_ptr,op2+1);
       toam_UPDATE_FIRST(dv_ptr,op2);
     } else if (op2==DV_last(dv_ptr)){
       op2 = domain_prev_bv(dv_ptr,op2-1);
       toam_UPDATE_LAST(dv_ptr,op2);
     } else {
       SAVE_H;SAVE_HB; SAVE_T;
       i = exclude_one_elm_bv(dv_ptr,op2);
       RESTORE_H_T;
     }
   }
   if (i==1){
     toam_UPDATE_SIZE(dv_ptr,op1-1);
   }
   CONTCASE_R;
 }

  /*------------------------------------------------------------------*/
 domain_region_1:{
   dv_ptr = (LONG_PTR)UNTAGGED(op1); /* dv_ptr */
/*
   printf("domain_region %x:%d(%d,%d) in %d %d\n",op1,DV_type(dv_ptr),DV_first(dv_ptr),DV_last(dv_ptr),op2,op3);
*/
   if (op2 <= DV_first(dv_ptr) && op3 >= DV_last(dv_ptr))
     CONTCASE_R;
   
   if (IS_IT_DOMAIN(dv_ptr)){
     op1 = DV_first(dv_ptr); /* update first */
     if (op2>op1){
       if (op2>DV_last(dv_ptr)) BACKTRACK;
       toam_UPDATE_FIRST(dv_ptr,op2);       
     } else op2 = op1;
     
     op1 = DV_last(dv_ptr); /* update last */
     if (op3<op1){
       if (op3<op2) BACKTRACK;
       toam_UPDATE_LAST(dv_ptr,op3);
     } else op3 = op1;

     arity = op3-op2+1; /* new size */
   } else {
     arity = 0;
     op1 = DV_first(dv_ptr); /* update first */
     if (op2>op1){
       if (op2>DV_last(dv_ptr)) BACKTRACK;
       arity = count_domain_elms(dv_ptr,op1,op2-1); /* borrow arity here */
       op2 = domain_next_bv(dv_ptr,op2);
       toam_UPDATE_FIRST(dv_ptr,op2);       
     } else op2 = op1;

     op1 = DV_last(dv_ptr); 
     if (op3<op1){
       if (op3 < op2) BACKTRACK;
       arity += count_domain_elms(dv_ptr,op3+1,op1);
       op3 = domain_prev_bv(dv_ptr,op3);
       toam_UPDATE_LAST(dv_ptr,op3);
     } 

     arity = DV_size(dv_ptr)-arity;
     if (arity==0) BACKTRACK;
   }
   if (arity == 1){
     BIND_DVAR_VALUE(dv_ptr,MAKEINT(op2));
     CONTCASE_R;
   }
   toam_UPDATE_SIZE(dv_ptr,arity); /* new count */
   CONTCASE_R;
 }

 domain_region_min_1:{
   dv_ptr = (LONG_PTR)UNTAGGED(op1); /* dv_ptr */
   if (op2<=DV_first(dv_ptr)) CONTCASE_R;
   if (op2>DV_last(dv_ptr)) BACKTRACK;
   if (IS_IT_DOMAIN(dv_ptr)){
     arity = DV_last(dv_ptr)-op2+1; /* new size */
   } else {
     op1 = DV_first(dv_ptr); /* update first */
     arity = count_domain_elms(dv_ptr,op1,op2-1); /* borrow arity here */
     op2 = domain_next_bv(dv_ptr,op2);
     arity = DV_size(dv_ptr)-arity;
   }
   if (arity==1){
     BIND_DVAR_VALUE(dv_ptr,MAKEINT(op2));
     CONTCASE_R;
   }
   toam_UPDATE_FIRST(dv_ptr,op2); 
   toam_UPDATE_SIZE(dv_ptr,arity); /* new count */
   CONTCASE_R;
 }

 domain_region_max_1:{
   dv_ptr = (LONG_PTR)UNTAGGED(op1); /* dv_ptr */
   if (op3>=DV_last(dv_ptr)) CONTCASE_R;
   if (op3 < DV_first(dv_ptr)) BACKTRACK;
   if (IS_IT_DOMAIN(dv_ptr)){
     arity = op3-DV_first(dv_ptr)+1; /* new size */
   } else {
     op1 = DV_last(dv_ptr); 
     arity = count_domain_elms(dv_ptr,op3+1,op1);
     op3 = domain_prev_bv(dv_ptr,op3);
     arity = DV_size(dv_ptr)-arity;
   }
   if (arity==1){
     BIND_DVAR_VALUE(dv_ptr,MAKEINT(op3));
     CONTCASE_R;
   }
   toam_UPDATE_LAST(dv_ptr,op3);
   toam_UPDATE_SIZE(dv_ptr,arity); /* new count */
   CONTCASE_R;
 }


}                             /* end toam.c */


    


