/* unify.c */

#include "basic.h"
#include "term.h"
#include "macro.h"
#include "event.h"

#define UNIFY_DVAR_SUSP(op1,op2) bind_susp_susp(op1,op2)

#define CHECK_INTEGER(op) \
  if (!ISINT(op)){ \
     return 0; \
  }

extern double floatval();

unify(op1, op2)
LONG op1, op2;
{
   register LONG_PTR top;
   register WORD     arity, i;
   LONG tmp_op;

   SWITCH_OP(op1,nunify,
	     {goto unify_v_d;},
	     {goto unify_atomic_d;},
	     {goto unify_lst_d;},
	     {goto unify_str_d;},
	     {goto unify_susp_d;});
 unify_v_d: 
   if (ISVAR(op2)){
     NDEREF(op2,unify_v_d);
     if (op1 != op2) {
       if (op1 < op2) {
	 if (op1 < (LONG)hreg) {	/* op1 not in loc stack */
	   FOLLOW(op2) = op1;
	   PUSHTRAIL(op2);
	 } else {		/* op1 points to op2 */
	   FOLLOW(op1) = op2;
	   PUSHTRAIL(op1);
	 }
       } else {		/* op1 > op2 */
	 if (op2 < (LONG) hreg) {
	   FOLLOW(op1) = op2;
	   PUSHTRAIL(op1);
	 } else {
	   FOLLOW(op2) = op1;
	   PUSHTRAIL(op2);
	 }
       }
     }
     return 1;
   } if (IS_SUSP_VAR(op2)){
     FOLLOW(op1) = UNTAGGED(op2);
     PUSHTRAIL(op1);
     return 1;
   } else {
     FOLLOW(op1) = op2;
     PUSHTRAIL(op1);
     return 1;
   }
    
/* unify_atomic_d: */
   SWITCH_OP_ATM(op2,unify_atomic_d,
		 {FOLLOW(op2) = op1;
		  PUSHTRAIL(op2);
		  return 1;},
		 {if (op1==op2) return 1;},
		 {goto bind_value_susp;});
   return 0;
	     
/* unify_lst_d: */
   SWITCH_OP_LST(op2,unify_lst_d,
		 {FOLLOW(op2) = op1;
		  PUSHTRAIL(op2);
		  return 1;},
		 
		 {if (op1 != op2) {
		   UNTAG(op1);
		   UNTAG(op2);
		   if (!unify(*(LONG_PTR) op1, *(LONG_PTR) op2) ||
		       !unify(*((LONG_PTR) op1 + 1), *((LONG_PTR) op2 + 1)))
		     return 0;
		 }
		  return 1;},
		 
		 {goto bind_value_susp;});
   return 0;

/* unify_str_d: */
   SWITCH_OP_STRUCT(op2,unify_str_d,
		 {FOLLOW(op2) = op1;
		  PUSHTRAIL(op2);
		  return 1;},

		 {if (op1 != op2) {
		   UNTAG(op1);
		   UNTAG(op2);
		   if (FOLLOW(op1) != FOLLOW(op2))
		     return 0;
		   else {
		     arity = GET_ARITY((SYM_REC_PTR)FOLLOW(op1));
		     for (i = 1; i <= arity; i++)
		       if (!unify(*((LONG_PTR) op1 + i), *((LONG_PTR) op2 + i))) {
			 return 0;
		       }
		   }
		 }
		  return 1;},
		 
		 {goto bind_value_susp;});
   return 0;
   
/* unify_susp_d: */
   SWITCH_OP_VAR(op2,unify_susp_d,
		 {FOLLOW(op2) = UNTAGGED(op1); 
		  PUSHTRAIL(op2); 
		  return 1;},
		 {return unify_suspvar_suspvar(op1,op2);},
		 {goto bind_susp_value;});


 bind_value_susp:
   tmp_op = op1; op1 = op2; op2 = tmp_op;
 bind_susp_value:
   top = (LONG_PTR)(op1 & VAL_MASK); 
   if (!IS_UN_DOMAIN(top)){
     if (!ISINT(op2)) return 0;
     tmp_op = INTVAL(op2);
     if (tmp_op<DV_first(top) || tmp_op>DV_last(top)) return 0;
     if (!IS_IT_DOMAIN(top) && !domain_true_bv(top,tmp_op)) return 0;
     INSERT_TRIGGER_dvar_ins(top); 
   } else {
     INSERT_TRIGGER_var_ins(top); 
   }
   PUSHTRAILB(top,op1); 
   FOLLOW(top) = op2; 
   return 1;

 }

int is_IDENTICAL(op1, op2)
LONG op1, op2;
{
  return identical(op1, op2);
}

int identical(op1, op2)
register LONG op1, op2;
{
  register LONG_PTR top;
  register WORD     arity, i;
  
  DEREF(op2);
  SWITCH_OP(op1,nidentical,
	    {if (op1 == op2) return 1;},
	    {if (op1 == op2) return 1;},
	    {if (ISLIST(op2)) {
	      if (op1 != op2) {
		UNTAG(op1);
		UNTAG(op2);
		if (!identical(*(LONG_PTR)op1, *(LONG_PTR)op2) ||
		    !identical(*(((LONG_PTR)op1)+1), *(((LONG_PTR)op2)+1)))
		  return 0;
	      }
	      return 1;
	    }},
	    {if (ISSTRUCT(op2)){
	      if (op1 != op2) {              /* a != b */
		UNTAG(op1);
		UNTAG(op2);
		if (FOLLOW(op1) != FOLLOW(op2)) /* 0(a) != 0(b) */
		  return 0;
		else {
		  arity = GET_ARITY((SYM_REC_PTR)(FOLLOW(op1)));
		  for (i = 1; i <= arity; i++) 
		    if (!identical(*((LONG_PTR)op1 + i), *((LONG_PTR)op2 + i)))
		      return 0;
		}
	      }
	      return 1;
	    }},
	    {if (op1 == op2) return 1;});
  return 0;
}

is_UNIFIABLE(op1,op2)
TERM op1,op2;
{
  LONG tmp_op;
  LONG_PTR top2;

/* Logically equivalent to not(not(op1 = op2)).   */
  top2 = treg;
  op1 =unify(op1, op2);
  while (treg != top2) {    /* undo bindings, if any */
    POPTRAIL(treg);
  }
  return op1;
}
  
c_UNIFIABLE(){
  register LONG op1,op2;
  LONG tmp_op;
  LONG_PTR top2;
  op1 = ARG(1,2); op2 = ARG(2,2);
  return is_UNIFIABLE(op1,op2);
}

unify_suspvar_suspvar(op1,op2)
LONG op1,op2;
{
  LONG_PTR dv_ptr1,dv_ptr2;

  if (op1==op2) return 1;
  dv_ptr1 = (LONG_PTR)(op1 & VAL_MASK);
  dv_ptr2 = (LONG_PTR)(op2 & VAL_MASK);
  if (!IS_UN_DOMAIN(dv_ptr1))
    if (!IS_UN_DOMAIN(dv_ptr2)) {
      return unify_dvar_dvar(dv_ptr1,dv_ptr2);
    }  else {
      INSERT_TRIGGER_var_ins(dv_ptr1);
      return UNIFY_DVAR_SUSP(dv_ptr1,dv_ptr2);
    }
  else
    if (!IS_UN_DOMAIN(dv_ptr2)) {
      INSERT_TRIGGER_var_ins(dv_ptr2);
      return UNIFY_DVAR_SUSP(dv_ptr2,dv_ptr1);
    }
    else {
      INSERT_TRIGGER_var_ins(dv_ptr2);
      return unify_susp_susp(dv_ptr1,dv_ptr2);
    }
}

unify_susp_susp(dv_ptr1,dv_ptr2)
LONG_PTR dv_ptr1,dv_ptr2;
{
  if (dv_ptr1>dv_ptr2) 
    return bind_susp_susp(dv_ptr2,dv_ptr1);
  else
    return bind_susp_susp(dv_ptr1,dv_ptr2);
}


unify_dvar_dvar(dv_ptr1,dv_ptr2)
LONG_PTR dv_ptr1,dv_ptr2;
{
  long int first,last;
  LONG op1;
  
  if (!domain_intersect(dv_ptr1,dv_ptr2)) return 0;
  if (!domain_intersect(dv_ptr2,dv_ptr1)) return 0;
  return bind_susp_susp(dv_ptr1,dv_ptr2);
}

bind_susp_susp(dv_ptr1,dv_ptr2)
LONG_PTR dv_ptr1,dv_ptr2;
{
  LONG_PTR top,tmp_op;

  PUSHTRAILB((LONG)dv_ptr2,FOLLOW(dv_ptr2));
  FOLLOW(dv_ptr2) = (LONG)dv_ptr1;
  move_dv_cs(A_DV_ins_cs(dv_ptr1),DV_ins_cs(dv_ptr1),DV_ins_cs(dv_ptr2));
  if (!IS_UN_DOMAIN(dv_ptr1) && !IS_UN_DOMAIN(dv_ptr2)){
    move_dv_cs(A_DV_min_cs(dv_ptr1),DV_min_cs(dv_ptr1),DV_min_cs(dv_ptr2));
    move_dv_cs(A_DV_max_cs(dv_ptr1),DV_max_cs(dv_ptr1),DV_max_cs(dv_ptr2));
    move_dv_cs(A_DV_dom_cs(dv_ptr1),DV_dom_cs(dv_ptr1),DV_dom_cs(dv_ptr2));
  }
  return 1;
}


move_dv_cs(addr1,cs1,cs2)
LONG_PTR addr1,cs1,cs2;
{
  while (ISLIST((LONG)cs1)){ 
    cs1 = (LONG_PTR)UNTAGGED(cs1);
    addr1 = CS_NEXT_ADDR(cs1);
    cs1 = (LONG_PTR)FOLLOW(addr1); 
  } 
  PUSHTRAILB((LONG)addr1,(LONG)cs1); 
  FOLLOW(addr1) = (LONG)cs2;
}



